--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- Various AST checks 
--  o patterns
--  o value re-definitions

module ASTChecks (performASTChecks) 
where

import Ann_InputAST
import TD_Monad
import TD_General
import TD_StdDefs (stdTypes)
import Terminal (uncommas)
import Misc

import Pretty

import Char
import List
import Monad


--------------------------------------------------------------------------------

-- performs various checks on the structure of the program 
-- program is rejected if:
--  - an overloaded id is used as a function id
--  - a non-overloaded id is declared multiple times
--  - all clauses of some declaration don't have the same number of patterns
performASTChecks :: Prog -> TD Bool
performASTChecks p@(Prog adecs) = do
	err <- sequence [idErrRep, repErrRep, npatErrRep, instErrRep]
	if any (==False) err 
	 then return False
	 else -- report warnings
	      return True
    where
    
	-- find val ids which are also over ids
	idErrRep  = report idErr $
		     puts ("ERROR: following value ids are also " ++
			   "overloaded ids: " ++ 
			   uncommas [ id | (b, id) <- idOvers, b ])
	    where idErr = any fst idOvers
		  idOvers = zip (map (`elem`oids) vids) vids


	-- find val decs that share an id
    	repErrRep = report repErr $
		     puts ("ERROR: following values are declared " ++
			   "more than once: " ++ uncommas (nub repDef))
	    where repErr = not (null repDef)
		  repDef = [ id | (d, ods) <- allLeft evals, let id = decId d,
						   any ((id==) . decId) ods ]

	-- find vals with defs of varying numbers of patterns
	npatErrRep = report npatErr $
		      puts ("ERROR: following functions have varying numbers" ++
			    " of patterns: " ++ uncommas (nub npatDec))
	    where npatErr = not (null npatDec)
		  npatDec = [ decId d | d <- ovals, 
					let fs = decDefs d
					    ns = map (defNumPats.item) fs,
					not (allEqual ns) ]

	-- find instance decs where the constraint name does not match the
	-- function definition
	instErrRep = report instErr $
		      puts ("ERROR: instance predicate name doesn't match " ++
			    "definition name:\n" ++ instTxt)
	    where instErr = not (null mixedInst)
		  mixedInst = [ id | id@(InstDec s ds) <- ids,
				    let sid = instSigId (item s)
					did = defId (item (head ds)),
				    not (null ds),
				    sid /= up did ]
		  instTxt = pprintPlain (head mixedInst)
		  up "" = ""
		  up (c:cs) = toUpper c : cs
				     

	report cond io | cond	   = io >> return False
		       | otherwise = return True

	------------ 
	
	(_,_,ods,o0ds,vds,ads,ids) = splitDecs (items adecs)
	vals  = vds ++ ads
	evals = vals ++ items (externDecs adecs) ++ 
		map (ExternDec . Right . nullAnn . show) stdTypes 
	ovals= ods ++ vals 
	vids = nub (map decId (vds ++ ads))
	oids = nub (map decId (ods ++ o0ds))


--------------------------------------------------------------------------------

-- given a list of patterns (from the same column of a group of clauses),
-- returns those which are redundant (ie. will never be matched)
-- note: types are not considered
redundantPatterns :: [Pat] -> [Pat]
redundantPatterns ps = proc ps
    where 

    proc [] = []
    proc (p:ps) = filter (red p) ps ++ proc ps
  
    -- true if second pattern is made redundant by the first
    red (JokPat)   p2 = True
    red (VarPat _) p2 = True
    red (LitPat l1) (LitPat l2) = l1 == l2
    red (ConPat c1 ps1) (ConPat c2 ps2) = item c1 == item c2 &&
				      and (zipW red (items ps1) (items ps2))
    red (TupPat ps1)  (TupPat ps2)  = and (zipW red (items ps1) (items ps2))
    red (ListPat ps1) (ListPat ps2) = and (zipW red (items ps1) (items ps2))
    red (TupPat  _) _ = False
    red (ListPat _) _ = False
    red (LitPat _) _ = False
    red (ConPat _ _) p2 = False

    zipW f xs ys | length xs /= length ys = [False]
	         | otherwise = zipWith f xs ys



