--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------



-- simplified versions of the input abstract syntax tree

module ASTs (VarRef(..), VarTree(..), 
	     vtProg, vtDec, parseVarRef, lookupVarTree,
	     vtDefLocs,
	     
	     LocTree(..), ltProg, nodeLoc, findLocNode, 
	     nodeSigType, nodeSigId, 
	     commonParentLoc, parentLoc, grandParentLoc)
where

import Ann_InputAST

import Char
import Maybe
import IOExts

type IdLoc = (Id, Loc)

--------------------------------------------------------------------------------

-- references to arbitrary variables in a program
-- encodes regular language: ((\;)|,)var(#n)?)+
data VarRef = NoRef
	    | LetRef Id RefNum VarRef
	    | VarRef Id RefNum
     deriving Show

type RefNum = Int

-- attempts to parse a string representing a VarRef
parseVarRef :: String -> Maybe VarRef
parseVarRef [] = Nothing
parseVarRef cs@(c:_) | c /= ';'  = p (';':cs)
		     | otherwise = p cs
    where p :: String -> Maybe VarRef
	  p (';':';':cs) = let (id, cs0) = spanId cs
			       (num,cs1) = n cs0
			   in  return (VarRef id num)

	  p (';':cs) = do let (id, cs0) = spanId cs
			      (num,cs1) = n cs0
			  vr <- p cs1
		          return $ LetRef id num vr

	  p [] = return NoRef
	  p _  = fail "haha"

	  n ('#':cs) = let (m, cs') = spanNum cs
		       in  (read m, cs')
	  n cs = (0, cs)    -- zero signifies that no number was specified
			    -- it will be treated as a 1 essentially

	  spanId  cs = span isAlphaNum cs
	  spanNum cs = span isDigit cs

--------------------------------------------------------------------------------

-- looks up the given ref. in the tree, returning either 
-- the id/loc that it represents (Left), or the Id that wasn't found (Right)
lookupVarTree :: VarRef -> VarTree -> Either IdLoc Id 
lookupVarTree vr vt = lookup vr vt 
    where 
	lookup :: VarRef -> VarTree -> Either IdLoc Id
	lookup (VarRef id n)    (VT _ _ idls _)     = 
		let ls = [ idl | idl@(id',_) <- idls, id == id' ]
		    n' = if n==0 then 1 else n
		in  if n' > length ls 
		    then Right id
		    else Left (ls!!(n'-1))
			
	-- final LetRef, if its a 1-ref, it refers to the entire declaration
	-- (not just the first clause)
	lookup (LetRef id 0 NoRef) (VT _ _ _ vts) =
		let vts' = filter (`vtHasId`id) vts
		    VT ld _ _ _ = head vts'
		in  if null vts'
		    then Right id
		    else Left (id,ld)
			
	lookup (LetRef id n vr) (VT _ _ _ vts) =
		let vts' = filter (`vtHasId`id) vts
		    n' = if n==0 then 1 else n
		in  if n' > length vts'
		    then Right id
		    else lookup vr (vts'!!(n'-1))

	lookup NoRef (VT _ idl _ _) = Left idl
			
    

--------------------------------------------------------------------------------
-- for each definition, contains a list of the variables occurring within
-- (in order of their appearance), as well as a list of sub-definitions (also
-- in order of appearance.)
data VarTree = VT Loc		-- location of the declaration it belongs to
		  IdLoc		-- current definition/binding
		  [IdLoc]	-- local variables
		  [VarTree]	-- sub-definitions
     deriving Show

vt l = VT l ("", noLoc) 
emptyVT = vt noLoc [] []


addVT (VT dl d1 ils1 vts1) 
      (VT _  _  ils2 vts2) = VT dl d1 (ils1++ils2) (vts1++vts2)

sumVTs = foldr addVT emptyVT

sumMap f = sumVTs . map f


vtHasId :: VarTree -> Id -> Bool
vtHasId (VT _ (id1,_) _ _) id2 = id1 == id2

-- given a var tree, returns the locations of all definitions within it
vtDefLocs :: VarTree -> [(Id,Loc)]
vtDefLocs (VT l (id,_) _ vts) = (id,l) : concatMap vtDefLocs vts

----------------------------------------

vtProg :: Prog -> VarTree
vtProg (Prog decs) = vt noLoc [] (vtDecs decs)

vtDecs :: [Ann Dec] -> [VarTree]
vtDecs = concatMap vtDec
vtDec dec =
    let a  = ann dec
	l  = ixLoc a
	af = (a, False)
    in  case item dec of
	    ValDec defs	   -> vtDefs l defs
	    AValDec s defs -> vtDefs l defs
	    OverDec s defs -> vtDefs l defs
	    InstDec s defs -> vtDefs l defs
	    _		   -> []

--vtDefs :: [Ann Def] -> [VarTree]
vtDefs ld = map (vtDef ld)
--vtDef :: Ann Def -> VarTree
vtDef ld def =
    case item def of
	 Def id ps rhs w -> 
		let vt' = VT ld (item id, ixLoc (ann def)) 
		    VT _ _ l x = vtPats ld ps `addVT` vtRHS ld rhs `addVT` 
				 vt ld [] (vtDecs w)
		in  vt' l x

vtRHS ld (RHS1 ae)  = vtExp ld ae
vtRHS ld (RHSG ges) = vtGEs ld ges

vtGEs ld = sumMap (vtGE ld)
vtGE ld (g, e) = vtExp ld g `addVT` vtExp ld e


vtExps ld = sumMap (vtExp ld)
--vtExp :: Ann Exp -> VarTree
vtExp ld exp =
    let lf id = (id, ixLoc (ann exp))
	vt' = vt ld
	vtExp'  = vtExp ld
	vtExps' = vtExps ld
    in	case item exp of
	    VarExp id	    -> vt' [lf id] []
	    ConExp id	    -> vt' [lf id] []
	    AbsExp ps e1    -> vtPats ld ps `addVT` vtExp' e1
	    AppExp e1 e2    -> vtExp' e1 `addVT` vtExp' e2
	    IfxAppExp e1 e2 e3 -> vtExp' e1 `addVT` vtExp' e2 `addVT` vtExp' e3
	    IfExp  e1 e2 e3 -> vtExp' e1 `addVT` vtExp' e2 `addVT` vtExp' e3
	    AnnExp e1 t     -> vtExp' e1 
	    TAddExp e1 t    -> vtExp' e1
	    TupExp  es      -> vtExps' es
	    ListExp es      -> vtExps' es
	    SecExpLeft  e1 e2 -> vtExp' e1 `addVT` vtExp' e2 
	    SecExpRight e1 e2 -> vtExp' e1 `addVT` vtExp' e2
	    EnumExpFrom       e1	-> vtExp' e1
	    EnumExpFromTo     e1 e2	-> vtExp' e1 `addVT` vtExp' e2
	    EnumExpFromThen   e1 e2	-> vtExp' e1 `addVT` vtExp' e2
	    EnumExpFromThenTo e1 e2 e3  -> vtExp' e1 `addVT` vtExp' e2 `addVT`	
					   vtExp' e3					   
	    ListCompExp e ss -> vtExp' e `addVT` vtStmts ld ss
	    CaseExp e as     -> vtExp' e `addVT` vtAlts  ld as
	    LetExp ds e      -> vtExp' e `addVT` vt ld [] (concatMap vtDec ds)
	    
	    _ -> emptyVT

vtAlts ld = sumMap (vtAlt ld)
vtAlt ld (CaseAlt p e w) = vtPat ld p `addVT` vtExp ld e  `addVT` 
			   sumVTs (vtDecs w)
vtAlt ld (CaseAltGuard p ges w) = vtPat ld p `addVT` sumVTs (vtDecs w) `addVT` 
				  vtGEs ld ges

vtStmts ld = sumMap (vtStmt ld)
vtStmt ld stmt = case item stmt of
	    GenStmt p e -> vtPat ld p `addVT` vtExp ld e
	    QualStmt e  -> vtExp ld e
	    LetStmt ds  -> sumVTs (vtDecs ds)


vtPats ld = sumMap (vtPat ld)
vtPat ld pat =
    let lf id = (id, ixLoc (ann pat))
    in  case item pat of
            VarPat id	 -> vt ld [lf id] []
            ConPat id ps -> vt ld [(item id, ixLoc (ann id))] [] `addVT` 
			    vtPats ld ps
	    TupPat  ps	 -> vtPats ld ps
            ListPat ps	 -> vtPats ld ps
	    _		 -> emptyVT


--------------------------------------------------------------------------------
-- mirrors the hierarchy of program locations present in the input AST
-- all nodes being of the same type, it's easier to search for related locations
data LocTree = LocNode Loc		-- current location
		       [LocTree]	-- sub-trees of locations
     deriving Show

lt = LocNode
emptyLT = []
nodeLoc (LocNode l _) = l

----------------------------------------

-- returns the node of the given location
findLocNode :: LocTree -> Loc -> Maybe LocTree
findLocNode lt l = if null found then Nothing
				 else Just (head found)
    where found = find lt
	  find lt@(LocNode m lts) | l == m    = [lt]
				  | otherwise = concatMap find lts

nodeSubLocs :: LocTree -> [Loc]
nodeSubLocs (LocNode l lts) = l : concatMap nodeSubLocs lts


-- finds the first node in the tree that satisfies the given condition
findNodeShallow :: (LocTree -> Bool) -> LocTree -> Maybe LocTree
findNodeShallow p lt = if null found then Nothing 
				     else Just first
    where
	first = head found
	found = find lt
	find lt@(LocNode l lts) | p lt	    = [lt]
				| otherwise = concatMap find lts

-- finds the deepest node in the tree that satisfies the given condition
findNodeDeep :: (LocTree -> Bool) -> LocTree -> Maybe LocTree
findNodeDeep p lt = if isJust found then Just (snd (fromJust found))
				    else Nothing
    where
	found = find 1 lt

	find n lt@(LocNode _ []) | p lt	     = Just (n, lt)
				 | otherwise = Nothing
	find n lt@(LocNode l lts) =
	    -- try children
	    let dlts = filter isJust $ (if p lt
					then Just (n,lt)
					else Nothing    ) : map (find (n+1)) lts
	    in  deepest (map fromJust dlts)
	    
	    where
		deepest [] = Nothing
		deepest (x:xs) = Just (deepest' x xs)
		deepest' nlt [] = nlt
		deepest' nlt1@(n1,lt1) (nlt2@(n2,lt2):nlts) 
		    | n2 > n1	= deepest' nlt2 nlts
		    | otherwise	= deepest' nlt1 nlts
	    


-- finds the deepest node in the tree which is a parent of all the given 
-- locations 
-- note: pretty damn slow, I'd expect
commonParentLoc lt ls = findNodeDeep pred lt
    where
	pred lt = all (`elem`(nodeSubLocs lt)) ls

   

grandParentLoc :: Loc -> LocTree -> Maybe LocTree
grandParentLoc l lt = do
    lt'  <- parentLoc l lt
    let l' = nodeLoc lt'
    parentLoc l' lt
    
parentLoc :: Loc -> LocTree -> Maybe LocTree
parentLoc l lt = let lts = pl lt
		 in  if null lts then Nothing
				 else Just (head lts)
    where 
	pl (LocNode _ []) = []
	pl lt@(LocNode _ lts)  
	    | any (\n -> nodeLoc n == l) lts = [lt]
	    | otherwise	= concatMap pl lts

----------------------------------------

ltProg :: Prog -> LocTree
ltProg (Prog decs) = lt noLoc (ltDecs decs)

ltDecs = map ltDec
ltDec :: Ann Dec -> LocTree
ltDec dec = lt (annl dec) $ case item dec of
	    ValDec defs	   -> ltDefs defs
	    AValDec s defs -> ltSig s : ltDefs defs
	    OverDec s defs -> ltSig s : ltDefs defs
	    InstDec s defs -> ltInstSig s : ltDefs defs
	    _		   -> emptyLT

ltDefs :: [Ann Def] -> [LocTree]
ltDefs = map ltDef
ltDef def =
    lt (annl def) $ case item def of
	 Def id ps rhs w -> ltId id : ltPats ps ++ ltRHS rhs ++ ltDecs w

ltId :: Ann Id -> LocTree
ltId id = lt (annl id) []

ltRHS :: RHS -> [LocTree]
ltRHS (RHS1 ae)  = [ltExp ae]
ltRHS (RHSG ges) = ltGEs ges

ltGEs :: [GuardedExp] -> [LocTree]
ltGEs = concatMap ltGE
ltGE (g, e) = [ltExp g, ltExp e]


ltExps :: [Ann Exp] -> [LocTree]
ltExps = map ltExp
ltExp :: Ann Exp -> LocTree
ltExp exp =
    lt (annl exp)  $ case item exp of
	    VarExp id	    -> []
	    ConExp id	    -> []
	    AbsExp ps e1    -> ltPats ps ++ [ltExp e1]
	    AppExp e1 e2    -> [ltExp e1, ltExp e2]
	    IfxAppExp e1 e2 e3 -> [ltExp e1, ltExp e2, ltExp e3]
	    IfExp  e1 e2 e3 -> [ltExp e1, ltExp e2, ltExp e3]
	    AnnExp e1 t     -> [ltExp e1, ltTSchm t]
	    TAddExp e1 t    -> [ltExp e1, ltTSchm t]
	    TupExp  es      -> ltExps es
	    ListExp es      -> ltExps es
	    SecExpLeft  e1 e2 -> [ltExp e1, ltExp e2]
	    SecExpRight e1 e2 -> [ltExp e1, ltExp e2]
	    EnumExpFrom       e1	-> [ltExp e1]
	    EnumExpFromTo     e1 e2	-> [ltExp e1, ltExp e2]
	    EnumExpFromThen   e1 e2	-> [ltExp e1, ltExp e2]
	    EnumExpFromThenTo e1 e2 e3  -> [ltExp e1, ltExp e2, ltExp e3]
	    ListCompExp e ss -> ltExp e : ltStmts ss
	    CaseExp e as     -> ltExp e : ltAlts  as
	    LetExp ds e      -> ltExp e : ltDecs ds
	    
	    _ -> emptyLT

ltAlts :: [CaseAlt] -> [LocTree]
ltAlts = concatMap ltAlt
ltAlt (CaseAlt p e w)	     = ltPat p : ltExp e  : ltDecs w
ltAlt (CaseAltGuard p ges w) = ltPat p : ltDecs w ++ ltGEs ges

ltStmts = map ltStmt
ltStmt stmt = lt (annl stmt) $ case item stmt of
	    GenStmt p e -> [ltPat p, ltExp e]
	    QualStmt e  -> [ltExp e]
--	    LetStmt ds  -> ltDecs ds


ltPats = map ltPat
ltPat pat = lt (annl pat) $ case item pat of
            VarPat id	 -> []
            ConPat id ps -> ltId id : ltPats ps
	    TupPat  ps	 -> ltPats ps
            ListPat ps	 -> ltPats ps
	    _		 -> emptyLT


ltSig :: Ann Sig -> LocTree
ltSig sig = lt (annl sig) $ case item sig of
	    Sig _ id ctxt typ -> ltId id : ltCtxt ctxt : [ltTyp typ]

ltInstSig :: Ann InstSig -> LocTree
ltInstSig isig = lt (annl isig) $ case item isig of
	    InstSig ctxt pred -> ltCtxt ctxt : [ltPred pred]

ltCtxt :: Ann Ctxt -> LocTree
ltCtxt ctxt = lt (annl ctxt) $ case item ctxt of
	    Ctxt preds -> ltPreds preds 

ltPreds = map ltPred 
ltPred :: Ann Pred -> LocTree
ltPred pred = lt (annl pred) $ case item pred of
	    Pred _ t -> [ltTyp t]

ltTSchm :: Ann TSchm -> LocTree
ltTSchm ts = lt (annl ts) $ case item ts of
		(TSchm ctxt typ) -> ltCtxt ctxt : [ltTyp typ]

ltTyps = map ltTyp
ltTyp :: Ann Typ -> LocTree
ltTyp typ = lt (annl typ) $ case item typ of
	    VarTyp _ -> emptyLT
	    ConTyp _ -> emptyLT
	    AppTyp t1 t2 -> ltTyp t1 : [ltTyp t2]
	    TupTyp ts -> ltTyps ts
	    ListTyp t -> [ltTyp t]

-- given a node representing a Sig location, return the node representing the
-- signature's type
nodeSigType (LocNode _ [_lid, _lc, lt]) = lt	-- for Sigs
nodeSigType (LocNode _ [_le, lt]) = lt          -- for annon. anns  
nodeSigType l = trace ("nodeSigType: " ++ show l) lt 1 []

nodeSigId (LocNode _ [lid, _lc, _lt]) = lid	-- for Sigs
nodeSigId (LocNode _ [le, _lt]) = le		-- for annon. anns
nodeSigId l = trace ("nodeSigId: " ++ show l) lt 1 []

annl = ixLoc . ann
