--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- contains type annotation checks:
--  o (un)ambiguity check
--  o subsumption check

module AnnChecks (checkAnnotations)
where

import Core
import Justify
import SolverIO
import Chr
import Misc
import TD_Monad
import CHRChecks (equiv2, entail2, entail2DontSimplify, entail3)
import Pretty (pprintLoc)
import OptFlags (noChecks, noCkUnambigFlag, compileFlag, verbose, Flag)
import Ann_InputAST (Id)
import TD_StdDefs (tupleType)

import TDInference hiding (tdDeriver)
import TypeScheme

import ErrorMsg (errorSubsumption, errorAmbiguous, SubFailure)
import ConsReason (entailSubsetExp, entailSubsetSub)

import List
import Maybe
import Monad
import IOExts


--------------------------------------------------------------------------------

-- performs subsumption and unambiguity checks
-- returns False if there were any failures
-- the list of flags contains command-line options
checkAnnotations :: [Flag] -> TD Bool
checkAnnotations fs = do
    chrs <- getCHRs

    -- get all chrs generated from annotations, and their mdef. counterparts
    let achrs = filter isCheckableAnnCHR chrs
	-- ps is a list of matching ann/mdef chrs
	ps = [ (a, i) | a <- achrs, let is = filter (`fitsAnnCHR`a) chrs
					i  = head is,
					not (null is) ]

    -- pick the right subsumption test, based on whether we're compiling
    let (subsumptionTest, substr) = if compileFlag fs
				    then (checkSubsumption2Fast, "strong")
				    else (checkSubsumptionFast, "weak")

    -- all the "inferred-type" CHRs 
    let all_ichrs = filter isJoinCHR chrs 

    -- check that annotated types schemes are unambiguous
    (ann_unambig, ambig_achrs) <- checkAllUnambiguous False achrs

    let chk_ichrs = [ chr | chr <- all_ichrs, 
			    not $ any (chr`fitsAnnCHR`) ambig_achrs ]

    -- check that inferred types are unambiguous
    (inf_unambig, ambig_ichrs) <- checkAllUnambiguous True chk_ichrs

    -- check subsumption of provided type schemes
    verb (puts ("  | Checking subsumption of declared types (" ++ substr ++")"))
    bs <- mapM subsumptionTest ps
    -- subf is a list of the locations of annotations and defs. which fail
    let subf = [ (chrs, tss, fromJust msf) | 
		 ((b, tss, msf), chrs) <- zip bs ps, not b ]
	sub_okay = null subf
    reportSubsumptionFailed subf
  
  
    return (inf_unambig && ann_unambig && sub_okay)
    
    where -- do the action only if we're in verbose mode
	  verb :: TD () -> TD ()
	  verb = when (verbose fs)   

	  -- given a list of CHRs representing type schemes, run the
	  -- unambiguity check for each - report failure if necessary
	  checkAllUnambiguous :: Bool -> [CHR] -> TD (Bool, [CHR])
	  checkAllUnambiguous inf chrs =
	    -- if user has disabled checks, just return
	    if noCkUnambigFlag fs || noChecks fs
	       then return (True, [])
	       else do verb (puts ("  | Checking for ambiguous type schemes " ++
			           if inf then "(inf.)" else "(ann.)"))
		       -- check each chr
		       avss <- mapM checkUnambiguousCheap chrs
		       if all (null.fst) avss
			  -- if all checks return True, just return
			  then return (True, [])
			  -- else report failure where it occurred
			  else do let bad_chrs = 
				       [ chr | ((avs,_),chr) <- zip avss chrs,
					       not (null avs) ]
--				  puts ("bad_chrs: " ++ show bad_chrs)
				  avss <- mapM checkUnambiguous bad_chrs
				  prog <- getProg
				  let ls = [ x | x@(c,avs) <- zip bad_chrs avss,
						 not (null (fst avs)) ]
				      msgs = map (errorAmbiguous prog inf) ls
				  mapM puts msgs
				  return (False, map fst ls)

	  -- report error, given the locations of a pair of declared/inferred 
	  -- types which failed the subsumption test
	  reportSubsumptionFailed :: 
		[((CHR,CHR), (TypeScheme,TypeScheme), SubFailure)]-> TD ()
	  reportSubsumptionFailed cts = do
		prog <- getProg
		mapM_ (puts.(errorSubsumption prog)) cts

--------------------------------------------------------------------------------

-- new version, calls buildTypeSchemeSub (as per inference paper)
-- returns True if test succeeded, also returns 
-- a) the declared and inferred type schemes (in that order)
-- b) an error diagnosis (if there is one):
--	- Left (annotation too polymorphic): returns the set of locations for 
--		each subsumption error explanation
--	- Right (unmatched inferred constraints): returns each unmatched user
--		constraint paired with its source location


checkSubsumptionFast chrs = do res@(b,_,_) <- checkSubsumptionCheap chrs
			       if b then return res -- passed
				    else checkSubsumption chrs -- do diagnosis

checkSubsumption = checkSubsumptionConf False
checkSubsumptionCheap = checkSubsumptionConf True

checkSubsumptionConf :: Bool -> (CHR, CHR) -> 
			TD (Bool, (TypeScheme, TypeScheme), Maybe SubFailure)
checkSubsumptionConf cheap (achr, ichr) = do
    mts_a <- buildTypeSchemeSubCheap cheap True achr  -- buildTypeSchemeSub True achr
    mts_i <- buildTypeSchemeSubCheap cheap True ichr -- buildTypeSchemeSub True ichr
    if isNothing mts_a || isNothing mts_i
     then return (True, (nullTS, nullTS), Nothing)
     else checkSubsumptionTypeschemes cheap (fromJust mts_a, fromJust mts_i)

checkSubsumptionTypeschemes :: Bool -> (DebugTypeScheme, DebugTypeScheme) -> 
		    TD (Bool, (TypeScheme, TypeScheme), Maybe SubFailure)
checkSubsumptionTypeschemes cheap (ts_a0, ts_i0) = do
    let (var_a,(orig_cons_a, cons_a),type_a) = ts_a0
	(var_i,(orig_cons_i, cons_i),type_i) = ts_i0
	[t_a, vg_a] = fromJust (tupleTypes type_a)
	[t_i, vg_i] = fromJust (tupleTypes type_i)

	ts_a = (var_a,cons_a,t_a)
	ts_i = (var_i,cons_i,t_i)
	tss = (ts_a, ts_i)
		 
    	eqs = vg_a `njEq` vg_i

	-- we need to unify the global type variables in the inferred
	-- and declared types (else they're distinct, and the
	-- equivalence test at the end will fail)
	phi = unSucc $ unify [] idSubst (buildUP [eqs])
		 
	cons_a' = apply phi cons_a
	type_a' = apply phi t_a
	cons_i' = apply phi cons_i
	type_i' = apply phi t_i

--    puts ("1: ts_a: " ++ show ts_a)
--    puts ("1: ts_i: " ++ show ts_i)
--  puts ("eqs : " ++ show eqs)

    a  <- newVar
    a' <- newVar
    -- f_a :: forall var_a. cons_a => type_a
    -- f_i :: forall var_i. cons_i => type_i
		     
    -- c1 = cons_a' /\ type_a' = a' /\ a = a'
    let c1 = cons_a' `cand` bToCons (type_a' `njEq` a') `cand` 
	     bToCons (a `njEq` a')
				  
    -- c2 = c1 /\ type_i' = a /\ cons_i'
	c2 = c1 `cand` bToCons (type_i' `njEq` a) `cand` cons_i'

	vs = nub (fv c1 ++ (fv (cons_i', type_i') `without` var_i))
{-
    puts ("c1: " ++ show c1)
    puts ("c2: " ++ show c2)
    puts ("vs: " ++ show vs)
-}
    chrs <- getCHRs
    -- note that the order of the constraints passed to equiv2
    -- is significant! (equiv2 call matchCons to match user
    -- constraints. matchCons only attempts to match in one
    -- direction. i.e. c1 to c2!)
    res  <- equiv2 chrs vs (c1,c2)
    if res || cheap
     then return (res, tss, Nothing)
     else do -- error dianosis
	     --NOTE: the check can fail for one of two reasons:
	     --
	     --FIRST:
	     -- Project c1 and c2 onto annotated variables -- then explain each
	     -- constraint from c2 that is different (for the same variable) as
	     -- that in c1  i.e. we're finding those variables which are more
	     -- instantiated in c2.
	     -- Then, for each such variable, we find a minimal set of
	     -- constraints from c2 that implies this.
	     -- 
	     -- NOTE: we only explain ONE such constraint (ignore the rest)
	     --       else explanations get too difficult to understand
	     --       (this is in the same spirit as reporting one type error
	     --        at a time)
	     --
	     --SECOND:
	     --	There are inferred constraints which don't match annotated
	     --	constraints. That's because either:
	     --	    1) their predicate symbol is different
	     --	    
	     --	    2) or 1 holds, but for each annotation constraint that
	     --	       matches, the type of the annotation constraint is more
	     --	       general than the corresponding inferred constraint
	let avs = nub (fv (type_a', cons_a'))
	    c1_ts = projBConsKeepIdentities avs (consToBs c1)
	    c2_ts = projBConsKeepIdentities avs (consToBs c2)
		 
	    -- bc_exps is a list of pairs; left is the type variable that
	    --	differs (in ann/inf), right are the differing RHSs
	    bc_exps = [ (l2,(r1,r2)) | (bc1, bc2) <- zip c1_ts c2_ts, 
			let r1 = bcRight bc1
			    r2 = bcRight bc2
			    l2 = bcLeft bc2, r1 /= r2 ]
			    
	    -- inferred constraints, plus one linking the two types
	    exp_cons = orig_cons_i `cand`
		       bToCons (type_a `njEq` type_i)

	    -- explains equation represented by a pair
	    explain (l,(r1,r2)) = do
		r <- return r1
		let eq  = l `njEq` r
		    exp = entailSubsetSub {-orig_cons_i-} exp_cons eq
		return exp
	    
	-- locations explaining more-instantiated inferred type 
	diag <- case bc_exps of
		    []    -> return []
		    (e:_) -> do exp <- explain e
				return [unJust . bconsLocs $ exp]

	-- now to explain any non-matching user constraints
	-- (but only if there's no existing diagnosis)
	let diag' = if not (null diag) then Just (Left diag)
		     else let ucs_a = consToUs cons_a'
			      ucs_i = consToUs cons_i'
			      unmatched = [ uc_i | uc_i <- ucs_i, 
					    not.fst $ matchCons [] ucs_a [uc_i]]
			      ps = zip ucs_i (map (head.unJust) unmatched)
			  in  Just (Right ps)

	return (res, tss, diag')



checkSubsumption2Fast chrs = do res@(b,_,_) <- checkSubsumption2Cheap chrs
			        if b then return res -- passed
				     else checkSubsumption2 chrs -- do diagnosis

checkSubsumption2 = checkSubsumption2Conf False
checkSubsumption2Cheap = checkSubsumption2Conf True

-- as above, but does NOT simplify the declared type scheme
-- this version of the check is performed for the evidence translation
checkSubsumption2Conf :: Bool -> (CHR, CHR) -> 
			 TD (Bool, (TypeScheme, TypeScheme), Maybe SubFailure)
checkSubsumption2Conf cheap (achr, ichr) = do
    -- don't simplify constraints
    mts_a <- buildTypeSchemeSubConf cheap True False achr 

    -- simplified (for hconstraints)
    mts_b <- buildTypeSchemeSubConf cheap True True  achr 

    -- inferred
    mts_i <- buildTypeSchemeSubConf cheap True True ichr

    if isNothing mts_i || isNothing mts_b || isNothing mts_a 
     then return (True, (nullTS, nullTS), Nothing)
     else do -- we do: - strong check for Chameleon constraints
	     --        - weak check (simplified ann) for Haskell constraints
	     ts_a1 <- omit snd (fromJust mts_a)
	     ts_a2 <- omit fst (fromJust mts_b)

	     ts_i1 <- omit snd (fromJust mts_i)
	     ts_i2 <- omit fst (fromJust mts_i)

	     r1@(b1,_,_) <- checkSubsumptionTypeschemes2 cheap (ts_a1, ts_i1)
	     r2@(_,_,_)  <- checkSubsumptionTypeschemes  cheap (ts_a2, ts_i2)
	     if b1 then return r2
		   else return r1

    where omit :: ((Constraint,Constraint) -> Constraint) -> 
		  DebugTypeScheme -> TD DebugTypeScheme
	  omit f (v,(c,d),t) = do
	    c' <- splitCons c
	    d' <- splitCons d
	    return (v, (f c', f d'), t)
    

checkSubsumptionTypeschemes2 :: Bool -> (DebugTypeScheme, DebugTypeScheme) -> 
		    TD (Bool, (TypeScheme, TypeScheme), Maybe SubFailure)
checkSubsumptionTypeschemes2 cheap (ts_a0, ts_i0) = do
    let (var_a,(orig_cons_a, cons_a),type_a) = ts_a0
	(var_i,(orig_cons_i, cons_i),type_i) = ts_i0
        [t_a, vg_a] = fromJust (tupleTypes type_a)
	[t_i, vg_i] = fromJust (tupleTypes type_i)

	ts_a = (var_a,cons_a,t_a)
	ts_i = (var_i,cons_i,t_i)
	tss = (ts_a, ts_i)

	eqs = vg_a `njEq` vg_i

	-- we need to unify the global type variables in the inferred
	-- and declared types (else they're distinct, and the
	-- equivalence test at the end will fail)
	phi = unSucc $ unify [] idSubst (buildUP [eqs])
		 
	cons_a' = apply phi cons_a
	type_a' = apply phi t_a
	cons_i' = apply phi cons_i
	type_i' = apply phi t_i
{-    
    puts ("2: ts_a: " ++ show ts_a)
    puts ("2: ts_i: " ++ show ts_i)
-}
    a  <- newVar
    a' <- newVar
    -- f_a :: forall var_a. cons_a => type_a
    -- f_i :: forall var_i. cons_i => type_i
		     
    -- c1 = cons_a' /\ type_a' = a' /\ a = a'
    let c1 = cons_a' `cand` bToCons (type_a' `njEq` a') `cand` 
	     bToCons (a `njEq` a')
				  
    -- c2 = c1 /\ type_i' = a /\ cons_i'
	c2 = c1 `cand` bToCons (type_i' `njEq` a) `cand` cons_i'

	vs = nub (fv c1 ++ (fv (cons_i', type_i') `without` var_i))
{-    
    puts ("c1: " ++ show c1)
    puts ("c2: " ++ show c2)
    puts ("vs: " ++ show vs)
-}  
    chrs <- getCHRs
    res@(r,_) <- entail2DontSimplify chrs vs (c1,c2)
    if r || cheap
     then return (r, tss, Nothing)
     else do -- error diagnosis
    
	-- finds source locations with more specific types than declared
	-- (copied from above, `checkSubsumption')
	let avs = nub (fv (type_a', cons_a'))
	    c1_ts = projBConsKeepIdentities avs (consToBs c1)
	    c2_ts = projBConsKeepIdentities avs (consToBs c2)
	    
	    -- bc_exps is a list of pairs; left is the type variable that
	    -- differs (in ann/inf), right are the differing RHSs
	    bc_exps = [ (l2,(r1,r2)) | (bc1, bc2) <- zip c1_ts c2_ts, 
			let r1 = bcRight bc1
			    r2 = bcRight bc2
			    l2 = bcLeft bc2, r1 /= r2 ]
	    
	    -- inferred constraints, plus one linking the two types
	    exp_cons = orig_cons_i `cand`
		       bToCons (type_a `njEq` type_i)
	    
	    explain (l,(r1,r2)) = do
		r <- return r2
		let eq  = l `njEq` r
		    exp = entailSubsetSub exp_cons eq
		return exp
{-	
	puts ("exp_cons: " ++ show exp_cons)
	puts ("bc_exps: " ++ show bc_exps)
-}	
	-- diagnose `too-ground' inferred type 
	diag <- case bc_exps of
		[]    -> return []
		(e:_) -> do exp <- explain e
			    return [unJust . bconsLocs $ exp]

	-- diagnose unmatched inferred constraints
	-- (but only if there's no existing diagnosis)
	let diag' = if not (null diag) then Just (Left diag)
		     else let ucs_a = consToUs cons_a'
			      ucs_i = consToUs cons_i'
			      unmatched = [ uc_i | uc_i <- ucs_i, 
					    not.fst $ matchCons [] ucs_a [uc_i]]
			      ps = zip ucs_i (map (head.unJust) unmatched)
			  in  Just (Right ps)

	return (r, tss, diag')



-- takes a constraint and returns a version containing only Haskell ucons, and
-- one containing only Chameleon ucons (in that order)
splitCons :: Constraint -> TD (Constraint, Constraint)
splitCons (C ucs bcs) = do
    hcns <- getHaskellConNames
    let (ucs_h, ucs_c) = partition ((`elem`hcns).uconsPSym) ucs
    return (C ucs_h bcs, C ucs_c bcs)

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
-- returns the list of type variables within ambiguous constraints
-- (empty if the type represented by the given CHR is unambiguous)
-- if flag is True, we perform a cheap test 
--  (an error will be reported, but the details of the diagnosis will be wrong)
-- else we perform a full, correct diagnosis
checkUnambiguousConf :: Bool -> CHR -> TD ([TVar], DebugTypeScheme)
checkUnambiguousConf cheap chr = do
	-- f :: forall v. c => t
	mts <- if cheap then buildTypeSchemeCheap chr
			else buildDebugTypeScheme chr
	if isNothing mts
	 then return ([], nullTS)
	 else do let ts@(v,(orig_c,c),t) = fromJust mts
	
		 chrs <- getCHRs

		 -- rho = renaming on v
		 erho <- doTP (renameSubst v)

		 let rho = unSucc erho
		 -- c /\ rho c /\ t = rho t
		     con = c `cand` apply rho c `cand` 
				    bToCons (t `njEq` apply rho t)
				    
		 -- solve first, this is important
		     goal = convGoal con

		 store <- tdSolver chrs goal
		 let bcs = consToBs (storeToCons (fromJust store))

		 -- bail out if we got a type error
		 if isNothing store
		  then return ([], nullTS)
		  else -- check then
		       do avs <- check [] bcs rho v
			  return (avs, ts)
	  
    where   -- check that bcs -> v = rho v, for all vs
	    check avs bcs rho [] = return avs
	    check avs bcs rho (v:vs) =
		let -- eq: v = rho v
		    eq  = [Var v `njEq` apply rho (Var v)]
		    ent = entailBCons2 bcs eq
		in  if ent then do -- puts (show (Var v) ++ "==" ++ 
				   -- 	 show (apply rho (Var v)))
				   check avs bcs rho vs
			   else do -- puts (show (Var v) ++ "/=" ++ 
				   -- 	 show (apply rho (Var v)))
				   check (v:avs) bcs rho vs

checkUnambiguous = checkUnambiguousConf False
checkUnambiguousCheap = checkUnambiguousConf True

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

{-
buildTypeScheme :: Bool -> CHR -> TD (Maybe ([TVar], Constraint, Type))
buildTypeScheme simplify chr = do
    let uc = simpUCons chr
	[t,l,v] = ucTypes uc
    -- chr is uc(t,l,s) <==> ...
    chrs <- getCHRs
    mst  <- tdSolver chrs (consToGoal (uToCons uc))
    if isNothing mst 
     then -- type error
	  return Nothing
     else     -- uc(t,l,s) -->* cons
	      -- build mgu phi of cons
	      -- avec = fv(phi cons) \ ( (fv (phi l)) cup (fv (phi sg)))
	      --        (where sg are global scoped vars.)
	      -- ts = forall avec. phi cons => phi t

	  let cons = if simplify 
		     then storeToCons (fromJust mst)
		     else rhsToCons (chrRhs chr)

	      bcs  = consToBs cons
	      -- phi is mgu of cons
	      phi  = unSucc (unify [] idSubst (buildUP bcs))
	      t' = apply phi t
	      l' = apply phi l
	      v' = apply phi v
	      vl = locals v'
	      vg = globals v'
	      lg = globals l'
	      cons'  = apply phi cons
	      cons'' = remRedundantCons cons'

	      vs = -- (fv cons' `without` fv (v', l')) ++ fv vl
		   fv cons' `without` fv (vg, lg)

	      -- throw out qualified vars not appearing in type scheme
	      vs' = [ v | v <- vs, v `elem` fv (cons'', t') ]

	  in  return (Just (nub vs', cons'', t'))

-}

{-   
-- builds a type scheme for the ambiguity check 
-- also returns the unsimplified constraint component, for improved
-- error reporting
buildTypeSchemeAmbigDebug :: CHR -> 
			     TD (Maybe ([TVar], (Constraint, Constraint), Type))
buildTypeSchemeAmbigDebug chr = do
    let uc = simpUCons chr
	[t,l,v] = ucTypes uc
	rhs = consToGoal (rhsToCons (chrRhs chr))
    -- chr is uc(t,l,s) <==> ...
    chrs <- getCHRs
    mst  <- do opt <- getSolvOpt
	       setSolvOpt SDebug
	       r <- tdDeriver GlobalJust chrs rhs
	       setSolvOpt opt
	       return r
	       
    -- attempt to build the mgu of the result
    let cons = storeToCons (fromJust mst)
	bcs  = consToBs cons
	-- phi is mgu of cons
	ephi  = unify [] idSubst (buildUP bcs)
	    
    if isNothing mst || isErr ephi
     then -- type error
	  return Nothing
     else     -- uc(t,l,s) -->* cons
	      -- build mgu phi of cons
	      -- avec = fv(phi cons) \ ( (fv (phi l)) cup (fv (phi sg)))
	      --        (where sg are global scoped vars.)
	      -- ts = forall avec. phi cons => phi t

	  let phi = unSucc ephi
	      t' = apply phi t
	      l' = apply phi l
	      v' = apply phi v
	      vl = locals v'
	      vg = globals v'
	      lg = globals l'
	      cons'  = apply phi cons
	      cons'' = remRedundantCons cons'

	      vs = fv cons' `without` fv (vg, lg)

	      -- throw out qualified vars not appearing in type scheme
	      vs' = [ v | v <- vs, v `elem` fv (cons'', t') ]

	  in  return (Just (nub vs', 
			   (cons, remRedundantCons cons''), t'))


-}

	

--------------------------------------------------------------------------------

-- version of the solver which maintains the unique number store
tdSolver = solverSt

solverSt chrs goal = do st <- solve chrs goal
			return $ case st of
			    Sat   s -> Just s
			    Unsat s -> Nothing



--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

{-
	  -- True if the annotated type (represented by the first chr) is
	  -- subsumed by the inferred type (second chr)
          -- more restricted version of checkSubsumption
	  checkSubsumption2 :: (CHR, CHR) -> TD Bool
	  checkSubsumption2 (achr, ichr) = do
{-
	    ts_a <- buildSubTypeScheme achr
	    ts_i <- buildSubTypeScheme ichr
	    puts "checkSubsumption2: "
	    puts ("ann. type scheme: " ++ show ts_a)
	    puts ("inf. type scheme: " ++ show ts_i)
-}	  
	    ti <- newVar
	    li <- newVar
	    vi <- newVar
            {-  achr equals f_a(t,v) <==> rhs
                c2 = rhs , t=t_i, v=v_i
                c1 = c2, f_i(t_i,v_i)
                vars = fv c2
                c1 -->* d1

                check for |= c2 -> proj_vars d1
                if successful return the list of matched pairs
            -}
	    let uci = simpUCons ichr
		uci' = newUCTypes [ti, li, vi] uci
                (Simp _ (UC _ _ [t,_l,v] _) _ rhs) = achr
                bti = t `njEq` ti
                bvi = v `njEq` vi
                c2 = (rhsToCons rhs) `cand` bsToCons [bti, bvi]
		c1 = c2 `cand` (uToCons uci') 
		vars  = fv c2 -- fv [va, vi, ta, ti]
	       
	    chrs <- getCHRs
	    res <- tdRunTP (entail2 chrs vars (c2, c1))
            case res of
             Succ (r,m) -> return r     -- m holds the list of matched pairs  
             _          -> return False
-}	     
{-
	  -- True if the annotated type (represented by the first chr) is
	  -- subsumed by the inferred type (second chr)
	  checkSubsumption :: (CHR, CHR) -> TD Bool
	  checkSubsumption (achr, ichr) = do
{-
	    ts_a <- buildSubTypeScheme achr
	    ts_i <- buildSubTypeScheme ichr
	    puts "checkSubsumption: "
	    puts ("ann. type scheme: " ++ show ts_a)
	    puts ("inf. type scheme: " ++ show ts_i)
-}	  
	    ta <- newVar 
	    ti <- newVar
	    la <- newVar
	    li <- newVar
	    va <- newVar
	    vi <- newVar
	    t' <- newVar
	    v' <- newVar
            l' <- newVar
	    let uca = simpUCons achr
		uci = simpUCons ichr
		uca' = newUCTypes [ta, la, va] uca
		uci' = newUCTypes [ti, li, vi] uci

		-- ann constraints
                -- c2 = fa(ta,la,va) /\ t'=ta /\ t'=ti /\ v'=va /\ v'=vi /\ l'=la /\ l'=li
		c2 = uToCons uca' `cand` 
		     bsToCons [t' `njEq` ta, t' `njEq` ti, 
			       v' `njEq` va, v' `njEq` vi,
                               l' `njEq` la, l' `njEq` li]
		-- ann + inf constraints
                -- c1 = c2 /\ fi(ti,li,vi)
		c1 = c2 `cand` uToCons uci' 
		
		vars  = fv [ta, ti, t', va, vi, v', la, li, l']
{-
            puts ("achr: " ++ show achr)
            puts ("ichr: " ++ show ichr)
	    puts ("c1: " ++ show c1)
	    puts ("c2: " ++ show c2)  
	    puts ("vars: " ++ show vars)
-}
	    chrs <- getCHRs
            -- order of c2 and c1 is VERY important here
            -- due to the call to matchCons in equiv2
	    res  <- tdRunTP (equiv2 chrs vars (c2,c1))

	    if isErr res then return False
			 else return (unSucc res) 

-}

{-	    
	  -- subsumption test, as per "Theory of Overloading"
	  -- we first build type schemes (buildSubTypeScheme): 
	  --	    C => (t,l,s) and C' => (t',l',s')
	  -- then unify, l = l', s = s'
	  -- returns True if subsumption holds
	  checkSubsumption :: (CHR, CHR) -> TD Bool
	  checkSubsumption (achr, ichr) = do
	    mts_a <- buildDebugSubTypeScheme True achr
	    mts_i <- buildDebugSubTypeScheme True ichr

--	    puts ("\n\nchr_a: " ++ show achr)
--	    puts ("chr_i: " ++ show ichr)
	    
	    let (var_a,cons_a,type_a) = fromJust mts_a
		(var_i,cons_i,type_i) = fromJust mts_i
	    if isNothing mts_i || isNothing mts_a
	     then return True  -- was a type error
	     else do let [ta,la,sa] = fromJust (tupleTypes type_a)
			 [ti,li,si] = fromJust (tupleTypes type_i)
			 eql = la `njEq` li
		     let eqs = fromJust (globalParam sa) `njEq` 
			       fromJust (globalParam si)
			 phi = unSucc $ unify [] idSubst (buildUP [eqs]) -- [eql,eqs])
			 
			 cons_a' = apply phi cons_a
			 cons_i' = apply phi cons_i
			 type_a' = apply phi ta
			 type_i' = apply phi ti
			 
			 
                     puts ("\n\nts_a: forall " ++ show var_a ++ " . " ++ 
                           show cons_a ++ " => " ++ show type_a)
                     puts ("ts_i: forall " ++ show var_i ++ " . " ++ 
                           show cons_i ++ " => " ++ show type_i)
                     puts ("ts_a': forall " ++ show var_a ++ " . " ++ 
                           show cons_a' ++ " => " ++ show type_a')
                     puts ("ts_i': forall " ++ show var_i ++ " . " ++ 
                           show cons_i' ++ " => " ++ show type_i')

		     a  <- newVar
                     a' <- newVar
                     let -- f_a :: forall var_a'. cons_a' => type_a'
                         -- f_i :: forall var_i'. cons_i' => type_i'
                             
                         -- c1 = cons_a' /\ type_a' = a' /\ a = a'
                         c1 = cons_a' `cand` bToCons (type_a' `njEq` a') 
				      `cand` bToCons (a `njEq` a')
					  
                         -- c2 = c1 /\ type_i' = a /\ cons_i'
                         c2 = c1 `cand` bToCons (type_i' `njEq` a) 
				 `cand` cons_i'
			 vs = nub (fv c1 ++ 
				  (fv (cons_i', type_i') `without` var_i))

		     puts ("c1: " ++ show c1)
		     puts ("c2: " ++ show c2)
		     puts ("vs: " ++ show vs)

		     chrs <- getCHRs
		     -- note that the order of the constraints passed to equiv2
		     -- is significant! (equiv2 call matchCons to match user
		     -- constraints. matchCons only attempts to match in one
		     -- direction. i.e. c1 to c2!)
		     res  <- tdRunTP (equiv2 chrs vs (c1,c2))
		     if isErr res then return False
				  else return (unSucc res) 


-}

{-
	  checkSubsumption :: (CHR, CHR) -> TD Bool
	  checkSubsumption (achr, ichr) = do
	    mts_a <- buildSubTypeScheme True achr
	    mts_i <- buildSubTypeScheme True ichr
	    let (var_a,cons_a,type_a) = fromJust mts_a
		(var_i,cons_i,type_i) = fromJust mts_i
	    if isNothing mts_i || isNothing mts_a
	     then return True  -- was a type error
	     else do let [ta,la,sa] = fromJust (tupleTypes type_a)
			 [ti,li,si] = fromJust (tupleTypes type_i)
			 eql = la `njEq` li
		     let eqs = fromJust (globalParam sa) `njEq` 
			       fromJust (globalParam si)
			 phi = unSucc $ unify [] idSubst (buildUP [eqs])
			 
			 cons_a' = apply phi cons_a
			 cons_i' = apply phi cons_i
			 type_a' = apply phi ta
			 type_i' = apply phi ti
{-			 
                     puts ("\n\nts_a: forall " ++ show var_a ++ " . " ++ 
                           show cons_a ++ " => " ++ show type_a)
                     puts ("ts_i: forall " ++ show var_i ++ " . " ++ 
                           show cons_i ++ " => " ++ show type_i)
                     puts ("ts_a': forall " ++ show var_a ++ " . " ++ 
                           show cons_a' ++ " => " ++ show type_a')
                     puts ("ts_i': forall " ++ show var_i ++ " . " ++ 
                           show cons_i' ++ " => " ++ show type_i')
-}
		     a  <- newVar
                     a' <- newVar
                     let -- f_a :: forall var_a'. cons_a' => type_a'
                         -- f_i :: forall var_i'. cons_i' => type_i'
                             
                         -- c1 = cons_a' /\ type_a' = a' /\ a = a'
                         c1 = cons_a' `cand` bToCons (type_a' `njEq` a') 
				      `cand` bToCons (a `njEq` a')
					  
                         -- c2 = c1 /\ type_i' = a /\ cons_i'
                         c2 = c1 `cand` bToCons (type_i' `njEq` a) 
				 `cand` cons_i'
			 vs = nub (fv c1 ++ 
				  (fv (cons_i', type_i') `without` var_i))
{-
		     puts ("c1: " ++ show c1)
		     puts ("c2: " ++ show c2)
		     puts ("vs: " ++ show vs)
-}
		     chrs <- getCHRs
		     -- note that the order of the constraints passed to equiv2
		     -- is significant! (equiv2 call matchCons to match user
		     -- constraints. matchCons only attempts to match in one
		     -- direction. i.e. c1 to c2!)
		     res  <- tdRunTP (equiv2 chrs vs (c1,c2))
		     if isErr res then return False
				  else return (unSucc res) 
	    
-}



{-
	  checkSubsumption2 :: (CHR, CHR) -> TD Bool
	  checkSubsumption2 (achr, ichr) = do
	    mts_a <- buildSubTypeScheme False achr
	    mts_i <- buildSubTypeScheme True ichr

--	    puts ("\n\nchr_a: " ++ show achr)
--	    puts ("chr_i: " ++ show ichr)
--	    puts ("ts_a: " ++ show mts_a)
--	    puts ("ts_i: " ++ show mts_i)
	    
	    let (var_a,cons_a,type_a) = fromJust mts_a
		(var_i,cons_i,type_i) = fromJust mts_i
	    if isNothing mts_i || isNothing mts_a
	     then return True  -- was a type error
	     else do let [ta,la,sa] = fromJust (tupleTypes type_a)
			 [ti,li,si] = fromJust (tupleTypes type_i)
			 eql = la `njEq` li
			 eqs = fromJust (globalParam sa) `njEq` 
			       fromJust (globalParam si)
			 phi = unSucc $ unify [] idSubst (buildUP [eqs]) -- [eql,eqs])
			 
			 cons_a' = apply phi cons_a
			 cons_i' = apply phi cons_i
			 type_a' = apply phi ta
			 type_i' = apply phi ti

		     a  <- newVar
                     a' <- newVar
                     let -- f_a :: forall var_a'. cons_a' => type_a'
                         -- f_i :: forall var_i'. cons_i' => type_i'
                             
                         -- c1 = cons_a' /\ type_a' = a' /\ a = a'
                         c1 = cons_a' `cand` bToCons (type_a' `njEq` a') 
				      `cand` bToCons (a `njEq` a')
					  
                         -- c2 = c1 /\ type_i' = a /\ cons_i'
                         c2 = c1 `cand` bToCons (type_i' `njEq` a) 
				 `cand` cons_i'
			 vs = nub (fv c1 ++ 
				  (fv (cons_i', type_i') `without` var_i))
{-
		     puts ("c1: " ++ show c1)
		     puts ("c2: " ++ show c2)
		     puts ("v: " ++ show vs)
-}
		     chrs <- getCHRs
		     res  <- tdRunTP (entail2 chrs vs (c1,c2))
	             case res of
			Succ (r,_) -> return r 
			_          -> return False

-}
  

{-	  
-- returns the list of type variables within ambiguous constraints
-- (empty if the type represented by the given CHR is unambiguous)
checkUnambiguous :: CHR -> TD ([TVar], Constraint)
checkUnambiguous chr = do

		-- f :: forall v. c => t
		mts <- buildTypeSchemeAmbigDebug chr
		if isNothing mts
		 then return ([], trueCons)
		 else do let (v,(orig_c,c),t) = fromJust mts
			 chrs <- getCHRs
			 -- rho = renaming on v
			 erho <- doTP (renameSubst v)
			 let rho = unSucc erho
			 -- c /\ rho c /\ t = rho t
			     con = c `cand` apply rho c `cand` 
					    bToCons (t `njEq` apply rho t)
					    
                    -- solve first, this is important
			     goal = convGoal con

			 store <- tdSolver chrs goal
			 let bcs = consToBs (storeToCons (fromJust store))

                    -- check then
			 avs <- check [] bcs rho v
{-
			 puts ("built type scheme: forall "++show (map Var v) ++
			       ". " ++ show c ++ " => " ++ show t)
			 puts ("OR: forall "++show (map Var v) ++ 
			       ". " ++ show orig_c ++ " => " ++ show t)
			 puts ("from chr: " ++ show chr)
			 puts ("con: " ++ show con)
			 puts ("bcs: " ++ show bcs)
			 puts ("v: " ++ show v)
			 puts ("rho: " ++ show rho)
			 puts ("unambiguous?: " ++ show avs ++ "\n")
-}
			 return (avs, orig_c)
		  
	    where   -- check that bcs -> v = rho v, for all vs
		    check avs bcs rho [] = return avs
		    check avs bcs rho (v:vs) =
			let -- eq: v = rho v
			    eq  = [Var v `njEq` apply rho (Var v)]
			    ent = entailBCons2 bcs eq
			in  if ent then check avs bcs rho vs
			           else check (v:avs) bcs rho vs

-}
	  
{-	  
-- returns the list of type variables within ambiguous constraints
-- (empty if the type represented by the given CHR is unambiguous)
checkUnambiguous :: CHR -> TD [TVar]
checkUnambiguous chr = do

		-- f :: forall v. c => t
		mts <- buildTypeScheme chr 
		       -- buildDebugTypeScheme chr
		if isNothing mts
		 then return []
		 else do let (v,c,t) = fromJust mts
		
			 chrs <- getCHRs

			 -- rho = renaming on v
			 erho <- doTP (renameSubst v)

			 let rho = unSucc erho
			 -- c /\ rho c /\ t = rho t
			     con = c `cand` apply rho c `cand` 
					    bToCons (t `njEq` apply rho t)
					    
                    -- solve first, this is important
			     goal = convGoal con

			 store <- tdSolver chrs goal
			 let bcs = consToBs (storeToCons (fromJust store))

                    -- check then
			 avs <- check [] bcs rho v

			 puts ("built type scheme: forall "++show (map Var v) ++
			       ". " ++ show c ++ " => " ++ show t)
			 puts ("from chr: " ++ show chr)
--			 puts ("con: " ++ show con)
			 puts ("bcs: " ++ show bcs)
			 puts ("v: " ++ show v)
			 puts ("rho: " ++ show rho)
			 puts ("unambiguous?: " ++ show avs ++ "\n")

			 return avs
		  
	    where   -- check that bcs -> v = rho v, for all vs
		    check avs bcs rho [] = return avs
		    check avs bcs rho (v:vs) =
			let -- eq: v = rho v
			    eq  = [Var v `njEq` apply rho (Var v)]
			    ent = entailBCons2 bcs eq
			in  if ent then check avs bcs rho vs
			           else check (v:avs) bcs rho vs
-}


{-
	  -- given a simp. CHR, builds the type scheme it represents
	  buildTypeScheme :: CHR -> TD (Maybe ([TVar], Constraint, Type))
	  buildTypeScheme chr = do
	    let uc = simpUCons chr
		[t,l,s] = ucTypes uc
	    -- chr is uc(t,l,s) <==> ...
	    chrs <- getCHRs
	    mst  <- tdSolver chrs (consToGoal (uToCons uc))
	    if isNothing mst 
	     then -- type error
		  return Nothing
	     else     -- uc(t,l,s) -->* cons
                      -- build mgu phi of cons
                      -- avec = fv(phi cons) \ ( (fv (phi l)) cup (fv (phi sg)))
		      --        (where sg are global scoped vars.)
                      -- ts = forall avec. phi cons => phi t

		  let cons = storeToCons (fromJust mst)
		      bcs  = consToBs cons
		      -- phi is mgu of cons
		      phi  = unSucc (unify [] idSubst (buildUP bcs))
		      t' = apply phi t
		      l' = apply phi l
		      s' = apply phi s
		      sg = globals s'

		      cons' = -- remRedundantCons (apply phi cons)
			      (apply phi cons)
		      vs = fv cons' `without` (fv (apply phi l)  ++ 
					      fv (apply phi sg) ++
					      fv [l, s])
		      ret = tupleType [t',l',s']

		  in{-  do puts "building type scheme:"
		         puts ("  uc:  " ++ show uc)
			 puts ("  cons:" ++ show cons)
			 puts ("  cons':" ++ show cons')
			 puts ("  phi: " ++ show phi)
		         puts ("  s':  " ++ show s')
		         puts ("  sg:  " ++ show sg)
			 puts ("  vs:  " ++ show (fv cons') ++ " without " 
					 ++ show (fv (apply phi (l,sg))))
		      -}
			 return (Just (nub vs, remRedundantCons cons', t'))

-}


{-	     
	  -- given a simp. CHR, builds the *unsimplified* type scheme it 
	  -- represents, that is, we do not solve the constraints on the rhs
	  buildUnsimpTypeScheme :: CHR -> TD (Maybe ([TVar], Constraint, Type))
	  buildUnsimpTypeScheme chr = do
	    let uc = simpUCons chr
		[t,l,v] = ucTypes uc
		rhs = consToGoal (rhsToCons (chrRhs chr))
	    -- chr is uc(t,l,s) <==> ...
	    chrs <- getCHRs
	    mst  <- -- tdSolver chrs (consToGoal (uToCons uc))
		    do opt <- getSolvOpt
		       setSolvOpt SDebug
		       r <- tdDeriver GlobalJust chrs rhs -- (consToGoal (uToCons uc))
		       setSolvOpt opt
		       return r
	    if isNothing mst 
	     then -- type error
		  return Nothing
	     else     -- uc(t,l,v) -->* cons
                      -- build mgu phi of cons
                      -- avec = fv(phi cons) \ ( (fv (phi l)) cup (fv (phi vg)))
		      --        (where sg are global scoped vars.)
                      -- ts = forall avec. phi cons => phi t

		  let cons = storeToCons (fromJust mst)
		      bcs  = consToBs cons
		      -- phi is mgu of cons
		      phi  = unSucc (unify [] idSubst (buildUP bcs))
		      t' = apply phi t
		      l' = apply phi l
		      v' = apply phi v
		      vg = globals v'

		      cons' = apply phi cons
		      vs = fv cons `without` (fv (apply phi l)  ++ 
					      fv (apply phi vg) ++
					      fv [l, v])
		      ret = tupleType [t',l',v']

		  in  do puts "building type scheme:"
		         puts ("  uc:  " ++ show uc)
			 puts ("  cons:" ++ show cons)
			 puts ("  phi: " ++ show phi)
		         puts ("  s':  " ++ show v')
		         puts ("  sg:  " ++ show vg)
		      
			 return -- (Just (nub vs, cons', t'))
				(Just (nub vs, cons, t))
-}

{-
	  -- given a simp. CHR, builds the type scheme it represents
	  -- (where the type component is actually a triple combining the types
	  -- of the t,v,l components in the chr head)
	  -- Only simplify the CHR (run through solver) if flag is True
	  buildSubTypeScheme :: Bool -> CHR -> 
				TD (Maybe ([TVar], Constraint, Type))
	  buildSubTypeScheme simplify chr = do
	    let uc = simpUCons chr
		[t,l,s] = ucTypes uc
	    -- chr is uc(t,l,s) <==> ...
	    chrs <- getCHRs
	    mst  <- tdSolver chrs (consToGoal (uToCons uc))
	    if isNothing mst 
	     then -- type error
		  return Nothing
	     else     -- uc(t,l,s) -->* cons
                      -- build mgu phi of cons
                      -- avec = fv(phi cons) \ ( (fv (phi l)) cup (fv (phi sg)))
		      --        (where sg are global scoped vars.)
                      -- ts = forall avec. phi cons => phi t

		  let -- if we're not allowed to simplify (e.g. in the case of
		      -- an annotated CHR in checkSubsumption2), just use the
		      -- CHR's body
		      cons = if simplify 
			     then storeToCons (fromJust mst)
			     else rhsToCons (chrRhs chr)

		      bcs  = consToBs cons
		      phi  = unSucc (unify [] idSubst (buildUP bcs))
		      t' = apply phi t
		      l' = apply phi l
		      s' = apply phi s
		      sg = globals s'

		      cons' = -- remRedundantCons (apply phi cons)
			      apply phi cons
		      cons'' = remRedundantCons cons'
		      vs = fv cons' `without` (fv (apply phi l)  ++ 
					      fv (apply phi sg))
		      ret = tupleType [t',l',s']

		  in  return (Just (nub vs, cons'', ret))
	


	  buildDebugSubTypeScheme :: Bool -> CHR -> 
				TD (Maybe ([TVar], Constraint, Type))
	  buildDebugSubTypeScheme simplify chr = do
	    let uc = simpUCons chr
		[t,l,s] = ucTypes uc
		rhs = consToGoal (rhsToCons (chrRhs chr))
	    -- chr is uc(t,l,s) <==> ...
	    chrs <- getCHRs
	    mst  <- do opt <- getSolvOpt
		       setSolvOpt SDebug
		       r <- tdSolver chrs rhs -- (consToGoal (uToCons uc))
		       setSolvOpt opt
		       return r
	    if isNothing mst 
	     then -- type error
		  return Nothing
	     else     -- uc(t,l,s) -->* cons
                      -- build mgu phi of cons
                      -- avec = fv(phi cons) \ ( (fv (phi l)) cup (fv (phi sg)))
		      --        (where sg are global scoped vars.)
                      -- ts = forall avec. phi cons => phi t

		  let cons = if simplify 
			     then storeToCons (fromJust mst)
			     else rhsToCons (chrRhs chr)

		      bcs  = consToBs cons
		      phi  = unSucc (unify [] idSubst (buildUP bcs))
		      t' = apply phi t
		      l' = apply phi l
		      s' = apply phi s
		      sg = globals s'

		      cons'  = apply phi cons
		      cons'' = remRedundantCons cons'
		      vs = fv cons' `without` (fv (apply phi l)  ++ 
					      fv (apply phi sg))
		      ret = tupleType [t',l',s']

		  in  return (Just (nub vs, cons, ret))

-}



