--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- Annotated version of the grammar in InputAST

module Ann_InputAST (module Ann_InputAST) where

import Monad 

import Char
import qualified InputAST as I
import Misc (SrcPos, noSrcPos)
import Name

data Flag   = Unknown | Pass | Fail  deriving Show
type Loc    = Int
data Index = Ix Loc Flag SrcPos | NoIx

-- for references of form: (f;(n;)?)+
type NestedRef = [(Id, Maybe Int)]

idToNestedRef id = [(id, Nothing)]

noLoc    = (-1)	:: Loc -- never part of an AST
unkIx  n = Ix n Unknown noSrcPos
unkIxPos n p = Ix n Unknown p
passIx n = Ix n Pass noSrcPos
failIx n = Ix n Fail noSrcPos	
noIx     = NoIx


annNewLoc x l = Ann (unkIx l) (item x)
newSrcPos p (Ix a b _) = Ix a b p
newSrcPos _ _	       = error "Ann_InputAST.newSrcPos"

withAnnOf (Ann _ i) (Ann x _) = Ann x i


isIxLoc (Ix _ _ _) = True
isIxLoc _	   = False
isNoIx NoIx = True
isNoIx _    = False

ixLoc' _ (Ix n _ _)  = n
ixLoc' n x	     = error ("Ann_InputAST.ixLoc: " ++ show (x,n))


ixLoc (Ix n _ _)  = n
ixLoc x	          = error ("Ann_InputAST.ixLoc: " ++ show x)
ixNum (Ix n _ _)  = n
ixNum x	          = error ("Ann_InputAST.ixNum: " ++ show x)
ixFlag (Ix _ f _) = f
ixFlag _          = error "Ann_InputAST.ixFlag"
ixSrcPos (Ix _ _ p) = p
ixSrcPos _	    = error "Ann_InputAST.ixSrcPos" 

changeAnnFlag :: (Ann a) -> Flag -> (Ann a)
changeAnnFlag (Ann NoIx p) f = error "Ann_InputAST.changeAnnFlag: NoIx"
changeAnnFlag (Ann ix p)   f = Ann (changeFlag ix f) p
    where changeFlag NoIx       f = NoIx
	  changeFlag (Ix n _ p) f = Ix n f p

	
isUnknown f = case f of Unknown -> True; _ -> False
isPass    f = case f of Pass -> True; _ -> False
isFail    f = case f of Fail -> True; _ -> False

instance Read Index where
--    readsPrec _ ('*':_) = [(NoIx, "")]
--    readsPrec _ num	= [(Ix (read num), "")]
    readsPrec _ cs = --let xxs@(x:_) = dropWhile isSpace cs
		     case dropWhile isSpace cs of
			[]	    -> error "Read.Index"
			xxs@(x:_)   -> case x of 
			                  '*' -> [(NoIx, "")]
			                  _   -> case reads xxs of
			                  	  []	  -> []
			                  	  [(n,_)] -> [(unkIx n, "")]
			                  	  _	  -> error "Read Index"
			    

instance Show Index where
    show (Ix n _ p) = show n
    show NoIx       = "*"


instance Eq Index where
    NoIx        == NoIx        = True
    (Ix n1 _ _) == (Ix n2 _ _) = n1 == n2
    _	        == _	       = False

data Ann a = Ann Index a    deriving Eq

ann  (Ann a _) = a
item (Ann _ p) = p

nullAnn a = Ann (noIx) a

-- annl xs ix = map (Ann ix) xs
anns  xs = map ann xs
items xs = map item xs

instance Functor Ann where
    fmap f (Ann ix p) = Ann ix (f p)

instance Show a => Show (Ann a) where
    showsPrec p (Ann _ a) s = showsPrec p a s

annLoc = ixLoc . ann

zipAnns []	= []
zipAnns (p:ps)	= (ann p, item p) : zipAnns ps 

isVarTyp (Ann _ (VarTyp _)) = True
isVarTyp _ = False
isConTyp (Ann _ (ConTyp _)) = True
isConTyp _ = False


--------------------------------------------------------------------------------

multiTypeName n = name ("{" ++ show n ++ "}") ((-10) - n)
stdMultiTypeNames = map multiTypeName [1..9]   

--------------------------------------------------------------------------------
-- source-based debugger commands

data SrcCommand = SCTypeLoc Loc		    -- type command
		| SCTypeVar Id Loc
		| SCExplainLoc Loc TSchm    -- explain command
		| SCExplainVar Id Loc TSchm

instance Show SrcCommand where 
    show (SCTypeLoc l)	      = "SCTypeLoc " ++ show l
    show (SCTypeVar i l)      = "SCTypeVar " ++ show (i, l)
    show (SCExplainLoc l t)   = "SCExplainLoc " ++ show l
    show (SCExplainVar i l t) = "SCExplainVar " ++ show (i, l)


--------------------------------------------------------------------------------
--newtype Id = Id (Ann String)
type Id = String
type Guard      = Exp
type GuardedExp = (Ann Guard, Ann Exp)
type WhereDecs  = [Ann Dec]

data AnnType = Univ    -- ::
	     | Exist   -- :::
	       deriving Eq

data Prog = Prog [Ann Dec]
data Dec  = DatDec (Ann Dat)			    -- data type 
	  | RuleDec (Ann Rule)			    -- rule
	  | OverDec (Ann Sig) [Ann Def]		    -- overload
    
	  | InstDec (Ann InstSig) [Ann Def]	    -- instance declaration
	  
	  | OverDec0 Id
	  | ValDec [Ann Def]			    -- function binding
	  | AValDec (Ann Sig) [Ann Def]		    -- annotated function bind.
	  
	  | PatDec (Ann Pat) RHS WhereDecs
	  | HConDec [Id]			    -- Haskell constraint
	  | ExternDec (Either (Ann Sig) (Ann Id))   -- extern 
	  | InterfaceDec (Ann Id)		    -- interface

data Dat  = Dat (Ann Id) ([Ann Id]) ([Ann Con])
data Con  = Con (Ann Id) ([Ann Typ])

data Def  = Def (Ann Id) ([Ann Pat]) RHS WhereDecs

data RHS  = RHS1 (Ann Exp)		    -- single expression
	  | RHSG [GuardedExp]		    -- guarded expression(s)

data Exp  = VarExp Id
          | ConExp Id
	  | LitExp Lit
	  | AbsExp [Ann Pat] (Ann Exp)	  
	  | AppExp (Ann Exp) (Ann Exp)
	  | IfxAppExp  (Ann Exp) (Ann Exp) (Ann Exp)
	  | PIfxAppExp (Ann Exp) (Ann Exp) (Ann Exp)	-- parens
	  | LetExp [Ann Dec] (Ann Exp)
	  | CaseExp (Ann Exp) [CaseAlt]
	  | IfExp  (Ann Exp) (Ann Exp) (Ann Exp)
	  | AnnExp (Ann Exp)  (Ann TSchm) -- (Ann Typ)
	  | TAddExp (Ann Exp) (Ann TSchm) -- (Ann Typ)
	  | TupExp  [Ann Exp]
	  | ListExp [Ann Exp]
	  | ListCompExp (Ann Exp) [Ann Stmt]
	  | SecExpLeft  (Ann Exp) (Ann Exp)			-- (e op)
	  | SecExpRight (Ann Exp) (Ann Exp)			-- (op e)
	  | EnumExpFrom        (Ann Exp)			-- [n..]
	  | EnumExpFromTo      (Ann Exp) (Ann Exp)		-- [n..l]
	  | EnumExpFromThen    (Ann Exp) (Ann Exp)		-- [n,m..]
	  | EnumExpFromThenTo  (Ann Exp) (Ann Exp) (Ann Exp)	-- [n,m..l]

data Stmt = GenStmt  (Ann Pat) (Ann Exp)
	  | QualStmt (Ann Exp)
	  | LetStmt  [Ann Dec]	

data CaseAlt = CaseAlt (Ann Pat) (Ann Exp) WhereDecs
	     | CaseAltGuard (Ann Pat) [GuardedExp] WhereDecs

data Lit  = IntLit  Integer String	-- value and string representation
	  | FloatLit Double String	-- value and string representation
	  | StrLit  String
	  | CharLit Char	deriving Eq

data Pat  = JokPat			-- deprecated
          | VarPat Id
	  | LitPat Lit
	  | ConPat (Ann Id) ([Ann Pat])
	  | TupPat ([Ann Pat])
	  | ListPat ([Ann Pat])	deriving Eq
	  
data Sig  = Sig AnnType (Ann Id) (Ann Ctxt) (Ann Typ)	-- regular type sig
data InstSig = InstSig (Ann Ctxt) (Ann Pred)		-- instance dec sig

data Typ  = VarTyp Id
          | ConTyp Id
	  | AppTyp (Ann Typ) (Ann Typ)
	  | TupTyp ([Ann Typ])
	  | ListTyp (Ann Typ)			    

data Ctxt = Ctxt ([Ann Pred])			    
data Cnst = Cnst ([Ann Prim])			    
data Rule = PropRule (Ann Ctxt) (Ann Cnst)
	  | SimpRule (Ann Ctxt) (Ann Cnst)
data Pred = Pred Id (Ann Typ)		    
data Prim = PredPrim (Ann Pred)
          | EqPrim (Ann Typ) (Ann Typ)		    

data TSchm = TSchm (Ann Ctxt) (Ann Typ)

nullCtxt (Ctxt ps) = null ps
emptyCtxt = Ctxt []
sigPreds (Sig _ _ ctxt _) = case item ctxt of Ctxt aps -> aps

rhsExps (RHS1 e)   = [e]
rhsExps (RHSG ges) = let (gs, es) = unzip ges in gs ++ es

isUnivSig (Sig st _ _ _) = st == Univ
univSig  = Sig Univ
existSig = Sig Exist

tschmToSig (TSchm c t) id = Sig Univ id c t
makeSigUniv  (Sig _ id c t) = Sig Univ id c t
makeSigExist (Sig _ id c t) = Sig Exist id c t


patAVars :: Ann Pat -> [Ann Id]
patAVars ap = case item ap of
    VarPat id -> [Ann (ann ap) id]
    ConPat _ ps -> concatMap patAVars ps
    TupPat ps   -> concatMap patAVars ps
    ListPat ps  -> concatMap patAVars ps
    _ -> []

--------------------------------------------------------------------------------
-- strips all annotations from the program

strip :: Prog -> I.Prog
strip = stProg

stProg (Prog decs)	= I.Prog (concatMap stDec (map item decs))

stId  a 		= a

stDec (DatDec  a)	= [I.DatDec  (stDat  (item a))]
stDec (ValDec a)	= [I.ValDec (map stDef (items a))]
stDec (AValDec a b)	= if isUnivSig (item a)
			  then [I.AValDec(stSig (item a)) (map stDef (items b))]
			  else [I.ValDec (map stDef (items b))]
stDec (InterfaceDec a)	= [I.ImportDec (item a)]
-- other declarations have no Haskell equivalent and should be left out
stDec _ = []	    

stDat (Dat a b c)	= I.Dat (stId (item a)) (map stId (items b))
				(map stCon (items c))

stCon (Con a b)		= I.Con (stId (item a)) (map stTyp (items b))

stDef (Def a b c d)	= I.Def (stId (item a)) (map stPat (items b)) 
				(stRHS c) (concatMap (stDec . item) d)

stRHS (RHS1 a)		= I.RHS1 (stExp (item a))
stRHS (RHSG a)		= I.RHSG (map (\(g,e) -> (stExp (item g), 
						  stExp (item e))) a)

stExp :: Exp -> I.Exp
stExp (VarExp a)	= I.VarExp (stId a)
stExp (ConExp a)	= I.ConExp (stId a)
stExp (LitExp a)	= I.LitExp (stLit a)
stExp (AbsExp a b)	= I.AbsExp (map stPat (items a)) (stExp (item b))
stExp (AppExp a b)	= I.AppExp (stExp (item a)) (stExp (item b))
stExp (IfxAppExp a b c)	= I.IfxAppExp (stExp (item a)) (stExp (item b))
				      (stExp (item c))
stExp (LetExp a b)	= I.LetExp (concatMap (stDec .item) a) (stExp (item b)) 
stExp (CaseExp a b)	= I.CaseExp (stExp (item a)) (map stAlt b)
stExp (IfExp  a b c)	= I.IfExp  (stExp (item a)) (stExp (item b)) 
				   (stExp (item c))
stExp (AnnExp a b)	= I.AnnExp (stExp (item a)) (stTSchm (item b))
stExp (TAddExp a b)	= stExp (item a)
stExp (TupExp a)	= I.TupExp (map stExp (items a))
stExp (ListExp a)	= I.ListExp (map stExp (items a))

stExp (ListCompExp a b)	= I.ListCompExp (stExp (item a)) (map stStmt (items b))

stExp (SecExpLeft  a b)	= I.SecExpLeft  (stExp (item a)) (stExp (item b))
stExp (SecExpRight a b)	= I.SecExpRight (stExp (item a)) (stExp (item b))

stExp (EnumExpFrom e1)  = I.EnumExpFrom (stExp (item e1))
stExp (EnumExpFromTo e1 e2) = I.EnumExpFromTo (stExp(item e1)) (stExp(item e2))
stExp (EnumExpFromThen e1 e2) = I.EnumExpFromThen (stExp(item e1)) 
						  (stExp(item e2))
stExp (EnumExpFromThenTo e1 e2 e3) = I.EnumExpFromThenTo (stExp(item e1))
					    (stExp(item e2)) (stExp(item e3))					    

stStmt (GenStmt a b)	= I.GenStmt (stPat (item a)) (stExp (item b))
stStmt (QualStmt a)	= I.QualStmt (stExp (item a))
stStmt (LetStmt ds)	= I.LetStmt (concatMap stDec (items ds))

stAlt (CaseAlt a b c)	   = I.CaseAlt (stPat (item a)) (stExp (item b))
				       (concatMap stDec (items c))
stAlt (CaseAltGuard a b c) = I.CaseAltGuard (stPat (item a)) (map stGdExp b)
					    (concatMap stDec (items c))

stGdExp (g, e) = (stExp (item g), stExp (item e))

stLit (IntLit  a b)	= I.IntLit a
stLit (FloatLit a b)	= I.FloatLit a
stLit (StrLit  a)	= I.StrLit a
stLit (CharLit a)	= I.CharLit a

stPat (JokPat)		= I.JokPat
stPat (VarPat a)	= I.VarPat (stId a)
stPat (LitPat a)	= I.LitPat (stLit a)
stPat (ConPat a b)	= I.ConPat (stId (item a)) (map stPat (items b))
stPat (TupPat a)	= I.TupPat (map stPat (items a))
stPat (ListPat a)	= I.ListPat (map stPat (items a))

stSig (Sig _ a b c)	= I.Sig (stId (item a)) (stCtxt (item b)) 
				(stTyp (item c))				

stTSchm (TSchm a b)	= I.TSchm (stCtxt (item a)) (stTyp (item b))

stTyp (VarTyp a)	= I.VarTyp (stId a)
stTyp (ConTyp a)	= I.ConTyp (stId a)
stTyp (AppTyp a b)	= I.AppTyp (stTyp (item a)) (stTyp (item b))
stTyp (TupTyp a)	= I.TupTyp (map stTyp (items a))
stTyp (ListTyp a)	= I.ListTyp (stTyp (item a))

stCtxt (Ctxt a)		= map stPred (items a)
stCnst (Cnst a)		= map stPrim (items a)

stRule (PropRule a b)	= I.PropRule (stCtxt (item a)) (stCnst (item b))
stRule (SimpRule a b)	= I.SimpRule (stCtxt (item a)) (stCnst (item b))

stPred (Pred a b)	= I.Pred (stId a) (stTyp (item b))
stPrim (PredPrim a)	= I.PredPrim (stPred (item a))
stPrim (EqPrim a b)	= I.EqPrim (stTyp (item a)) (stTyp (item b))
 
--------------------------------------------------------------------------------

newIxType ix (Ann _ t@(VarTyp _)) = Ann ix t
newIxType ix (Ann _ t@(ConTyp _)) = Ann ix t
newIxType ix (Ann _ (AppTyp t1 t2)) = Ann ix (AppTyp (newIxType ix t1)
						     (newIxType ix t2))
newIxType ix (Ann _ (TupTyp ts)) = Ann ix (TupTyp (map (newIxType ix) ts))
newIxType ix (Ann _ (ListTyp t)) = Ann ix (ListTyp (newIxType ix t))


--------------------------------------------------------------------------------

{-
-- apply an expression-transforming function to each expression in the program
-- (including sub-expressions of transformed expressions)
appExpProg :: Monad m => (Ann Exp -> m (Ann Exp)) -> Prog -> m Prog
appExpProg f (Prog ads) = do ads' <- mapM (appExpDec f) ads
			     return (Prog ads')

appExpDec :: Monad m => (Ann Exp -> m (Ann Exp)) -> Ann Dec -> m (Ann Dec)
appExpDec f adec = 
    let ix  = ann adec
	dec = item adec
	f' ads = mapM (appExpDef f) ads
    in  do dec <- case dec of
		    ValDec ads	  -> f' ads >>= \ads' -> return (ValDec ads')
		    AValDec s ads -> f' ads >>= \ads' -> return (AValDec s ads')
		    OverDec s ads -> f' ads >>= \ads' -> return (OverDec s ads')
		    d		  -> return d
	   return (Ann ix dec)

--appExpDef = undefined

-- apply the given function to the top-level expression in the def
appExpDef :: Monad m => (Ann Exp -> m (Ann Exp)) -> Ann Def -> m (Ann Def)
appExpDef f def = chDefM def
    where
	chRHS :: Monad m => RHS -> m RHS
	chRHS (RHS1 ae)  = liftM RHS1 (chExp ae)
	chRHS (RHSG ges) = liftM RHSG [ (chExp g, chExp e) | (g,e) <- ges ]

	chDecs ads = map chDec ads
	chDec  ad  = let d  = item ad
			 ix = ann ad
		     in  case d of
			ValDec afs    -> Ann ix (ValDec (chDefs afs))
			AValDec s afs -> Ann ix (AValDec s (chDefs afs))
			_	      -> ad
	
	chDefM = liftM chDef
	chDefsM = liftM chDefs
	chDefs ads = map chDef ads
	chDef  ad  = let d  = item ad 
			 ix = ann ad
		     in  case d of
			Def id ps rhs w -> Ann ix (Def id ps (chRHS rhs)
							     (chDecs w))
	chExp :: Monad m => Ann Exp -> m (Ann Exp)
	chExp = undefined
			
{-	chStmt stmt = let ix = ann stmt
		      in  Ann ix $ case item stmt of
	    GenStmt p e -> GenStmt p (chExp e)
	    QualStmt e  -> QualStmt e
	    LetStmt ds  -> LetStmt (chDecs ds)

	chAlt (CaseAlt p e w) = CaseAlt p (chExp e) (chDecs w)
	chAlt (CaseAltGuard p ges w) = CaseAltGuard p (map chGE ges) (chDecs w)
	    where chGE (g,e) = (chExp g, chExp e)
-}
	

{-	
	chExp aexp = do 
	  aexp' <- f aexp
	  aexp'' <- case item aexp' of
	    
	    AbsExp id e1  -> AbsExp id (chExp e1)
	    AppExp e1 e2  -> AppExp (chExp e1) (chExp e2)
	    IfxAppExp  e1 e2 e3 -> IfxAppExp  (chExp e1) (chExp e2) (chExp e3)
	    PIfxAppExp e1 e2 e3 -> PIfxAppExp (chExp e1) (chExp e2) (chExp e3)
	    IfExp  e1 e2 e3 -> IfExp (chExp e1) (chExp e2) (chExp e3)
	    AnnExp e1 t	  -> AnnExp (chExp e1) t
	    TupExp es	  -> TupExp (map chExp es)
	    LetExp ds e1  -> LetExp (chDecs ds) (chExp e1)
	    CaseExp e as  -> CaseExp (chExp e) (map chAlt as)
	    ListExp es	     -> ListExp (map chExp es)
	    ListCompExp p ss -> ListCompExp p (map chStmt ss)
	    SecExpLeft  e1 e2 -> SecExpLeft  (chExp e1) (chExp e2)
	    SecExpRight e1 e2 -> SecExpRight (chExp e1) (chExp e2)
	    EnumExpFrom e1 -> EnumExpFrom (chExp e1)
	    EnumExpFromTo e1 e2	-> EnumExpFromTo (chExp e1) (chExp e2)
	    EnumExpFromThen e1 e2 -> EnumExpFromThen (chExp e1) (chExp e2)
	    EnumExpFromThenTo e1 e2 e3 -> EnumExpFromThenTo (chExp e1)
						 (chExp e2) (chExp e3)
	    e -> e	-- no subexpressions
	      
	  return (Ann (ann aexp') aexp'')
-}
-}

appExpProg :: (Ann Exp -> Ann Exp) -> Prog -> Prog
appExpProg f (Prog ads) = Prog (map (appExpDec f) ads)

appExpDec :: (Ann Exp -> Ann Exp) -> Ann Dec -> Ann Dec
appExpDec f adec = 
    let ix  = ann adec
	dec = item adec
    in  Ann ix $ case dec of
		    ValDec ads	    -> ValDec (map (appExpDef f) ads)
		    AValDec s ads -> AValDec s (map (appExpDef f) ads)
		    OverDec s ads -> OverDec s (map (appExpDef f) ads)
		    InstDec s ads -> InstDec s (map (appExpDef f) ads)
		    d		  -> d

-- apply the given function to the top-level expression in the def
appExpDef :: (Ann Exp -> Ann Exp) -> Ann Def -> Ann Def
appExpDef f def = chDef def
    where
	chRHS (RHS1 ae)  = RHS1 (chExp ae)
	chRHS (RHSG ges) = RHSG [ (chExp g, chExp e) | (g,e) <- ges ]

	chDecs ads = map chDec ads
	chDec  ad  = let d  = item ad
			 ix = ann ad
		     in  case d of
			ValDec afs    -> Ann ix (ValDec (chDefs afs))
			AValDec s afs -> Ann ix (AValDec s (chDefs afs))
			_	      -> ad
			
	chDefs ads = map chDef ads
	chDef  ad  = let d  = item ad 
			 ix = ann ad
		     in  case d of
			Def id ps rhs w -> Ann ix (Def id ps (chRHS rhs)
							     (chDecs w))
	chStmt stmt = let ix = ann stmt
		      in  Ann ix $ case item stmt of
	    GenStmt p e -> GenStmt p (chExp e)
	    QualStmt e  -> QualStmt e
	    LetStmt ds  -> LetStmt (chDecs ds)

	chAlt (CaseAlt p e w) = CaseAlt p (chExp e) (chDecs w)
	chAlt (CaseAltGuard p ges w) = CaseAltGuard p (map chGE ges) (chDecs w)
	    where chGE (g,e) = (chExp g, chExp e)
	
	chExp aexp = let aexp' = f aexp
		     in  Ann (ann aexp') $ case item aexp' of
	    
	    AbsExp id e1  -> AbsExp id (chExp e1)
	    AppExp e1 e2  -> AppExp (chExp e1) (chExp e2)
	    IfxAppExp  e1 e2 e3 -> IfxAppExp  (chExp e1) (chExp e2) (chExp e3)
	    PIfxAppExp e1 e2 e3 -> PIfxAppExp (chExp e1) (chExp e2) (chExp e3)
	    IfExp  e1 e2 e3 -> IfExp (chExp e1) (chExp e2) (chExp e3)
	    AnnExp e1 t	  -> AnnExp (chExp e1) t
	    TupExp es	  -> TupExp (map chExp es)
	    LetExp ds e1  -> LetExp (chDecs ds) (chExp e1)
	    CaseExp e as  -> CaseExp (chExp e) (map chAlt as)
	    ListExp es	     -> ListExp (map chExp es)
	    ListCompExp e ss -> ListCompExp (chExp e) (map chStmt ss)
	    SecExpLeft  e1 e2 -> SecExpLeft  (chExp e1) (chExp e2)
	    SecExpRight e1 e2 -> SecExpRight (chExp e1) (chExp e2)
	    EnumExpFrom e1 -> EnumExpFrom (chExp e1)
	    EnumExpFromTo e1 e2	-> EnumExpFromTo (chExp e1) (chExp e2)
	    EnumExpFromThen e1 e2 -> EnumExpFromThen (chExp e1) (chExp e2)
	    EnumExpFromThenTo e1 e2 e3 -> EnumExpFromThenTo (chExp e1)
						 (chExp e2) (chExp e3)
	    e -> e	-- no subexpressions

