--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- contains code to perform structural CHR checks
-- ie. confluence, termination and range-restrictedness
--

module CHRChecks (performCHRChecks, equiv2, entail2, entail2DontSimplify, 
		  entail3)
where

import Core
import OptFlags
import Misc
import TD_Monad
import Chr
import Solver (Goal, goalToCons)
import SolverIO

import Monad
import Maybe
import List
import IOExts

-- returns success, with chrs from completion procedure, if all checks passed
-- else, returns error with a message from the first failed test
performCHRChecks :: [Flag] -> [CHR] -> TD (E [CHR])
performCHRChecks fs chrs = do 
    t <- term
    if isErr t
       then return (Err (unErr t)) 
       else do c <- conf
	       if isErr c
		  then return (Err (unErr c)) 
		  else do r <- rr
			  if isErr r
			     then return (Err (unErr r)) 
			     else return (Succ (unSucc c)) 

    where term = if NoCkTerm `elem` fs
		    then return success
		    else if H98Term `elem` fs
			    then do verb $ 
					puts "  | Haskell98 Termination Check"
				    return (checkH98Termination chrs)
			    else do verb (puts "  | Termination Check")
				    return (checkTermination chrs)

	  conf = if NoCkConf `elem` fs
		    then return (Succ [])
		    else let comp = Complete `elem` fs
			     olap = CkSimpOverlap `elem` fs
			 in do verb $ puts "  | Confluence Check"
			       checkConfluence [] chrs comp olap

	  rr   = if NoCkRR `elem` fs
		    then return success
		    else do verb $ puts "  | Range-Restrictedness Check"
			    return (checkRangeRestricted chrs)

	  verb = when (verbose fs)


	  
--------------------------------------------------------------------------------
-- Confluence test
-- 
-- We consider a set of CHRs to be confluent iff:
--  1) for existing overlapping simp. rules: U t1 <==> ... , U t2 <==> ...
--     t1 and t2 are NOT unifiable
--  2) for each simp, prop pair: 
--		U t1 <==> ds , U t2, C ==> D
--     if t1 and t2 are unifiable, a derivation using either rule leads to an
--     equivalent result

-- NOTE: I implicitly assume that predicates only have a single argument
--	 (pattern-matching issues will arise if this is not that case.)

-- top-level interfaces to the two components of the confluence check
-- (if complete == True, then we attempt to complete the rules)
-- (if overlap == True, then we check for overlapping simps)
checkConfluence :: [CHR] -> [CHR] -> Bool -> Bool -> TD (E [CHR])
checkConfluence env chrs complete overlap = do
    let c1 = noSimpOverlap chrs
    case c1 of
	Err e -> return (err e)
	_     -> equivDerivs env chrs complete

-- checks condition (2) above
-- (if cmp== True, then we attempt to complete the rules)
-- 
-- NOTE: the code has been written to facilitate multiple passes over the list
-- of critical pairs (sp_pairs). In the current version, however, only one
-- pass is performed. Take note of the comments below in order to revert to
-- the multiple-pass version.
-- 
equivDerivs :: [CHR] -> [CHR] -> Bool -> TD (E [CHR])
equivDerivs env chrs cmp =
    do	e_simps <- doTP $ rename (simpRules chrs)
	let simps = unSucc e_simps
	    props = propRules chrs	

	-- list of (simp., [matching pred.]) pairs
	    matches = zip simps (map (inPropLhs props) simps)
	    sp_pairs = [(s,p1,p2,phi) | (s, ps) <- matches, (p1, phi, p2) <- ps]
	    ss_pairs = allMatchingSimps simps
	chr' <- checkAllSPDerivs cmp [] [] (nub sp_pairs)
	case chr' of
	    (Left (e:es)) -> return (err ("ERROR: rules not confluent:\n" ++ 
				      (confError (nub (e:es)))))


-- NOTE: To perform multiple passes of the list of pairs, replace the following
-- case option with the two immediately following it (commented out.)

	    (Right chrs') -> do let chrs'' = elimEquivProps chrs'
				let env' = chrs'' ++ env -- foldl insertCHR env chrs''
			        checkAllSSDerivs ss_pairs
				return2 env'
				
--	    (Right [])    -> do checkAllSSDerivs ss_pairs
--			        return env
--
--	    (Right chrs') -> let env' = foldl insertCHR env chrs'
--			     in  equivDerivs env' chrs cmp
			     
    where

	confError []	      = ""
	confError ((s,p):sps) = 
	    let err = ("  " ++ show (presentableVars s)) ++ "\n" ++ 
		      ("  " ++ show (presentableVars p))
	    in  err ++ "\n" ++ (confError sps)

	-- calls checkSPDerivs for each crtical pair
	-- if any succeed with new rules... we return those rules
	-- if any fail, and no new rules are suggested... we fail
	checkAllSPDerivs :: Bool -> [(CHR, CHR)] -> [CHR] ->
			    [(CHR, CHR, CHR, Subst)] ->
			    TD (Either [(CHR, CHR)] [CHR])

	checkAllSPDerivs cmp []		 chrs []   = return (Right chrs)
	checkAllSPDerivs cmp errs@(e:es) []   []   = return (Left errs)
	checkAllSPDerivs cmp errs@(e:es) chrs []   = return (Right chrs)
	checkAllSPDerivs cmp es          chrs (p:ps) =
	    do res  <- checkSPDerivs cmp p
	       case res of
		   (Right chrs') -> checkAllSPDerivs cmp es (chrs'++chrs) ps
		   (Left e)	 -> -- NOTE: replace following line with
				    -- commented line for multi-pass version
				    return (Left [e])
				    --checkAllSPDerivs cmp (e:es) chrs ps
	

	-- takes a (simp,prop',prop,phi) tuple (where the first prop is without 
	-- the matching pred, and the second is the original prop. rule.)... 
	-- causes an error iff application of simp or prop lead to 
	-- non-equivalent constraints
	-- returns (Left (s,p) on failure ... non-confluent/couldn't complete
	--	   (Right [CHR]) on success, with CHRs added
	checkSPDerivs :: Bool -> (CHR, CHR, CHR, Subst) -> 
			 TD (Either (CHR, CHR) [CHR])
	checkSPDerivs cmp spp@(s,p2,p,phi) = 
	    do let spp = (s,p2,p,phi)
	           c1 = applySimp spp
	           c2 = applyProp spp
	           c  = orig spp
	           vars = nub $ fv c
	       d1 <- solver (env ++ chrs) c1
	       d2 <- solver (env ++ chrs) c2
	       eq <- equiv2 (env ++ chrs) vars (d1, d2) 
	       let err = return (Left (s,p))
{-
	       let eq' = trace ("\ncmp: " ++ show cmp ++
				"\ns: " ++ show s ++
				"\np: " ++ show p ++
				"\nc: " ++ show c ++
				"\nc1: " ++ show c1 ++
				"\nc2: " ++ show c2 ++
				"\nd1: " ++ show d1 ++
				"\nd2: " ++ show d2 ++
                                "\nphi: " ++ show phi ++ 
				"\neq: " ++ show eq) eq
-}	       
	       if eq then return (Right [])
		     else if cmp 
			     then do comp <- complete (env ++ chrs) vars (d1,d2)
				     case comp of 
					    (Just chr) -> return (Right [chr])
					    Nothing    -> err
			     else err

	-- simulate simp. rule firing
	applySimp ((Simp _ _ _ (suc,sbc)), (Prop _ _ puc pbc _), _, phi) = 
	    let c = C (suc ++ puc) (sbc ++ pbc)
	    in  apply phi c
		
	-- simulate prop. rule firing
	applyProp ((Simp _ suc _ _), (Prop _ _ puc pbc (puc',pbc')), _, phi) =
	    let	suc' = apply phi suc
		lhs  = apply phi (C puc  pbc)
		rhs  = apply phi (C puc' pbc')
		c = (C [suc'] []) `addCons` (lhs `addCons` rhs)
	    in  c

	orig ((Simp _ suc _ _), (Prop _ _ puc pbc _), _, phi) = 
	    apply phi (addCons (C [suc] []) (C puc pbc))


	checkAllSSDerivs [] = return2 ()
	checkAllSSDerivs ((s1@(Simp _ (UC _ _ [t1] _) _ (u1, b1)), 
			   s2@(Simp _ (UC _ _ [t2] _) _ (u2, b2)), phi):ss) =
	    do	let vars = fv $ apply phi t1
		let c1   = apply phi (C u1 b1)
		let c2   = apply phi (C u2 b2)
		d1 <- solver (env ++ chrs) {-trace ("\ns1: "++(show s1))-} c1
		d2 <- solver (env ++ chrs) {-trace ("s2: "++(show s2))-} c2
		eq <- equiv2 (env ++ chrs) vars (d1, d2)
		if eq 
		    then checkAllSSDerivs ss
		    else return (err ("rules not confluent\n" ++ (show s1) ++
				  "\n" ++ (show s2)))
		

    ---- finding matching simps
	allMatchingSimps :: [CHR] -> [(CHR, CHR, Subst)]
	allMatchingSimps simps = concat $ allMatchingSimps' simps simps []
	allMatchingSimps' []     _      _  = []
	allMatchingSimps' (s:ss) (a:as) bs = 
	    let m  = matchingSimps s (as++bs) 
		m' = map (\(a,b) -> (s,a,b)) m
	    in  m' : (allMatchingSimps' ss as (a:bs))

	matchingSimps :: CHR -> [CHR] -> [(CHR, Subst)]
	matchingSimps _  []      = []
	matchingSimps s1 (s2:ss) = case simpMatch s1 s2 of
					Just mgu    -> (s2, mgu) : 
							matchingSimps s1 ss
					Nothing	    -> matchingSimps s1 ss

	simpMatch :: CHR -> CHR -> Maybe Subst
	simpMatch (Simp _ (UC _ p1 ts1 _) _ _) (Simp _ (UC _ p2 ts2 _) _ _)
	    | p1 == p2	= case unify [] idSubst (zip ts1 ts2) of
				Succ mgu    -> Just mgu
				Err  _	    -> Nothing
	    | otherwise	= Nothing
	    

    ---- finding matching simp, prop pairs
	inPropLhs :: [CHR] -> CHR -> [(CHR, Subst, CHR)]
	inPropLhs ps (Simp _ suc _ _) = 
	    let pucs = map (\(Prop _ _ x _ _) -> x) ps
		ps'  = zip ps (map (getSubsts suc) pucs)
		ps'' = filter (\ps@(_,xs)-> xs /= []) ps'
		ps3  = concat [ map (newp p) us | (p, us) <- ps'' ]
	    in  ps3
	    where newp p@(Prop n ix _ pbc rhs) (uc, s) = 
				(Prop n ix uc pbc rhs,s,p)

	-- tries to match c with each element of us... if it matches, generates
	-- a tuple containing the unifier and all the *other* elements of us
	getSubsts :: UCons -> [UCons] -> [([UCons], Subst)]
	getSubsts c us = getSubsts' c us []
	getSubsts' c [] as = []
	getSubsts' c (u:us) as = 
	    case getSubst c u of
		Just s	-> -- [ (as ++ us, s) ] 
			   (as ++ us, s) : getSubsts' c us (u:as)
		Nothing	-> getSubsts' c us (u:as)
	
	getSubst (UC _ ps ts _) (UC _ ps' ts' _)
	        | ps == ps' = case unify [] idSubst (zip ts ts') of
				Succ s -> Just s
				Err  _ -> Nothing
		| otherwise = Nothing



-- checks for overlapping simplification rules
-- causes an error if any found
noSimpOverlap :: [CHR] -> E ()
noSimpOverlap chrs =
    do	let simps = simpRules chrs
	    pairs = pairUp simps
	-- a list of (r, [rs]) pairs; where rs overlap with r
	    allOverlaps = map (simpOverlaps) pairs
	    allTypes = map (justTypes) allOverlaps
	    anyUnify = or [or $ map (unifiable t) ts | (t,ts) <- allTypes]
	if anyUnify
	 then err "rules not confluent: overlapping simp. rules"
	 else success
    where
	
	justTypes (c,cs) = (simpType c, map (simpType) cs)
	simpType (Simp _ (UC _ _ [t] _) _ _) = t
	
	lhsPSym (Simp _ (UC _ ps _ _) _ _) = ps

	overlap s1 s2   = s2 `hasPSym` (lhsPSym s1)
	
	-- takes a simp. rule and finds all those with overlapping head preds.
	simpOverlaps (s,ss) = (s, filter (overlap s) ss)
	
	hasPSym (Simp _ (UC _ ps' _ _) _ _) ps
	    | ps == ps'	= True
	    | otherwise = False

	
-- rewrite these two functions as a single "split" function
-- given a list of CHRs, returns a list of those which are simplification rules
simpRules :: [CHR] -> [CHR]
simpRules chrs = filter (isSimp) chrs
    where   isSimp (Simp _ _ _ _) = True
	    isSimp _            = False

-- as above, but for prop. rules
propRules :: [CHR] -> [CHR]
propRules chrs = filter (isProp) chrs
    where   isProp (Prop _ _ _ _ _) = True
	    isProp _		  = False


-- takes a list and returns a list pairing each element with all the others
pairUp xs = pairUp' [] xs
pairUp' _  []     = []
pairUp' as (x:xs) = (x, as++xs) : pairUp' (x:as) xs


unifiable :: Type -> Type -> Bool
unifiable t1 t2 = isSucc $ unify [] idSubst [(t1,t2)]


----------------------------------------

-- given a list of CHRs (props), eliminates those which are functionally 
-- equivalent 
elimEquivProps :: [CHR] -> [CHR]
elimEquivProps chrs = 
    let props = filter isProp chrs
	simps = filter isSimp chrs
	ms'   = findMatches props
	ms    = {-trace ("matches: " ++ show ms' ++ "\n")-} ms'
	res   = concatMap (\(p1,p2,s)->if rhsEquiv p1 p2 s then [p2] else []) ms
	ixs   = map (\(Prop _ ix _ _ _) -> ix) res
	ixs'  = nub ixs
	props'= removePropsWithIxs props ixs'
	chrs' = simps ++ props'
    in  chrs'

    where

	removePropsWithIxs :: [CHR] -> [Index] -> [CHR]
	removePropsWithIxs [] _ = []
	removePropsWithIxs ((p@(Prop _ ix _ _ _)):ps) ixs
	    | ix `elem` ixs = removePropsWithIxs ps ixs
	    | otherwise	    = p : removePropsWithIxs ps ixs

	-- takes two props and a subst (unifier of preds. on lhs) and tests if 
	-- the rhss are equivalent
	rhsEquiv :: CHR -> CHR -> Subst -> Bool
	rhsEquiv (Prop _ _ _ _ (ucs1, bcs1)) (Prop _ _ _ _ (ucs2, bcs2)) phi =
		let b = bcEquiv [] (apply phi bcs1) bcs2
		    (u,_) = matchCons [] (apply phi ucs1) ucs2
		in  b && u

	-- takes a list of props and returns a list of (prop1, prop2, mgu_lhs) 
	-- triples where prop1 and prop2 have an "equivalent" lhs, and mgu_lhs 
	-- is the mgu of the lhss
	findMatches :: [CHR] -> [(CHR, CHR, Subst)]
	findMatches []     = []
	findMatches (p:ps) = let ps' = filter (sameLhs p) ps
				 ms  = map (\x -> (p, x)) ps'
				 ms' = [(p1, p2, unifyLhs p1 p2)|(p1, p2) <-ms ]
			     in  ms' ++ findMatches ps

	sameLhs  (Prop _ _ ucs1 _ _) (Prop _ _ ucs2 _ _) = 
		isJust (sameUCs ucs1 ucs2 [])
	unifyLhs (Prop _ _ ucs1 _ _) (Prop _ _ ucs2 _ _) = 
		let ts = fromJust (sameUCs ucs1 ucs2 [])
		in  unSucc (unify [] idSubst ts)

	sameUCs :: [UCons] -> [UCons] -> [(Type, Type)] -> Maybe [(Type, Type)]
	sameUCs []       []   []  = Nothing
	sameUCs []       []   acc = Just acc
	sameUCs []       _    _   = Nothing
	sameUCs (uc:ucs) ucs' acc = 
	      case getMatch uc ucs' of
		  Nothing	-> Nothing
		  Just (ix, ts) -> let ucs'' = removeUCWithIx ix ucs'
				   in  sameUCs ucs ucs'' (ts:acc)

	getMatch _  []		 = Nothing
	getMatch uc@(UC _ p1 [t1] _) ((UC ix p2 [t2] _):ucs)
	  | p1 == p2	= Just (ix, (t1, t2))
	  | otherwise	= getMatch uc ucs

	removeUCWithIx _  [] = []
	removeUCWithIx ix ((uc@(UC ix' _ _ _)):ucs)
	  | ix == ix'	= ucs
	  | otherwise	= uc : removeUCWithIx ix ucs


----------------------------------------
-- complete env vars (d1,d2)
-- Assumptions: 
-- d1 and d2 are the final stores from the confluence test, i.e.
--  c -> c1 ->^* d1   and c -> c2 ->^* d2
-- vars = fv(c)  we assume CHRs are range-restricted
-- we check whether d1 entails d2 or d2 entails d1
-- if not (conjecture) CHRs are not completable
-- otherwise assume d1=C uc1 bc1  d2=C uc2 bc2
-- check whether uc1 subsetof uc2 or uc2 subsetof uc1
-- if not (conjecture) CHRs are not completable
-- in the first case (d1 entails d2 and uc1 subsetof uc2), 
-- add Prop  uc2 ==> bc1
-- second case, add Prop uc1 ==> bc2
complete :: [CHR] -> [TVar]-> (Constraint, Constraint) -> TD (Maybe CHR)
complete env vars (c1@(C uc1 bc1), c2@(C uc2 bc2)) =
       do b <- equiv2 env vars ((C (removedups2 (uc1++uc2)) (bc1++bc2)), c2)
               -- check c2 entails c1
          if b then checkCompletion (c2, c1)
           else do b' <- equiv2 env vars (c1, (C (removedups2 (uc1++uc2)) (bc1++bc2)))
                         -- check c1 entails c2
                   if b' then checkCompletion (c1, c2)
                    else return Nothing

-- add Prop uc2 ==> bc1 if uc1 subsetof uc2
-- add Prop uc1 ==> bc2 if uc2 subsetof uc1
checkCompletion :: (Constraint, Constraint) -> TD (Maybe CHR)
checkCompletion (C uc1 bc1, C uc2 bc2) =
      do index <- next
         if subsetCons uc1 uc2 
	  then return (Just (Prop noInfo index uc2 [] ([],bc1)))
          else return Nothing



--------------------------------------------------------------------------------
-- Termination checks (includes Haskell98-style check)


checkTermination :: [CHR] -> E ()
checkTermination rules = let c1 = doEs $ map (check rules) rules
			     c2 = ckRuleCycles rules
			 in  doEs [c1,c2]
    where   check rs r
		| isProp r  = ckProp r
		| isSimp r  = ckSimp r
		| isMSimp r = ckMSimp r

	    
checkH98Termination :: [CHR] -> E ()
checkH98Termination rules = 
    let c1 = ckRuleCycles  props
        c2 = doEs $ map ckInPropCycle props
        c3 = doEs $ map ckInMSimp msimps
    in  doEs [c1,c2,c3]
    where 
	  (s, m, props) = groupCHRs rules
	  msimps = m ++ map simpToMSimp s

	  ckInPropCycle (Prop _ _ lucs _ (rucs, _)) =
	    let lps = map uconsPSym lucs
		rps = map uconsPSym rucs
	    in  if any (`elem`rps) lps
		then err ("ERROR: rules fail termination criteria " ++
			      "(cycle detected!)")
		else success
		    
	  ckInMSimp r@(MSimp _ lucs _ _) =
	    let ts = map uconsType lucs
		isVar t = case t of (Var _) -> True ; _ -> False
	    in  if any isVar ts
		then err ("ERROR: rule fails termination criteria:\n  " ++
			       (show (presentableVars r)))
		else success


-- checks that the given prop. rule satisfies the termination criteria.
-- ie. is of form:
--	    C ==> C' (where C' contains only equality constraints)
--	OR  C ==> C' (where C' contains no equality constraints)
ckProp :: CHR -> E ()
ckProp (Prop _ _ _ _ ([],  []))  = success
ckProp (Prop _ _ _ _ (ucs, []))  = success
ckProp (Prop _ _ _ _ ([], bcs))  = success
ckProp r@(Prop _ _ _ _ (ucs, bcs)) = 
	    err ("ERROR: rule fails termination criteria:" ++ 
		      "\n  " ++ show (presentableVars r))


ckRuleCycles :: [CHR] -> E ()
ckRuleCycles rs =
    let (s, m, props) = groupCHRs rs
	msimps  = m ++ map simpToMSimp s
	psyms = map uconsPSym
	shbs = map (\(MSimp _ lucs _ (rucs, _))->(psyms lucs,psyms rucs)) msimps
	phbs = map (\(Prop _ _ lucs _ (rucs, _))->(psyms lucs,psyms rucs)) props
	hbs  = [ (h, filter (/=h) bs) | (hs, bs) <- shbs ++ phbs, h <- hs ]
    in	if isCycle hbs 
	then err ("ERROR: rules fail termination criteria" ++
				     " (cycle detected!)")
	else success


-- True if there is a cycle in the graph
isCycle :: Eq a => [(a, [a])] -> Bool
isCycle rs = any (\(h,b) -> isCycle' [] rs h) rs
    where isCycle' vs rs h
	    | h `elem` vs   = True
	    | otherwise     = let bs = concat $ lookupAll h rs
			      in  any (isCycle' (h:vs) rs) bs

lookupAll x [] = []
lookupAll x ((y,z):ys)
    | x == y	= z : lookupAll x ys
    | otherwise	= lookupAll x ys


-- Checks that the given simp. rule satisfies the termination criteria.
-- ie. if U t <==> ... , U t', ...       t > t' (for some '>')
-- Will want to extend this by tracing all the constraints on the lhs to make
-- sure they don't further introduce some circularity.
ckSimp :: CHR -> E ()
ckSimp (Simp n luc lbcs rhs) = ckMSimp (MSimp n [luc] lbcs rhs)

ckMSimp :: CHR -> E ()
ckMSimp r@(MSimp _ lucs _ (rucs, _)) =
	if all (noSimpLoop rucs) lucs
	    then success
	    else err ("ERROR: rule fails termination criteria:\n" ++
			    ("  " ++ show (presentableVars r)))

	where gt t1 t2 = (weight t1) > (weight t2)
	      weight (TApp t1 t2) = 1 + weight t1 + weight t2
	      weight _		  = 1

	      noSimpLoop :: [UCons] -> UCons -> Bool
	      noSimpLoop ucs uc = let id = uconsPSym uc
				      idMatches uc = id == uconsPSym uc
				      ucs'   = filter idMatches ucs
				      t      = uconsType uc
				      ucs_ts = map uconsType ucs'
				  in  all (gt t) ucs_ts



--------------------------------------------------------------------------------
-- Range-restrictedness

checkRangeRestricted :: [CHR] -> E ()
checkRangeRestricted chrs = 
    let nonrr = filter (not.isRangeRestricted) chrs
    in  case nonrr of
	(c:_)	-> err ("ERROR: rule not range-restricted:\n" ++ 
			    ("  " ++ show (presentableVars c) ++ "'\n"))
	[]	-> success

-- succeeds if for: Cs ==> Ds, or Cs <=> Ds, instantiating the variables in the
-- rule's head causes all variables on the right to also become ground
-- procedure: find the mgu of variables in the body; apply it to the head and 
-- body; generate a substitution which grounds all variables on left; apply to
-- right; if any variables are not ground -> not rr!
isRangeRestricted :: CHR -> Bool 
isRangeRestricted chr =
    let lhs = ruleHead chr
	rhs = ruleBody chr
	fvr = fv rhs
	eq  = consToBs rhs
	eq' = buildUP eq
	mmgu = unify fvr idSubst eq'
	mgu   = unSucc mmgu
	rhs'  = apply mgu rhs
	lhs'  = apply mgu lhs
	fvl'  = fv lhs'
	typ   = TyCon boolName -- arbitrary
	subst = foldl (\s v -> updateSubst s v typ) idSubst fvl'
	rhs'' = apply subst rhs'
    in  isErr mmgu || fv rhs'' == [] 

    where
	ruleHead, ruleBody :: CHR -> Constraint
	ruleHead (Simp _ uc bcs _)    = C [uc] bcs
	ruleHead (MSimp _ ucs bcs _)  = C ucs  bcs
	ruleHead (Prop _ _ ucs bcs _) = C ucs  bcs

	ruleBody (Simp _ _ _ (ucs, bcs))   = C ucs bcs
	ruleBody (MSimp _ _ _ (ucs, bcs))  = C ucs bcs
	ruleBody (Prop _ _ _ _ (ucs, bcs)) = C ucs bcs

--------------------------------------------------------------------------------
-- helper functions that were in Infer, that the above code depends on

{-
solver :: [CHR] -> Constraint -> TP Constraint
solver chrs (C ucs bcs) =
    let goal = (ucs,bcs)
    in do (([],[]),ucs',bcs',_,_,_s) <- solve chrs goal
	  return (C ucs' bcs')
-}



solver :: [CHR] -> Constraint -> TD Constraint
solver chrs (C ucs bcs) = do
    let goal = (ucs,bcs)
    res <- solve chrs goal
    return $ case res of
	Sat s	-> mixedStoreToCons s
	Unsat s -> mixedStoreToCons s
 

----------------------------------------
-- equivalence (wrt. variables in vars) tester among constraints
-- note we return true if both constraints are inconsistent
equiv2 :: [CHR] -> [TVar] -> (Constraint,Constraint) -> TD Bool
equiv2 env vars (c1 , c2) = 
     let consistent bc = case (unify [] idSubst (buildUP bc)) of
                           Succ _ -> True
                           Err _  -> False
     in  do sc1@(C uc1 bc1) <- solver env c1
            sc2@(C uc2 bc2) <- solver env c2
              -- both could be inconsistent
            if (not (consistent bc1)) && (not (consistent bc2))
             then return True
	     else if not (consistent bc1) || not (consistent bc2)
		then return False
                -- we first check whether the builtins
                -- entail each other
		else
                 if bcEquiv vars bc1 bc2
                  then case (unify [] idSubst (buildUP (bc1 ++ bc2))) of
                        Succ s1 -> case (unify [] idSubst (buildUP (bc1 ++ bc2))) of
						
                                    Succ s2 -> -- here we check for
					       -- the user-defined constraints
					       do let uc1' = apply s1 uc1
						      uc2' = apply s2 uc2
						      (r,m) = matchCons (calculate vars s1) uc1' uc2' 
{-                                                  trace ("\nuc1: " ++ show uc1 ++ "\t uc1'" ++ show uc1' ++
							 "\nuc2: " ++ show uc2 ++ "\t uc2'" ++ show uc2' ++
							 "\nm: " ++ show m ++
							 "\nr: " ++ show r) $ -}
						  return r
						  -- return r
                                    _       -> -- trace "B" $ 
					       return False  -- either one or
                        _       -> -- trace "C" $ 
				   return False              -- the other is false
                  else -- trace "D" $ 
		       return False


-- always simplify the second constraint
entail2 = entail2Conf True

entail2DontSimplify = entail2Conf False

-- entailment (wrt. variables in vars) tester among constraints
-- note we return true if both constraints are inconsistent
-- check |= c1 -> proj_vars c2
-- NOTE: flag indicates whether second argument is simplified or not
--	 (if True, then it is)
entail2Conf :: Bool -> [CHR] -> [TVar] -> (Constraint,Constraint) -> TD (Bool, [(UCons,UCons)])
entail2Conf simplify env vars (c1, c2) = 
     let  consistent bc = case (unify [] idSubst (buildUP bc)) of
                           Succ _ -> True
                           Err _  -> False
     in
      do sc1@(C uc1 bc1) <- return c1 -- solver env c1
         sc2@(C uc2 bc2) <- if simplify then solver env c2
					else return c2
         if (not (consistent bc1)) 
          then return (True,[])
          else if (not (consistent bc2))
                then let res = -- trace ("\n Something unusal happened: True -> False, best to inform Chameleon team: " ++ show bc2) 
			       False
                     in return (res,[])
                     --return (False,[])
                else  
		 -- we first check whether the builtins
		 -- entail each other	 	 
                 if entailBCons bc1 vars bc2
                  then case (unify [] idSubst (buildUP (bc1 ++ bc2))) of
                        Succ s1 -> case (unify [] idSubst (buildUP (bc1 ++ bc2))) of
						
                                    Succ s2 -> do -- here we check for
					        -- the user-defined constraints
                                                let -- vars' = trace ("\n vars=" ++ show vars) vars
                                                    -- vars'' = calculate vars s1
                                                    -- vars''' = trace ("\n vars'" ++ show vars'') vars''

					        return (matchCons {-vars'''-} (calculate vars s1) --(vars) 
                                                        (apply s1 uc1)
                                                        (apply s2 uc2))
                  else return (False,[])


-- calculates the proper set of variables on which we project on
calculate vs subst = let res = map (\ a -> lookup a subst) vs
                         res' = zip res vs
                         res'' = [ (case x of
                                     Just t -> fv t
                                     Nothing -> [v])    | (x,v) <- res']
                         res''' = concat res''
                   in res'''
 


-- check |= c1 -> proj_vars c2
entail3 :: [CHR] -> [TVar] -> (Constraint,Constraint) -> TD (Bool, [(UCons,UCons)])
entail3 env vars (c1, c2) = 
     let  consistent bc = case unify [] idSubst (buildUP bc) of
                           Succ _ -> True
                           Err _  -> False
     in
      do sc1@(C uc1 bc1) <- return c1 -- solver env c1
         sc2@(C uc2 bc2) <- return c2 -- solver env c2
         if (not (consistent bc1)) 
          then return (True,[])
          else if (not (consistent bc2))
                then let res = False
                     in return (res,[])
                else  
		 -- we first check whether the builtins
		 -- entail each other	 	 
                 if entailBCons bc1 vars bc2
                  then case (unify [] idSubst (buildUP (bc1 ++ bc2))) of
                        Succ s1 -> case (unify [] idSubst (buildUP (bc1 ++ bc2))) of
						
                                    Succ s2 -> do -- here we check for
					        -- the user-defined constraints
                                                let -- vars' = trace ("\n vars=" ++ show vars) vars
                                                    -- vars'' = calculate vars s1
                                                    -- vars''' = trace ("\n vars'" ++ show vars'') vars''

					        x <- return (matchConsEnt (calculate vars s1)
                                                        (apply s1 uc1)
                                                        (apply s2 uc2))
						-- seq x (return x)
						return x
                  else return (False,[])


matchConsEnt vs uc1 uc2 = let uc1s = powerset uc1 -- [] : permute uc1
			      ms  = [ matchCons vs uc1 uc2 | uc1 <- uc1s ]
			      ms' = filter fst ms
			  in  {- trace ("\nvars: " ++ show vs ++ 
				     "\nuc2:" ++ show uc2 ++ 
				     "\nuc1s:" ++ show uc1s ++ 
				     "\nms: " ++ show ms ++ 
				     "\nms':" ++ show ms') $ -}
			       (if null ms' then {- trace "HERE" $ -} (False,[])
					    else {- trace "THERE" $ -} head ms')

 

-- we assume bc1, bc2 are consistent
-- we test whether proj vs. bc1 iff proj vs bc2
bcEquiv :: [TVar] -> [BCons] -> [BCons] -> Bool
bcEquiv vs bc1 bc2 =
      let bc1' = projBCons vs bc1
          bc2' = projBCons vs bc2
	  b1 = entailBCons bc1' vs bc2'
	  b2 = entailBCons bc2' vs bc1'
      in {- 
          trace ("vs: " ++ show vs ++ 
		 "\n bc1: " ++ show bc1  ++ "\n bc2: " ++ show bc2 ++
		 "\n bc1':" ++ show bc1' ++ "\n bc2':" ++ show bc2' ++
		 "\n (b1, b2): " ++ show (b1,b2) ) $
  -}
		b1 && b2    
--		entailBCons bc1' vs bc2' && entailBCons bc2' vs bc1'
      


--------------------------------------------------------------------------------
{-
doEs :: [E ()] -> E ()
doEs [] = success
doEs es | any isErr es = case filter isErr es of
			    (e:_) -> e
			    _	  -> success
	| last es
-}

doEs :: [E a] -> E a
doEs [] = success
doEs (e@(Err _):es) = e
doEs [Succ x] = Succ x
doEs (_:es) = doEs es

err txt = Err txt

return2 :: (Monad m, Monad n) => v -> m (n v)
return2 = return . return
