--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


module Control (start)
where

import OptFlags
import TD_Monad
import TD_General (decToSig, sigAId)
import TD_Translate
import TD_StdDefs
import Misc
import qualified Parser (parseInit)
import Operators (fixOperators)
import Justify (JustProp(..))
import Ann_InputAST hiding (Flag)
import InitChecks (initChecks)
import Compiler (compile, compileChecks)
import TypeDebug (debugger, doSrcCommands)
import qualified Version


import IO
import System
import List
import Monad
import Solver (genSolverOpt)

-- import IOExts

--------------------------------------------------------------------------------

-- call immediately after command-line options have been read
--
-- fs : list of the command-line flags parsed earlier
-- srs: list of the filenames to read from (we only do the first)
--
-- works as follows:
--  - parses program source
--  - calls CHR translator
--  - calls CHR checks (for rules arising from rule and overload decls)
--  - calls typechecker, which goes through all value declarations in the
--    program, inferring a type for each and checking the consistency of any
--    annotations present (subsumption and unambiguity checks for OverDecs and
--    AValDecs)
--    report all errors found
--  - interactive mode
start :: ([Flag], [String]) -> IO ()
start (fs, strs) = do
--    inp <- getInput strs    
--    let res = Parser.parse inp
    res <- parseAllFiles strs    
    if isErr res
       then fatalError (unErr res)
       else do let -- ix is the next label, we need to thread through each 
		   -- process the unique number store
		   (ix, cmds, aprog) = unSucc res
		   cmds' = reverse cmds
		
		   -- correct operator precedence/associativity in program
	           aprog' = fixOperators aprog
	           n = fromIntegral (ixLoc ix) 
	       
	       -- check for use of undeclared constraints 
	       undec <- checkUndeclaredConstraints aprog'
	       when undec exitFailure

		   -- process hconstraint declarations
	       (hcs, aprog'') <- progHaskellCons aprog'
	       let hns = map snd hcs

	       -- check for annotations without definitions
	       prog <- unpackMismatchedDecsProg aprog''
	       checkDanglingAnnotations prog
		
	       -- merge multiple clauses of the same def into individual
	       -- declarations
	       prog' <- mergeDefsProg prog
		   
	       -- translate program to CHRs, env is the translator's
	       -- environment (required later if we need to invoke it again)
	       -- st is its state (we need it to build the inital TR state)
	
	       -- putStrLn (show prog')	   
	       
	       ((env, tchrs), st) <- progToCHRs n hcs prog'

	       sopt <- genSolverOpt fs
	       let jp   = genJustProp fs
	           chrs  = tchrs ++ stdCHRs
		   schrs = simplifyCHRs chrs

		    -- initial state for the TD monad
		   fn = case strs of 
			 (n:_) -> basename n
			 _     -> "Prelude"
		   tdstate = genTDState prog' chrs schrs env st stdNames 
					sopt jp fs fn hns
		   
		   -- do the source commands first since `initChecks' will add
		   -- annotations to unannotated decs
		   stage1 = unless (null cmds) (TypeDebug.doSrcCommands cmds')
		   stage2 = initChecks fs

		   (stage0, stage3) = 
			if Debug `elem` fs
			then (displayHeader,
			      TypeDebug.debugger)
			else (Compiler.compileChecks fs, Compiler.compile fs)
		   
		   -- performs all initial checks
		   td = runTD (stage0 >> stage1 >> stage2 >> stage3)

	       -- run the checker
	       td tdstate
	       return ()

----------------------------------------

displayHeader = puts $ unlines [" _                         ",
		 		"/ |_  _.._ _  _ | _  _ ._  ",
		 		"\\_| |(_|| | |(/_|(/_(_)| | ",
		 		"                           ",
				Version.str,
				"--------------------------------------" ]


----------------------------------------

-- beginning with the given, top-level file, parse it and all imports
parseAllFiles :: [String] -> IO (E (Index, [SrcCommand], Prog))
parseAllFiles [f]  = do 
    e <- processImports 1 (intProg f)
    case e of
	Succ (ix,ids,cmds,p@(Prog adecs)) ->
	    -- check that the Prelude was included
	    if "Prelude.ch" `elem` ids
	    then return $ Succ (ix,cmds,p)
	    else do prel <- processImports (ixLoc ix) (intProg "Prelude.ch")
		    case prel of
			Succ (ix',_,_,Prog adecs') -> 
			     return $ Succ (ix',cmds,Prog (adecs'++adecs))
			Err _ -> -- couldn't find it, print a warning?
			     do putStrLn "Warning: Could not find `Prelude.ch'"
				return $ Succ (ix,cmds,p)

	Err msg -> return $ Err msg

    where intProg f  = Prog [nullAnn (InterfaceDec (nullAnn f))]

parseAllFiles (f:_) = do putStr $ "warning: only reading first file ("++f++")\n"
			 parseAllFiles [f]

-- default to just loading the Prelude
parseAllFiles [] = parseAllFiles ["Prelude.ch"]

{-
parseAllFiles [] = do putStr "Please enter your program on stdin:\n"
	 	      str <- readInput (nullAnn "stdin") stdin
		      let ep = Parser.parseInit "stdin" str 1
			  (_, cmds, ix, prog) = unSucc ep
			  n = ixLoc ix 
		      p <- processImports n prog
		      case p of
			e@(Err _) -> return e
			Succ (n', cmds', prog') ->
			    return (Succ (n', reverse cmds ++ cmds', prog'))
-}

-- Process all the import/interface declarations. 
-- This means: 1) find all the interfaceDecs
--	       2) for each, read the corresponding file
--	       3) parse it
--	       4) include the result into the AST in place of the interfaceDec
--	       5) recurse until no interfaceDecs left
-- `n' is the initial label number 
-- Returns: - the index containing the unique number we're up to
--	    - a list of the files which were opened
--	    - a list of the source commands
--	    - the complete parsed program
processImports :: Int -> Prog -> IO (E (Index, [Id], [SrcCommand], Prog))
processImports n (Prog adecs) = do 
	p <- processDecs True n [] adecs
	case p of 
	    Err e -> return (Err e)
	    Succ (n', ids, adecs', cmds) ->
		    return (Succ (unkIx n', ids, cmds, Prog adecs'))

    where processDecs :: Bool -> Int -> [Id] -> [Ann Dec] -> 
			 IO (E (Int, [Id], [Ann Dec], [SrcCommand]))
	  processDecs _ n ids [] = return (Succ (n, ids, [], []))
	  processDecs first n ids (ad:ads) = do 
	    p1 <- processDec first n ids ad
	    case p1 of
		Err e -> return (Err e)
		Succ (n1, ids, ads1, cmds1) -> do
		    p2 <- processDecs False n1 ids ads
		    case p2 of
			Err e -> return (Err e)
			Succ (n2, ids, ads2, cmds2) ->
			    return (Succ (n2, ids, ads1++ads2, cmds1++cmds2))
		
	  processDec :: Bool -> Int -> [Id] -> Ann Dec -> 
			IO (E (Int, [Id], [Ann Dec], [SrcCommand]))
	  processDec first n ids adec = case item adec of
	  
		(InterfaceDec id) -> do 
			let fn = if first 
				 then id 
				 else (Ann (ann id) (item id ++ ".ch"))
		    	estr <- getInput [fn]
			case estr of
			    Err e -> return (Err e)
			    Succ str ->
			     let ep = Parser.parseInit (item id) str n
			     in  case ep of
				Err e -> return (Err e)
				Succ (_, cmds, ix, Prog adecs) -> do
				    let n1 = ixLoc ix
				        cmds' = reverse cmds
				-- now process any interface decs amongst these
				    ep2 <- processDecs False n1 (item fn:ids) adecs
				    case ep2 of
					Err e -> return (Err e)
					Succ (n2, ids, adecs', cmds'') -> 
					 return (Succ (n2, ids, 
							if first then adecs'
							else adec:adecs',
						   cmds' ++ cmds''))

		_    -> return (Succ (n, ids, [adec], []))
	    
----------------------------------------
   
genJustProp :: [Flag] -> JustProp 
genJustProp fs | jpGlobalFlag fs = GlobalJust
	       | otherwise       = LocalJust

----------------------------------------

-- reads input from either the first named file, or stdin if none
getInput :: [Ann Id] -> IO (E String)
--getInput []     = do putStr "Please enter your program on stdin:\n"
--	 	     readInput (nullAnn "stdin") stdin
getInput [f]    = do let f' = item f
		     eh <- (eOpenFile f' ReadMode) `catch` (handleFileError f)
		     case eh of
			Succ h -> readInput f h
			Err e  -> return $ Err e

    where eOpenFile f m = do h <- openFile f m
			     return (Succ h)

getInput (f:fs) = do putStr $"warning: only reading first file ("++item f++")\n"
		     getInput [f]

readInput :: Ann Id -> Handle -> IO (E String)
readInput f h = (eHGetContents h) `catch` (handleFileError f)
    where eHGetContents h = do x <- hGetContents h
			       return (Succ x)

-- a simple exception handler for file operations, takes the name of the file 
-- being worked on and in the event of an error prints an appropriate message
-- and exits (failure)
handleFileError :: Ann Id -> IOError -> IO (E a)
handleFileError aid x 
    | isAlreadyInUseError x = err ("`" ++ f ++ "' is already in use")
    | isDoesNotExistError x = err ("`" ++ f ++ "' does not exist")
    | isPermissionError   x = err ("you do not have permission to " ++
					   "access `" ++ f ++ "'")
    | otherwise             = err "miscellaneous file error"
    where   f  = item aid
	    ix = ann aid
	    pos | isNoIx ix = ""
		| otherwise = (formatSrcPosLine $ ixSrcPos ix) ++ " "
	    err str = return (Err (pos ++ "ERROR: " ++ str))
	    
	    
fatalError s = do  
    putStrLn s
    exitFailure

--------------------------------------------------------------------------------

checkDanglingAnnotations :: Prog -> IO ()
checkDanglingAnnotations prog = 
    case findDanglingAnnotations prog of
	[]  -> return ()
	ads -> let ixs = map (ann . sigAId . item . decToSig . item) ads
		   pls = map (\ix -> (ixSrcPos ix, ixLoc ix)) ixs
	           msg (p,l) = "ERROR: dangling type annotation at " ++ 
				formatSrcPos' p 
	           msgs = unlines (map msg pls)
	       in  fatalError msgs
