--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------



-- pretty prints error messages

module ErrorMsg (errorBasic, 
		 errorSubsumption, errorAmbiguous, errorInference,
		 SubFailure, showTypeScheme)
where

import Pretty
import Terminal

import Ann_InputAST (noLoc, Loc)
import Core
import Chr  (chrSrcPos)
import Misc (SrcPos, formatSrcPosLine, unSucc)

import List (nub)


--------------------------------------------------------------------------------

-- a subsumption error diagnosis (if there is one):
--  - Left (annotation too polymorphic): returns the set of locations for 
--              each subsumption error explanation
--  - Right (unmatched inferred constraints): returns each unmatched user
--		constraint paired with its source location
type SubFailure = Either [[Loc]] ([(UCons,Loc)])

--------------------------------------------------------------------------------

errorBasic :: SrcPos -> PrettyDoc -> PrettyDoc
errorBasic pos ptxt = textPD (formatSrcPosLine pos ++ " ERROR: ") <||> ptxt


showTypeScheme (v,c,t) = let phi = simpTVarSubst (nub (v ++ fv (c,t)))
			     (c',t') = apply phi (c,t) -- presentableVars (c,t)
			     v' = apply phi (map Var v)
			     paren d | length (consToUs c) == 1 = d
				     | otherwise = textPD "(" <||> d <||>
						     textPD ")"
			     dv = if null v
				  then emptyPD
				  else let str = "forall " ++	
						 uncommas (map show v') ++ ". "
				       in  textPD str
				      
			     dc = if c == trueCons 
				  then emptyPD 
				  else paren (pprint c') <||> textPD " => "
				      
			     dt = pprint t'
			     
			 in  (dv <||> dc <||> dt, phi)

--------------------------------------------------------------------------------

--errorSubsumption :: ((CHR, CHR), (TypeScheme, TypeScheme)) -> String

--errorSubsumption :: Prog -> ((CHR, CHR), (TypeScheme, TypeScheme), SubFailure) 
--		    -> String

errorSubsumption prog ((chr_a, chr_i), (ts_a, ts_i), sf) = 
    let pos = chrSrcPos chr_a
	hd  = errorBasic pos (textPD ("Inferred type does not subsume " ++ 
				      "declared type"))
	dec = textPD "Declared: " <||> (fst (showTypeScheme ts_a))
	(phi,inf) = let (text,phi) = showTypeScheme ts_i
		    in  (phi, textPD "Inferred: " <||> text)
	diag = case sf of
		-- the "annotation-too-polymorphic" case
		Left  lss  -> if null lss || null (head lss) 
			      then {- trace "LEFT" -} emptyPD
			      else vcat [textPD "problematic locations:",
					 textPD (show lss),
				         textPD (printProgHL (concat lss) prog)]
		-- the "unmatched inferred constraints" case
		Right cons -> if null cons then {- trace "RIGHT" -} emptyPD
			      else let cons' = [(apply phi c,l) | (c,l) <- cons]
				       ctxt = uncommas (map (show.fst) cons')
				   in textPD $
					"following inferred " ++ 
					plural cons' "constraint" ++ 
					" unmatched: " ++ ctxt
			      
    in  toString (vcat [hd, dec, inf, diag, nothingPD])


----------------------------------------

-- gets passed: the CHR that was tested, the ambiguous variables, the
-- constraint from the type scheme
-- if the flag is True, the ambiguous type is an inferred type
--errorAmbiguous :: Bool -> Prog -> (CHR, ([TVar], Constraint)) -> String
errorAmbiguous prog inf (chr, (tvs, ts@(vs,(orig_c,c),t))) =
	  let (dts,phi) = showTypeScheme (vs,c,t)
	      vls = reportVars (tvs, orig_c) 
	      dvls = vcat [textPD "Ambiguous variables appear at these locations:",
		           vcat [ vcat [dv <||> textPD ": ", dls] | (v,ls) <- vls, 
			          v `elem` vs,
			          let dv  = textPD (show (apply phi (Var v)))
				      dls = textPD (printProgHL ls prog) ] ]
		     
	  in  toString $ if inf
		then vcat [msg, dts, dvls, nothingPD]
		     -- don't print any explanation if it's a declared type
		else vcat [msg, dts, nothingPD]

  where msg = errorBasic pos (textPD (kind ++ " type scheme is ambiguous:"))

	pos = chrSrcPos chr
	kind | inf = "Inferred"
	     | otherwise = "Declared"
		

        -- given a list of "ambiguous" variables, and a constraint,
        -- finds the locations whose type contains any of the variables
        reportVars (avs, c) = map reportVar [ (v,c) | v <- avs ]
        reportVar (v, c) = 
          let bcs = consToBs c
	      mgu = unSucc $ unify [v] idSubst (buildUP bcs)
--	      vs  = [ tv | (tv,t) <- mgu, isVar t, let v' = unVar t, v == v' ]
	      vs  = [ tv | (tv,t) <- mgu, v `elem` fv t ]
	      ls  = filter (/=noLoc) (map varLoc (v:vs))
          in  (v, ls)

--------------------------------------------------------------------------------

-- just prints a common header, doesn't do the full-blown error reporting
-- (which is handled by TypeDebug.reportConflicts)
errorInference :: SrcPos -> String
errorInference sp = toString $ errorBasic sp (textPD "Type error")
