--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------

-- evidence translation

module EvidenceTrans (evidenceTranslation)
where

import Core
import Chr
import Ann_InputAST
import TD_Monad 
import TD_General
import Justify
import TD_StdDefs (trueCName, haskellUCNames)
import TDInference
import Misc

import Pretty

import List
import Maybe
import IOExts


-- transforms program in global state
evidenceTranslation :: TD ()
evidenceTranslation = do

    -- for each fun id location, find if its an overloaded id or not
    lmap <- classifyIds

    -- use set of CHRs with True user constraint
    chrs   <- getCHRs
    chrs'  <- removeCHRHaskellCons chrs
    chrs'' <- addTrueToCHRs chrs'
    setCHRs chrs''
  
    p0 <- getProg
    p1 <- transAnnDecs False p0
    let Prog adecs0 = p1

--    puts ("p0: " ++ pprintAnn p0)
--    puts ("p1: " ++ pprintAnn p1)

    adecs1 <- processDecs lmap adecs0
    let p2 = Prog adecs1

--    puts ("p2: " ++ pprintAnn p2)

    p3 <- transAnnDecs True p2

    setProg p3
    removeRuleDecs

    where
	
	-- need a version that also looks into the RHS of each def. and
	-- processes the decs therein
	processDecs :: [LocMap] -> [Ann Dec] -> TD [Ann Dec]
	processDecs lmap [] = return []
	processDecs lmap (ad:ads) = do
	    ad' <- processEachSubDec procOne ad
	    let d = item ad
		mlec | isAValDec d = avalDecEvs lmap ad'
		     | isValDec  d = valDecEvs lmap ad'
		     | otherwise   = return ad'
		     
	    lec  <- mlec
	    lecs <- processDecs lmap ads
	    return (lec:lecs)

	    where
		procOne d = do ds <- processDecs lmap [d]
			       return (head ds)

	-- translates an unannotated definition, simply by demanding and using
	-- the most-specific necessary evidence
	-- e.g. say the function demands Eq [a] - we simply require that this
	--	is constructed externally and passed in, in its entirety
	--	(as opposed to, say, requesting Eq a, and building the list
	--	version internally)
	valDecEvs :: [LocMap] -> Ann Dec -> TD (Ann Dec)
	valDecEvs lmap ad@(Ann ix (ValDec ads)) = do
	    let l = ixLoc ix
	    ucs <- getUnannotatedEvidenceCons l
	    let ls  = sort $ map unJust ucs  -- just want the justifications
		ids = map locsToId ls
		gs  = group (zip ls ids)
		es  = map build gs
	    
		pats = map (nullAnn . VarPat) ids
		ads1 = map (addPatsToADef pats) ads
		ads2 = map (applyExpUpdates es) ads1
	    return (nullAnn (ValDec ads2))
	    
	    where
		locsToId ls = 'e' : (concat $ intersperse "_" (map show ls))
		
		-- groups evidence by leading location
		group :: [([Loc], Id)] -> [(Loc, [Id])]
		group [] = []
		group (x@(l:_,id):lids) = 
		    let (xs,ys) = span (\(n:_,_) -> l == n) lids
			g = (l, map snd (x:xs))
		    in  g : group ys

		-- builds translated expression for location l, out of evidence
		-- represented by ids
		build :: (Loc, [Id]) -> (Loc, Exp)
		build (l, ids) = case lookup l lmap of
		    Nothing -> error "valDecEvs: zat should not be!"
		    Just OverID	     -> (l, cons ids)
		    Just (FuncID id) -> (l, cons (id:ids))
		    where
			cons ids = let (v:vs) = map (nullAnn . VarExp) ids
				       app e f = nullAnn (AppExp e f)
				   in  item $ foldl app v vs

        -- for each AValDec, go through all defs, generating constraints and
	-- building evidence values
	avalDecEvs :: [LocMap] -> Ann Dec -> TD (Ann Dec)
	avalDecEvs lmap adec = 
	    case item adec of
		AValDec sig ads -> do
		    let l = ixLoc $ ann adec
		    ev_ucs <- getEvidenceCons l
		    hcs <- getHaskellConNames
		    let ev_ucs' = removeHaskellCons hcs ev_ucs
			ucs_i = ev_ucs'
{-		    
			-- split constraints into inferred and annotated
			(ucs_i, ucs_a) = 
			    let pls  = sigPredLocs (item sig)
				pred = \u-> head (unJust (ucIx u)) `notElem` pls
			    in  partition pred ev_ucs'
-}
		        evs = map ucToEv ucs_i

		    -- now get rid of the "global" constraints
		    -- when there are multiple, equivalent constraints arising
		    -- form the same initial location, keep only the one with
		    -- the smallest justification 
		    -- (assumption is that constraints from an outer scope will
		    --  have a lower number)
			evs' = removeGlobalEvidence evs
		        ces  = constructEvidence ecPrefix epPrefix lmap evs

			id = sigId (item sig)
		    

		    -- find the chr corresponding to the type annotation
		    -- get this type context, and generate the mapping from
		    -- provided evidence to inferred evidence
		    chrs <- getCHRs
		    let id = adefId (head ads)
			achr = head [ c | c <- chrs, 
					  let t = chrType c,
					  (isValueAnn t && chrTypeId t == id) ||
					  (chrLoc c == ixLoc (ann sig)) ]

		    -- add additional function parameters to each def
		    -- first remove True ucs
		    let (ucs_a,_) = chrRhs achr

		    let	ucs_i' = filter (\uc -> uconsPSym uc /= trueCName) ucs_i
			ucs_a' = filter (\uc -> uconsPSym uc /= trueCName) ucs_a
			ucs_a''= removeHaskellCons hcs ucs_a'

		    -- now remove all global constraints from annotation
			-- local constraint locations
			luc_ls = map ixLoc $ anns $ sigPreds $ item sig
			ucs_a'''= [ uc | uc <- ucs_a'', 
					 any (`elem`luc_ls) (unJust uc) ]
		    

			match = snd (matchCons [] ucs_a'' ucs_i') 
				-- ucs_i' ucs_a'') -- ucs_a'' ucs_i')

			pids = map ucToId ucs_a'''
			pats = map (nullAnn . VarPat) (nub pids)
		    let ads1 = map (addPatsToADef pats) ads
		    -- now insert evidence where needed
			ads2 = map (applyExpUpdates ces) ads1

		    -- we have pairs of matching constraints, generate the 
		    -- let-expression as so: [(uc1, uc1'), ..., (ucn, ucn')]
		    -- becomes: let suc1 = suc1' in ... let sucn = sucn' in ...
			ads3 = addEvMap match ads2

		        avaldec = nullAnn (AValDec sig ads3)
                    -- remove all type annotations found in subexpressions
                    -- necessary because hugs/ghci don't support scoped type 
		    -- variables (in a way compatible with our approach)
		    avaldec' <- rmAValAnnot avaldec
{-
		    puts ("-----------------------")
		    puts ("id: " ++ show id)
		    puts ("getEvidenceCons: " ++ show ev_ucs)
		    -- puts ("achr: " ++ show achr)
		    puts ("ucs_i: " ++ show ucs_i)
		    puts ("ucs_i': " ++ show ucs_i') -- (ucs_i, ucs_i'))
		    puts ("ucs_a'': " ++ show ucs_a'') -- (ucs_a, ucs_a',ucs_a''))
		    puts ("ucs_a''': " ++ show ucs_a''')
		    puts ("luc_ls: " ++ show luc_ls)
		    puts ("evs: " ++ show evs)
		    puts ("evs': " ++ show evs')
		    puts ("ces: " ++ show ces)
		    puts ("match: " ++ show match)
		    puts ("pats: " ++ show pats)
		    puts ("ad:\n" ++ pprintAnn adec)
{-
		    puts ("ads1:\n" ++ show ads1)
		    puts ("ads2:\n" ++ show ads2)
		    puts ("ads3:\n" ++ show ads3)
	    
		    puts ("av:\n" ++ show avaldec)
		    puts ("av':\n" ++ show avaldec')
-}		    
		    puts ("-----------------------")
-}
		    return avaldec'


-- removes all annotations from expressions in the body of the dec
-- ALSO, replaces all occurrences of anonymous type variables (*), by unique
-- named variables (since Haskell doesn't support them)
rmAValAnnot :: Ann Dec -> TD (Ann Dec)
rmAValAnnot ad = 
    let ret x = return $ Ann (ann ad) x
    in  case item ad of
	    AValDec s defs -> do s' <- nameAnonVars s 
				 ret (AValDec s' (map rmAnnot defs))

rmAnnot :: Ann Def -> Ann Def
rmAnnot = rmAnnotDef
    where
	rmAnnotDef ad = 
	    let ix = ann ad
		d  = item ad
	    in  case d of 
		Def id ps rhs w -> Ann ix (Def id ps (rmAnnotRHS rhs)
						     (rmAnnotDecs w))

	rmAnnotDecs = map rmAnnotDec
	rmAnnotDec ad = 
	    let ix = ann ad
		d  = item ad
	    in  Ann ix $ case d of 
			    ValDec fs    -> ValDec (map rmAnnotDef fs)
			    AValDec s fs -> -- remove annotation on definitions
					    -- AValDec s (map rmAnnotDef fs)
					    ValDec (map rmAnnotDef fs)
			    OverDec s fs -> OverDec s (map rmAnnotDef fs)
			    _		 -> d

	rmAnnotGEs ges = [ (rm g, rm e) | (g,e) <- ges ]

	rmAnnotRHS (RHS1 ae)  = RHS1 (rmAnnotExp ae)
	rmAnnotRHS (RHSG ges) = RHSG (rmAnnotGEs ges)

	rmAnnotAlt (CaseAlt ap ae w) = CaseAlt ap (rm ae) (rmAnnotDecs w)
	rmAnnotAlt (CaseAltGuard ap ges w) = CaseAltGuard ap (rmAnnotGEs ges)
							     (rmAnnotDecs w) 

	rm = rmAnnotExp
	rmAnnotExp :: Ann Exp -> Ann Exp
	rmAnnotExp ae =
	    let ix = ann ae
		e  = item ae
	    in  Ann ix $ case e of
			    AbsExp id e1 -> AbsExp id (rm e1)
			    AppExp e1 e2 -> AppExp (rm e1) (rm e2)
			    IfxAppExp e1 e2 e3 -> IfxAppExp(rm e1)(rm e2)(rm e3)
			    SecExpLeft  e1 e2 -> SecExpLeft  (rm e1) (rm e2)
			    SecExpRight e1 e2 -> SecExpRight (rm e1) (rm e2)
			    LetExp ds e1 -> LetExp (rmAnnotDecs ds) (rm e1)
			    CaseExp ae as -> CaseExp (rm ae) (map rmAnnotAlt as)
			    IfExp e1 e2 e3 -> IfExp (rm e1) (rm e2) (rm e3)
			    AnnExp e1 _	-> item (rm e1)
			    TupExp es -> TupExp (map rm es)
			    ListExp es -> ListExp (map rm es)
			    EnumExpFrom e1 -> EnumExpFrom (rm e1)
			    EnumExpFromTo e1 e2 -> EnumExpFromTo (rm e1) (rm e2)
			    EnumExpFromThen e1 e2 -> EnumExpFromThen (rm e1) 
								     (rm e2)
			    EnumExpFromThenTo e1 e2 e3 -> EnumExpFromThenTo 
							(rm e1) (rm e2) (rm e3)
			    _ -> e

-- names anonymous variables in the given sig
nameAnonVars :: Ann Sig -> TD (Ann Sig)
nameAnonVars as = 
    case item as of
     Sig st id ac at -> do ac' <- nameCtxt ac 
			   at' <- nameVars at
			   return $ Ann (ann as) $ Sig st id ac' at'
    
    where nameVars :: Ann Typ -> TD (Ann Typ)
	  nameVars at = do
	    t <- case item at of
		    v@(VarTyp id) | id == "*" -> do n <- next
						    let t = '_' : show n
						    return (VarTyp t)
				  | otherwise -> return v
		    AppTyp t1 t2 -> do t1' <- nameVars t1
				       t2' <- nameVars t2  
				       return (AppTyp t1' t2')
		    TupTyp ts	 -> do ts' <- mapM nameVars ts
				       return (TupTyp ts')
		    ListTyp t	 -> do t' <- nameVars t
				       return (ListTyp t')
		    t		 -> return t
	    return (Ann (ann at) t)	

	  nameCtxt :: Ann Ctxt -> TD (Ann Ctxt)
	  nameCtxt ac = 
	    case item ac of
		Ctxt aps -> do aps' <- mapM namePred aps
			       return $ Ann (ann ac) $ Ctxt aps'
			       
	  namePred :: Ann Pred -> TD (Ann Pred)
	  namePred ap = 
	    case item ap of
		Pred id t -> do t' <- nameVars t
			        return $ Ann (ann ap) $ Pred id t'




addEvMap :: [(UCons, UCons)] -> [Ann Def] -> [Ann Def]
addEvMap pucs ads = addEvMap' (buildEvMap pucs) ads
    where
	addEvMap' _ [] = []
        addEvMap' bevm (ad:ads) =
	    let rhs = defRHS ad
		rhs'= appEvMapRHS bevm rhs		
		ad' = newDefRHS ad rhs' -- rhs 
		ads'= addEvMap' bevm ads
	    in  (ad':ads')

-- applies an evmap functio (as per buildEvMap) to all expressions in the rhs
appEvMapRHS :: (Ann Exp -> Ann Exp) -> RHS -> RHS
appEvMapRHS f (RHS1 ae)  = RHS1 (f ae)
appEvMapRHS f (RHSG ges) = RHSG [ (f g, f e) | (g, e) <- ges ]
    
-- returns a function which should be applied to the existing expression in the
-- body of the function definition
-- NOTE: pairs are backwards (relative to what I'd expect)
buildEvMap :: [(UCons, UCons)] -> (Ann Exp -> Ann Exp)
buildEvMap [] = \ae -> ae
buildEvMap ((uc1, uc2):pucs) = \ae -> 
    let id1 = ucToId uc1
	id2 = ucToId uc2
	
	id = nullAnn id2
	e1 = nullAnn (VarExp id1)

	e2 = buildEvMap pucs ae
	dec = [nullAnn (ValDec [nullAnn (Def id [] (RHS1 e1) [])])]
	-- test whether it's a redundant binding, which might end up
	-- introducing non-termination eg. let f = f in ... f ...
	-- if so, ignore it
    in  if id1 == id2
	   then e2
	   else nullAnn (LetExp dec e2)


-- sorts given evidence by locations
sortEvidence :: [Evidence] -> [Evidence]
sortEvidence evs = sortBy ord evs
    where ord (Ev _ ls1) (Ev _ ls2) = compare ls1 ls2

ucToEv :: UCons -> Evidence
ucToEv uc = let ps = uconsPSym uc
		ts = ucTypes uc
		ls = unJust uc
		tag = if trueCName == ps then TRUE 
					 else USER ps ts
	    in  Ev tag ls

----------------------------------------

-- removes rule (and overload0) declarations from program in state
removeRuleDecs :: TD ()
removeRuleDecs = do
    (Prog adecs) <- getProg
    let rem ad = let d = item ad in not (isRuleDec d || isOver0Dec d)
	adecs' = filter rem adecs
	p' = Prog adecs'
    setProg p'

----------------------------------------

-- 1) translates all the annotated declarations in the given program, such that
--    the context constraints result in function parameters
--    eg. from:  f :: (C t1, ... , C tn) => t
--        to:  f :: t1 -> ... tn -> t
-- 2) renames ids defined by overload decs
-- 3) converts overload decs into avaldecs
--
-- NOTE: evidence constructor prefix below must be used consistently elsewhere!
-- ALSO: we skip over constraints which are native to Haskell
--
-- NOTE: type schemes are only transformed if flag is True!
transAnnDecs :: Bool -> Prog -> TD Prog
transAnnDecs trans (Prog adecs) = do adecs' <- mapM processAll adecs
				     return (Prog adecs')
	
    where processAll adec = do adec' <- processDec adec
			       processSubs adec'
	  processSubs = processEachSubDec processDec 
	  processDec adec
	    | isAValDec (item adec) = transAVal adec
	    | isOverDec (item adec) = transOver adec
	    | otherwise		    = return adec
	    
	  transAVal aaval =
	    let aval = item aaval
		ix   = ann aaval
	    in  case aval of
		    AValDec sig ds -> do sig' <- transSigAval sig
					 return $ Ann ix (AValDec sig' ds)

	  transOver aover =
	    let over = item aover
		ix   = ann aover
	    in  case over of
		    OverDec sig ds -> do (id, sig') <- transSigOver sig
					 let ds' = map (newADefId id) ds
					 return $ Ann ix (AValDec sig' ds')
	    
	  -- also returns the new name for this identifier
	  transSigOver asig =
	    let sig = item asig
		ix  = ann asig
	    in  case sig of
		    (Sig st aid ac at) -> do
			-- if ls empty then use ix of sig
			(ac', cts) <- ctxtTypes ac
		        let (ls, ts) = unzip cts
			    ls' = if null ls then [ix] else ls
			    at' = foldr arr at ts
			    id' = ecPrefix ++ unwordsUS (map show ls')
			    aid'= Ann (ann aid) id'
			return$if trans then (id', Ann ix (Sig st aid' ac' at'))
					else (id', Ann ix (Sig st aid' ac  at))

	  transSigAval asig =
	    let sig = item asig
		ix  = ann asig
	    in  case sig of
		    (Sig st id ac at) -> do 
			(ac', cts) <- ctxtTypes ac
			let ts  = map snd cts
			    at' = foldr arr at ts
			return $ if trans then Ann ix (Sig st id ac' at')
					  else Ann ix (Sig st id ac  at)
					  
	  arr t1 t2 = let arrCon = nullAnn (ConTyp "->")
		      in  nullAnn (AppTyp (nullAnn (AppTyp arrCon t1)) t2)
		
	  ctxtTypes actxt = 
	    case item actxt of 
	     (Ctxt aps) -> do
		xs <- mapM predType aps
		let (aps', ts) = splitEither xs
		return $ (Ann (ann actxt) (Ctxt (concat aps')), ts)
				    
	  predType apred =
	    case item apred of    
		(Pred id at) -> do
		    hns <- haskellUCStrs
		    return $ 
		     if id `elem` hns
		     then if isConTyp at	     -- throw away
			  then Left []		     -- tycon types
			  else Left [apred]
		     else Right (ann apred, at)
	
	  haskellUCStrs = do
		hns <- getHaskellConNames
		return $ map nameStr (haskellUCNames ++ hns)
	    

unwordsUS []  = ""
unwordsUS [x] = x
unwordsUS (x:xs) = x ++ "_" ++ unwordsUS xs   
	

----------------------------------------

ecPrefix = "ec"
epPrefix = "p"

-- we abstract from the actual user-defined constraints (including the true constraint)
data EvTag    = TRUE | USER PredSym [Type] deriving (Eq, Show)
data Evidence = Ev EvTag [Loc] deriving (Eq, Show)

isUSERTag (USER _ _) = True
isUSERTag _ = False

removeGlobalEvidence es = rem [] es
    where 
      rem as [] = as
      rem as (e@(Ev TRUE ls1):es) = rem (e:as) es
      rem as (e@(Ev t1 (l1:ls1)):es)   = 
	 let (ms,es') = partition (\(Ev t2 (l2:ls2)) -> t1 == t2 && l1 == l2) es
	     m = min (e:ms)
	 in  rem (m:as) es'

      min (e:es) = m e es
	    where m a [] = a
		  m a@(Ev _ ls1) (e@(Ev _ ls2):es) | ls2 < ls1 = m e es
						   | otherwise = m a es

instance Ord Evidence where
   (<) (Ev _ l1) (Ev _ l2) = l1 < l2

data LocKind = FuncID String | OverID deriving (Eq, Show)

type LocMap = (Loc, LocKind) 

-- assume evidence constraints stored in ucs are already sorted
-- ec is the name we give to the evidence constructors, will be indexed by locations
-- evp is the name we give to evidence parameters, indexed by predicate symbol and type
-- lmap tells whether evidence corresponds to a regular function or overloaded defintion application
--constructEvidence :: String -> String -> [LocMap] -> [Evidence] -> [(Loc,String)]
constructEvidence :: String -> String -> [LocMap] -> [Evidence] -> [(Loc, Exp)]
constructEvidence ec evp lmap ucs = constructEvidence2 ec evp lmap (msort ucs)
constructEvidence2 _ _ _ [] = []
constructEvidence2 ec evp lmap ucs =
   let  -- collect all evidences according to their leading location
        collect :: [Evidence] -> [[Evidence]]
        collect [] = []
        collect [x] = [[x]] 
        collect ((x@(Ev _ l1)):((y@(Ev _ l2)):ys)) =
            if ((head l1) == (head l2)) 
            then  let z:zs = collect (y:ys)in (x:z):zs
            else [x]:(collect (y:ys)) 

	lookup lm l = Prelude.lookup l lm

	mshow xs = unwordsUS (map show xs)

        -- generate evidence for (U1 t1)_J1, ..,(Un tn)_Jn where Ji= m,Ji'
        genEv :: [Evidence] -> (Loc,Exp)
        genEv [Ev TRUE [m,l]] = (m, VarExp (ec ++ (mshow [l])))
        genEv ((Ev TRUE [m,l]):xs) = error "Should never happen !"
          -- cause we exclude a rule such as UC _ <==> True, UC' _
        genEv [Ev (USER p ts) [m]] = 
	    let ep = evp ++ predNameAlpha (show p) ++ concatMap typeToId ts
	    in  (m, VarExp ep)
	
        genEv (ev@(Ev (USER p ts) [m]):xs) = error ("Should never happen !!" ++ show (ev:xs))
        genEv (xs@((Ev _ (m:_)):_)) = 
             let -- remove m from all evidence lists
		 xs' = map (\x -> case x of (Ev evTag ls) -> (Ev evTag (tail ls))) xs
                 evidence = constructEvidence2 ec evp lmap xs'
                 -- evidence = [(l1,e1), ..., (ln,en)]
                 app c c' = AppExp (nullAnn c) (nullAnn c')
                 -- c, c' represent code, we generate c @ c'
             in case lookup lmap m of
                  Just (FuncID f) -> 
			      let f' = VarExp f
				  code = foldl (\c -> \x -> case x of (l,c') -> app c c') f' evidence
                              in  (m,code) 
                               -- build (..(f e1)... en)
--                  Just (OverID) -> 
		  _ ->
			      let idx = nub (map (\x -> case x of (Ev evTag ls) -> head (tail ls)) xs)
                                  -- important to remove redundant idxs
                                  f  = ec ++ (mshow idx)
				  f' = VarExp f
                                  code = foldl (\c -> \x -> case x of (l,c') -> app c c') f' evidence
                              in (m, code)
                              -- build  (.. (ec(nub idx) e1)...en)
--		  Nothing -> error ("genEv: lookup failed on " ++ show m ++
--			        "\nlmap: " ++ show lmap)
        genEv xs = error ("Should never happen !!!: " ++ show xs)
   in  map genEv (collect ucs)



-- takes a type and generates a string representation that can be used as an
-- indentifier in Haskell (no guarantee about case of first letter)
-- WARNING: ids generated could clash in the presence of type constructors
-- called "ARR", "LST, etc.
typeToId :: Type -> String
typeToId (Var (x, i)) = show x
typeToId (TApp t1 t2) = "I" ++ typeToId t1 ++ "_" ++ typeToId t2 ++ "I"
typeToId (TyCon x)
    | x == arrowName = "ARR"
    | x == nilName   = "LST"
    | otherwise = let tup = tupleArity x
		  in  if isJust tup
			 then "T" ++ show (fromJust tup)
			 else show x


ucToId uc = let id = ucNameStr uc
		ts = ucTypes uc
		tids = unwordsUS (map typeToId ts)
	    in  epPrefix ++ predNameAlpha id ++ tids

-- converts the given predicate name to one containing only alphanumeric
-- characters
predNameAlpha :: String -> String
predNameAlpha xs = concatMap conv xs
    where conv c = case c of 
		    ':' -> "COLON"
		    '!' -> "BANG"
		    '#' -> "HASH"
		    '$' -> "DOLLAR"
		    '%' -> "PERCENT"
		    '&' -> "AMPERSAND"
		    '*' -> "ASTERISK"
		    '+' -> "PLUS"
		    '.' -> "DOT"
		    '/' -> "FSLASH"
		    '<' -> "LTHAN"
		    '=' -> "EQUALS"
		    '>' -> "GTHAN"
		    '?' -> "QMARK"
		    '@' -> "AT"
		    '\\'-> "BSLASH"
		    '^' -> "HAT"
		    '|' -> "BAR"
		    '-' -> "DASH"
		    '~' -> "TILDE"
		    _   -> [c]
	  

--------------------------------------------------------------------------------

-- replace "blank" body of an overload simp. CHR with the True token
addTrueToCHRs :: [CHR] -> TD [CHR]
addTrueToCHRs chrs = do
    let (chrs1, chrs2) = partition (isOverloadRule . chrType) chrs
    chrs1' <- mapM replaceRHS chrs1
    return (chrs1' ++ chrs2)

    where replaceRHS chr = 
	    case chrRhs chr of
		([], bcs) -> do n <- next
				let j = unJust (simpUCons chr)
				-- note, True should be justified by same
				-- location as head ucons
				let uc = UC n trueCName [] (J j)
				return (chrNewRhs ([uc], bcs) chr)
		rhs 	  -> return chr



-- NOTE: we ignore literals (Haskell will handle their overloading)
classifyIds :: TD [LocMap]
classifyIds = do
    chrs <- getCHRs
    (Prog adecs) <- getProg
    let -- names of overloaded ids
	overIds = overDecsToIds0 adecs
	lm1 = concatMap (genLMDec [] overIds) (items adecs)
	-- now do this for locations appearing on the rhs of simplification
	-- rules (arising from overload decs)
	ochrs = filter (isOverloadRule . chrType) chrs
	rhss  = map chrRhs ochrs
	lm2   = [ (l, OverID) | (ucs,_) <- rhss, l <- concatMap unJust ucs ]

    return (lm1 ++ lm2)
	
    where 

	genLMDec :: [Id] -> [Id] -> Dec -> [LocMap]
	genLMDec vs mp dec = case dec of
	    ValDec  ads	  -> concatMap (genLMDef vs mp) (items ads)
	    AValDec _ ads -> concatMap (genLMDef vs mp) (items ads)
	    OverDec _ ads -> concatMap (genLMDef vs mp) (items ads)
	    _		  -> []
	    
	genLMDef :: [Id] -> [Id] -> Def -> [LocMap]
	genLMDef vs mp (Def _ ps rhs w) =
	    let vs' = genLMPats ps ++ vs
		ws  = concatMap (genLMDec vs' mp) (items w)
		rs  = genLMRHS vs' mp rhs
	    in  rs ++ ws

	genLMRHS vs mp (RHS1 ae)  = genLMExp vs mp ae
	genLMRHS vs mp (RHSG ges) = 
		concat [ genLMExp vs mp g ++ genLMExp vs mp e | (g,e) <- ges ]

	genLMCA vs mp (CaseAlt ap ae w) = 
	    let vs' = genLMPat (item ap) ++ vs
		ws  = concatMap (genLMDec vs' mp) (items w)
		rs  = genLMExp vs' mp ae
	    in  rs ++ ws
	genLMCA vs mp (CaseAltGuard ap ges w) = 
	    let vs' = genLMPat (item ap) ++ vs
		ws  = concatMap (genLMDec vs' mp) (items w)
		rs  = concat [genLMExp vs mp g ++ genLMExp vs mp e |(g,e)<-ges]
	    in  rs ++ ws


	genLMStmt vs mp (GenStmt ap ae) = 
	    let vs' = genLMPat (item ap) ++ vs
	    in  genLMExp vs' mp ae
	genLMStmt vs mp (QualStmt ae) = genLMExp vs mp ae
	genLMStmt vs mp (LetStmt ads) = concatMap (genLMDec vs mp) (items ads)

	genLMExp :: [Id] -> [Id] -> Ann Exp -> [LocMap]
	genLMExp vs mp aexp =
	    let exp = item aexp
		ix  = ixLoc (ann aexp)
		gen = genLMExp vs mp
	    in  case exp of
		    VarExp id	-> if id `elem` vs
		       	      then []
		       	      else if id `elem` mp
				   then [(ix, OverID)]
				   else [(ix, FuncID id)]
				    		    
		    AbsExp aps ae -> genLMExp (genLMPats aps ++ vs) mp ae

		    AppExp ae1 ae2 -> gen ae1 ++ gen ae2

		    IfxAppExp ae1 ae2 ae3 -> gen ae1 ++ gen ae2 ++ gen ae3

		    LetExp ads ae -> concatMap (genLMDec vs mp) (items ads) ++ 
				     gen ae
				     
		    CaseExp ae cas -> gen ae ++ concatMap (genLMCA vs mp) cas
		    
		    ListCompExp ae ss -> gen ae ++ 
					 concatMap (genLMStmt vs mp) (items ss)
		    
		    SecExpLeft ae1 ae2 -> gen ae1 ++ gen ae2
		    
		    SecExpRight ae1 ae2 -> gen ae1 ++ gen ae2
		    
		    EnumExpFrom ae -> gen ae
		    
		    EnumExpFromTo ae1 ae2 -> gen ae1 ++ gen ae2
		    
		    EnumExpFromThen ae1 ae2 -> gen ae1 ++ gen ae2

		    EnumExpFromThenTo ae1 ae2 ae3 ->gen ae1 ++ gen ae2++ gen ae3
		    
		    IfExp ae1 ae2 ae3 -> gen ae1 ++ gen ae2 ++ gen ae3

		    AnnExp ae _ -> gen ae

		    TAddExp ae _ -> gen ae

		    TupExp aes -> concatMap gen aes

		    ListExp aes -> concatMap gen aes

		    _		-> []	-- no sublocations, and not VarExp

	genLMPats ps = concatMap genLMPat (items ps)
	genLMPat (VarPat id)   = [id]
	genLMPat (ConPat _ ps) = genLMPats ps
	genLMPat (TupPat ps)   = genLMPats ps
	genLMPat (ListPat ps)  = genLMPats ps
	genLMPat _             = []


--------------------------------------------------------------------------------
-- takes a program location and returns all the user constraints in its type,
-- ordered by justification, and with prop. and global constraints removed
--
-- NOTE: we assume that administrative user constraints have a leading
-- justification location as passed to the function - these constraints must be
-- removed, they play  no part in evidence construction
getEvidenceCons :: Loc -> TD [UCons]
getEvidenceCons l = do
    opt <- getSolvOpt
    setSolvOpt SEvidence
    setJProp GlobalJust
    chrs <- getCHRs
    let chrs' = map stripGlobalUCons chrs
    setCHRs chrs'   -- get rid of global constraints from CHRs
    inferTypeLoc l (False, True, False) undefined
    setCHRs chrs
    setSolvOpt opt
    res <- lastResult
    if isResult res
       then do chrs <- getCHRs
	       let (_, _, _, (_, scons), _) = unResult res
	           ucs = consToUs scons
		   ucs' = dropAdmin ucs
	       ucs''  <- dropProps ucs'
	       return (sortBy ord ucs'')
       else trace ("lastResult: " ++ show res) $ return []

    where ord uc1 uc2 = compare (unJust uc1) (unJust uc2)

	  dropAdmin ucs = filter (not.isAdmin) ucs
	    where isAdmin uc = case unJust (ucJust uc) of
		    []	  -> False
		    (m:_) -> m == l

	  dropProps ucs = do	
	    pls <- propRuleLocs
	    let ucs' = filter (\uc -> not (any (`elem`pls) (unJust uc))) ucs
	    return ucs'

-- as above, but omits overload simplification rules from the derivation
-- this is used for unannotated definitions, since we want the most specific
-- evidence constraints
getUnannotatedEvidenceCons :: Loc -> TD [UCons]
getUnannotatedEvidenceCons l = do
    chrs <- getCHRs
    let chrs' = filter (not . isOverloadCHR) chrs
    setCHRs chrs'
    ucs <- getEvidenceCons l
    setCHRs chrs
    return ucs
    

-- given a list of user constraints, removes those which are native to haskell
-- (eg. Num, Show, Read etc.) uses those in StdDefs
-- first list is generated from 'hconstraint' declarations
removeHaskellCons :: [Name] -> [UCons] -> [UCons]
removeHaskellCons ns = 
    filter (\uc -> ucName uc `notElem` (haskellUCNames ++ ns))


removeCHRHaskellCons :: [CHR] -> TD [CHR]
removeCHRHaskellCons chrs = do
    ns <- getHaskellConNames
    return (map (remRHS ns) chrs)

    where remRHS ns chr = let (ucs,bcs) = chrRhs chr
			      ucs' = removeHaskellCons ns ucs
			      rhs' = (ucs',bcs)
			  in  chrNewRhs rhs' chr
	
--------------------------------------------------------------------------------

