--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


module InitChecks (initChecks)
where

import Core
import Chr
import Justify
import TD_Monad
import TD_General (withVar)
import ASTs
import Misc
import qualified TD_Translate as Trans
import OptFlags as Opt
import OptFlags (verbose, compileFlag)
import Ann_InputAST
import TD_StdDefs
import TDInference (inferTypeLoc, reportConflicts)

import Pretty

import ErrorMsg (errorInference)

import CHRChecks (performCHRChecks)
import ASTChecks (performASTChecks)
import AnnChecks (checkAnnotations)

import Maybe
import Monad

import IOExts

--------------------------------------------------------------------------------
-- performs all the initial checks
-- 1) structural CHR checks (confluence, termination, range-restrictedness) for
--    CHRs arising from overload and rule decls
-- 2) inference for value decls
-- 3) check annotations - unambiguity and subsumption

initChecks :: [Opt.Flag] -> TD ()
initChecks fs = do 
    prog    <- getProg
    allChrs <- getCHRs
    let dchrs = ruleOverCHRs allChrs
    verb (puts "> structural program checks")
    progOkay <- performASTChecks prog
    verb (puts "> structural CHR checks")
    chrCks <- performCHRChecks fs dchrs
    if not progOkay || isErr chrCks
       then if not progOkay then bailToOS "" else bailToOS (unErr chrCks)
       else -- checks passed, add the chrs from completion, and do inference
	    do let chrs' = allChrs ++ (unSucc chrCks)
	       setCHRs chrs'

	       verb (puts "> type inference")
	       -- if we were planning to compile, but inference failed, stop
	       mtmap <- typeInference fs
	       let tmap = fromJust mtmap

	       let finish = bailToOS "Cannot compile"

	       ann_okay <- if isNothing mtmap && compileFlag fs
			    then finish
			    else do verb (puts "> checking annotations")
				    checkAnnotations fs

	       comp (if not ann_okay
			then bailToOS "Cannot Compile"
			else -- if we're compiling, transform the program,
			     -- adding all top-level type annotations
			     do 
				verb (puts "> re-annotating all definitions")

				p <- getProg

--				puts "Ann."
--				puts (pprintAnn p ++ "\n")
				
--				puts "BEFORE"
--				puts (pprintPlain p ++ "\n")
				 
				addFunAnns tmap
				-- now regenerate CHRs
				p'@(Prog decs) <- getProg

--				puts "\nAFTER"
--				puts (pprintPlain p')

--				puts "Ann."
--				puts (pprintAnn p' ++ "\n")
				
--				error "stop"
				r_chrs <- regenerateCHRs p'
				setCHRs r_chrs

		    )
		       
    where -- returns chrs belonging to rule and overload decls
	  ruleOverCHRs :: [CHR] -> [CHR]
	  ruleOverCHRs chrs = filter isStructCHR chrs

	  verb = when (verbose fs)
	  comp = when (compileFlag fs)


-- after adding type annotations to all fun. defs, regenerate their CHRs
regenerateCHRs :: Prog -> TD [CHR]
regenerateCHRs p@(Prog decs) = do
    n <- next
    hns <- getHaskellConNames
    let hcs = [ (nameStr nm, nm) | nm <- hns ] 
    ecs <- doIO (Trans.progToCHRs (fromIntegral n) hcs p) 
    let ((_, chrs), m) = ecs
    setCounter (fromIntegral (Trans.num m))
    return (chrs ++ stdCHRs)


--------------------------------------------------------------------------------

-- go through the parsed program, attempt to infer a type for each value
-- declaration (ValDec, AValDec and OverDec)
-- returns a mapping from function id to type (in the form of a 
-- type var, constraint pair) if successful, otherwise Nothing
-- NOTE: also returns a type representing the LT parameter of the lambda-bound
--       variables
typeInference :: [Opt.Flag] -> TD (Maybe [((Id, Loc), (TVar,Type,Constraint))])
typeInference fs = do
    chrs <- getCHRs
    let -- look at join CHRs to find declaration ids and locations
	-- NOTE: ignore CHRs arising from universal annotations `::'
	jchrs = filter (\c -> (isValJoinCHR c || isOverloadJoinCHR c) &&
			      (not ("ann:" `isSublist` chrId c)) ) chrs
	lis  = reverse $ [ (chrLoc chr, chrId chr) | chr <- jchrs ]	
	locs = map fst lis
	msg  = if null locs
	       then "Nothing to type-check"
	       else "Type checking" ++ 
		    if verbose fs then " locations: " ++ show locs ++ "\n"
				  else " : "

    -- do this in two phases:
    --  - for each location, do inference in solver mode 0
    --  - then for each failed location, repeat in mode 2
    sopt <- getSolvOpt
    setSolvOpt SInference
    mapM_ (\l -> inferTypeLoc l (False, True, False) GlobalJust) locs
    ress <- getResults
    clrResults
    let lres = zip locs (reverse ress)
	-- locations with type errors, where inference needs to be repeated for
	-- the error report
	lbad = [ l | (l,r) <- lres, isConflict r ]
    if not (null lbad) -- any isConflict ress
       then do puts (msg ++ "errors found!\n")
	       setSolvOpt SDebug
	       mapM_ (\l -> inferTypeLoc l (False,True,True) GlobalJust) lbad
	       ress <- getResults
	       let cons = [ (sp,c,cs) | r <- ress, isConflict r, 
				        let Conflict (sp,_,cs,c) = r ]

		   report (sp,c,cs) = do puts (errorInference sp)
					 reportConflicts True c cs
					 puts ""
	       mapM_ report cons	       
	       setSolvOpt sopt
	       return Nothing
	    
       else do setSolvOpt sopt
	       when (not (null locs)) (puts (msg ++ "well-typed\n"))
	       let tvcs = map resToTVarCons (reverse ress) -- stored last-first
		   mp = [ ((id, l), tvc) | ((l,id), tvc) <- zip lis tvcs ]
	       return (Just mp)

    where resToTVarCons (Result (((_, tv),(_,lv),_), _, _, (cons, scons), _)) = 
	     -- trace ("(tv,lv,scons,cons): " ++ show (tv,lv,scons,cons)) 
	     -- (tv, scons)
	    let bcs = consToBs cons
		phi = unSucc $ unify [] idSubst (buildUP bcs)
		lv' = apply phi (Var lv)

		rho = unSucc $ unify (fv lv') idSubst (buildUP bcs)
		tv' = apply rho (Var tv)
		
	    in  -- trace  ("(tv,lv,tv',lv',scons,cons): " ++ show (tv,lv,tv',lv',scons,cons))    
		(tv, lv', scons)

	  resToTVarCons r = error ("resToTVarCons: " ++ show r)
	    

--------------------------------------------------------------------------------


{-

-- 1) updates the program by adding type annotations to all unannotated
--    functions - the function type is given by the corresponding (tvar, cons) 
--    pair (ie. translates ValDecs to AValDecs)
--    NOTE: the additional type passed in represents the LT parameter of the
--	    lambda-bound variables. This is necessary because our type
--	    variables are scoped. (so must be named consistently)
addFunAnns :: [((Id, Loc), (TVar, Type, Constraint))] -> TD ()
addFunAnns mp = do

    chrs <- getCHRs

    (Prog decs) <- getProg
    decs' <- processDecs mp decs

    -- p is the transformed program
    let p = Prog decs'
    setProg p
    
    return ()
    
    where processDecs :: [((Id, Loc), (TVar, Type, Constraint))] -> 
			 [Ann Dec] -> TD [Ann Dec]
	  processDecs _ [] = return []
	  processDecs mp (ad:ads) = do
	    let dec = item ad
	    ad' <- case dec of
		    (ValDec defs) -> 
			let idl@(id,_) = (adefId (head defs), ixLoc (ann ad))
			
			    mtvc = lookup idl mp
			    (tv, ls, c) = fromJust mtvc
			in  if isNothing mtvc
			    then error ("failed to add an annotation for `" ++
					show idl ++ "'\n" ++ 
					"couldn't find a type for it in: " ++ 
					showlist mp)
			    else do n1 <- next
				    sig <- genSig id tv c
				    let dec' = AValDec sig defs
				        ix1 = unkIx (fromIntegral n1)
					ad'  = Ann ix1 dec'

				    return ad' -- ad''
		    _	      -> return ad 
	    
	    let pd d = do d' <- processDecs mp [d]
			  return (head d')
	    ad'' <- processEachSubDec pd ad'
				    
	    ads' <- processDecs mp ads
	    return (ad'' : ads')

-}


-- 1. group locations by the top-level definition that they fall under
-- 2. then for each group, find all the inferred type information, and build a
--    substitution for unifying all of their lambda-bound variable components
addFunAnns :: [((Id, Loc), (TVar, Type, Constraint))] -> TD ()
addFunAnns mp = do
--    puts ("mp:\n" ++ unlines (map show mp))
    prog@(Prog adecs) <- getProg
    let vts = concatMap vtDec adecs
	dls = map vtDefLocs vts
	dlts = map getTypes dls
	phis = map unifyLambdas dlts
	dlts'= concat [ map (applyPhi phi) dlt | (phi,dlt) <- zip phis dlts ]
    
	sigInfo = [ (l,(id,tv,c)) | ((id,l),(tv,t,c)) <- dlts' ]

    lsigs <- mapM (\(l,(id,tv,c)) -> do sig <- genSig id tv c
--					puts ("id/c: " ++ show id ++ "/" ++ show c)
				        return (l,sig)) sigInfo

    let adecs1 = map (update lsigs) adecs
	adecs2 = map (runIdentity . processEachSubDec 
				    (return . update lsigs)) adecs1

    -- puts ("lsigs: " ++ show lsigs)

    setProg (Prog adecs2)

    where
    
	update :: [(Loc, Ann Sig)] -> Ann Dec -> Ann Dec
	update lsigs adec = case lookup (ixLoc (ann adec)) lsigs of
	    Nothing  -> adec
	    Just sig -> case item adec of
		    AValDec _ ads -> Ann (ann adec) (AValDec sig ads)

		    OverDec _ ads -> Ann (ann adec) (OverDec sig ads)
				       
		    ValDec ads -> Ann (ann adec) (AValDec sig ads)

		    _	       -> adec

    
	getTypes :: [(Id,Loc)] -> [((Id,Loc),(TVar, Type, Constraint))]
	getTypes idls = [ idt | idt@(idl, tc) <- mp, idl `elem` idls ]

	applyPhi phi (idl,(tv,t,c)) = (idl,(tv,t, apply phi c))
	
	unifyLambdas lts = let (v:vs) = map getV lts
			       eqs = map (v`njEq`) vs
			   in  unSucc (unify [] idSubst (buildUP eqs))
	    where
		getV (_,(_,v,_)) = v
	

   

-- takes an id, a type variable and a simplified constraint, and generates a
-- syntactic type signature from it
genSig :: Id -> TVar -> Constraint -> TD (Ann Sig)
genSig id tv c =
    let ucs = consToUs c
	bcs = consToBs c
	teq = withVar tv bcs
    in if not (singleton teq)
	  then error "genSig: error getting type component"
	  else do let t    = bcRight (head teq)
		      ucs' = -- filter out the administrative constraints
			     -- (those with an admin. justification)
			     filter ((adminLoc`notElem`) .unJust) ucs
		  typ  <- genTyp t
		  ctxt <- genCtxt ucs'
		  id' <- an id
		  an (univSig id' ctxt typ)
	      
    where
	an :: a -> TD (Ann a)
	an x = do n <- next
		  return $ Ann (unkIx (fromIntegral n)) x

	-- generates an AST Ctxt from a list of UCons
	genCtxt :: [UCons] -> TD (Ann Ctxt)
	genCtxt ucs = do ps <- mapM genPred ucs
			 an (Ctxt ps)

	-- generates an AST Pred from a UCons
	genPred :: UCons -> TD (Ann Pred)
	genPred (UC _ ps ts _) = do let id  = nameStr ps
				    typ <- genTyp (head ts)
				    if not (singleton ts)
					then error "genPred: context pred has >1 arg!"
					else an (Pred id typ)

	-- converts from internal Type format, back to external syntax format
	-- for types (as per AST)
	genTyp :: Type -> TD (Ann Typ)
	genTyp (Var tv) = an (VarTyp (show (fst tv)))

	genTyp (TyCon tc) | tc == tupName 0 = an (TupTyp [])
			  | otherwise	    = an (ConTyp (nameStr tc))
	
	genTyp ta@(TApp (TyCon tc) t1)
	    | tc == listName = do t1' <- genTyp t1
				  an (ListTyp t1')
	    | otherwise	     = genTypTApp ta
	
	genTyp ta@(TApp t1 t2) = genTypTApp ta

	genTypTApp ta@(TApp t1 t2) =
	    let mts = tupleTypes ta
		ts  = fromJust mts
	    in  if isNothing mts
		   then -- not part of a tuple constructor application
			do t1' <- genTyp t1
			   t2' <- genTyp t2
			   an (AppTyp t1' t2')
		   else -- generate a TupTyp
			do ts' <- mapM genTyp ts
			   an (TupTyp ts')



