--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------

-- This module defines the program structure as returned by the parser.


-- prog ::= [dec ";"]
-- dec  ::= "data" dat
--          "rule" rule
--          "overload" id
--          "overload" sig ";" def
--          def
-- def  ::= id [pat] "=" exp
-- exp  ::= id
--          conid
--          "\" id "->" exp
--          exp exp
--          exp id exp
--          "if" exp "then" exp "else" exp
--          exp "::" typ
--          "(" <exp ["," exp]> ")"
--          "[" <exp ["," exp]> "]"
-- pat  ::= "_"
--          id
--          conid [pat]
--          "(" <pat ["," pat]> ")"
--          "[" <pat ["," pat]> "]"
-- sig  ::= id "::" <"(" ctxt ")" "=>"> typ
-- rule ::= ctxt "==>" cnst
-- ctxt ::= pred ["," pred]
-- cnst ::= prim ["," prim]
-- pred ::= predid typ
-- prim ::= pred
--          typ "=" typ
-- typ  ::= id
--          conid
--          typ typ
--          typ conid typ
--          "(" <typ ["," typ]> ")"
--          "[" typ "]"
-- dat  ::= conid [id] <"=" con ["|" con]>
-- con  ::= conid [typ]

module InputAST where

import Misc
import Char

type Id   = String
type Guard      = Exp
type GuardedExp = (Guard, Exp)
type WhereDecs  = [Dec]


data Prog = Prog [Dec]	
data Dec  = DatDec Dat | RuleDec Rule | OverDec Sig Def
	  | OverDec0 Id | ValDec [Def]  | AValDec Sig [Def]
	  | ImportDec Id

data Dat  = Dat Id [Id] [Con]	    
data Con  = Con Id [Typ]	    

data Def  = Def Id [Pat] RHS [Dec]

data RHS  = RHS1 Exp
	  | RHSG [GuardedExp]

data Exp  = VarExp Id
          | ConExp Id
	  | LitExp Lit
	  | AbsExp [Pat] Exp
	  | AppExp Exp Exp
	  | IfxAppExp Exp Exp Exp 
	  | LetExp [Dec] Exp
	  | CaseExp Exp [CaseAlt]
	  | IfExp Exp Exp Exp
	  | AnnExp Exp TSchm -- Typ
	  | TupExp [Exp]
	  | ListExp [Exp]
	  | ListCompExp Exp [Stmt]
	  | SecExpLeft  Exp Exp
	  | SecExpRight Exp Exp
	  | EnumExpFrom Exp
	  | EnumExpFromTo Exp Exp
	  | EnumExpFromThen Exp Exp
	  | EnumExpFromThenTo Exp Exp Exp

data Stmt = GenStmt  Pat Exp
	  | QualStmt Exp
	  | LetStmt  [Dec]	

data CaseAlt = CaseAlt Pat Exp WhereDecs
	     | CaseAltGuard Pat [GuardedExp] WhereDecs


data Lit  = IntLit  Integer
	  | FloatLit Double
	  | StrLit  String
	  | CharLit Char

data Pat  = JokPat
          | VarPat Id
	  | LitPat Lit
	  | ConPat Id [Pat]
	  | TupPat [Pat]
	  | ListPat [Pat]	    

data TSchm = TSchm Ctxt Typ

data Sig  = Sig Id Ctxt Typ	   
data Typ  = VarTyp Id
          | ConTyp Id
	  | AppTyp Typ Typ
	  | TupTyp [Typ]
	  | ListTyp Typ		  

type Ctxt = [Pred]
type Cnst = [Prim]
data Rule = PropRule Ctxt Cnst	    
	  | SimpRule Ctxt Cnst	    deriving (Show)
data Pred = Pred Id Typ		    
data Prim = PredPrim Pred
          | EqPrim Typ Typ	    deriving (Show)

--------------------------------------------------------------------------------
-- methods below are defined purely for the purpose of detecting multiple
-- conflicting uses of a single id (eg. in the same pattern)

instance Eq Dec where
    (==) (DatDec x) (DatDec y) = x == y
    (==) (ValDec x) (ValDec y) = x == y
    (==) _          _          = False

instance Eq Def where
    (==) (Def id _ _ _) (Def id' _ _ _) = id == id'

instance Eq Dat where
    (==) (Dat id _ _) (Dat id' _ _) = id == id'

instance Eq Con where
    (==) (Con id _) (Con id' _) = id == id'

instance Eq Pat where
    (==) (VarPat id) (VarPat id') = id == id'
    (==) _	     _		  = False

{-
instance Show Pat where
    show (VarPat id) = id
    show _           = ""

instance Show Con where
    show (Con id _) = id
-}

--------------------------------------------------------------------------------
-- show returns a program suitable for Haskell 

-- show for lists; one item per line 
showl xs = unlines (map show xs)

-- show for lists; wraps items in parens, separated by a space
shows_ xs = unwords (map showp xs)
showsc xs = unwordsDiv "," (map showp xs)
showsc' xs = unwordsDiv "," (map show xs)
showsb xs = unwordsDiv "|"  (map show xs)
showssc xs = unwordsDiv ";" (map show xs)

showp x = "(" ++ show x ++ ")" 
showb x = "{" ++ show x ++ "}"

unwordsDiv d [] = ""
unwordsDiv d [x] = x
unwordsDiv d (x:xs) = x ++ d ++ unwordsDiv d xs


instance Show Prog where
    show (Prog ds) = showl ds

instance Show Dec where
    show (DatDec d) = show d
    show (AValDec s ds) = show s ++ "\n" ++ showl ds
    show (ValDec ds) = showssc ds -- showl ds
    show (ImportDec id) = "import " ++ id ++ "\n"
    show _  = "{- MISSING SHOW INSTANCE FOR DEC -}"

instance Show Dat where
    show (Dat i is cs) = "data " ++ i ++ " " ++ unwords is ++ 
			 if null cs 
			 then ""
			 else " = " ++ showsb cs ++ " deriving Show" 

instance Show Con where
    show (Con i ts) = i ++ " " ++ shows_ ts

instance Show Def where
    show (Def i ps rhs w) = infixParens i ++ " " ++ shows_ ps ++ show rhs ++
			    if null w then ""
				      else " where {" ++ showssc w ++ "}"

instance Show RHS where
    show (RHS1 e) = " = " ++ show e
    show (RHSG []) = ""
    show (RHSG ((g,e):ges)) = "| " ++ show g ++ " = " ++ show e ++ 
			      show (RHSG ges)

instance Show Pat where
    show (JokPat) = "_"
    show (VarPat i) = i
    show (LitPat l) = show l
    show (ConPat i ps)
    	| isInfixId i = "(" ++ i ++ ") " ++ shows_ ps
	| otherwise   = i ++ " " ++ shows_ ps
    show (TupPat ps) = "(" ++ showsc ps ++ ")"
    show (ListPat ps) = "[" ++ showsc ps ++ "]"

instance Show Exp where
    show (VarExp i) = i 
    show (ConExp i) = i
    show (LitExp l) = show l
    show (AbsExp ps e) = "\\ " ++ shows_ ps ++ " -> " ++ showp e
    show (AppExp e f)  = showp e ++ " " ++ showp f
    show (IfxAppExp e f g) = "(" ++ show f ++ ") " ++ showp e ++ " " ++ showp g

    show (LetExp ds e)  = "let {" ++ showssc ds ++ "} in  " ++ showp e
    show (CaseExp e as) = "case " ++ show e ++ " of {" ++ showssc as ++ "}"
    show (IfExp e f g)  = "if " ++ showp e ++ " then " ++ showp f ++
					     " else " ++ showp g
    show (AnnExp e t) = showp e ++ " :: " ++ showp t
    show (TupExp es)  = "(" ++ showsc es ++ ")"
    show (ListExp es) = "[" ++ showsc es ++ "]"

    show (ListCompExp e ss)  = "[" ++ show e ++ "|" ++ showsc' ss ++ "]"

    show (SecExpLeft  e f) = "(" ++ showp e ++ showp f ++ ")"
    show (SecExpRight e f) = "(" ++ showp e ++ showp f ++ ")"
    
    show (EnumExpFrom e) = "[" ++ showp e ++ "..]"
    show (EnumExpFromTo e f) = "[" ++ showp e ++ ".." ++ showp f ++ "]"
    show (EnumExpFromThen e f) = "[" ++ showp e ++ "," ++ showp f ++ "..]"
    show (EnumExpFromThenTo e f g) = "[" ++ showp e ++ "," ++ showp f ++ 
				       ".." ++ showp g ++ "]"

instance Show Stmt where
    show (GenStmt p e) = show p ++ "<-" ++ show e
    show (QualStmt e)  = show e
    show (LetStmt ds)  = "let {" ++ showssc ds ++ "}"

instance Show CaseAlt where
    show (CaseAlt p e w) = show p ++ "->" ++ show e ++ 
			   if null w then "" 
			   else " where {" ++ showsc w ++ "}"
    show (CaseAltGuard p ges w) = "guarded case alt"


instance Show Lit where
    show (IntLit  i) = show i
    show (FloatLit f)= show f
    show (StrLit  s) = show s
    show (CharLit c) = show c

instance Show TSchm where
    show (TSchm c t) = (if null c then ""
				  else "(" ++ showsc c ++ ") => ") ++ show t

instance Show Typ where
    show (VarTyp i) = i
    show (ConTyp i) = i
    show (AppTyp t s) = showp t ++ " " ++ showp s
    show (TupTyp ts) = "(" ++ showsc ts ++ ")"
    show (ListTyp t) = "[" ++ showp t ++ "]"

instance Show Sig where
    show (Sig i c t) = infixParens i ++ " :: " ++ 
		       (if null c then "" else "(" ++ showsc' c ++ ") => ") ++ 
		       show t

instance Show Pred where
    show (Pred i t) = i ++ " " ++ showp t
