--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------

-- lexer for Chameleon

module Lexer (lexer, Token(..), idStr, srcPos)
where

import ParserMisc
import Char
import Misc
import Tokens
import Maybe



{-
test  s = runP s (lexAll s)
test2 s = runP s (do ts <- lexAll s
		     return (addLayoutTokens ts))

test3 s = runP s (do ts1 <- lexAll s
		     let ts2 = addLayoutTokens ts1
			 ts3 = removeLayout ts2
		     return ts3)


-}
{-
lexer :: (Token -> P a) -> P a
lexer cont = do input <- getInput
		token <- lexer' input
		let token' = trace ("token: " ++ show token) token
		cont token
-}

lexer :: (Token -> P a) -> P a
lexer cont = do done <- tokensAvailable
		if done then do token <- nextToken
				cont token
			else -- do the actual work
			     do input <- getInput
			        ts <- lexAll input
				lay <- ckHandleLayout
				let ts' = if lay then transformLayout ts
						 else ts
--				    ts'' = trace ("stream: " ++ show ts') ts'
				setStream ts'
				
				tokensDone
				lexer cont

-- turn a stream of characters into a stream of tokens
lexAll :: String -> P [TokenInfo]
lexAll input = do setInput input
		  lex
    where lex = do input <- getInput
		   token <- lexer' input
		   row	 <- getTokRow
		   col   <- getTokCol
		   updateTokPos
		   let tinfo = (token, row, col)
		   if token == Teof then return [tinfo]
				    else do tokens <- lex
					    return (tinfo:tokens)

lexer' :: String -> P Token
lexer' "" = returnPI 0 Teof ""
lexer' ('\'':c:'\'':xs) = returnPI 3 (TChar c) xs
lexer' ('"':xs) = do (str, rst) <- getStrLit xs
		     addColNum (length str + 1)
		     returnPI 1 (TStr str) rst

lexer' ('-':'-':xs) = do xs' <- dropLine xs
			 lexer' xs'
			 
lexer' ('{':'-':xs) = do dropBlock 1 xs
			 addColNum 2
	                 s <- getInput
	                 lexer' s

lexer' xxs@(x:xs)
    | isSpace x          = do xs' <- skipSpace xxs
			      setInput xs'
			      lexer' xs'

    | x == '`' = do pos <- getPos 
		    m <- takeInfixNonSymbol xxs
		    let (str, rst) = fromJust m
		    if isNothing m 
		     then fail ("Badly formed operator at " ++ show pos ++
				" - no closing '")
		     else returnPI 0 (TIId (str, pos)) rst

    | isPunct x || isSymbol x = 
		  do pos <- getPos
		     st  <- getState	-- kinda hacky
		     (pstr, prst) <- takeToken isPunct xxs
		     (sstr, srst) <- takeToken isSymbol xxs
		     setState st
		     let m = prefixKeySymbol pstr
			 (et,(s,r)) = head m
		     if null m || length sstr > length s
		      then do takeToken isSymbol xxs
			      returnPI 0 (TIId (sstr, pos)) srst
		      else do takeToken isPunct xxs
			      let t = either id ($pos) et
			      returnPI (-(length r)) t (r ++ prst)
		     
    | isAlphaChar x = 
		  do pos <- getPos
		     (str, rst) <- takeToken isIdChar xxs
		     let tid    x = TId (x, pos)
			 tconid x = TConId (x, pos)
		     let con = if isLowerChar x then tid else tconid
		     case lookup str keywords of
			Nothing  -> returnPI 0 (con str) rst
			Just tok -> returnPI 0 tok rst
    
    -- a numeric literal - int or float
    | isDigit x	= do (t, rst) <- getNumLit xxs
		     returnPI 0 t rst


    | otherwise		= fail $ "unrecognised token `" ++ 
			    (takeWhile (/='\n') xxs) ++ "'"


-- standard symbols (NOTE: arranged by length of string)
keySymbols = [	("<==>", Left TSimpArrow),("==>", Left TPropArrow),
		(":::", Right TEAnn),	    ("::?", Left TQuery),
		("=>", Left TImpl),	    ("<-", Left TLArrow),
		("->", Left TArrow),	    ("..", Left TCont),
		("::", Right TAnn),
		("=", Left TEquals),	    ("\\", Left TBackslash), 
		("(", Left TLParen),	    (")", Left TRParen),
		("[", Left TLBracket),	    ("]", Left TRBracket),
		{-("_", Left TUnderscore),-}(",", Left TComma),
		("|", Left TBar),	    ("{", Left TLBrace),
		("}", Left TRBrace),	    (";", Left TSemicolon) ]


-- map from keyword strings to tokens
keywords = [("rule", TRule), ("overload", TOverload), ("data", TData), 
	    ("instance", TInstance), ("hconstraint", THConstraint),
	    ("if", TIf), ("then", TThen), ("else", TElse), 
	    ("let", TLet), ("in", TIn), ("where", TWhere),
	    ("case", TCase), ("of", TOf),
	    ("extern", TExtern), ("interface", TInterface),
	    ("True", TTrue), ("False", TFalse)]

-- keywords that introduce an indented scope
indentTokens = [TLet, TWhere, TOf]

-- if some prefix of the given string is a key symbol, return the token it
-- represents, and a pair containing the matched prefix and the rest of the 
-- string (for all key symbols, in order of length of symbol - use head)
-- note that the "token" may need to be applied to a source position!
prefixKeySymbol :: String -> [(Either Token (SrcPos -> Token),(String, String))]
prefixKeySymbol str = [ (t,(s,r)) | (s,t) <- keySymbols, 
				    let m = s `pre` str,	
				    isJust m, 
				    let r = fromJust m ]
    where   pre []     ys = Just ys
	    pre (x:xs) (y:ys) | x == y	= pre xs ys
	    pre _ _ = Nothing


-- skips over whitespace, incrementing position counters
skipSpace :: String -> P String
skipSpace "" = updateTokPos >> return ""
skipSpace (' ':xs)  = incColNum >> skipSpace xs
skipSpace ('\n':xs) = nextLinePos >> skipSpace xs
skipSpace ('\f':xs) = nextLinePos >> skipSpace xs
skipSpace ('\r':xs) = nextLinePos >> skipSpace xs
skipSpace ('\t':xs) = tabColNum >> skipSpace xs
skipSpace xs = updateTokPos >> return xs
    

nextLinePos = resetColNum >> incLineNum >> updateTokPos

-- returns prefix of string which satisfies the pred, as well as the rest
-- NOTE: do not use this to skip over whitespace! (it won't account for 
--	 newlines properly)
takeToken :: (Char -> Bool) -> String -> P (String, String)
takeToken _ ""    = return ("", "")
takeToken p xxs@(x:xs) = if p x then do (ts, rs) <- takeToken p xs
					incColNum
					return (x:ts, rs)
				else return ("", xxs)

-- takes a variable which has been applied infix (eg. `mod`)
takeInfixNonSymbol :: String -> P (Maybe (String, String))
takeInfixNonSymbol ('`':xs) = do 
    (str, rst) <- takeToken isIdChar xs
    case rst of 
	('`':rst') -> addColNum 2 >> return (Just (str, rst'))
	_	   -> return Nothing
takeInfixNonSymbol _ = return Nothing

isIdChar x = isAlphaNum x || x == '\'' || x == '_'
isLowerChar c = Char.isLower c || c == '_'
isAlphaChar c = Char.isAlpha c || c == '_'

-- True if the given character can appear within an operator name
isSymbol :: Char -> Bool
isSymbol c = c `elem` cs    where cs = ":!#$%&*+./<=>?@\\^|-~"

-- True if the given char appears within a key symbol (punctuation)
isPunct :: Char -> Bool
isPunct c = c `elem` cs	    where cs = "?(){}:;,.=<->|[]*\\"

-- assumes the given string begins a string literal
-- returns the literal, and the following list of chars
getStrLit :: String -> P (String, String)
getStrLit ('\\':'"':cs) = getStrLit cs >>= \(s, r) -> return ('\\':'"':s, r)
getStrLit ('"':cs)	= return ("", cs)
getStrLit (c:cs)	= getStrLit cs >>= \(s, r) -> return (c:s, r)
getStrLit ""		= fail "non-terminated string literal" 


-- reads the next number - int or float
-- also returns the remainder of the string
getNumLit :: String -> P (Token, String)
getNumLit ('0':b:cs) 
    | b == 'x' || b == 'X' = do let (num, rst) = span isHexDigit cs
				    hex = '0':b:num
			        addColNum (length hex)
				return (TInt (read hex, hex), rst)

    | b == 'o' || b == 'O' = do let (num, rst) = span isOctDigit cs
				    oct = '0':b:num
			        addColNum (length oct)
				return (TInt (read oct, oct), rst)
getNumLit cs = 
    case getDecimal cs of
	 (dec1, rst@('.':cs0)) -> -- float
		case getDecimal cs0 of
		 ([], cs0) -> -- no number after the decimal point, it's an int
			      do addColNum (length dec1)
				 return (TInt (read dec1, dec1), rst)
			     
		 (dec2, e:cs1)  
		    | e == 'e' || e == 'E' -> case cs1 of
			(s:cs2) | s == '+' || s == '-' -> do
			      let (dec3, cs3) = getDecimal cs2
				  f = dec1 ++ "." ++ dec2 ++ [e] ++ [s] ++ dec3
			      addColNum (length f)
			      return (TFloat (read f, f), cs3)

			cs2 -> do let (dec3, cs3) = getDecimal cs2
				      f = dec1 ++ "." ++ dec2 ++ [e] ++ dec3
				  addColNum (length f)
				  return (TFloat (read f, f), cs3)

		 (dec2, cs1) -> do let f = dec1 ++ "." ++ dec2
				   addColNum (length f)
				   return (TFloat (read f, f), cs1)
		      
	 (dec, cs0) -> -- decimal
		       do addColNum (length dec)
			  return (TInt (read dec, dec), cs0)
	 
    where getDecimal = span isDigit 



-- throws away the rest of the line (including the newline character)
dropLine :: String -> P String
dropLine "" = return ""
dropLine ('\n':xs) = nextLinePos >> return xs
dropLine (x:xs)	   = incColNum >> dropLine xs

-- discards characters from the given string until we leave the outer-most
-- comment block
dropBlock :: Int -> [Char] -> P ()
dropBlock 1 ('-':'}':xs) = addColNum 2 >> setInput xs
dropBlock n ('-':'}':xs) = addColNum 2 >> dropBlock (n-1) xs
dropBlock n ('{':'-':xs) = addColNum 2 >> dropBlock (n+1) xs
dropBlock n ('\n':xs)    = do incLineNum
			      resetColNum
			      dropBlock n  xs
dropBlock n (_:xs)       = incColNum >> dropBlock n xs
dropBlock _ []		 = fail "comment block not closed"


-- Like return, but update the pending input stream as well.
-- Also increments the column number by the given ammount
returnPI :: Int -> Token -> String -> P Token
returnPI n t s = do addColNum n
		    setInput s
		    return t


--------------------------------------------------------------------------------
-- deal with layout

transformLayout :: [TokenInfo] -> [TokenInfo]
transformLayout = removeLayout . addLayoutTokens


-- transforms a Chameleon program into an equivalent layout-insensitive version
removeLayout :: [TokenInfo] -> [TokenInfo]
removeLayout ts = lay (init ts) [] ++ [teof]

    where
    
    lay tts@((TLayoutFirst n,r,c):ts) mms@(m:ms) 
	| m == n    = sc r c : lay ts mms
	| n <  m    = rb r c : lay tts ms

    lay tts@((TLayoutFirst n,r,c):ts) ms = lay ts ms


    lay ((TLayoutBlock n,r,c):ts) mms@(m:ms) 
	| n > m	    = lb r c : lay ts (n:mms)
	| otherwise = lb r c : rb r c : lay ((TLayoutFirst n,r,c):ts) ms

    lay ((TLayoutBlock n,r,c):ts) [] = lb r c : lay ts [n]


    lay ((TRBrace,r,c):ts) (0:ms) = rb r c : lay ts ms

    lay ((TRBrace,r,c):ts) ms = exit "removeLayout: parse error 1"

    lay ((TLBrace,r,c):ts) ms = lb r c : lay ts (0:ms)

    -- missing the case where we insert '}' in the event of a parse error

    lay [] [] = []
  
    lay [] (m:ms) 
	| m /= 0    = rb' m : lay [] ms
	| otherwise = exit "Parse error: outer-most declaration block not \
			     \closed (try adding a '}' to the end)"

    lay (t:ts) ms = t : lay ts ms


    rb' m = (TRBrace, -1, m)
    sc r c = (TSemicolon, r, c)
    lb r c = (TLBrace, r, c)
    rb r c = (TRBrace, r, c)


-- adds layout tokens to the token stream
addLayoutTokens :: [TokenInfo] -> [TokenInfo]
addLayoutTokens ts = let ts' = init (first ts)
		     in  add ts' ++ [teof]
    where

    -- if the program doesn't start with a '{', note the indentation of the
    -- first token
    first tts@((t,r,c):_) | t /= TLBrace = (TLayoutBlock c, r, c) : tts
			  | otherwise    = tts
    
    add [x@(t,r,c)]
	-- token introduces a new block, but is not followed by anything
	| t `elem` indentTokens = x : [(TLayoutBlock 0, r, c)]
	| otherwise		= [x]

    add (x1@(t1,r1,c1):x2@(t2,r2,c2):ts)
	-- token t1 introduces a new block
	| t1 `elem` indentTokens && t2 /= TLBrace = 
			x1 : (TLayoutBlock c2,r2,c2) : x2 : add ts
	-- check if t2 is first on its line
	| r2 > r1   = x1 : (TLayoutFirst c2, r2, c2) : add (x2:ts)
	| otherwise = x1 : add (x2:ts)
    
