--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- module for miscellaneous functions which don't fit elsewhere (but probably
-- should)... none of these functions should require any other module to be
-- imported

module Misc where

import Char
import System
import IOExts
import List


--------------------------------------------------------------------------------
type SrcPos = (String, Int, Int)    -- module name, row and column of source

noSrcPos :: SrcPos
noSrcPos = ("",-1,-1)

formatSrcPos (mod, x, y) = let pos = show (x,y)
			   in  if null mod then pos ++ ":"
					   else mod ++ ":" ++ pos ++ ":"

formatSrcPos' (mod, x, y) = let pos = show (x,y)
			    in  if null mod then pos ++ ":"
					    else mod ++ ":" ++ pos

-- just the line number
formatSrcPosLine (mod, x, _) = let pos = show x
			       in  if null mod then pos
					       else mod ++ ":" ++ pos ++ ":"
--------------------------------------------------------------------------------

-- functions for testing lists as sets

superset, subset, sameSet :: Eq a => [a] -> [a] -> Bool
subset   xs ys = all (`elem` ys) xs
sameset  xs ys = sameSet xs ys
sameSet  xs ys = subset xs ys && subset ys xs
superset xs ys = all (`elem`xs) ys


powerset :: [a] -> [[a]]
powerset []     = [[]]
powerset (x:xs) = let ps = powerset xs
		  in  ps ++ [ x : ys | ys <- ps ]

without :: Eq a => [a] -> [a] -> [a]
without xs ys = [ x | x <- xs, x `notElem` ys ]

-- true if first list is longer than second
longer :: [a] -> [a] -> Bool
longer (x:xs) (y:ys) = longer xs ys
longer []     _	     = False
longer _      []     = True

allEqual :: Eq a => [a] -> Bool
allEqual [] = True
allEqual (x:xs) = all (x==) xs

elimSupersets :: Eq a => [[a]] -> [[a]]
elimSupersets xss = 
    [ ys | (ys, yss) <- allLeft xss, not $ any (ys`superset`) yss ]


-- first list appears in second
isSublist :: Eq a => [a] -> [a] -> Bool
isSublist xs ys = any (uncurry eq) (zip (repeat xs) (tls ys))
    where 
	tls []	       = []
	tls xxs@(_:xs) = xxs : tls xs
    
	eq []	    _  = True
	eq _	    [] = False
	eq (x:xs) (y:ys) | x == y    = eq xs ys
			 | otherwise = False


fst3 (a,_,_) = a
snd3 (_,b,_) = b
thd3 (_,_,c) = c

lookupBy :: (a -> b -> Bool) -> a -> [(b,c)] -> Maybe c
lookupBy eq a [] = Nothing
lookupBy eq a ((b,c):bcs) | a `eq` b  = Just c
			  | otherwise = lookupBy eq a bcs


-- partitions xs into elements which satisfy p and those which don't. Also,
-- applies g to x in either case
partitionF :: (a -> Bool) -> (a -> a) -> [a] -> ([a], [a])
partitionF p g [] = ([], [])
partitionF p g (x:xs) = let (t, f) = partitionF p g xs
			    x' = g x
		        in  if p x then (x':t, f)
				   else (t, x':f)

-- pads out a string to length n by adding spaces to its end
pad n str = let len = length str
	    in  if len < n
		   then str ++ replicate (n - len) ' '
		   else str

splitEither :: [Either a b] -> ([a], [b])
splitEither [] = ([], [])
splitEither ((Left l):es)  = let (ls, rs) = splitEither es in (l:ls, rs)
splitEither ((Right r):es) = let (ls, rs) = splitEither es in (ls, r:rs)
    
singleton [x] = True
singleton _   = False

notNull [] = False
notNull _  = True

conId "" = ""
conId (c:cs) = toUpper c : cs

varId "" = ""
varId (c:cs) = toLower c : cs

showlist :: Show a => [a] -> String
showlist xs = unlines (map show xs)

predIdStr cs = not (null cs) && isUpper (head cs)


permute [] = [[]] 
permute (x:xs) = [ take i ys ++ [x] ++ drop i ys 
                       | ys <- permute xs, i <- [0..length ys] ] 


-- given a list, returns a list of pairs where for all elements in the list,
-- each pair contains a list element along with all the *other* elements
allLeft xs = allLeft' xs []
    where allLeft' []     _  = []
          allLeft' (x:xs) as = (x, as ++ xs) : allLeft' xs (x:as)
	      
-- returns a list which contains elements which appear in all of the input
-- lists
commonElems :: Eq a => [[a]] -> [a]
commonElems []	     = []
commonElems [xs]     = xs
commonElems (xs:xss) = foldl intersect xs xss
    where intersect []	   _  = []
	  intersect (x:xs) ys
	    | x `elem` ys   = x : intersect xs ys
	    | otherwise	    = intersect xs ys

heady msg [] = error $ "lack of heady goodness: " ++ msg
heady _   xs = head xs

filler str = error ("don't look at the: " ++ str)

-- finds elements of list which occur more than once
multiples :: Eq a => [a] -> [a]
multiples xs = nub $ mults xs
    where mults [] = []
	  mults (x:xs) | x `elem` xs = x : mults xs
		       | otherwise   = mults xs

-- returns string suffix following the last '/'
basename :: String -> String
basename str = reverse (takeWhile (/='/') (reverse str)) 

-- change nth list element (from 1)
changeN :: Int -> [a] -> a -> [a]
changeN 1 (x:xs) y = y : xs
changeN n (x:xs) y = x : changeN (n-1) xs y
changeN _ [] _ = error "changeN"

-- returns nth element of list (from 1, unlike !!)
nthElem :: Int -> [a] -> a
nthElem 1 (x:xs) = x
nthElem n (x:xs) = nthElem (n-1) xs
nthElem _ []     = error "nthElem"

exit :: String -> a
exit str = unsafePerformIO $ do let str' = unlines $ indent $ lines str
			        putStrLn str'
			        exitFailure 
    where indent :: [String] -> [String]
	  indent xs = let s  = hd xs
		          ss = tl xs
		      in  (prefix ++ s) : map (space++) ss
	  prefix = "ERROR: "
	  space  = replicate (length prefix) ' '

	  tl [] = []
	  tl xs = tail xs
	  hd [] = error "CALL exit WITH A NON-EMPTY STRING, FOOLISH PROGRAMMER!"
	  hd xs = head xs

-- indent second string, and place first, as a title, to the left of its first
-- line
titleLeft :: String -> String -> String
titleLeft hdg str = let ls = lines str
			w  = length hdg
			ls'= indentRest w ls
		    in  unlines ls'
			
    where indentRest w [] = []
	  indentRest w (l:ls) = (hdg ++ l) : map (indent w) ls

	  indent w l = replicate w ' ' ++ l

--------------------------------------------------------------------------------
-- error monad (just like Maybe, but can carry around error information)

data E a = Succ a | Err String deriving Show

instance Monad E where
  m >>= k = case m of
            Succ a -> k a
            Err s -> Err s
  return x = Succ x
  fail s = Err s

isSucc, isErr :: E a -> Bool
isSucc (Succ _) = True
isSucc _        = False
isErr (Err _) = True
isErr _       = False

unSucc :: E a -> a
unSucc (Succ s) = s
unSucc _	= error "unSucc: not a Succ"

unErr :: E a -> String
unErr  (Err e)  = e
unErr  _        = error "unErr: not an Err"

unSucc' s (Succ x) = x
unSucc' s _ = error ("not a succ!: " ++ s)


success = Succ (error "success!")

--------------------------------------------------------------------------------

-- identity monad 

data Identity a = Identity { runIdentity :: a }

instance Monad Identity where
    return v = Identity v
    (Identity a) >>= f = f a

--------------------------------------------------------------------------------
isInfixId ""	= False
isInfixId (c:_) = not ((isAlphaNum c) || (isSpace c) || c == '(' || c == ')')

infixParens id  = if isInfixId id
                  then ("(" ++ id ++ ")")
		  else id

--------------------------------------------------------------------------------

-- solver options
data SOpt = SInference | SEvidence | SDebug 	deriving (Show, Eq, Enum)


--------------------------------------------------------------------------------

fj str (Just x) = x
fj str Nothing  = error ("fromJust: " ++ str)
