--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------

-- contains functions for transforming expressions to reflect precedence and
-- associativity rules

module Operators where

import Misc (runIdentity)
import Ann_InputAST


data Assoc = LA | RA | NA   deriving Eq -- right/left/not associative
type Prec = Int
type OpInfo = (Id, PAInfo)
type PAInfo = (Prec, Assoc)

--------------------------------------------------------------------------------
-- given a program, fixes operator precedence
fixOperators :: Prog -> Prog
fixOperators p = -- runIdentity $ 
		    appExpProg (--return . 
				 removeParens . fixLeft . fixRight) p

--------------------------------------------------------------------------------

removeParens :: Ann Exp -> Ann Exp
removeParens ae = Ann (ann ae) $ case item ae of
	PIfxAppExp e1 e2 e3 -> IfxAppExp e1 e2 e3
	e -> e


-- first fix the children, then fix this expression
fixLeft :: Ann Exp -> Ann Exp
fixLeft ae@(Ann ix e@(IfxAppExp ae1' ae2 ae3')) =
    case (fixLeft ae1', fixLeft ae3') of
	(ae1, ae3) ->
	    let a1 = ann ae1
		e1 = item ae1
		pe = prec e
		pe1 = prec e1
		id = ifxId e
	    in if pe > pe1 || pe == pe1 && idRA id
		then let left  = ifxLeft e1
			 mid   = ifxMid e1
			 right = Ann ix (IfxAppExp (ifxRight e1) ae2 ae3)
		     in  Ann a1 (IfxAppExp left mid right)
		else ae
fixLeft ae = ae

fixRight :: Ann Exp -> Ann Exp
fixRight ae@(Ann ix e@(IfxAppExp ae1' ae2 ae3')) =
    case (fixRight ae1', fixRight ae3') of
	(ae1, ae3) -> 
	    let a3 = ann ae3
		e3 = item ae3
		pe = prec e
		pe3 = prec e3
		id = ifxId e
	    in if pe > pe3 || pe == pe3 && idLA id
		then let left  = Ann ix (IfxAppExp ae1 ae2 (ifxLeft e3))
			 mid   = ifxMid e3
			 right = ifxRight e3
		     in  Ann a3 (IfxAppExp left mid right)
		else ae
fixRight ae = ae


--------------------------------------------------------------------------------
prec (IfxAppExp _ (Ann _ (VarExp id)) _)  = fst (findOpInfo standardOps id)
prec _ = 11

assoc (IfxAppExp _ (Ann _ (VarExp id)) _) = snd (findOpInfo standardOps id)
assoc _	= NA

idLA id = snd (findOpInfo standardOps id) == LA
idRA id = snd (findOpInfo standardOps id) == RA
idNA id = snd (findOpInfo standardOps id) == NA

ifxId	 (IfxAppExp _ (Ann _ (VarExp id)) _) = id
ifxId    _ = "NOT AN OPERATOR APPLICATION"
ifxLeft  (IfxAppExp e1 _ _) = e1
ifxMid   (IfxAppExp _ e2 _) = e2
ifxRight (IfxAppExp _ _ e3) = e3

isIfx (IfxAppExp _ _ _) = True
isIfx _ = False

lowestPA, highestPA :: PAInfo
lowestPA  = (0, NA)
highestPA = (11, NA)

-- given a map of known operator information, find the precedence and
-- associativity of the named operator
-- if operator is not listed, use default (highest precedence, left associative)
findOpInfo :: [OpInfo] -> Id -> PAInfo
findOpInfo is id = find is
    where find [] = (10, LA)
	  find (i@(s,pa):is) | id == s	 = pa
			     | otherwise = find is 

--------------------------------------------------------------------------------
standardOps = 
    [ (".", (9, RA)), ("!!", (9, LA)),
      ("^", (8, RA)), ("^^", (8, RA)), ("**", (8, RA)),
      ("*", (7, LA)), ("/", (7, LA)), ("quot", (7, LA)), ("rem", (7, LA)), 
	("div", (7, LA)), ("mod", (7, LA)),
      ("+", (6, LA)), ("-", (6, LA)),
      (":", (5, RA)), ("++", (5, RA)),
      ("==", (4, NA)), ("/=", (4, NA)), ("<", (4, NA)), ("<=", (4, NA)), 
	(">=", (4, NA)), (">", (4, NA)), ("elem", (4, NA)), 
	("notElem", (4, NA)),
      ("&&", (3, RA)),
      ("||", (2, RA)),
      (">>", (1, LA)), (">>=", (1, LA)),
      ("=<<", (1, RA)),
      ("$", (0, RA)), ("$!", (0, RA)), ("seq", (0, RA)) ]
      
      
--------------------------------------------------------------------------------

{- TESTs
 
e = plus (gt (n vb vy) vx) vz


plus x y = nullAnn (IfxAppExp x (nullAnn (VarExp "+")) y)
gt   x y = nullAnn (IfxAppExp x (nullAnn (VarExp ">")) y)
n    x y = nullAnn (IfxAppExp x (nullAnn (VarExp "&&")) y)

pplus x y = nullAnn (IfxAppExp x (nullAnn (VarExp "+")) y)
pgt   x y = nullAnn (IfxAppExp x (nullAnn (VarExp ">")) y)
pand  x y = nullAnn (IfxAppExp x (nullAnn (VarExp "&&")) y)

vx = nullAnn (VarExp "x")
vy = nullAnn (VarExp "y")
vz = nullAnn (VarExp "z")
vb = nullAnn (VarExp "b")

iterateUntilFixed f x = fixed (iterate f x)
    where fixed (x:y:xs) | x == y    = x
			 | otherwise = fixed (y:xs)

-}
