--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
----
-- Parser for Chameleon
-- (now returns annotated AST... as per Ann_InputAST)
--
----
-- Notes:
--  - infix function identifiers cannot contain '(', ')', '[', or ']'
--  
--  - to do:
--	- shouldn't require parens around a context if it contains only a
--	  single constraint. eg should be able to write: C a => a
--					     instead of: (C a) => a
----
-------------------------------------------------------------------
{

module Parser (parse, parseInit, parseTyp, parseExp, parseTSchm)
where

-------------------------------------------------------------------
-- Dependencies.
-------------------------------------------------------------------
import Lexer
import ParserMisc
import Ann_InputAST
import Misc (noSrcPos, SrcPos)

}

-------------------------------------------------------------------
-- Happy parser dec's.
-------------------------------------------------------------------
%name parserTyp Typ
%name parserExp Exp
%name parserSigTyp SigTyp
%name parserNestedRef NRef
%name parser

%monad {P}
%lexer {lexer} {Teof}
%tokentype { Token }
%token
    id          { TId $$ }
    conid	{ TConId $$ }
    iid		{ TIId $$ }
    rule        { TRule }
    overload    { TOverload }
    instance	{ TInstance }
    data	{ TData }
    hconstraint	{ THConstraint }
    ';'         { TSemicolon }
    '{'		{ TLBrace }
    '}'		{ TRBrace }
    '='         { TEquals }
    '\\'        { TBackslash }
    "->"        { TArrow }
    if          { TIf }
    then        { TThen }
    else        { TElse }
    let		{ TLet }
    in		{ TIn }
    case	{ TCase }
    of		{ TOf }
    where	{ TWhere }
    extern	{ TExtern }
    interface	{ TInterface }
    '('         { TLParen }
    ')'         { TRParen }
    '['		{ TLBracket }
    ']'		{ TRBracket }
    ".."	{ TCont }
    '_'         { TUnderscore }
    "=>"	{ TImpl }
    "==>"       { TPropArrow }
    "<==>"	{ TSimpArrow }
    ','         { TComma }
    '|'		{ TBar }
    "<-"	{ TLArrow }
    "::"        { TAnn $$ }
    ":::"	{ TEAnn $$ }
    "True"	{ TTrue }
    "False"	{ TFalse }
    "::?"	{ TQuery }
    '*'		{ TAsterisk }
    int		{ TInt $$ }
    float	{ TFloat $$ }
    char	{ TChar $$ }
    str		{ TStr $$ }

%% 

-------------------------------------------------------------------
-- Productions.
-------------------------------------------------------------------
Prog    :: { Prog }
	: Decs			    {(Prog (reverse $1))}

Decs	:: { [Ann Dec] }
	: {- empty -}		    {[]}
	| '{' SDecs DecsEnd	    {concat $2}

DecsEnd :: { () }  -- dummy (value unused)
	: '}'			    {()}
	| error			    {% dropContextEnd } 
{-
SDecs	:: { [[Ann Dec]] }
	: Dec			    {% do l <- newLabel;
					  if null $1 
					   then return [] 
					   else return [[Ann l (head $1)]]}
	| SDecs ';' Dec  	    {% do { l <- newLabel;
					    let { d = if null $3 
						    then []
						    else [Ann l (head $3)] };
								

					    return (d:$1)} }  -- intoDecs
-}
{-					    
SDecs	:: { [[Ann Dec]] }
	: AnnDec		    {% if null $1 
					  then return [] 
					  else return $1}
	| SDecs ';' AnnDec  	    {% do { let { d = if null $3 
							then []
							else $3 };
								
					    return (d:$1)} }  -- intoDecs
-}

SDecs  :: { [[Ann Dec]] }
       : AnnDec			    {[$1]}
       | SDecs ';' AnnDec	    {$3 : $1}

AnnDec	:: { [Ann Dec] }
	: Dec			    {% if null $1
					then return []
					else do { l <- newLabel;
						  return [Ann l (head $1)] } }
				    
	| Var "::?" ';' Def	    {% do{l1 <- newLabel;
					  l2 <- newLabel;
					  let {sc = SCTypeVar (idStr $1)
						     (ixLoc l2);};
					  addCommand sc;
					  return [Ann l2 (ValDec [Ann l1 $4])]}}

	| Var "::?" SigTyp ';' Def  {% do{l1 <- newLabel;
					  l2 <- newLabel;
					  let {(c,t) = $3;
					       ts = TSchm c t;
					       sc = SCExplainVar (idStr $1)
						     (ixLoc l2) ts};
					  addCommand sc;
					  return [Ann l2 (ValDec [Ann l1 $5])]}}

Dec     :: { [Dec] }
	: data Dat		    {% do l <- newLabel;
					  return [DatDec (Ann l $2)]}
	| rule Rule		    {% do l <- newLabel;
					  return [RuleDec (Ann l $2)]}
        | overload Sig where Defs   {% do l1 <- newLabel;
					  l2 <- newLabel;
					  return [OverDec (Ann l1 $2) $4]} 
	| instance InstSig where Defs 
				    {% do l1 <- newLabel;
					  l2 <- newLabel;
					  return [InstDec (Ann l1 $2) $4]}

	| instance InstSig	    {% do l1 <- newLabel;
					  let {dec =InstDec (Ann l1 $2) []};
					  dec' <- addUndefinedInstanceDef dec;
					  return [dec']}
					  
	| overload Var		    {% do l <- newLabel;
					  return [OverDec0 (idStr $2)]}
					  
        | Def			    {% do l <- newLabel;
					  return [ValDec [Ann l $1]]}

	| PatDec		    {[$1]}

	| ADef			    {[$1]}

	| HCons			    {[$1]}

	| Extern		    {[$1]}

	| IFace			    {[$1]}

	| {- empty -}		    {[]}

HCons	:: { Dec }
	: hconstraint conid	    {% return (HConDec [idStr $2]) }
	| hconstraint '(' HCIds	')' {% return (HConDec (reverse $3))}

HCIds	:: { [Id] }
	: conid			    {[idStr $1]}
	| HCIds ',' conid	    {idStr $3 : $1}

IFace	:: { Dec }
	: interface conid	    {% do l <- newLabelPos (srcPos $2);
					return (InterfaceDec(Ann l(idStr $2)))}

Extern	:: { Dec }
	: extern AnySigCon	    {% do l <- newLabel;
					  return (ExternDec (Left (Ann l $2)))}
	| extern conid		    {% do l <- newLabel;
					  return (ExternDec 
						    (Right (Ann l (idStr $2))))}

Defs	:: { [Ann Def] }
	: '{' DefBlk '}'	    {reverse $2}

DefBlk	:: { [Ann Def] }
	: Def			    {% do l <- newLabel;
					  return [Ann l $1]}
	| DefBlk ';' Def	    {% do l <- newLabel;
					  return ((Ann l $3) : $1)}


ADef	:: { Dec }
	: AnySig ';' Def	    {% do l1 <- newLabel;
					  l2 <- newLabel;
					  return (AValDec (Ann l1 $1)
							  [Ann l2 $3])} 

PatDec	:: { Dec }
	: BPat RHS Where	    {% do l1 <- newLabel;
					  return (PatDec (Ann l1 $1) $2 $3)}
	

Def     :: { Def }
	: id Pats RHS Where	    {% do let {p = srcPos $1};
					  l1 <- newLabelPos p;
				          return (Def (Ann l1 (idStr $1))
						      (reverse $2) $3 $4)}
	| '(' iid ')' Pats RHS Where {% do let {p = srcPos $2};
					   l1 <- newLabelPos p;
				           return (Def (Ann l1 (idStr $2)) 
						       (reverse $4) $5 $6)}
	| Pat iid Pat RHS Where	    {% do {let {p = srcPos $2};
					   l1 <- newLabelPos p;
					   l2 <- newLabel;
					   l3 <- newLabel;
					   return (Def (Ann l1 (idStr $2))
						    [Ann l2 $1, Ann l3 $3]
						    $4 $5)}}
						       

RHS	:: { RHS }
	: '=' Exp		    {RHS1 $2}
	| GdOpts		    {RHSG (reverse $1)}

GdOpts  :: { [GuardedExp] }
	: {- empty -}		    {[]}
	| GdOpts GdOpt		    {$2 : $1}

GdOpt	:: { GuardedExp }
	: '|' Exp '=' Exp	    {($2, $4)}
	
Where	:: { [Ann Dec] }
	: {- empty -}		    {[]}
	| where Decs		    {reverse $2}
	
Exp	:: { Ann Exp }
	: Exp1			    {$1}
	| Exp1 "::?"		    {% do {let {l  = ixLoc (ann $1)};
					   let {sc = SCTypeLoc l};
					   addCommand sc;
					   return $1} }
	| Exp1 "::?" SigTyp	    {% do {let {l  = ixLoc (ann $1)};
					   let {(c,t) = $3};
					   let {ts = TSchm c t};
					   let {sc = SCExplainLoc l ts};
					   addCommand sc;
					   return $1} }

	| Exp1 "::" SigTyp	    {% do{l1 <- newLabelPos $2;
					  l2 <- newLabel;
					  let {(c,t) = $3;
					       ts = TSchm c t};
					  return (Ann l1 (AnnExp $1 
							    (Ann l2 ts)))}}
	| Exp1 ":::" SigTyp	    {% do{l1 <- newLabelPos $2;
					  l2 <- newLabel;
					  let {(c,t) = $3;
					       ts = TSchm c t};
					  return (Ann l1 (TAddExp $1 
							    (Ann l2 ts)))}}


Exp1	:: { Ann Exp }
	: Exp0			    {$1}
	| IfxExp		    {$1}

IfxExp	:: { Ann Exp }
	: Exp iid Exp0		    {% do{let {p = srcPos $2};
					  l1 <- newLabel;
					  l2 <- newLabelPos p;
					  l3 <- newLabel;

					  return $
					   Ann l1 (IfxAppExp $1 
						    (Ann l2 (VarExp (idStr $2)))
						    $3) }}
						    
	| '(' Exp iid Exp0 ')'	    {% do{let {p = srcPos $3};
					  l1 <- newLabel;
					  l2 <- newLabelPos p;
					  l3 <- newLabel;

					  return $
					   Ann l1 (PIfxAppExp $2
						    (Ann l2 (VarExp (idStr $3)))
						    $4) }}

Exp0	:: { Ann Exp }
	: if Exp then Exp else Exp  {% do l <- newLabel;
					  return (Ann l (IfExp $2 $4 $6))}
{-
	| '\\' id "->" Exp          {% do{let {p = srcPos $2};
					  l1 <- newLabelPos p;
					  l2 <- newLabel;
					  return (Ann l2 
						    (AbsExp (Ann l1 (idStr $2)) $4))}}
-}
    	| '\\' Pats "->" Exp        {% do l <- newLabel;
					  return (Ann l (AbsExp (reverse $2) $4))}

	| App			    {$1}
	| let Decs in Exp	    {% do l <- newLabel;
					  return (Ann l
						    (LetExp (reverse $2) $4))}
	| case Exp of '{' Alts '}'  {% do l <- newLabel;
					  return (Ann l (CaseExp $2 (concat $5)))}



Alts	:: { [[CaseAlt]] }
	: Alt			    {[$1]}
	| Alts ';' Alt		    {$3 : $1}

Alt	:: { [CaseAlt] }
	: Pat "->" Exp AltWhere	    {% do l1 <- newLabel;
					  l2 <- newLabel; 
				          return [CaseAlt (Ann l2 $1) $3 $4]}
	| Pat GdPat AltWhere	    {% do l1 <- newLabel;
					  l2 <- newLabel;
					  return [CaseAltGuard
						  (Ann l2 $1) $2 $3]}
	| {- empty -}		    {[]}

GdPat	:: { [GuardedExp] } 
	: GdAlt "->" Exp GdPats	    {($1,$3) : $4}

GdPats  :: { [GuardedExp] }
	: {- empty -}		    {[]}
	| GdPat			    {$1}

GdAlt   :: { Ann Exp }
        : '|' Exp		    {$2}
	

AltWhere:: { WhereDecs }
	: {- empty -}		    {[]}
	| where Decs		    {$2}






App	:: { Ann Exp }
	: App AExp		    {% do l <- newLabel;
					  return (Ann l (AppExp $1 $2))}
	| AExp			    {$1}
	

AExp	:: { Ann Exp }
	: id                        {% do{let {p = srcPos $1};
					  l <- newLabelPos p;
					  return (Ann l (VarExp (idStr $1)))}}
	| "True"		    {% do l <- newLabel;
					  return (Ann l (ConExp "True"))}
	| "False"		    {% do l <- newLabel;
					  return (Ann l (ConExp "False"))}
	| conid                     {% do{let {p = srcPos $1};
				          l <- newLabelPos p;
					  return (Ann l (ConExp (idStr $1)))}}
	| '(' iid ')'		    {% do{let {p = srcPos $2};
                                          l <- newLabelPos p;
                                          return (Ann l (VarExp (idStr $2)))}}
	| Lit			    {% do l <- newLabel;
					  return (Ann l (LitExp $1))}
	| '(' CExps ')'             {% do l <- newLabel;
					  case $2 of {
				    [x] -> return (addInfixParens x);
				    xs  -> return (Ann l(TupExp (reverse $2)))}}
				    
	| '(' iid Exp ')'	    {% do{let {p = srcPos $2};
					  l1 <- newLabel;
					  l2 <- newLabelPos p;
					  return (Ann l1 (SecExpRight 
						  (Ann l2 (VarExp (idStr $2)))
						  $3))}}
	 | '(' Exp iid ')'	    {% do{let {p = srcPos $3}; 
					  l1 <- newLabel;
					  l2 <- newLabelPos p;
					  return (Ann l1 (SecExpLeft
						 $2 
						 (Ann l2 (VarExp (idStr $3)))))}}
				    
	| '(' ')'                   {% do l <- newLabel;
					  return (Ann l (TupExp []))}
	| '[' ListExp		    {$2}

	| '[' ']'		    {% do l <- newLabel;
					  return (Ann l (ListExp []))}

ListExp	:: { Ann Exp }
--	: CExps	']'		    {% do l <- newLabel;
--					  return (Ann l (ListExp (reverse $1)))}
	: Exp ']'		    {% do l <- newLabel;
					  return (Ann l (ListExp [$1]))}
	| Exp ',' Exp ']'	    {% do l <- newLabel;
                                          return (Ann l (ListExp [$1,$3]))}
	| Exp ',' Exp ',' CExps ']' {% do l <- newLabel;
                                          return (Ann l (ListExp 
						    ([$1,$3] ++ reverse $5)))}

	| Exp ".." ']'		    {% do l <- newLabel;
					  return (Ann l (EnumExpFrom $1))}
	| Exp ',' Exp ".." ']'	    {% do l <- newLabel;
					  return (Ann l (EnumExpFromThen $1 $3))}
	| Exp ".." Exp ']'	    {% do l <- newLabel;
					  return (Ann l (EnumExpFromTo $1 $3))}
						    
	| Exp ',' Exp ".." Exp ']'  {% do l <- newLabel;
					  return (Ann l (EnumExpFromThenTo 
							$1 $3 $5))}
	| Exp '|' CStmts	    {% do l <- newLabel;
					  return (Ann l (ListCompExp $1 $3))}


CStmts	:: { [Ann Stmt] }
	: Stmt ']'		    {[$1]}
	| Stmt ',' CStmts	    {$1 : $3}

Stmt	:: { Ann Stmt }
	: Pat "<-" Exp		    {% do l1 <- newLabel;    
					  l2 <- newLabel;
					  return (Ann l1 (GenStmt (Ann l2 $1)
								  $3))}
--	| Exp			    {% do l <- newLabel;
--					  return (Ann l (QualStmt $1))}
	| let Decs 		    {% do l <- newLabel;
					  return (Ann l (LetStmt $2))}

CExps	:: { [Ann Exp] }
	: Exp			    {[$1]}
	| CExps ',' Exp		    {$3 : $1}


Lit	:: { Lit }
	: int			    {% return (uncurry IntLit $1)}
	| float			    {% return (uncurry FloatLit $1)}
	| str			    {% return (StrLit $1)}
	| char			    {% return (CharLit $1)}


{------------------------------------------------------------------------------}
Pat	:: { Pat }
	: ConPat		    {$1}
	| AtomPat		    {$1}

ConPat	:: { Pat }
	: '(' conid Pats ')'	    {% do let {p = srcPos $2};
					  l1 <- newLabelPos p;
					  return (ConPat (Ann l1 (idStr $2)) 
							 (reverse $3))}
	| '(' IfxPat ')'	    {$2}
	| conid			    {% do let {p = srcPos $1};
					  l1 <- newLabelPos p;
					  return (ConPat (Ann l1 (idStr $1))[])}

IfxPat  :: { Pat }
	: Pat iid Pat		    {% do let {p = srcPos $2};
					  l1 <- newLabelPos p;
					  l2 <- newLabel;
					  l3 <- newLabel;
					  return (ConPat (Ann l1 (idStr $2))
							 [Ann l2 $1,Ann l3 $3])}
							 
	| Pat iid IfxPat	    {% do let {p = srcPos $2};
					  l1 <- newLabelPos p;
					  l2 <- newLabel;
					  l3 <- newLabel;
					  return (ConPat (Ann l1 (idStr $2))
							 [Ann l2 $1,Ann l3 $3])}

AtomPat :: { Pat }
--	: '_'			    {JokPat}
	: id			    {% return (VarPat (idStr $1))}
	| Lit			    {% return (LitPat $1)}
	| "True"		    {% do l <- newLabel;
					  return ((ConPat (Ann l "True") []))}
	| "False"		    {% do l <- newLabel;
					  return ((ConPat (Ann l "False") []))}
	| '[' ']'		    {ListPat []}
	| '[' CPats ']'		    {ListPat $2}
	| '(' CPats ')'		    {TupPat $2}
	| '(' AtomPat ')'	    {$2}
	| '(' ')'		    {TupPat []}
	| '[' Pat ']'		    {% do l <- newLabel;
					  return (ListPat [Ann l $2])}


Pats	:: { [Ann Pat] }
	:			    {[]}
	| Pats Pat		    {% do l <- newLabel;
					    return ((Ann l $2) : $1)} -- into

CPats	:: { [Ann Pat] }
	: Pat2 ',' Pat2		    {% do l1 <- newLabel;
					  l2 <- newLabel;
					  return [Ann l1 $1, Ann l2 $3]}
	| Pat2 ',' CPats	    {% do l <- newLabel;
					    return ((Ann l $1) : $3)}

Pat2	:: { Pat }
	: ConPat2		    {$1}
	| AtomPat		    {$1}

ConPat2	:: { Pat }
	: '(' conid Pats ')'        {% do l1 <- newLabel;
					  return (ConPat (Ann l1 (idStr $2)) 
							 (reverse $3))}
	| '(' IfxPat ')'            {$2}
	| conid Pats		    {% do l1 <- newLabel;
					  return (ConPat (Ann l1 (idStr $1)) 
						    	 (reverse $2))}

{------------------------------------------------------------------------------}
-- non-simple patterns - can appear in bindings

BPat	:: { Pat }
	: BConPat		    {$1}
	| BAtomPat		    {$1}

BConPat	:: { Pat }
	: '(' conid Pats ')'	    {% do let {p = srcPos $2};
					  l1 <- newLabelPos p;
					  return (ConPat (Ann l1 (idStr $2)) 
							 (reverse $3))}
	| '(' IfxPat ')'	    {$2}
	| conid			    {% do let {p = srcPos $1};
					  l1 <- newLabelPos p;
					  return (ConPat (Ann l1 (idStr $1))[])}

BAtomPat:: { Pat }
	: '_'			    {JokPat}
	| Lit			    {% return (LitPat $1)}
	| "True"		    {% do l <- newLabel;
					  return ((ConPat (Ann l "True") []))}
	| "False"		    {% do l <- newLabel;
					  return ((ConPat (Ann l "False") []))}
	| '[' ']'		    {ListPat []}
	| '[' CPats ']'		    {ListPat $2}
	| '(' CPats ')'		    {TupPat $2}
	| '(' AtomPat ')'	    {$2}
	| '(' ')'		    {TupPat []}
	| '[' Pat ']'		    {% do l <- newLabel;
					  return (ListPat [Ann l $2])}

{------------------------------------------------------------------------------}


-- Signature for functions and data constructors
AnySigCon :: { Sig }
	  : Sig				    {$1}
	  | ConSig			    {$1}

ConSig	:: { Sig }
	: conid "::" SigTyp		    {% do l1 <- newLabel;
						  let (c,t) = $3 in
						  return (Sig Univ
							      (Ann l1(idStr $1))
							      c t)}
AnySig :: { Sig }
       : Sig				    {$1}
       | ESig				    {$1}


-- signature for instance declarations
InstSig	:: { InstSig }
	: Typ "=>" Pred			    {% do l1 <- newLabel;
						  l2 <- newLabel;
						  ps <- tupToCtxt $1;
						  let {ctxt = Ann l1 (Ctxt ps)};
						  return (InstSig ctxt $3)}
						  
	| Typ				    {% do l <- newLabel;
						  let {ctxt = (Ctxt [])};
						  ps <- tupToCtxt $1;
						  return (InstSig (Ann l ctxt) (head ps))}


InstCtxt:: { Ann Ctxt }
	: Typ				    {% do l <- newLabel;
						  ps <- tupToCtxt $1;
						  return (Ann l (Ctxt ps)) }

Sig	:: { Sig }
	: Var "::" SigTyp		    {% do l1 <- newLabelPos (srcPos $1);
						  let (c,t) = $3 in
						  return (Sig Univ
							      (Ann l1(idStr $1))
							      c t)}
ESig	:: { Sig }
	: Var ":::" SigTyp		    {% do l1 <- newLabelPos (srcPos $1);
						  let (c,t) = $3 in
						  return (Sig Exist
							      (Ann l1(idStr $1))
							      c t)}
Var	:: { (String, SrcPos) }
	: id				    {$1}
	| '(' iid ')'			    {$2}
--	| '(' '*' ')'			    {("*", noSrcPos)} -- Hacky

-- same hack as in hsparser, parse context as a tuple type, then convert
{-
SigTyp  :: { (Ann Ctxt, Ann Typ) }
	: Typ				    {% do l <- newLabel;    
					          return (Ann l (Ctxt []), $1)}
	| ATyp "=>" Typ			    {% do l <- newLabel;
						  ps <- tupToCtxt $1;
						  return (Ann l (Ctxt ps), $3)}
-}

SigTyp  :: { (Ann Ctxt, Ann Typ) }
	: Typ				    {% do l <- newLabel;    
					          return (Ann l (Ctxt []), $1)}
	| Typ "=>" Typ			    {% do l <- newLabel;
						  ps <- tupToCtxt $1;
						  return (Ann l (Ctxt ps), $3)}


Rule	:: { Rule }
	: PropRule		    {$1}
	| SimpRule		    {$1}

PropRule :: { Rule }
	 : CtxtA "==>" CnstA	    {% do l1 <- newLabel;
					  l2 <- newLabel;
					  return (PropRule (Ann l1 $1)
						           (Ann l2 $3))}
SimpRule :: { Rule }
	 : CtxtA "<==>" CnstA	    {% do l1 <- newLabel;
					  l2 <- newLabel;
					  return (SimpRule (Ann l1 $1)
							   (Ann l2 $3))}
CtxtA	:: { Ctxt }
	: Ctxt			    {Ctxt (reverse $1)}

Ctxt    :: { [Ann Pred] }
	: Pred			    {[$1]}
        | Ctxt ',' Pred		    {($3 : $1)}

CnstA	:: { Cnst }
	: Cnst			    {% do l <- newLabel;
					  return (Cnst (reverse $1))}
Cnst    :: { [Ann Prim] }
	: Prim			    {% do l <- newLabel;
					  return [Ann l $1]}
        | Cnst ',' Prim		    {% do l <- newLabel;
					  return ((Ann l $3) : $1)}

Pred    :: { Ann Pred }
	: "True"		    {% do{l <- newLabelPos noSrcPos;
					  return (Ann l truePred)}}
	| "False"		    {% do{l <- newLabelPos noSrcPos;
					  return (Ann l falsePred)}}
	
{-				    -- convert it to a type, and 
				    -- invoke typToPred, as per Ctxt preds
	| conid Typs		    {% do let {p = srcPos $1};
					  l1 <- newLabelPos p;
					  l2 <- newLabel;
					  l3 <- newLabel;
					  let {fold t [] = return t;
					       fold t (x:xs) = do {
						l <- newLabel;
						let {t' = Ann l (AppTyp t x)};
						fold t' xs };
					       c = Ann l2 (ConTyp (idStr $1))};
					  t <- fold c (reverse $2);
					  typToPred t}
-}
	| Typ			    {% typToPred $1}

	| '(' iid ')' Typ	    {% do{let {p = srcPos $2};
					  l1 <- newLabelPos p;
                                          return (Ann l1 (Pred (idStr $2) $4))}}
	| '(' Pred ')'		    {$2}

Prim    :: { Prim }
	: Pred			    {let {pred = item $1}
				     in  if isFalsePred pred
				         then falseEqPrim
				         else if isTruePred pred
					      then trueEqPrim
					      else (PredPrim $1)}
        | Typ '=' Typ		    {% do l1 <- newLabel;
					  l2 <- newLabel;
					  return (EqPrim $1 $3)}


Typ	:: { Ann Typ }
	: Typ0			    {$1}
	| InTyp			    {$1}


InTyp	:: { Ann Typ }
	: Typ iid Typ0		    {% do{let {p = srcPos $2};
					  l1 <- newLabel;
					  l2 <- newLabelPos p;
					  l3 <- newLabel;
					  l4 <- newLabel;
					  l5 <- newLabel;
					  l6 <- newLabel;
					  return (Ann l6 (AppTyp 
						  (Ann l1 
						   (AppTyp 
						    (Ann l2 
						     (ConTyp (idStr $2)))
						    $1)) $3))}}

	| Typ0 "->" Typ		    {% do l1 <- newLabel;
					  l2 <- newLabel;
					  l3 <- newLabel;
					  l4 <- newLabel;
					  l5 <- newLabel;
					  l6 <- newLabel;
					  return (Ann l6
						 (AppTyp 
						  (Ann l1 
						   (AppTyp 
						    (Ann l2 
						     (ConTyp ({-Ann l3-} "->")))
						    $1)) $3))}    
	
Typ0	:: { Ann Typ }
	: TApp			    {$1}

TApp	:: { Ann Typ }
	: TApp ATyp		    {% do l1 <- newLabel;
					  return (Ann l1 (AppTyp $1 $2))}
	| ATyp			    {$1}

ATyp	:: { Ann Typ }
	: id			    {% do{let {p = srcPos $1};
					  l <- newLabelPos p;
					  return (Ann l (VarTyp (idStr $1)))}}
--	| '*'			    {% do l <- newLabel;
--					  return (Ann l (VarTyp "*"))}
	| '(' iid ')'		    {% do{let {p = srcPos $2};
					  l <- newLabelPos p;
				          let{id = idStr $2};
				          if id == "*" 
					   then return (Ann l (VarTyp id))
					   else return (Ann l (ConTyp id))}}
					-- error "parse error in type"}}
					   
	| '(' ')'		    {% do l <- newLabel;
					  return (Ann l (TupTyp []))}
        | conid			    {% do{let {p = srcPos $1};
					  l <- newLabelPos p;
					  return (Ann l (ConTyp (idStr $1)))}}
	| '(' CTyps ')'		    {% do l <- newLabel;
					  return (Ann l (TupTyp (reverse $2)))}
	| '[' Typ ']'		    {% do l <- newLabel;
					  return (Ann l (ListTyp $2))}
	| '[' ']'		    {% do l <- newLabel;
					  return (Ann l (ConTyp "[]"))}
	| '(' Typ ')'		    {$2}

CTyps	:: { [Ann Typ] }
	: Typ			    {[$1]}
	| CTyps ',' Typ		    {($3 : $1)}

ATyps   :: { [Ann Typ] }
	:                           {[]}
        | ATyps ATyp                {$2 : $1}

Typs    :: { [Ann Typ] }
	: Typ			    {[$1]}
	| Typs Typ		    {$2 : $1}

Dat	:: { Dat }
	: conid Ids DatCons	    {% do l <- newLabel;
					  return (Dat (Ann l (idStr $1))
						      (reverse $2)
						      (reverse $3))}

DatCons	:: { [Ann Con] }
	:			    {[]}
	| '=' Cons		    {$2}

Cons	:: { [Ann Con] }
	: Con			    {% do l <- newLabel;
					  return [Ann l $1]}
	| Cons '|' Con		    {% do l <- newLabel;
					  return ((Ann l $3) : $1)}
	
Con	:: { Con }
	: conid ATyps		    {% do l1 <- newLabel;
					  return (Con (Ann l1 (idStr $1)) 
							    (reverse $2))}

Ids	:: { [Ann Id] }
	:			    {[]}
	| Ids id		    {% do l <- newLabel;
					  return ((Ann l (idStr $2)): $1)}


NRef	:: { NestedRef}
	: Var			    {[(idStr $1, Nothing)]}
	| Var ';' NRef		    {(idStr $1, Nothing) : $3}
	| Var ';' int ';' NRef	    {(idStr $1, Just (fromIntegral (fst $3))) : $5}

-------------------------------

{


parse s = runP s parser
parseInit fn s n = runPIx fn s parser n

interactiveParser = dontHandleLayout >> dontExitOnFail

parseTyp s = runP s $ do interactiveParser
			 typ <- parserTyp
			 l   <- newLabel
			 return (Ann l typ)

parseExp n s = flip (runPIx' s) n $ do interactiveParser
				       parserExp

parseTSchm n s = flip (runPIx' s) n $ do interactiveParser
					 (c, t) <- parserSigTyp
					 l1 <- newLabel
					 return (Ann l1 (TSchm c t))

parseNestedRef n s = flip (runPIx' s) n $ do interactiveParser
					     parserNestedRef


happyError :: P a
happyError = do 
    l <- getLineNum
    c <- getColNum
    fn <- getFileName
    s <- getPrevInput
    (t,_,_) <- getPrevToken
    let pos = show (l,c)
	posStr = if null fn then pos
			    else fn ++ ":" ++ pos

    let msg = "Parse error at " ++ posStr ++ ", token: `" ++ show t ++ "'"

    setInput ""
    fail msg
 
}
