--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------

-- miscellaneous functions for parsing Chameleon (includes parser monad)

module ParserMisc where

import Misc
import Ann_InputAST
import Tokens
import List

import IOExts
import Pretty

--------------------------------------------------------------------------------

-- input: current stream
-- prevInput: stream before last token was removed (used for error reports)
-- noToken: True if no token has yet been read on the current line
-- layoutCols: stack of the layout contexts
-- indentToken: True if the previous token introduces a new layout block
-- exitOnFail: if True, bail to OS in event of a parse error
-- lineNum: current line number of stream
-- colNum:  current column number of stream
-- handleLayout: if True, the lexer will do a layout-removing transformation on
--		 the input stream
-- prevColNum: column of previous token in stream (used for error reports)
-- label: next label for AST annotation
data State = MkState {input :: String, prevInput :: String,
		      noToken :: Bool, layoutCols :: [Int], indentToken :: Bool,
		      exitOnFail :: Bool,
		      linenum :: Int, 
		      colnum :: Int,
		      tokRow :: Row, tokCol :: Col,
		      label :: Index,
		      handleLayout :: Bool,
		      tokensReady :: Bool, prevToken :: TokenInfo,
		      stream :: [TokenInfo],
		      commands :: [SrcCommand],
		      fileName :: String}	deriving Show

newtype P a = P (State -> E (State, a))

type Row = Int
type Col = Int
type TokenInfo = (Token, Row, Col)

instance Monad P where
    -- return :: a -> P a
    return a = P (\st -> return (st, a))
    -- (>>=) :: P a -> (a -> P b) -> P b
    (P a) >>= f = P (\st -> do (st', a') <- a st
			       let P b = f a'
			       b st')
    fail p = P (\st -> if exitOnFail st 
		       then exit p
		       else fail "")
    
runP :: String -> P a -> E (Index, [SrcCommand], a)
runP s (P a) = case a initState of
		Err e	     -> Err e
		Succ (st, p) -> Succ (label st, commands st, p)
    where initState = MkState {input = s, prevInput = s,
			       noToken = True, layoutCols = [], 
			       indentToken = True,
			       exitOnFail = True,
			       linenum = 1, 
			       colnum = 1, tokCol = 1, tokRow = 1,
			       label = unkIx 0,
			       handleLayout = True,
			       tokensReady = False, 
			       prevToken = teof,
			       stream = [],
			       commands = [],
			       fileName = ""}

-- as above, but takes an initial value for label index
-- (also returns the remaining string input)
runPIx fn s (P a) n = case a initState of
		    Err e	 -> Err e
		    Succ (st, p) -> Succ (input st, commands st, label st, p)
    where initState = MkState {input = s, prevInput = s,
			       noToken = True, layoutCols = [],
			       indentToken = True,
			       exitOnFail = True,
			       linenum = 1, 
			       colnum = 1, tokCol = 1, tokRow = 1,
			       label = unkIx n,
			       handleLayout = True,
			       tokensReady = False, 
			       prevToken = teof,
			       stream = [],
			       commands = [],
			       fileName = fn}
runPIx' = runPIx ""


-- returns row,column position of parser
--getPos :: P (Int, Int)
getPos = P (\st -> return (st, ("noModule", linenum st, colnum st)))

getFileName = P (\st -> return (st, fileName st))

getColNum :: P Col
getColNum = P (\st -> return (st, colnum st))

getLineNum :: P Row
getLineNum = P (\st -> return (st, linenum st))

incLineNum :: P ()
incLineNum = P (\st -> return (st{linenum = linenum st + 1}, ()))

setLineNum :: Row -> P ()
setLineNum r = P (\st -> return (st{linenum = r}, ()))

setColNum :: Col -> P ()
setColNum c = P (\st -> return (st{colnum = c}, ()))

incColNum :: P ()
incColNum = addColNum 1


addColNum :: Int -> P ()
addColNum n = P (\st -> return (st {colnum = colnum st + n}, ()))


resetColNum :: P ()
resetColNum = P (\st -> return (st {colnum = 1, noToken = True}, ()))

tabColNum :: P ()
tabColNum = P (\st -> return (st {colnum = tab (colnum st)}, () ))
    where tab n = ((n `div` 8) + 1) * 8 + 1

newLabel = P (\st -> return (st 
	{label = let (Ix n _ _) = label st in unkIx (n+1)}, label st))

newLabelPos pos = P (\st -> return $
	let fn  = fileName st
	    (_,x,y) = pos
	    pos' = (fn,x,y)
	in  (st {label = let (Ix n _ _) = label st in unkIxPos (n+1) pos'}, 
	     newSrcPos pos' (label st)))


getInput :: P String
getInput = P (\st -> return (st, input st))

getPrevInput :: P String
getPrevInput = P (\st -> return (st, prevInput st))



getTokCol :: P Col
getTokCol = P (\st -> return (st, tokCol st))

getTokRow :: P Row
getTokRow = P (\st -> return (st, tokRow st))

updateTokPos :: P ()
updateTokPos = P (\st -> return (st {tokCol = colnum st, 
				     tokRow = linenum st} ,()))

setInput :: String -> P ()
setInput s = P (\st -> return (st {prevInput = input st, input = s}, ()))

ckHandleLayout :: P Bool
ckHandleLayout = P (\st -> return (st, handleLayout st))

doHandleLayout :: P ()
doHandleLayout = P (\st -> return (st {handleLayout = True}, ()))

dontHandleLayout :: P ()
dontHandleLayout = P (\st -> return (st {handleLayout = False}, ()))

dontExitOnFail :: P ()
dontExitOnFail = P (\st -> return (st {exitOnFail = False}, ()))

----------------------------------------
pushCol :: Int -> P ()
pushCol n = P (\st -> return (st {layoutCols = n : layoutCols st}, ()))

initLayoutCols :: Int -> P ()
initLayoutCols n = P (\st -> return (st {layoutCols = [n]}, ()))

emptyCols :: P Bool
emptyCols = P (\st -> return (st, null (layoutCols st)))

popCol :: P Int
popCol = topColF tl
	  
topCol :: P Int
topCol = topColF id

-- manipulate the stack, also return the top element
topColF :: ([Int] -> [Int]) -> P Int
topColF f = P (\st -> let cs = layoutCols st
		      in  return (st {layoutCols = f cs}, hd cs))
    where hd []	    = error "ParserMisc.popCol.hd"
	  hd (x:_)  = x


-- remove from the token stream the right brace which would match the last read
-- left brace (closing the block)
dropContextEnd :: P ()
dropContextEnd = do ts <- getStream
		    setStream (rem ts 0)
    where rem [] _ = []
	  rem (i@(t,_,_):is) c | t == TLBrace = i : rem is (c+1)
			       | t == TRBrace = if c == 0 then is
							  else i : rem is (c-1)
			       | otherwise    = i : rem is c

----------------------------------------

sawToken :: P ()
sawToken = setNoToken False

clrNoToken = sawToken

setNoToken :: Bool -> P ()
setNoToken b = P (\st -> return (st {noToken = b}, ()))

getNoToken :: P Bool
getNoToken = P (\st -> return (st, noToken st))

setIndentToken, clrIndentToken :: P ()
setIndentToken = P (\st -> return (st {indentToken = True}, ()))
clrIndentToken = P (\st -> return (st {indentToken = False}, ()))

wasIndentToken :: P Bool
wasIndentToken = P (\st -> return (st, indentToken st))


----------------------------------------

getState :: P State
getState = P (\st -> return (st, st))

setState :: State -> P ()
setState st = P (\_ -> return (st, ()))

----------------------------------------

tokensAvailable :: P Bool
tokensAvailable = P (\st -> return (st, tokensReady st))

tokensDone :: P ()
tokensDone = P (\st -> return (st {tokensReady = True}, ()))

getPrevToken :: P TokenInfo
getPrevToken = P (\st -> return (st, prevToken st))

setStream :: [TokenInfo] -> P ()
setStream ts = P (\st -> return (st {stream = ts}, ()))

getStream :: P [TokenInfo]
getStream = P (\st -> return (st, stream st))

popToken :: P TokenInfo
popToken = P (\st -> let ts = stream st
			 t  = hd ts
		     in  return (st {stream = tl ts, prevToken = t}, t))
tl [] = trace "EMPTY CONTEXT STACK" []
tl xs = tail xs
hd [] = error "ParserMisc.hd"
hd xs = head xs

-- also updates the parsers position state
nextToken :: P Token
nextToken = do (t,r,c) <- popToken
	       setLineNum r
	       setColNum c
	       return t

----------------------------------------


falseEqPrim = EqPrim (nullAnn $ TupTyp []) 
		     (nullAnn $ ListTyp (nullAnn (TupTyp [])))
isFalsePred pred = case pred of Pred "False" _ -> True; _ -> False
falsePred = Pred "False" (error "no need to look at False pred's type")

trueEqPrim = EqPrim (nullAnn $ TupTyp []) (nullAnn $ TupTyp [])
isTruePred pred = case pred of Pred "True" _ -> True; _ -> False
truePred = Pred "True" (error "no need to look at True pred's type")


----------------------------------------

addCommand :: SrcCommand -> P ()
addCommand sc = P (\st -> return (st { commands = sc : commands st }, ()))

--------------------------------------------------------------------------------

tupToCtxt :: Ann Typ -> P [Ann Pred]
tupToCtxt at = case item at of
	       (TupTyp ts)  -> mapM (cToPred) ts
	       (AppTyp b c) -> do p <- typToPred at
				  return [p]
{-	       
			      case item b of
				(ConTyp d) -> return ([Ann (ann b) (Pred d c)])
				_	   -> fail "Parse error in context (1a)"
-}				
	       _	    -> fail "Parse error in context (1b)"

cToPred :: Ann Typ -> P (Ann Pred)
cToPred at = case item at of
	     (AppTyp b c) -> typToPred at
{-	     
			     case item b of
			      (ConTyp d) -> let l = ann b
					    in  return (Ann l (Pred d c))
			      _		 -> fail "Parse error in context (2a)"
-}			      
	     _		  -> fail "Parse error in context (2b)"
	        

-- takes a type of form (T t1 ... tn), and turns it into a 
-- predicate (P [t1, ..., tn]) - where the list type is unique to
-- multi-parameter type classes 
typToPred :: Ann Typ -> P (Ann Pred)
typToPred t = do let res@(id,ix,ts) = undo t []
		     -- ts = reverse rts
		 ts' <- combine ts
	         return (Ann ix (Pred id ts'))
    where 

	-- recombines individual parameters into one type
	-- combine :: [Ann Typ] -> P Typ
	combine ts = do
	    ft <- fold ts
	    l1 <- newLabel
	    l2 <- newLabel
	    return $ if len == 1 
	     then ft
	     else Ann l1 (AppTyp (Ann l2 tcon) ft)
	    where
		len  = length ts
		tcon = ConTyp ("{" ++ show len ++ "}")
	    
		fold [t]    = return t
		fold (t:ts) = do ft <- fold ts
				 l  <- newLabel
				 return (Ann l (AppTyp t ft))
    
	-- undoes nested applications
	undo t as = case item t of
	    AppTyp t1 t2 -> case item t1 of 
		AppTyp _ _ -> undo t1 (t2:as)
		ConTyp id  -> (id, ann t1, t2:as)
	    _ -> error ("malformed type class constraint: " ++ show t)

----------------------------------------

-- adds an undefined definition for the instance declaration given
-- (only if there are no existing definitions for that instance)
-- e.g.	    instance C a => C [a]
--
-- BECOMES  instance C a => C [a] where
--		c = undefined
addUndefinedInstanceDef :: Dec -> P Dec
addUndefinedInstanceDef dec = case dec of 
    InstDec asig [] -> case item asig of
	    InstSig ctxt (Ann x (Pred id t)) -> do
		x1 <- newLabel
		x2 <- newLabel
		x3 <- newLabel
		let rhs = RHS1 (Ann x1 (VarExp "undefined"))
		    def = Def (Ann x2 id) [] rhs []
		    adef= Ann x3 def
		return (InstDec asig [adef])
    _ -> return dec


--------------------------------------------------------------------------------

addInfixParens (Ann x (IfxAppExp e1 e2 e3)) = Ann x (PIfxAppExp e1 e2 e3)
addInfixParens e = e

--------------------------------------------------------------------------------

teof = (Teof, -1, -1)

