--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- pretty printer for all things Chameleon
--
-- unifies the previously approach of having separate implementations for each
-- kind of output (plain, highlighted, annotated)
--
-- features:
--  - optional annotations/highlighting
--  - optional indentation (layout vs. non-layout)
--  - abbreviated output
--  - optional colour (except for highlighting, probably)

-- NOTE: Do not use <+> from PrettyPrint, as it will insert spaces of the
--	 current colour (rather than always-default-colour spaces).
--	 The effect is that spaces besides highlighted terms are also
--	 highlighted - an ugly effect, and potentially ambiguous, since it
--	 might look like we're highlighting an application (which we never do.)

module Pretty (PPrint, kw, showSemiList,
	       pprint, pprintPlain, pprintHL, pprintAnn, 
	       pprintLocs, pprintLoc, pprintAndHLLocs, pprintAndHLLoc,
	       pprintCommas, catListWith, pprintPlainList,
	       retsep, text, sep, (<>), (<+>),
	       showCommaList, semiList, showAnnList, 

	       Which(..), Style(..), PrettyDoc,
	       pprintJustLocation, pprintInStyle, toString,
	       emptyPD, textPD, vcat, nothingPD, (<++>), (<||>),
	       
	       printProgLoc, printProgHL, printProgHLGreen, printProgHLBlue, 
	       pplProg)
where

import qualified PrettyPrint as PP
import Terminal
import Ann_InputAST
import Misc
import Name
import Operators (OpInfo, PAInfo, lowestPA)
import ASTs (ltProg, commonParentLoc, nodeLoc, 
	     parentLoc, grandParentLoc, findLocNode)


import Char
import Monad
import Maybe

import IOExts

--------------------------------------------------------------------------------

type Doc = PP.Doc

text = PP.text
empty = PP.empty
(<>) = (PP.<>)
($$) = (PP.$$)
nest = PP.nest
space = PP.space
ghost = PP.ghost
parens = PP.parens
render = PP.render

textPD  = return . text
textPD' = neverColour . return . text
emptyPD = return empty
ghostPD = return . ghost


{-
-- generate a PrettyDoc containing the two arguments one space apart
class Spaceable a where
    (<+>) :: a -> a -> PrettyDoc

instance Spaceable a => Spaceable (Pretty a) where
    pd1 <+> pd2 = do d1 <- pd1
		     d2 <- pd2
		     c  <- topBGColour
		     dc <- getDefaultColour
		     return (d1 <> ghost (bg dc) <> space <> 
				   ghost (bg c) <> d2)
   
instance Spaceable Doc where
    d1 <+> d2 = do c  <- topBGColour
		   dc <- getDefaultColour
		   return (d1 <> ghost (bg dc) <> space <> 
				 ghost (bg c) <> d2)
-}
{-
(<+++>) :: Doc -> Doc -> PrettyDoc
d1 <+++> d2 = do c  <- topBGColour
	         dc <- getDefaultColour
	         return (d1 <> ghost (bg dc) <> space <> 
			       ghost (bg c) <> d2)
-}

(<++>) :: PrettyDoc -> PrettyDoc -> PrettyDoc
pd1 <++> pd2 = do d1 <- pd1
		  d2 <- pd2
		  c  <- topBGColour
		  dc <- getDefaultColour
		  return (d1 <> ghost (bg dc) <> space <> 
			        ghost (bg c) <> d2)

-- adds a default (neutral colour) space
(<+>) :: PrettyDoc -> Doc -> PrettyDoc
pd1 <+> d2 = do d1 <- pd1
		c  <- topBGColour
		dc <- getDefaultColour
		return (d1 <> ghost (bg dc) <> space <> 
			      ghost (bg c) <> d2)

-- no space
(<|>) :: PrettyDoc -> Doc -> PrettyDoc
pd1 <|> d2 = do d1 <- pd1
		return (d1 <> d2)

(<||>) :: PrettyDoc -> PrettyDoc -> PrettyDoc
pd1 <||> pd2 = do d1 <- pd1
		  d2 <- pd2
		  return (d1 <> d2)


($|$) :: PrettyDoc -> Doc -> PrettyDoc
pd1 $|$ d2 = do d1 <- pd1
		return (d1 $$ d2)

($||$) :: PrettyDoc -> PrettyDoc -> PrettyDoc
pd1 $||$ pd2 = do d1 <- pushBGColour Default >> 
			 (pd1 <++> (ghostPD defaultBG))
		  d2 <- popBGColour >> pd2
		  return (d1 $$ d2)

sep, vcat :: [PrettyDoc] -> PrettyDoc
retsep, retvcat :: [Doc] -> PrettyDoc -- also lifts the given docs into Pretty

sep []  = emptyPD
sep [x] = x
sep (x1:xxs@(x2:xs)) = x1 <++> sep xxs
vcat pds = do ds <- doPrettys pds
	      return (PP.vcat ds)

doPrettys :: [PrettyDoc] -> Pretty [Doc]
doPrettys [] = return []
doPrettys (pd:pds) = do d  <- pd
			ds <- doPrettys pds
			return (d:ds)

retsep  = sep . map return
retvcat = vcat . map return 

nestPD :: Int -> PrettyDoc -> PrettyDoc
nestPD n pd = do d <- pd
		 return (nest n d)

spacePD = do c  <- topBGColour
	     dc <- getDefaultColour
	     return (ghost (bg dc) <> space <> ghost (bg c))

nothingPD = textPD ""

--------------------------------------------------------------------------------
-- exported print functions

-- old Show
pprintPlain :: PPrint a => a -> String
pprintPlain a = toString $ pprint a

-- old ShowHL
pprintHL :: PPrint a => [Loc] -> a -> String
pprintHL ls a = toStringOpts opts $ pprint a
    where opts = defaultOptions { highlight = Only ls } -- ,
--				  locations = Only ls, -- }
--				  style = Children }

-- old ShowAnn
pprintAnn :: PPrint a => a -> String
pprintAnn a = toStringOpts opts $ pprint a
    where opts = defaultOptions { annotate = All }

-- first  arg: locations to display
-- second arg: locations to highlight (within those displayed)
pprintHLLocs ls hls a = pprintColHLLocs ls hls Red a

pprintAndHLLoc  l a  = pprintAndHLLocs [l] a
pprintAndHLLocs ls a = pprintColHLLocs ls ls Blue a

pprintColHLLocs :: PPrint a => [Loc] -> [Loc] -> Colour -> a -> String
pprintColHLLocs ls hls c a = toStringOpts opts $ pprint a
    where opts = defaultOptions { highlight = Only hls,
				  locations = Only ls,
				  highlightColour = c,
				  style = Children }

pprintLoc l = pprintLocs [l]
pprintLocs ls a = toStringOpts opts $ do doc0 <- pprint a
				         docs <- getDocStore
				         doc1 <- retvcat docs
				         retvcat [doc0,doc1]

    where opts = defaultOptions { locations = Only ls,
				  style = Children }
				  -- style = Parents }


pprintPlainList a = toString $ catListWith "," a

--------------------------------------------------------------------------------

-- qualify an option's use
data Which = All | None | Only [Loc]	deriving Show

-- two printing styles 
-- (for dealing with locations that aren't explicitly mentioned)
--	parents : display only named locations and their parent locs
--	children: display only named locations and their sublocations
data Style = Parents | Children

-- annotate  : display annotation for these locations
-- highlight : highlight these locations
-- locations : locations to display (including their sublocations)
-- style     : determines how non-nominated locations are dealt with
-- abbreviate: if True, abbreviate locations not in above list (else ignore)
-- opInfo    : operator precedence/associativity
-- useColour : useColour for highlighting (as per Terminal module)
-- bgColours : a stack of background colours, top is the last one used
-- showedLast : True if the last location was actually printed (not ...)
-- defaultColour   : colour to use when NOT highlighting
-- highlightColour : colour to use for background highlighting 
-- subLocations : a stack of the locations of all the sub-nodes 
--		  (built on the return)
-- allowOverstep : allow pretty printer to display locations one deeper than
--		   those explicitly requested
-- parents : the current node's parents (always initialise as a singleton list!)
-- itsInfix : True if the thing being displayed is an infix expression
-- docStore: holds disconnected strings
data Options = Options { annotate :: Which, highlight :: Which, 
			 locations :: Which, abbreviate :: Bool, 
			 style :: Style,
			 opInfo :: [OpInfo], useColour :: Bool,
			 highlightColour :: Colour, defaultColour :: Colour,
			 bgColours :: [Colour], subLocations :: [Loc],
			 allowOverstep :: Bool, parents :: [Loc],
			 showedLast :: Bool,
			 itsInfix :: Bool,
			 docStore :: [Doc] }

newtype Pretty a = Pretty (Options -> (Options, a))
type PrettyDoc = Pretty Doc
type PrettyString = Pretty String

instance Monad Pretty where

    return v = Pretty (\opt -> (opt, v))

    (Pretty a) >>= b = Pretty (\opt -> let (opt', str) = a opt
					   Pretty b' = b str
				       in  b' opt')

class PPrint a where

    pprint :: a -> PrettyDoc
    pprintPA :: PAInfo -> Loc -> a -> PrettyDoc

    pprint x = pprintPA lowestPA 0 x
    pprintPA pa l x = pprint x


------------------------------------------------------------


toString :: PrettyDoc -> String
toString pd = toStringOpts defaultOptions pd

toStringOpts :: Options -> PrettyDoc -> String
toStringOpts opts (Pretty g) = render $ (snd (g opts))

defaultOptions = Options { annotate = None, highlight = None,
			   locations = All, useColour = True,
			   style = Parents,
			   abbreviate = False, opInfo = [],
			   highlightColour = Red, defaultColour = Default,
			   bgColours = [Default], subLocations = [],
			   allowOverstep = True, parents = [noLoc], 
			   showedLast = False, itsInfix = False,
			   docStore = [] }

getStyle     = Pretty (\opt -> (opt, style opt))
getAnnotate  = Pretty (\opt -> (opt, annotate opt))
getHighlight = Pretty (\opt -> (opt, highlight opt))
getUseColour = Pretty (\opt -> (opt, useColour opt))
getLocations = Pretty (\opt -> (opt, locations opt))
getDocStore  = Pretty (\opt -> (opt, docStore opt))
getAbbreviate    = Pretty (\opt -> (opt, abbreviate opt))
getAllowOverstep = Pretty (\opt -> (opt, allowOverstep opt))
getDefaultColour   = Pretty (\opt -> (opt, defaultColour opt))
getHighlightColour = Pretty (\opt -> (opt, highlightColour opt))



topBGColour    = Pretty (\opt -> (opt, head (bgColours opt)))
popBGColour    = Pretty (\opt -> (opt {bgColours = tail (bgColours opt)}, ()))
pushBGColour c = Pretty (\opt -> (opt {bgColours = c : bgColours opt}, ()))

immediateParent = Pretty (\opt -> (opt, head (parents opt)))
popParent = Pretty (\opt -> (opt {parents = tail (parents opt)}, ()))
pushParent l = Pretty (\opt -> (opt {parents = l : parents opt}, ()))
getParents = Pretty (\opt -> (opt, parents opt))


getSubLocations = Pretty (\opt -> (opt, subLocations opt))
addSubLocation l =Pretty (\opt->(opt {subLocations = l : subLocations opt}, ()))
addSubLocations ls =Pretty (\opt->(opt {subLocations=ls++subLocations opt}, ()))
clearSubLocations = Pretty (\opt -> (opt { subLocations = []}, ())) 

didShowLast   = Pretty (\opt -> (opt { showedLast = True }, ()))
didntShowLast = Pretty (\opt -> (opt { showedLast = False }, ()))
getShowedLast = Pretty (\opt -> (opt, showedLast opt))

addToDocStore doc = Pretty (\opt -> (opt {docStore = doc : docStore opt}, ()))

isInfix   = Pretty (\opt -> (opt { itsInfix = True }, ()))
isntInfix = Pretty (\opt -> (opt { itsInfix = False }, ()))
isItInfix = Pretty (\opt -> (opt, itsInfix opt))
doInfix c = do isInfix 
	       dc <- c 
	       isntInfix
	       return dc

handleInfixId id = do ix <- isItInfix
		      return $ text $ if ix then if isInfixId id
						 then id
						 else ('`':id++"`")
					    else infixParens id

----------------------------------------
-- exported combinators

pprintMod :: Pretty a -> (a -> Pretty ()) -> PrettyDoc -> PrettyDoc
pprintMod get put pd = do a <- get
			  d <- pd
			  put a
			  return d

pprintJustLocation :: Which -> PrettyDoc -> PrettyDoc
pprintJustLocation l pd = 
    pprintMod getLocations setLocations (setLocations l >> pd)

pprintInStyle :: Style -> PrettyDoc -> PrettyDoc
pprintInStyle s pd = pprintMod getStyle setStyle (setStyle s >> pd)

setStyle s     = Pretty (\opt -> (opt { style = s }, ()))
setLocations l = Pretty (\opt -> (opt { locations = l }, ()))


neverColour d = do cin <- topBGColour
		   ghostPD defaultBG <||> d <||> ghostPD (bg cin)

----------------------------------------

-- where the magic happens
instance PPrint a => PPrint (Ann a) where
    pprintPA pa _ x = do 
	let l' = ixLoc' (show (ann x)) (ann x)
	aopt <- getAnnotate
	hopt <- getHighlight
	uc   <- getUseColour
	ls   <- getLocations
	sls  <- getSubLocations
	ab   <- getAbbreviate

	-- find, cin : the background colour going in to this text
	--	 cout: the background colour following this text
	--	 doc : the text itself
	(cin, doc, cout) <- 
	    if uc && hopt `includes` l'
	    then -- we can use colour, and this location is to be highlighted
		 do o <- topBGColour
		    i <- getHighlightColour
		    pushBGColour i
		    doc  <- pprintParent pa l' (item x)
		    popBGColour
	    	    return (bg i, doc, bg o)
	    else if not uc
		 then -- cannot use colour, just return the document itself
		      do doc  <- pprintParent pa l' (item x)
			 return ("", doc, "")
		 else -- can use colour, but don't highlight this location
		      do o <- topBGColour
			 d <- getDefaultColour
			 pushBGColour d
			 doc <- pprintParent pa l' (item x)
			 popBGColour
			 return (bg d, doc, bg o)

	-- strings to precede and follow this text
	-- (a location number, if annotation has been requested)
        let (pre, post) = if aopt `includes` l'
			  then (show l' ++ "{", "}")
			  else ("", "")
	addSubLocation l'
	sls' <- getSubLocations

	let disp = (ghost cin <> text pre <> doc <> text post <> ghost cout)
       
        ao <- getAllowOverstep
	popParent
	p  <- immediateParent
	ps <- getParents

	-- whether we display anything for this location depends on the style
	s <- getStyle
--	let doShow = case s of 
--		Parents  -> any (ls`includes`) sls' || ao && ls `includes` p
--		Children -> let okay = any (ls`includes`) (l':ps)
--			    in  trace ("\n(ls,l): " ++ show (ls,l') ++
--				   "\nps: " ++ show ps ++
--				   "\nshow?: " ++ show okay) okay
	
	let doShow = ls `includes` l'
					    

	let dispDoc = ghost cin<> text pre <> doc <> text post <> ghost cout
	
	if not $ isIxLoc (ann x)
	 then return empty   -- can't do anything if the ix has no location!
         else if doShow  -- check if this is one of the locations to display
	      then do	 -- it is, display it, and record that we did so
		 didShowLast
		 return dispDoc
	      else -- we can't, display abbreviation, and record the fact
	      do -- if we're in Parent mode then we definitely don't show
		 -- anything for this location 
		 -- however, if we're in Children mode, and one of the child
		 -- nodes is to be displayed then we store it
		 sl <- getShowedLast
--		 case s of Children -> when sl (addToDocStore dispDoc)
--			   _ -> return ()
		 if sl then do didShowLast
			       return dispDoc
		       else do didntShowLast
			       return (ghost defaultBG <> abbrev <> ghost cout)

	where -- like pprintPA, but tracks parents (superLocs) on the descent
	      pprintParent pa l x = do
		pushParent l
		pprintPA pa l x
		    

-- default printing of lists
-- use separator-specific version when necessary for presentation
instance PPrint a => PPrint [a] where
    pprint xs = pprintCommas xs

instance (PPrint a, PPrint b) => PPrint (a,b) where
    pprint (a,b) = do da <- pprint a
		      db <- pprint b
		      retsep [text "(" <> da <> text ",", db <> text ")"]

plain = bg Default

includes None l = False
includes All  l = True
includes (Only ls) l = l `elem` ls

abbrev = text "..."

ppar cdoc = do doc <- cdoc
	       return (parens doc)

-- if it's an operator, surround it with parens
parId cs@(c:_) | not (isAlpha c || c == '_') = '(' : cs ++ ")"
parId cs = cs

--retvcat :: [Doc] -> Pretty Doc
--retvcat = return . vcat

-- hack
pprintAId :: Ann Id -> PrettyDoc
pprintAId aid = pprint $ Ann (ann aid) (VarExp (item aid))


pprintBars   xs = pprintList "|" xs
pprintSemis  xs = pprintList ";" xs
pprintCommas xs = pprintList "," xs
pprintSpaces xs = pprintList "" xs
pprintList :: PPrint a => String -> [a] -> PrettyDoc
pprintList s xs = pprintListOpt sep s xs


pprintCatCommas xs = catListWith "," xs

catListWith :: PPrint a => String -> [a] -> PrettyDoc
catListWith s xs = pprintListOpt sep s xs


-- pretty print a list with the given element separator
pprintListOpt :: PPrint a => ([PrettyDoc] -> PrettyDoc) -> String -> [a] -> 
			     PrettyDoc
pprintListOpt comb s xs = if null xs
		  then return empty
		  else do (f, dxs) <- ppc xs
			  ab <- getAbbreviate
			  if f && ab then return abbrev
				     else comb (map return dxs) -- retsep dxs
    where   -- flag True if all elements were abbreviated
	    ppc :: PPrint a => [a] -> Pretty (Bool, [Doc])
	    ppc []  = return (True, [empty])
	    ppc [x] = do dx <- pprint x
			 f  <- getShowedLast
			 return (not f, [dx])
	    ppc (x:xs) = do dx  <- pprint x
			    f1 <- getShowedLast
			    (f2, dxs) <- ppc xs
			    return (not f1 && f2, ((dx <> text s):dxs))


pprintWhere [] = return empty
pprintWhere ds = do cs <- clauses
		    return (nest 2 (text "where" $$ (nest 6 cs)))
    where clauses = do dds <- mapM pprint ds
		       retvcat dds

pprintDecs :: [Ann Dec] -> Pretty [Doc]
pprintDecs ds = mapM (\d -> do clearSubLocations
			       dd <- pprint d
			       sl <- getShowedLast
			       -- skip abbreviated output ("...")
			       if sl then return dd	    
				     else return empty) ds

----------------------------------------

showVarExp :: Ann Exp -> String
showVarExp ve = case item ve of 
    VarExp id | all (not.isAlphaNum) id -> id
	      | otherwise	        -> "`" ++ id ++ "`"
    _ -> ""

--------------------------------------------------------------------------------


instance PPrint Prog where
    pprint (Prog ds) = do dds <- pprintDecs ds
			  retvcat dds

override :: PPrint a => PAInfo -> Ann a -> PrettyDoc
override pa x = pprintPA pa (annLoc x) (item x)

instance PPrint Dec where
    pprintPA pa@(p,a) l dec = do
     -- never attempt to highlight entire declarations!
     pushBGColour Default
     x <- case dec of
	
	ValDec ds    -> do dds <- mapM (override pa) ds
			   retvcat dds

	AValDec s ds -> do xs <- pprint s 
			   sl <- getShowedLast
			   subs <- getSubLocations
			   clearSubLocations
			   xds <- mapM (override pa) ds
			   addSubLocations subs
			   retvcat $ if not sl then xds else xs:xds
			
	OverDec0 id  -> textPD "overload" <+> text (parId id)
    
	OverDec s ds -> do xs <- pprint s 
			   sl <- getShowedLast
			   subs <- getSubLocations
			   clearSubLocations
	                   xds <- mapM (override pa) ds
			   addSubLocations subs
			   let xs'  = textPD "overload" <+> xs <+> text "where"
			       xds' = nestPD 2 (retvcat xds)
	                   if not sl then xds'
				     else vcat [xs', xds']

	InstDec s ds -> do xs <- pprint s
			   sl <- getShowedLast
			   subs <- getSubLocations
			   clearSubLocations
	                   xds <- mapM (override pa) ds
			   addSubLocations subs
			   let xs'  = textPD "instance" <+> xs <+> 
				      if null ds then text ""
						 else text "where"
			       xds' = nestPD 2 (retvcat xds)
	                   if not sl then xds'
				     else vcat [xs', xds']

	PatDec p rhs w -> do pushBGColour Default
			     dp <- pprint p
			     drhs <- pprint rhs
			     dwh  <- pprintWhere w
			     popBGColour
			     return (ghost defaultBG) <||>
				((return dp) <+> drhs) $|$ dwh
	
	DatDec d     -> pprint d
	
	RuleDec r    -> pprint r

	HConDec _    -> return empty

	ExternDec s  -> do xs <- either pprint (pprintAId) s -- return empty
			   textPD "extern" <+> xs

	InterfaceDec _ -> return empty
    
     popBGColour
     return (ghost defaultBG <> x)

instance PPrint Def where
    pprintPA _ l (Def id ps rhs w) = do 
		    did  <- pprintAId (Ann (unkIx l) (item id))
	-- never attempt to highlight entire definitions!
		    pushBGColour Default
		    dps  <- if null ps 
			    then emptyPD
			    else pprintSpaces ps <|> space
		    drhs <- pprint rhs
		    dwh  <- pprintWhere w
		    popBGColour
		    return (ghost defaultBG) <||> 
		     ((return did) <+> dps <|> drhs) $|$ dwh
		     
	       
instance PPrint RHS where
    pprint (RHS1 e)  = do de <- pprint e
			  textPD "=" <+> de
    pprint (RHSG []) = emptyPD
    pprint (RHSG ((g,e):ges)) = do
		dg <- pprint g
		let de = pprint e
		dr <- pprint (RHSG ges)
		sep [textPD "|" <+> dg <+> text "=", de] $|$ dr

instance PPrint Pat where

    pprintPA (p,a) l pat = case pat of
	
	JokPat	    -> return $ text "_"

	VarPat id   -> return $ text id

	LitPat lit  -> pprint lit

	TupPat ps   -> do dps <- pprintCommas ps
			  return (text "(" <> dps <> text ")")

	ListPat ps  -> do dps <- pprintCommas ps
			  return (text "[" <> dps <> text "]")

	ConPat id ps ->
	    if not (null ps) then ppar s else s
	    where s :: PrettyDoc
		  s  = do is <- -- {- doInfix $ -} pprintAId id
				pprint $ Ann (ann id) (VarPat (item id))
			  if ifx 
			   then do s1 <- pprintPA (2,a) l p1
                                   s2 <- pprintPA (2,a) l p2
				   return (s1 <> is <> s2)
			   else do ss <- if null ps then emptyPD
						    else spacePD <||> 
							 pprintSpaces ps
				   (return is) <|> ss

		  ifx = isInfixId (item id) && l2 ps
		  [p1, p2] = ps
		  l2 [_,_] = True
		  l2 _	   = False


instance PPrint Exp where
    -- prec 0: abs, ite, case, let
    -- prec 1: ::
    -- prec 2: infix
    -- prec 3: application
    -- prec 4: atomic
    pprintPA pa@(p,a) l exp = case exp of

	VarExp id   -> handleInfixId id

	ConExp id   -> handleInfixId id
	
	LitExp lit  -> pprint lit
	
	AbsExp ps e -> 
	    if p > 0 then ppar s else s
	    where s = do de  <- pprint e 
			 dps <- pprintSpaces ps	
			 retsep [text "\\" <> dps <> text "->", de]
	
	AppExp e1 e2 ->
	    if p > 3 then ppar s else s
	    where s = do de2 <- pprintPA (4,a) l e2
			 de1 <- pprintPA (3,a) l e1
			 (return de1) <+> de2


	IfxAppExp e1 e2 e3 ->
	    if p > 2 then ppar s else s
	    where s = do let de1 = pprintPA (3,a) l e1
			 de2 <- doInfix $ pprint e2 
			 let de3 = pprintPA (2,a) l e3
			 sep [de1 <+> de2, de3]
	    
	IfExp e1 e2 e3 ->
	    if p > 0 then ppar s else s
	    where s = do de1 <- pprint e1
			 de2 <- pprint e2
			 de3 <- pprint e3
			 sep [textPD "if" <+> de1, 
			      textPD "then" <+> de2, 
			      textPD "else" <+> de3]	    

	CaseExp e as ->
	    if p > 0 then ppar s else s
	    where s = do de  <- pprint e
			 das <- mapM pprint as
			 let as' = nestPD 2 (retvcat das)
			 (textPD "case" <+> de <+> text "of") $||$ as'
			 
	AnnExp e t ->
	    if p > 1 then ppar s else s
	    where s = do let de = pprintPA (1,a) l e
			     dt = pprint t
				  -- pprintPA pa (ixLoc (ann t)) t
			 sep [de <+> text "::", dt]

	TAddExp e t ->
	    if p > 1 then ppar s else s
	    where s = do de <- pprintPA (1,a) l e
			 dt <- pprint t
			 retsep [de, text ":::", dt]

	TupExp es  -> do des <- pprintCommas es
			 return (text "(" <> des <> text ")")

	ListExp es -> do des <- pprintCommas es
	                 return (text "[" <> des <> text "]")
	

	LetExp ds e -> 
	    if p > 0 then ppar s else s
	    where s = do dds <- mapM pprint ds
			 de  <- pprint e
			 let dds' = nestPD 4 (retvcat dds)
			 vcat [textPD "let", dds', textPD "in " <+> de]
	    
	ListCompExp e ss -> do de <- pprint e
			       dss <- pprintCommas ss
			       textPD "[" <+> de <+> text "|" <+> 
				      dss <+> text "]"
	
	SecExpLeft  e1 e2 -> do de1 <- pprint e1
				de2 <- return $ text (showVarExp e2)
				return (text "(" <> de1 <> de2 <> text ")")
	
	SecExpRight e1 e2 -> do de2 <- pprint e2
				de1 <- return $ text (showVarExp e1)
				return (text "(" <> de1 <> de2 <> text ")")
	
	EnumExpFrom e1 -> do de1 <- pprint e1
			     return (text "[" <> de1 <> text "..]")
	
	EnumExpFromTo e1 e2 -> do de1 <- pprint e1
				  de2 <- pprint e2
				  return (text "[" <> de1 <> text ".." <> 
						      de2 <> text "]")
	
	EnumExpFromThen e1 e2 -> do de1 <- pprint e1
				    de2 <- pprint e2
				    return (text "[" <> de1 <> text "," <> 
						        de2 <> text "..]")

	EnumExpFromThenTo e1 e2 e3 -> do 
			de1 <- pprint e1
			de2 <- pprint e2
			de3 <- pprint e3
			return (text "[" <> de1 <> text "," <>
				            de2 <> text ".." <>
					    de3 <> text "]")
	PIfxAppExp x e1 e2 -> error ("p infix!: " ++ show e1 ++ "|" ++ 
						     show x ++ "|" ++ show e2)
	
instance PPrint CaseAlt where
    
    pprint (CaseAlt p e w) = do
         	let dp = pprint p
         	    de = pprint e
         	dw <- pprintWhere w
         	sep [dp <+> text "->", de, return $ nest 2 dw]

    pprint (CaseAltGuard p ges w) =  do
		dp    <- pprint p
		dges  <- mapM pprintGE ges
		dw    <- pprintWhere w
		dges' <- nestPD 2 (retvcat dges)
		retsep [dp, dges', dw]

	where pprintGE (g,e) = do dg <- pprint g
				  let de = pprint e
				  sep [textPD "|" <+> dg <+> text "->", de]
    
	

instance PPrint Stmt where
    
    pprint stmt = case stmt of
	
	GenStmt	 p e -> do let dp = pprint p 
			   de <- pprint e
			   dp <+> text "<-" <+> de
	
	QualStmt e   ->	pprint e

	LetStmt  ds  ->	do dds  <- mapM pprint ds
			   dds' <- nestPD 4 (retvcat dds)
			   retvcat [text "let", dds']
	

instance PPrint Sig where
    pprintPA _ l (Sig st id ctxt t) = do 
		dctxt <- if nullCtxt (item ctxt)
			 then return empty
			 else let dctxt = pprint ctxt
			      in  textPD' "(" <||> dctxt <||> 
				  textPD' ")" <++> textPD' "=>" <||> spacePD
		dt  <- pprint t
		let did = pprintAId (Ann (unkIx l) (item id))
		    stt = case st of Univ  -> "::"
			             Exist -> ":::"
		did <+> text stt <+> dctxt <|> dt

instance PPrint InstSig where
    pprintPA _ l (InstSig ctxt p) =
		if nullCtxt (item ctxt)
		then pprint p
		else let dctxt = pprint ctxt
		     in  textPD' "(" <||> dctxt <||> 
			 textPD' ")" <++> textPD' "=>" <++> pprint p
    
instance PPrint Ctxt where
    pprint (Ctxt ps)	= pprintCommas ps

instance PPrint Cnst where
    pprint (Cnst ps)	= pprintCommas ps

instance PPrint Prim where
    pprint (PredPrim p)   = pprint p 
    pprint (EqPrim t1 t2) = case (item t1, item t2) of	
	    (TupTyp [], TupTyp []) -> return (text "True")
	    (TupTyp [], ListTyp t) -> case item t of 
			TupTyp []  -> return (text "False")
			_	   -> general
	    _			   -> general
	where general = do let dt1 = pprint t1 
			       dt2 = pprint t2
			   dt1 <++> textPD "=" <++> dt2

instance PPrint Pred where
    pprint (Pred id t)    = do dt <- pprint t 
			       return (text (parId id) <> text "(" <> dt <> 
							  text ")")

instance PPrint Typ where

    pprintPA (p,a) l typ = case typ of
    
	VarTyp id   -> return (text id)
    
	ConTyp id   -> return (text id)

	TupTyp ts   -> do dts <- pprintCatCommas ts
			  return (text "(" <> dts <> text ")")
				
	ListTyp t   -> do dt <- pprint t
			  return (text "[" <> dt <> text "]")

	AppTyp t1 t2 -> s

	    where pre1  = if p > 1 then text "(" else empty
		  post1 = if p > 1 then text ")" else empty
		  non_arrow = do dt1 <- pprintPA (1,a) l t1
				 dt2 <- pprintPA (2,a) l t2
				 (return pre1) <|> dt1 <+> dt2 <|> post1

		  pre0  = if p > 0 then text "(" else empty
		  post0 = if p > 0 then text ")" else empty

		  s = case item t1 of
			AppTyp t3 t4  -> 
			 case item t3 of
			    ConTyp "->" -> do arr <- pprint t3
					      dt4 <- pprintPA (1,a) l t4
					      dt2 <- pprintPA (0,a) l t2
					      (return pre0) <|> dt4 <+> arr <+>
					       dt2 <|> post0
{-					       
			    ConTyp mt | mt `elem` 
					map nameStr (stdMultiTypeNames) ->
				emptyPD
-}			    
			    _		-> non_arrow
			_		-> non_arrow


instance PPrint Rule where
    pprint rule = do
	    (dl,da,dr) <- case rule of
		    SimpRule l r -> do dl <- pprint l
				       dr <- pprint r
				       return (dl, text "<==>", dr)
		    PropRule l r -> do dl <- pprint l
                                       dr <- pprint r
	                               return (dl, text "<==>", dr)
	    sep [textPD "rule" <+> dl <+> da, return dr]
    
instance PPrint Dat where
    pprint (Dat id ids cs) = do
	    let did = pprintAId id
	    dids <- mapM pprintAId ids
	    dcs  <- pprintBars cs
	    let vars = foldl (<+>) emptyPD dids
		cons = if null cs
		       then emptyPD
		       else textPD "=" <+> dcs
	    textPD "data" <++> did <++> vars <++> cons
	    
instance PPrint Con where
    pprint (Con id ts) = do let did = pprintAId id
				dts = pprintSpaces ts
			    did <++> dts

instance PPrint Lit where
    pprint (IntLit l s)   = return $ text s
    pprint (FloatLit l s) = return $ text s
    pprint (StrLit l)  = return $ text (show l)
    pprint (CharLit l) = return $ text (show l)

instance PPrint TSchm where
    pprint (TSchm c t) = do let dc = pprint c 
				dt = pprint t
			    if nullCtxt (item c)
			       then dt
			       else dc <++> textPD "=>" <++> dt

{-			    dc <- pprint c 
			    dt <- pprint t
			    return $ (if nullCtxt c
				      then empty
				      else dc <+> text "=>") <+> dt
-}

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
-- generic formatting functions

kw = keyword

showCommaList es = showSepList "," es
showSpaceList es = showSepList " " es
showBarList   es = showSepList "| " es
showSemiList  es = showSepList ";" es
showSepList sep es = sepList sep (map show es)

showAnnList es = sepList " " (map pprintAnn es)

semiList es = sepList ";" es
sepList sep es = insertSep es
     where insertSep []     = ""
           insertSep [x]    = x
	   insertSep (x:xs) = x ++ sep ++ insertSep xs



----------------------------------------


instance Show Prog  where show x = pprintPlain x
instance Show Exp   where show x = pprintPlain x
instance Show Pat   where show x = pprintPlain x
instance Show Def   where show x = pprintPlain x
instance Show Rule  where show x = pprintPlain x
instance Show Dec   where show x = pprintPlain x
instance Show Sig   where show x = pprintPlain x
instance Show Typ   where show x = pprintPlain x
instance Show TSchm where show x = pprintPlain x


--------------------------------------------------------------------------------

printProgHL = printProgHLRed
printProgHLRed   = printProgHLWithCol Red
printProgHLGreen = printProgHLWithCol Green
printProgHLBlue  = printProgHLWithCol Blue

-- pprint from the given loc, highlighting those in list
printProgHLWithCol :: Colour -> [Loc] -> Prog -> String
printProgHLWithCol c ls p@(Prog decs) = 
    case loc of 
	Nothing -> ""
	Just l  -> toStringOpts opts $ pplProg l p
    where 
{-    
	-- remove locations associated with extern decs
	-- (we can't print them anyway)
	ls = let getLoc dec = case item dec of
		  ExternDec (Left (Ann _ (Sig _ _ _ at))) -> [ixLoc (ann at)]
		  _ -> []
		 ext_ls = concatMap getLoc decs
		 ls' = [ l | l <- ols, l `notElem` ext_ls ] 
	     in  trace ("\next_ls: " ++ show ext_ls ++
			"\nols: " ++ show ols ++
			"\nls': " ++ show ls') ls'
-}  
	loc = let cpl = if singleton ls
		      then -- find a slightly more distant parent
			   let l  = head ls
			       lt = ltProg p
			   in  case grandParentLoc (head ls) lt of
				j@(Just ls) -> j
				Nothing	    -> findLocNode lt l 
		      else -- find the lowest common parent
			   commonParentLoc (ltProg p) ls
			   
	      in  case cpl of
		    Nothing -> Nothing
		    Just l  -> Just (nodeLoc l) 

	opts = defaultOptions { highlight = Only ls,
				highlightColour = c }


-- finds the nominated location in the given program, and pretty prints it and
-- its children
printProgLoc :: Loc -> Prog -> String
printProgLoc l p = toStringOpts opts $ pplProg l p
    where opts = defaultOptions -- { locations = Only [l] }

-- do a pprint from the given location
pplProg :: Loc -> Prog -> PrettyDoc
pplProg l prog@(Prog decs) 
    | l == noLoc = pprint prog
    | otherwise  = pplDecs decs
    where
	annl = ixLoc . ann

	combine = foldr (<||>) emptyPD
    
	pplDecs decs = combine (map pplDec decs)
	pplDec dec | l == annl dec = pprint dec
		   | otherwise     = case item dec of
		    ValDec defs	   -> pplDefs defs
		    AValDec s defs -> pplSig s <||> pplDefs defs
		    OverDec s defs -> pplSig s <||> pplDefs defs
		    InstDec s defs -> pplInstSig s <||> pplDefs defs
		    _		   -> emptyPD

	pplDefs defs = combine (map pplDef defs)
	pplDef def | l == annl def = pprint def
		   | otherwise = case item def of
			Def id ps rhs w -> pplId id <||> pplPats ps <||> 
					   pplRHS rhs <||> pplDecs w
	
	pplId id | l == annl id = pprintAId id
		 | otherwise	= emptyPD

	pplRHS (RHS1 ae)  = pplExp ae
	pplRHS (RHSG ges) = pplGEs ges

	pplGEs ges = combine (map pplGE ges)
	pplGE (g, e) = pplExp g <||> pplExp e


	pplExps es = combine (map pplExp es)
	pplExp :: Ann Exp -> PrettyDoc
	pplExp exp | l == annl exp = pprint exp
		   | otherwise     = case item exp of
		    VarExp id	    -> emptyPD
		    ConExp id	    -> emptyPD
		    AbsExp ps e1    -> pplPats ps <||> pplExp e1
		    AppExp e1 e2    -> pplExp e1 <||> pplExp e2
		    IfxAppExp e1 e2 e3 -> pplExp e1 <||> pplExp e2 <||>pplExp e3
		    IfExp  e1 e2 e3 -> pplExp e1 <||> pplExp e2 <||> pplExp e3
		    AnnExp e1 t     -> pplExp e1 <||> pplTSchm t
		    TAddExp e1 t    -> pplExp e1 <||> pplTSchm t
		    TupExp  es      -> pplExps es
		    ListExp es      -> pplExps es
		    SecExpLeft  e1 e2 -> pplExp e1 <||> pplExp e2
		    SecExpRight e1 e2 -> pplExp e1 <||> pplExp e2
		    EnumExpFrom       e1	-> pplExp e1
		    EnumExpFromTo     e1 e2	-> pplExp e1 <||> pplExp e2
		    EnumExpFromThen   e1 e2	-> pplExp e1 <||> pplExp e2
		    EnumExpFromThenTo e1 e2 e3  -> pplExp e1 <||> pplExp e2 <||>
						   pplExp e3
		    ListCompExp e ss -> pplExp e <||> pplStmts ss
		    CaseExp e as     -> pplExp e <||> pplAlts  as
		    LetExp ds e      -> pplExp e <||> pplDecs ds
		    
		    _ -> emptyPD

	pplAlts alts = combine (map pplAlt alts)
	pplAlt (CaseAlt p e w)	      = pplPat p <||> pplExp e  <||> pplDecs w
	pplAlt (CaseAltGuard p ges w) = pplPat p <||> pplDecs w <||> pplGEs ges

	pplStmts stmts = combine (map pplStmt stmts)
	pplStmt stmt | l == annl stmt = pprint stmt
		     | otherwise      = case item stmt of
		    GenStmt p e -> pplPat p <||> pplExp e
		    QualStmt e  -> pplExp e
	--	    LetStmt ds  -> ltDecs ds


	pplPats pats = combine (map pplPat pats)
	pplPat pat | l == annl pat = pprint pat
		   | otherwise     = case item pat of
		    VarPat id	 -> emptyPD
		    ConPat id ps -> pplPats ps
		    TupPat  ps	 -> pplPats ps
		    ListPat ps	 -> pplPats ps
		    _		 -> emptyPD

	pplSig sig | l == annl sig = pprint sig
		   | otherwise	   = case item sig of
		    Sig _ id ctxt typ -> pplId id <||> pplCtxt ctxt <||> 
					 pplTyp typ

	pplInstSig isig | l == annl isig = pprint isig
			| otherwise      = case item isig of
			 InstSig ctxt pred -> pplCtxt ctxt <||> pplPred pred

	pplTSchm tschm | l == annl tschm = pprint tschm
		       | otherwise       = case item tschm of
			TSchm ctxt typ -> pplCtxt ctxt <||> pplTyp typ

	pplTyps typs = combine (map pplTyp typs)
	pplTyp typ | l == annl typ = pprint typ
		   | otherwise	   = case item typ of
		    VarTyp _ -> emptyPD
		    ConTyp _ -> emptyPD
		    AppTyp t1 t2 -> pplTyp t1 <||> pplTyp t2
		    TupTyp ts -> pplTyps ts
		    ListTyp t -> pplTyp t

	pplCtxt ctxt | l == annl ctxt = pprint ctxt
		     | otherwise      = case item ctxt of 
			Ctxt preds -> pplPreds preds

	pplPreds preds = combine (map pplPred preds)
	pplPred pred | l == annl pred = pprint pred
		     | otherwise      = case item pred of
			Pred _ t -> pplTyp t
