--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


module Solver (module Solver)
where

import Chr
import Core
import Justify
import Pretty
import TD_RuleMatch
import OptFlags
import Misc

import Char
import IOExts
import List
import Maybe

--------------------------------------------------------------------------------
-- CHR store(s)

-- data SOpt = SInference | SEvidence | SDebug 	deriving (Show, Eq, Enum)

dbgOpt SDebug = True
dbgOpt _      = False
notDbgOpt = not . dbgOpt

type Goal  = ([UCons],[BCons])
type Token = [Index]                     -- index of rule number and atoms    
                                         -- assumption: sorted in non-decreasing order 
type TokenSet = [Token] 
-- ucons must be sorted to handle commutativity
-- this is an invariant which needs to be preserved
type Store = (Goal,[UCons], [BCons],TokenSet, [TVar], Subst)

instance (Show a, Show b, Show c, Show d, Show e, Show f) => 
	    Show (a,b,c,d,e,f) where
    show (a,b,c,d,e,f) = "(" ++ show a ++ "," ++ show b ++ "," ++ show c ++ 
			 "," ++ show d ++ "," ++ show e ++ "," ++ show f ++ ")"

storeUCons (_, ucs, _, _, _, _) = ucs
storeBCons (_, _, bcs, _, _, _) = bcs
storeToCons (_, ucs, bcs, _, _, _) = C ucs bcs

consToGoal = convGoal
convGoal   (C ucs bcs) = (ucs, bcs)
goalToCons (ucs, bcs)  = C ucs bcs



--------------------------------------------------------------------------------
-- solver proper

-- removes first occurence of x from ys
remove _ [] = []
remove x (y:ys)
   | x == y = ys
   | otherwise = y:(remove x ys)

removedups [] = []
removedups (x:xs)
    | elem x xs = removedups xs
    | otherwise = x:(removedups xs)


formatStore :: Store -> String
formatStore ((gucs,gbcs), ucs, bcs, tokens, tvs, _subst) =
         ("\n **Store content**" ++
          "\n Goal: " ++ (format gucs) ++ ";" ++ (format gbcs) ++
          "\n User: " ++ (format ucs) ++
          "\n Builtins: " ++ (format bcs) ++
          "\n Tokens: " ++ (show tokens) ++
          "\n Variables: " ++ (show tvs))


-- we only rename indices of user-constraints
rnIdx :: CHR -> TP CHR
rnIdx (Simp n uc bcs1 (ucs,bcs2)) = do uc' <- rnUC uc
                                       ucs' <- rnUCs ucs
                                       return (Simp n uc' bcs1 (ucs',bcs2))
rnIdx (MSimp n ucs1 bcs1 (ucs2,bcs2)) = do ucs1' <- rnUCs ucs1
                                           ucs2' <- rnUCs ucs2
                                           return (MSimp n ucs1' bcs1 (ucs2',bcs2))
rnIdx (Prop n i ucs1 bcs1 (ucs2,bcs2)) = do ucs1' <- rnUCs ucs1
                                            ucs2' <- rnUCs ucs2
                                            return (Prop n i ucs1' bcs1 (ucs2',bcs2))

rnUC :: UCons -> TP UCons
rnUC (UC i p ts x) = do i' <- newInteger
                        return (UC i' p ts x)

rnUCs :: [UCons] -> TP [UCons]
rnUCs ucs = foldl f (return []) ucs
            where
              f ucstp uc = do ucs <- ucstp
                              uc' <- rnUC uc
                              return (ucs ++ [uc'])   -- keep original order


--------------------------
-- Constraint functions --
--------------------------


----------------
-- CHR solver --
----------------
 
-- normalization of CHR store

-- todo: normalization of tokens!

normalize  :: Store -> TP Store
normalize (store@((u1,b1),u2,b2,tokens,tvars,s)) =
                let u3 = u1 ++ u2                 -- all user constraints
                    b3 = b1 ++ b2		  -- all built-in constraints
                    eqp = buildUP b3              -- build all equality problems
                in -- build mgu s of equality problems
                   -- be aware of projected variables !!
                   do case unify tvars idSubst eqp of
                       Succ s -> let u3' = removedups2 (apply s u3)
				     b3' = removedups 
					    (projBConsWithUnifier s tvars b3)
				 in  return (([],[]), u3', b3',tokens,tvars,s)
                       Err _  -> errorTP ("CHR normalization step failed:"
					  -- ++ "\n"++ (show store) 
					  ++ "\n" ++ (show eqp)) store  

-- as per normalize, but calls td_projBCons instead of projBConsWithUnifier,
-- because we want to preserve identities (eg. v12 = v12)
td_normalize  :: Store -> TP Store
td_normalize (store@((u1,b1),u2,b2,tokens,tvars,s)) =
                let u3 = u1 ++ u2                 -- all user constraints
                    b3 = b1 ++ b2		  -- all built-in constraints
                    eqp = buildUP b3              -- build all equality problems
                in -- build mgu s of equality problems
                   -- be aware of projected variables !!
                   do case unify tvars idSubst eqp of
                       Succ s -> let u3' = removedups2 (apply s u3)
				     b3' = removedups 
					    (td_projBCons s tvars b3)
				 in  return (([],[]), u3', b3', tokens,tvars,s)
                       Err _  -> errorTP ("CHR normalization step failed:" 
					  -- ++ "\n" ++ (show store) 
					  ++ "\n" ++ (show eqp)) store  


td_normalizeOpt :: SOpt -> Store -> TP (Maybe Store)
td_normalizeOpt opt (store@((u1,b1),u2,b2,tokens,tvars,s)) =
    let u3 = u1 ++ u2               -- all user constraints
        b3 = b1 ++ b2		    -- all built-in constraints
        eqp = buildUP b3            -- build all equality problems
    in -- build mgu s of equality problems
       -- be aware of projected variables !!
       do case unify tvars s {-idSubst-} eqp of
            Succ s' -> let su3 = if opt == SEvidence
				   then removedups3 (apply s' u3)
				   else removedups2 (apply s' u3)
			   sb3 = removedups (td_projBCons s' tvars b3)
		           (u3', b3') = 
			    case opt of
				SDebug     -> (u3, b3)
				SEvidence  -> (su3, sb3)
				SInference -> (su3, sb3)
		      in  return (Just (([],[]), u3', b3', tokens,tvars,s'))
            Err _  -> return Nothing 

goalToStore ((u1,b1),u2,b2,tokens,tvars,s) =
	    (([],[]),u1++u2,b1++b2,tokens,tvars,s)

buildMGU :: Store -> TP Subst
buildMGU (store@((u1,b1),u2,b2,_,tvars,_s)) =
    let b3  = b1 ++ b2
	eqp = buildUP b2
	tvars' = []
	mgu = unify tvars' idSubst eqp
    in  case mgu of
	    Succ s -> return s
	    Err _  -> return idSubst



-- reduction of store, if rule is applicable

reduce :: CHR -> Store -> E Store 
-- simprule applicable only if goal is empty
reduce simp@(Simp _ (UC _ p ts _) [] (bodyuc,bodybc)) (store@(([],[]),ucs,bcs,tokens,tvs,s)) = 
     let -- find a particular match
         findmatch :: [UCons] -> E ([Type],[Type],Index)
         findmatch [] = Err ""
         findmatch ((UC i p1 ts1 _):ucs)
           | p == p1   = if (ruleapp (ts,ts1)) then return (ts,ts1,i)
                         else findmatch ucs
           | otherwise = findmatch ucs
         ruleapp :: ([Type],[Type]) -> Bool
         ruleapp (ts1,ts2) =  matchTypes ts1 ts2
         removeAtom :: Index -> [UCons] -> [UCons]
         removeAtom _ [] = []
         removeAtom i ((a@(UC j p ts _)):ucs)
            | i == j    = ucs
            | otherwise = a:(removeAtom i ucs) 
     in case (findmatch ucs) of
         Succ (ts1,ts2,idx) -> return ((bodyuc,bodybc), 
                                       removeAtom idx ucs,
                                       bcs++(buildEP (zip ts1 ts2)),   -- add matching substitution
                                       tokens,                      
                                       tvs, s)
         Err _          -> Err "rule not applicable" 
-- multi-headed simp rules
reduce msimp@(MSimp _ ucs1 guard (bodyuc,bodybc)) (store@(([],[]),ucs,bcs,tokens,tvs,s)) =
     let -- find a particular match
         -- build permutation of lhs-constraints
         -- assumption, ucs1 is not empty
         permute [] = [[]] 
         permute (x:xs) = [ take i ys ++ [x] ++ drop i ys 
                                | ys <- permute xs, i <- [0..length ys] ] 
         findmatch':: [UCons] -> [UCons] -> ([Type],[Type],[Index]) -> E ([Type],[Type],[Index])
         findmatch' ucs1 ucs (ts1,ts2,idxs) =
              let iterate [] = Err ""
                  iterate (ucs':ucss) = case (findmatch ucs' ucs (ts1,ts2,idxs)) of
                                         Succ s -> Succ s
                                         _      -> iterate ucss
              in iterate (permute ucs1) 
            
         findmatch :: [UCons] -> [UCons] -> ([Type],[Type],[Index]) -> E ([Type],[Type],[Index])
         findmatch [] _ (ts1,ts2,idxs) = if ruleapp(ts1,ts2)
                                         then return (ts1,ts2,idxs)
                                         else Err ""
         findmatch _ [] _ = Err ""
         findmatch (cs@((UC i1 p1 t1 _):cs1)) ((UC i2 p2 t2 _):ucs) (acc@(ts1,ts2,idxs))
           | p1 == p2  =  do if ruleapp (t1++ts1, t2++ts2)
                              then do case (findmatch cs1 ucs (t1++ts1,t2++ts2,i2:idxs)) of
                                       Succ res -> Succ res
                                       Err _    -> findmatch cs ucs acc
                              else findmatch cs ucs (ts1,ts2,idxs)    
                                      -- p1 p2 don't match, continue 
                             -- note that the above code reflects our assumption that
                             -- commutativity of atoms is handled by permuting heads in prop-rules  
           | otherwise = findmatch cs ucs acc
         ruleapp :: ([Type],[Type]) -> Bool
         ruleapp (ts1,ts2) =  matchTypes ts1 ts2 
         removeAtoms :: [Index] -> [UCons] -> [UCons]
         removeAtoms _ [] = []
         removeAtoms is ((a@(UC j p ts _)):ucs)
            | elem j is = removeAtoms is ucs 
            | otherwise = a:(removeAtoms is ucs) 
     in case (findmatch' ucs1 ucs ([],[],[])) of 
         Succ (ts1, ts2, idxs) -> return ((bodyuc,bodybc), 
                                         removeAtoms idxs ucs,
                                         bcs++(buildEP (zip ts1 ts2)), 
                                            -- add matching substitution
                                         tokens, tvs, s)
         Err _                 -> Err "rule not applicable" -- 
-- proprule only applicable if store is empty
reduce prop@(Prop _ i ucs1 [] (bodyuc,bodybc)) (store@(([],[]),ucs,bcs,tokens,tvs,s)) =
     let -- find a particular match
         -- build permutation of lhs-constraints
         -- assumption, ucs1 is not empty
         permute [] = [[]] 
         permute (x:xs) = [ take i ys ++ [x] ++ drop i ys 
                                | ys <- permute xs, i <- [0..length ys] ] 
         findmatch':: [UCons] -> [UCons] -> ([Type],[Type],[Index]) -> E ([Type],[Type],[Index])
         findmatch' ucs1 ucs (ts1,ts2,idxs) =
              let iterate [] = Err ""
                  iterate (ucs':ucss) = case (findmatch ucs' ucs (ts1,ts2,idxs)) of
                                         Succ s -> Succ s
                                         _      -> iterate ucss
              in iterate (permute ucs1) 
            
         findmatch :: [UCons] -> [UCons] -> ([Type],[Type],[Index]) -> E ([Type],[Type],[Index])
         findmatch [] _ (ts1,ts2,idxs) = if ruleapp(ts1,ts2,idxs)
                                         then return (ts1,ts2,idxs)
                                         else Err ""
         findmatch _ [] _ = Err ""
         findmatch (cs@((UC i1 p1 t1 _):cs1)) ((UC i2 p2 t2 _):ucs) (acc@(ts1,ts2,idxs))
           | p1 == p2  =  --error ("\n (p1,t1)=" ++ (show (p1,t1)) ++ "\n (p2,t2)=" ++ (show (p2,t2))) {-
                          do if matchTypes (t1++ts1) (t2++ts2)
                              then do case (findmatch cs1 ucs (t1++ts1,t2++ts2,i2:idxs)) of
                                       Succ res -> Succ res
                                       Err _    -> findmatch cs ucs acc
                              else findmatch cs ucs (ts1,ts2,idxs)    -- p1 p2 don't match, continue -- -}
                             -- note that the above code reflects our assumption that
                             -- commutativity of atoms is handled by permuting heads in prop-rules  
           | otherwise = findmatch cs ucs acc
         ruleapp :: ([Type],[Type],[Index]) -> Bool
         ruleapp (ts1,ts2,idxs) 
           | notElem (sort idxs) tokens                  -- rule index already included
                        = matchTypes ts1 ts2
           | otherwise                  = False
     in -- error ("\n ucs1=" ++ (show ucs1) ++ "\n ucs=" ++ (show ucs)) {-
       case (findmatch' ucs1 ucs ([],[],[i])) of         -- don't forget the index of the prop rule
         Succ (ts1, ts2, idxs) -> return ((bodyuc,bodybc), ucs,
                                         bcs++(buildEP (zip ts1 ts2)),   -- add matching substitution
                                         (sort idxs):tokens,             -- add tokens
                                         tvs, s)
         Err _                 -> Err "rule not applicable" -- -}


reduce _ _ = error "This case never happens"



----------------------------------------
-- The version of "reduce" used by the type debugger functions.
-- Key differences:
--  - takes a substitution as an additional argument
--  - does not expect store to be normalized
--  - instead, applies subst at when attempting to match
--  - moreover, the "matching substitution" (unifying the matching types) is
--    created in terms of the original types - without the substitution applied

-- the JustPolicy determines how new justifications are determined

td_reduce :: (SOpt, JustProp) -> CHR -> Store -> E Store 

-- simprule applicable only if goal is empty
td_reduce (opt, jp) simp@(Simp cinf yuc@(UC ucix p ts j1) [] (bodyuc,bodybc)) 
	      (store@(([],[]),ucs,bcs,tokens,tvs,s)) =
     let -- find a particular match
         findmatch :: [UCons] -> E ([Type],[Type],Index,Just)
         findmatch [] = Err "td_reduce: no ucons!"
         findmatch (uc@(UC i p1 ts1 _):ucs)
           | p == p1   = let (b,j) = ruleapp (ts,ts1)
			 in  if b then return (ts, ts1, i, j)
				  else findmatch ucs
           | otherwise = findmatch ucs
	   
         ruleapp :: ([Type],[Type]) -> (Bool, Just)
         ruleapp (ts1,ts2) =  let mt  = matchRule (notDbgOpt opt) s ts1 ts2 bcs
				  mtb = ruleMatches mt
				  mtj = if mtb then matchJust mt else noJust
				  mt' = trace ("\nsimp: " ++ show simp ++
					       "\nstore: " ++ show store ++
					       "\ntss: " ++ show (ts1, ts2) ++
					       "\nmt: " ++ show mt) (mtb, mtj)

			      in  (mtb, mtj)
	 
         removeAtom :: Index -> [UCons] -> ([UCons], Just)
         removeAtom _ [] = ([], noJust)
         removeAtom i ((a@(UC j p ts jst)):ucs)
            | i == j    = (ucs, jst)
            | otherwise = let (as, j) = removeAtom i ucs
			  in  (a:as, j)

	 isDefRule = isValCHR simp
	 isOverInf = isOverloadInf (chrType simp)
	 
     in case findmatch ucs of

	-- removeAtom should also return the justs on the removed cons... we 
	-- then add them to the constraints in the body (bodyuc, bodybc)
	-- justifications:
	--  j1: rule
	--  j2: user constraints
	--  j3: builtin constraints (implicants)
         Succ (ts1,ts2,idx,j3) -> 
	    let (ucs', j2) = removeAtom idx ucs
		(fu, fb, j) = 
		    if isDefRule 
		       then if isLocalJust jp
			       then (newUCJust j2, newBCJust j2, j2)
			       else (addUCJust j2, addBCJust j2, j2)
		       else case opt of
			     SDebug -> let j' = addJust j3 (addJust j2 j1)
				       in  (newUCJust j', newBCJust j', j')
				       
			     SEvidence -> let j' = j2 -- addJust j2 j1
					  in (addUCJust j', addBCJust j', j')
{-				       
			     SEvidence -> if isOverInf
					     then (newUCJust j2,newBCJust j2,j2)
					     else (catUCJust j2,catBCJust j2,j2)
-}					     

			     SInference -> let j' = noJust
					   in (newUCJust j', newBCJust j', j')
							    
		bodyuc' = map fu bodyuc
		bodybc' = map fb bodybc
						    
		matchSubst = buildEP (zip ts1 ts2)

		msg = "\nsimp. rule FIRED!" ++
		      "\n    rule: " ++ pprintAnn simp  ++
		      "\n      on: " ++ pprintAnn (ucs `without` ucs') ++
		      "\n   store: "++ show ucs ++ show bcs ++
		      "\n  adding: " ++ pprintAnn bodyuc' ++ 
					pprintAnn (bodybc' ++ matchSubst)++
		      "\n just: " ++ show j1 ++ "|" ++ show j2 ++ "|" ++ show j3-- ++
--		      "\nstore': " ++ show (((bodyuc',bodybc'), ucs', bcs ++
--					matchSubst, tokens, tvs))
	    in -- trace msg $
		return ((bodyuc',bodybc'), ucs', matchSubst ++ bcs, tokens, tvs,s)
         Err _          -> Err "rule not applicable"



-- proprule only applicable if store (goal?!) is empty
td_reduce (opt, jp) prop@(Prop _ i ucs1 [] (bodyuc,bodybc)) 
	      (store@(([],[]),ucs,bcs,tokens,tvs,s)) =
     let -- find a particular match
         -- build permutation of lhs-constraints
         -- assumption, ucs1 is not empty
         permute [] = [[]] 
         permute (x:xs) = [ take i ys ++ [x] ++ drop i ys 
                                | ys <- permute xs, i <- [0..length ys] ] 

         findmatch':: [UCons] -> [UCons] -> ([Type],[Type],[Index]) -> 
					  E ([Type],[Type],[Index],Just)
         findmatch' ucs1 ucs (ts1,ts2,idxs) =
              let iterate [] = Err ""
                  iterate (ucs':ucss) =			-- initially noJust
		    case (findmatch ucs' ucs (ts1,ts2,idxs,noJust)) of
                            Succ s -> Succ s
                            _      -> iterate ucss
              in iterate (permute ucs1) 
            

         findmatch :: [UCons] -> [UCons] -> ([Type],[Type],[Index],Just) -> 
					  E ([Type],[Type],[Index],Just)
         findmatch [] _ (ts1,ts2,idxs,j) = 
	    let (b,j') = ruleapp (ts1, ts2, idxs)
	    in if b then return (ts1,ts2,idxs,addJust j j')
		    else Err ""

         findmatch _ [] _ = Err ""
         findmatch (cs@((UC i1 p1 t1 j1):cs1)) ucz@((UC i2 p2 t2 j2):ucs) 
		   (acc@(ts1,ts2,idxs,j))
           | p1 == p2  = let ts1' = t1 ++ ts1
			     ts2' = t2 ++ ts2
			     j'   = addJust j2 (addJust j1 j)
			     fmMore = findmatch cs1 ucs (ts1',ts2',i2:idxs,j')
			    -- refine the condition below
			     dbMatch  = True 
					-- matchTypes ts1' (apply mgu ts2')
			     canMatch = notDbgOpt opt || dbMatch -- more convoluted test
			 in  if canMatch
				then case fmMore of
					Succ res -> Succ res
					Err _    -> findmatch cs ucs acc
			       -- p1 p2 don't match, continue
				else  findmatch cs ucs acc
	-- note that the above code reflects our assumption that
        -- commutativity of atoms is handled by permuting heads in prop-rules  
           | otherwise = findmatch cs ucs acc

         ruleapp :: ([Type],[Type],[Index]) -> (Bool, Just)
         ruleapp (ts1,ts2,idxs) 
					-- rule index already included
           | notElem (sort idxs) tokens = 
			    let mt  = matchRule (notDbgOpt opt) s ts1 ts2 bcs
				mtb = ruleMatches mt
				mtj = if mtb then matchJust mt else noJust
				mt' = trace ("\nprop: " ++ show prop ++
					       "\nstore: " ++ show store ++
					       "\ntss: " ++ show (ts1, ts2) ++
					       "\nflag: " ++ show (notDbgOpt opt) ++
					       "\nb: " ++ show (notDbgOpt opt)++
					       "\nmt: " ++ show mt) (mtb, mtj)
					       
			    in  (mtb,mtj)
           | otherwise = (False, noJust)
	   
	-- i: justification for rule
	-- j: justification on matching ucons (and rule?!)
	   
     in -- don't forget the index of the prop rule
	case (findmatch' ucs1 ucs ([],[],[i])) of 
	
         Succ (ts1, ts2, idxs, j') ->
	    let -- j = noJust -- some function of idxs and prop
		j = case opt of 
		       SDebug     -> j' `addJust` (locToJust i)
		       SEvidence  -> j' -- `catJust` (locToJust i)
		       SInference -> noJust
		
		bodyuc1 = map (newUCJust j) bodyuc
		bodybc1 = map (newBCJust j) bodybc
				
		matchSubst = buildEP (zip ts1 ts2)
		(bodyuc2, bodybc2) =  -- (bodyuc1, bodybc1)
			remRedundant' (ucs, matchSubst ++ bcs) 
				      (bodyuc1, bodybc1)
			       
			       
	        msg = "prop FIRED\n  rule: " ++ pprintAnn prop ++
		      "\n  with subst: " ++ show matchSubst ++
		      "\n  adding: "++ pprintAnn (bodyuc2, bodybc2) ++
		      "\n  to store: " ++ 
				     pprintPlainList (storeUCons store) ++", "++
				     pprintPlainList (storeBCons store) ++
		      "\n  i: " ++ show i ++
		      "\n  j: " ++ show j ++ "\n"
			      
			      -- msg = "prop"
	    in -- trace msg $
		return ((bodyuc2, bodybc2), ucs,      
                        matchSubst ++ bcs, (sort idxs):tokens, tvs, s) 

         Err _  -> {- trace ("prop not applicable:\n\t " ++ show prop ++
			  "\nto store:\n\t" ++ show (apply mgu ucs) ++
			  "\nmgu: \n\t" ++ show mgu)   $ -} Err "rule not applicable" 


td_reduce (opt, jp) msimp@(MSimp _ ucs1 [] (bodyuc,bodybc)) 
	      (store@(([],[]),ucs,bcs,tokens,tvs,s)) =
     let -- find a particular match
         -- build permutation of lhs-constraints
         -- assumption, ucs1 is not empty
         permute [] = [[]] 
         permute (x:xs) = [ take i ys ++ [x] ++ drop i ys 
                                | ys <- permute xs, i <- [0..length ys] ] 

         findmatch':: [UCons] -> [UCons] -> ([Type],[Type],[Index]) -> 
					  E ([Type],[Type],[Index],Just)
         findmatch' ucs1 ucs (ts1,ts2,idxs) =
              let iterate [] = Err ""
                  iterate (ucs':ucss) =			-- initially noJust
		    case (findmatch ucs' ucs (ts1,ts2,idxs,noJust)) of
                            Succ s -> Succ s
                            _      -> iterate ucss
              in iterate (permute ucs1) 
            

         findmatch :: [UCons] -> [UCons] -> ([Type],[Type],[Index],Just) -> 
					  E ([Type],[Type],[Index],Just)
         findmatch [] _ (ts1,ts2,idxs,j) = let (b,j') = ruleapp (ts1, ts2) -- , idxs)
					   in if b
					       then return 
						     (ts1,ts2,idxs,addJust j j')
					       else Err ""
         findmatch _ [] _ = Err ""
         findmatch (cs@((UC i1 p1 t1 j1):cs1)) ucz@((UC i2 p2 t2 j2):ucs) 
		   (acc@(ts1,ts2,idxs,j))
           | p1 == p2  = let ts1' = t1 ++ ts1
			     ts2' = t2 ++ ts2
			     j'   = addJust j2 (addJust j1 j)
			     fmMore   = findmatch cs1 ucs (ts1',ts2',i2:idxs,j')
			     -- refine this condition
			     dbMatch = True
			     canMatch = notDbgOpt opt || dbMatch
			 in  if canMatch
				then case fmMore of
					Succ res -> Succ res
					Err _    -> findmatch cs ucs acc
			       -- p1 p2 don't match, continue
				else  findmatch cs ucs acc
	-- note that the above code reflects our assumption that
        -- commutativity of atoms is handled by permuting heads in prop-rules  
           | otherwise = findmatch cs ucs acc

         ruleapp :: ([Type],[Type]) -> (Bool, Just)
         ruleapp (ts1,ts2) =
		let mt  = matchRule (notDbgOpt opt) s ts1 ts2 bcs
		    mtb = ruleMatches mt
		    mtj = if mtb then matchJust mt else noJust
		    msg = "\nrule: " ++ show msimp ++
					"\nmt: " ++ show mt ++
					"\nmtb: " ++ show mtb ++
					"\nmtj: " ++ show mtj

		in  -- trace msg 
		    (mtb,mtj)
	   

         removeAtom :: Index -> [UCons] -> ([UCons], Just)
         removeAtom _ [] = ([], noJust)
         removeAtom i ((a@(UC j p ts jst)):ucs)
            | i == j    = (ucs, jst)
            | otherwise = let (as, j) = removeAtom i ucs
			  in  (a:as, j)

	 removeAtoms :: [Index] -> [UCons] -> ([UCons], Just)
	 removeAtoms []       ucs = (ucs, noJust)
	 removeAtoms (ix:ixs) ucs = let (ucs', j) = removeAtom ix ucs
					(ucs'', j') = removeAtoms ixs ucs'
				    in  (ucs'', addJust j j')
	   
     in -- don't forget the index of the prop rule
	case (findmatch' ucs1 ucs ([],[],[])) of 
	
         Succ (ts1, ts2, idxs, j3) 
			-> let (ucs', j2) = removeAtoms idxs ucs
			       j1 = ucJust (head ucs1)			       

			       (fu, fb, j) = 
				    let j' = addJust j3 (addJust j2 j1)
				    in (newUCJust j', newBCJust j', j')
							    
			       bodyuc' = map fu bodyuc
			       bodybc' = map fb bodybc


			       matchSubst = buildEP (zip ts1 ts2)
			       msg = "msimp fired\n  rule: " ++ show msimp ++
				     "\n  with subst: " ++ pprintAnn matchSubst ++
				     "\n  adding: "++ pprintAnn (bodyuc', bodybc') ++
				     "\n  to store: " ++ 
				     pprintAnn (storeUCons store) ++ " " ++
				     pprintAnn (storeBCons store) ++
				     "\n  j: " ++ show j ++
				     "\n"
			   in --trace msg $
			       return ((bodyuc', bodybc'), ucs',
				      -- add matching substitution
                                         bcs ++ matchSubst,
				      -- add tokens
                                         (sort idxs):tokens, tvs, s) 
         Err _                 -> Err "rule not applicable" 


td_reduce _ r s = error ("td_reduce:\n" ++ show r ++ "\n")


-- takes two constraints, C1 and C2, and removes ucons from C2 which are 
-- redundant with respect to C1
-- call remRedundant with the current store and pending rhs constraints before
-- adding them to the store
-- this enforces set semantics for the solver
remRedundant' :: ([UCons], [BCons]) -> ([UCons], [BCons]) -> ([UCons], [BCons])
remRedundant' c1 c2 = consToGoal $ remRedundant (goalToCons c1) (goalToCons c2)

remRedundant :: Constraint -> Constraint -> Constraint
remRedundant c1 c2 =
    let bcs2 = consToBs c2
	bcs1 = consToBs c1
	bcs  = bcs2 ++ bcs1
	mmgu = unify [] idSubst (buildUP bcs)
    in  if isErr mmgu
	   then c2   -- will fail regardless of ucs
	   else let mgu = unSucc mmgu
		    ucs2  = consToUs c2
		    ucs1  = apply mgu (consToUs c1)
		    ucs2' = apply mgu ucs2
		    ucs2''= [ u | (u, u') <- zip ucs2 ucs2', u' `notIn` ucs1 ]
		    c2' = (usToCons ucs2'') `cand` (bsToCons bcs2)
		    msg = "\nc2: " ++ show c2 ++ "\nc2': " ++ show c2' ++
			  "\nucs1: " ++ pprintAnn ucs1 ++ "\nucs2': " ++ pprintAnn ucs2'
			  ++ "\nmgu: " ++ show mgu
		in -- trace msg 
		     c2'
    where -- elem, but ignores indices
	  notIn uc ucs = not $ any (uc`eq`) ucs
	  eq (UC _ p ts j) (UC _ p' ts' j') = 
	    p == p' && ts == ts' && ((unJust j) `sameSet` (unJust j'))
						


-- select one applicable rule, if any
-- rename rule before application
-- report applicable rule
reduces :: [CHR] -> Store -> TP (Store,Maybe CHR)
reduces [] s = return (s,Nothing)
reduces (chr:chrs) s = do chr' <- rename chr            -- rename type variables
                          chr'' <- rnIdx chr'           -- rename indices
                          case (reduce chr'' s) of
                            Succ s' -> return (s',Just chr'')
                            Err _   -> reduces chrs s



td_reduces:: (SOpt, JustProp) -> [CHR] -> Store -> TP (Store,Maybe CHR)
td_reduces opts [] s = return (s,Nothing)
td_reduces opts (chr:chrs) s = do 
    chr' <- rename chr
    chr'' <- rnIdx chr'
    let res = td_reduce opts chr'' s
    case res of
	 Succ s' -> return (s',Just chr'')
	 Err _   -> td_reduces opts chrs s


{-
solve :: [CHR] -> Goal -> TP Store
solve chrs (goal@(guc,gbc)) = 
        let init = (goal,[],[],[],removedups ((fv guc) ++ (fv gbc))) -- get rid of duplicate variables
            iterate :: Store -> TP Store
            iterate st = do st' <- normalize st
                            (st'',bchrs) <- reduces chrs st'
                            case bchrs of
                               Nothing     -> normalize st''
                               Just appchr -> iterate st''
        in iterate init
-}
solve = deriveInference LocalJust

td_solve :: [CHR] -> Goal -> TP Store
td_solve chrs goal@(guc, gbc) =
    let init = (goal,[],[],[],removedups ((fv guc) ++ (fv gbc)), idSubst)
        iterate :: Store -> TP Store
        iterate st = do st' <- td_normalize st
                        (st'',bchrs) <- reduces chrs st'
                        case bchrs of
                            Nothing     -> td_normalize st''
                            Just appchr -> iterate st''
    in iterate init
   

deriveDebug     = derive SDebug
deriveEvidence  = derive SEvidence
deriveInference = derive SInference

-- performs a derivation
derive :: SOpt -> JustProp -> [CHR] -> Goal -> TP Store
derive opt jp chrs (goal@(guc,gbc)) =
    let init = (goal,[],[],[],removedups ((fv guc) ++ (fv gbc)), idSubst)
        iterate :: Store -> TP Store
        iterate st = do 
	    st0 <- td_normalizeOpt opt st 
	    case st0 of
		Nothing	 -> return (goalToStore st)
		Just st1 -> do
		    (st2,bchrs) <- td_reduces (opt, jp) chrs st1
		    case bchrs of
			Nothing     -> return st2
			Just appchr -> iterate st2
   in  iterate init


-- same as normal solver, except it also takes the list of variables we're
-- interested in solving for (instead of the free var.s in the goal)
solveFor :: [CHR] -> Goal -> [TVar] -> TP Store
solveFor chrs (goal@(guc,gbc)) vs =
        let init = (goal,[],[],[],vs,idSubst)
            iterate :: Store -> TP Store
            iterate st = do st' <- td_normalize st
                            (st'',bchrs) <- reduces chrs st'
                            case bchrs of
                               Nothing     -> td_normalize st''
                               Just appchr -> iterate st''
        in iterate init   

--------------------------------------------------------------------------------
-- simplifies the RHS of the chrs (we ignore justs.)
-- can't fail (ie. Err)
simplifyCHRs :: [CHR] -> TP [CHR]
simplifyCHRs chrs = do 
    let (inf, rest) = List.partition (isDefCHR) chrs
    let fvs = map (fv . chrLhs) inf
    new <- mapM (simp chrs) (zip fvs inf)
    return chrs -- (rest ++ new) 
    
    where simp chrs (vs, chr) = do
		n <- newInteger
		let rhs = chrRhs chr
		    slv = solveFor chrs rhs vs
		    c   = case (unTP slv) n of
				Err e        -> falseCons
				Succ (_, st) -> storeToCons st
		    rhs' = (consToUs c, consToBs c)
		    chr' = chrNewRhs rhs' chr
		return chr'
	  
--------------------------------------------------------------------------------

-- takes a list of flags and returns the inital solver options
genSolverOpt :: [Flag] -> IO SOpt
genSolverOpt fs = do
    let ls = concatMap (\f -> case f of SolverOpt l -> [l]; _ -> []) fs
	l  = if null ls then 2 else head ls
        
    l' <- if l < 0 || l > 2 
	     then putStr ("WARNING: solver mode " ++ show l ++ " undefined" ++ 
			  "  (defaulting to mode 0)\n") >>
		  return 0
	     else return l 
    return (toEnum l')

