--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


module SolverIO 
		(derive, solve, 
		 FinalStore(..), mixedStoreToCons, fromSat,
		 storeToCons, convGoal, solveFor, consToGoal,
		 runTest)
where

import Chr
--import qualified Core as Std
import qualified UCons   as Std
import qualified Builtin as Std
import qualified Types   as Std
import qualified Constraint as Std
import qualified ConsReason as Std
import Name

import Justify
import Pretty
import OptFlags
import Misc

import TD_Monad

import Char
import IOExts
import List
import Maybe

import qualified Solver 

--------------------------------------------------------------------------------
-- CHR store(s)


dbgOpt SDebug = True
dbgOpt _      = False
notDbgOpt = not . dbgOpt

type StdGoal = ([Std.UCons], [Std.BCons])
type Goal  = ([UCons],[BCons])
type Token = [Index]            -- index of rule number and atoms    
                                -- assumption: sorted in non-decreasing order 

type TokenSet = [Token] 
-- ucons must be sorted to handle commutativity
-- this is an invariant which needs to be preserved

-- VarMap contains the variables we want to project on (in modes 0/1)
type Store = (Goal,[UCons], [BCons],TokenSet, [TVar], VarMap)

type MixedStore = 
    (Goal, [Std.UCons], [Std.BCons], TokenSet, [TVar], VarMap)

data FinalStore = Sat MixedStore
		| Unsat MixedStore

mixedStoreToCons (_, ucs, bcs, _, _, _) = Std.C ucs bcs

fromSat (Sat s) = s

formatStore Nothing = "unsatisfiable store!"
formatStore (Just ((gucs,gbcs), ucs, bcs, tokens, _, _)) =
         ("\n **Store content**" ++
          "\n Goal: " ++ (show gucs) ++ ";" ++ (show gbcs) ++
          "\n User: " ++ (show ucs) ++
          "\n Builtins: " ++ (show bcs) ++
          "\n Tokens: " ++ (show tokens)) --  ++
          -- "\n Variables: " ++ (show tvs))

storeUCons (_, ucs, _, _, _, _) = ucs
storeBCons (_, _, bcs, _, _, _) = bcs
storeToCons (_, ucs, bcs, _, _, _) = Std.C ucs bcs


consToGoal = convGoal
convGoal   (Std.C ucs bcs) = (ucs, bcs)
{-
goalToCons (ucs, bcs)  = C ucs bcs
-}

--------------------------------------------------------------------------------
-- solver proper

{-
formatStore :: Store -> String
formatStore ((gucs,gbcs), ucs, bcs, tokens, tvs, _subst) =
         ("\n **Store content**" ++
          "\n Goal: " ++ (format gucs) ++ ";" ++ (format gbcs) ++
          "\n User: " ++ (format ucs) ++
          "\n Builtins: " ++ (format bcs) ++
          "\n Tokens: " ++ (show tokens) ++
          "\n Variables: " ++ (show tvs))
-}

-- we only rename indices of user-constraints
rnIdx :: CHR -> TD CHR
rnIdx (Simp n uc bcs1 (ucs,bcs2)) = do uc' <- rnUC uc
                                       ucs' <- rnUCs ucs
                                       return (Simp n uc' bcs1 (ucs',bcs2))
rnIdx (MSimp n ucs1 bcs1 (ucs2,bcs2)) = do 
				ucs1' <- rnUCs ucs1
                                ucs2' <- rnUCs ucs2
				return (MSimp n ucs1' bcs1 (ucs2',bcs2))
rnIdx (Prop n i ucs1 bcs1 (ucs2,bcs2)) = do 
				ucs1' <- rnUCs ucs1
                                ucs2' <- rnUCs ucs2
                                return (Prop n i ucs1' bcs1 (ucs2',bcs2))

rnUC :: Std.UCons -> TD Std.UCons
rnUC (Std.UC i p ts x) = do i' <- next
			    return (Std.UC i' p ts x)

rnUCs :: [Std.UCons] -> TD [Std.UCons]
rnUCs = mapM rnUC


--------------------------------------------------------------------------------
-- internal versions of types and constraints (with IORefs)

type TypeRef = IORef Type
type TypeAndRef = (Type, TypeRef)

type TVar = (Name, Loc)
type TCon = Name

-- TApp needs both the original copy of each argument, as well as pointers to
-- them (so that when we unify variables with one of the args they points to
-- the same instance of the arg, not a copy!)

data Type = Var TVar TypeRef
	  | TyCon TCon
	  | TApp TypeAndRef TypeAndRef

data BCons = Eq TypeAndRef TypeAndRef Just
data UCons = UC Index PredSym [TypeAndRef] Just

data RCHR = RSimp  CHRInfo UCons [BCons] ([UCons], [BCons])
	  | RProp  CHRInfo Index [UCons] [BCons] ([UCons], [BCons])
	  | RMSimp CHRInfo [UCons] [BCons] ([UCons], [BCons])

instance Eq Type where
    (Var (n1,_) _) == (Var (n2,_) _) = n1 == n2
    (TyCon c1)     == (TyCon c2)     = c1 == c2
    (TApp t1 t2)   == (TApp t1' t2') = t1 == t1' && t2 == t2'
    _		   == _		     = False


instance Std.TC Type where
    fv (Var v _) = [v]
    fv (TyCon _) = []
    fv (TApp (t1,_) (t2,_)) = nub $ Std.fv t1 ++ Std.fv t2


instance Std.TC BCons where
    fv (Eq (t1,_) (t2,_) _) = Std.fv t1 ++ Std.fv t2

instance Std.TC UCons where
    fv (UC _ _ ts _) = concatMap (Std.fv . fst) ts

instance Show UCons where
    show uc = show (stripUC uc)

instance Show BCons where
    show bc = show (stripBC bc)

instance Show Type where
    show t = show (stripType (t,undefined))

instance Justifiable BCons where
    unJust (Eq  _ _ j) = unJust j
    reJust j (Eq a b _) = Eq a b j


newUCJust j (UC i p ts _) = UC i p ts j
newBCJust j (Eq t1 t2 _)  = Eq t1 t2 j

addUCJust j (UC i p ts k) = UC i p ts (addJust j k)
addBCJust j (Eq t1 t2 k)  = Eq t1 t2 (addJust j k)

nubUCJust (UC i p t j) = UC i p t (nubJust j)
nubBCJust (Eq t1 t2 j) = Eq t1 t2 (nubJust j)



buildEP = zipWith (\t1 t2 -> Eq t1 t2 noJust)

isValRCHR (RSimp i _ _ _) = isValCHR (Simp i undefined undefined undefined)

ucTypes (UC _ _ ts _) = ts
bcTypes (Eq t1 t2 _)  = [t1,t2]

setPropIndex x (RProp i _ ucs bcs rhs) = RProp i x ucs bcs rhs


formatTypeRef :: TypeRef -> IO String
formatTypeRef r = do
    t <- readIORef r
    case t of
        Var n r    -> do r' <- deRef r
                         t' <- readIORef r'
                         case t' of
                            Var n _ -> return (show (fst n))
                            _       -> do s <- formatTypeRef r'
                                          --return (show (fst n) ++ ")>" ++ s)
					  return (show (fst n))
        TyCon n    -> return (show n)
        TApp (_,r1) (_,r2) -> do s1 <- formatTypeRef r1
				 s2 <- formatTypeRef r2
				 return ("(" ++ s1 ++ " " ++ s2 ++ ")")
 


----------------------------------------

type VarMap = [((Name, Loc), TypeRef)]

emptyVarMap = []
insertVarMap = (:)


-- it's lookup, but ignores the variable location number, as it may vary for 
-- e.g. renamed vars
lookupVarMap (n,_) vm = let found = [ r | ((m,_),r) <- vm, n == m ]
			in  if null found then Nothing
					  else return (head found)
    

----------------------------------------

deRef :: TypeRef -> IO TypeRef
deRef r = do 
    t <- readIORef r
    case t of
	Var _ q | q /= r -> deRef q
	_   -> return r

-- goes into sub-types
deRefAll :: TypeRef -> IO TypeRef
deRefAll r = do
    t <- readIORef r
    case t of

	TApp (t1,r1) (t2,r2) -> do r1' <- deRefAll r1
				   r2' <- deRefAll r2
				   t1' <- readIORef r1'
				   t2' <- readIORef r2'
				   newIORef (TApp (t1',r1') (t2',r2'))

	Var n q | q /= r     -> do deRefAll q

	_		     -> return r


changeTVarRef r (Var n _) = Var n r
changeTVarRef _ t = t

newTVarRef v = let t = Var v undefined
               in  do r <- newIORef t
		      modifyIORef r (changeTVarRef r)
		      return r

newTConRef n = let t = TyCon n
	       in  newIORef t

newTAppRef r1 r2 = let t = TApp r1 r2
		   in  newIORef t


--------------------------------------------------------------------------------
-- unification with occurs check


-- returns True if unification was successful
unify :: TypeRef -> TypeRef -> IO Bool
unify r1 r2 = do
    t1 <- readIORef r1
    t2 <- readIORef r2
    case (t1,t2) of

	(Var n q, _) -> do 
		    oc <- occurs r1 r2
--		    putStrLn ("occurs: " ++ show t1 ++ "\t" ++ show t2 ++ "\t" 
--					 ++ show oc)
		    case oc of 
			True -> err "occurs check"
			_ -> do	
			    r1' <- deRef r1
			    r2' <- deRef r2
			    t1' <- readIORef r1'
			    case t1' of
				Var n q -> do modifyIORef r1' 
						    (changeTVarRef r2')
					      return True
				_	-> unify r1' r2'

	(_, Var _ _) -> unify r2 r1
	
	(TApp q1 q2, TApp s1 s2) -> do b1 <- unify (snd q1) (snd s1)
				       b2 <- unify (snd q2) (snd s2)
				       return (b1 && b2)

	(TyCon n1, TyCon n2) -> if n1 /= n2 then err "1"
					    else return True
	
	_		   -> err "2"

    where err str = {- do q1 <- deRef r1
		       q2 <- deRef r2
		       s1 <- readIORef q1
		       s2 <- readIORef q2
		       trace ("UNIFICATION ERROR: " ++ str ++ "\n" ++
		             show s1 ++ "\t\t" ++ show s2) $ 
		    -}
		       return False


occurs :: TypeRef -> TypeRef -> IO Bool
occurs r1 r2 = do
    r1' <- deRef r1
    r2' <- deRefAll r2
    t1 <- readIORef r1'
    t2 <- readIORef r2'
    case (t1, t2) of
	(Var n _, Var m _) -> return False
	ts -> nested r1' ts

	 where 
	    nested r1 (Var n _, Var m _) = return (fst n == fst m)
	    nested r1 (Var _ _, TyCon _) = return False
	    nested r1 (Var _ _, TApp q1 q2) = do b1 <- sub r1 (snd q1)
						 b2 <- sub r1 (snd q2)
						 return (b1 || b2)
	    nested r1 _ = do r1' <- deRef r1
			     if r1' == r1
				then return False 
				     -- error "occurs - not a var!"
				else sub r1' r2'

	    sub r1 r2 = do
		r2' <- deRef r2
		t1 <- readIORef r1
		t2 <- readIORef r2
		nested r1 (t1, t2)



----------------------------------------

-- returns the set represented by the given list of UCons
-- equality of two UCons is based on their pred.symbol and types
removedups2 :: [UCons] -> TD [UCons]
removedups2 ucs = removedupsEq eq ucs
    where 
	eq (Std.UC _ p1 ts1 j1) (Std.UC _ p2 ts2 j2) = p1 == p2 && ts1 == ts2

-- as above, but equality of UCons entails equality of justifications
removedups3 :: [UCons] -> TD [UCons]
removedups3 ucs = removedupsEq eq ucs
    where 
	eq (Std.UC _ p1 ts1 j1) (Std.UC _ p2 ts2 j2) = 
		p1 == p2 && ts1 == ts2 && j1 == j2

-- takes a function defining equality on UCons
removedupsEq :: (Std.UCons -> Std.UCons -> Bool) -> [UCons] -> TD [UCons]
removedupsEq eq ucs = do ucs' <- mapM derefUC ucs
			 let sucs  = map stripUC ucs'
			     ucs'' = rem sucs ucs
			 return ucs''
    where 
	derefUC (UC i p1 ts j) = do 
	    ts' <- mapM (\(t,r) -> do r' <- doIO $ deRefAll r
				      t' <- doIO $ readIORef r'
				      return (t',r')) ts
	    return (UC i p1 ts' j)
    
	rem [] [] = []
	rem (suc:sucs) (uc:ucs) | suc `uelem` sucs = rem sucs ucs
				| otherwise	   = uc : rem sucs ucs

--	uelem uc = any (eq uc)

	admin uc = adminLoc `elem` (unJust uc)

	uelem uc 
	    | admin uc = any (eq uc)
	    | otherwise = any (\uc2 -> eq uc uc2 && not (admin uc2))

   
--------------------------------------------------------------------------------

-- normalisation of CHR store
-- 1) add goal to store-proper
-- 2) perform unification
td_normalizeOpt :: SOpt -> Store -> TD (Maybe Store)
td_normalizeOpt opt (store@((u1,b1),u2,b2,tokens,tvars,vm)) = do
    -- unify Herbrand constraints

--    doIO $ showStore >> putStrLn ""
    b <- doIO $ unifyEqs b1

--    puts "\n(normalising store) NEW STORE:"
--    doIO $ showStore

    let b3 = b1 ++ b2
        u3 = u1 ++ u2

    -- remove redundant UCons
    u3' <- case opt of
	    SInference -> removedups2 u3 -- removedups3 u3 -- removedups2 u3
	    _	       -> removedups3 u3

    b3' <- case opt of
	    SDebug -> return b3		-- accumulate all eqs
	    _	   -> return []		-- already accounted for

    -- if unification succeeded, return the new store
    if b then return (Just (([],[]), u3', b3', tokens, tvars, vm))
	 else return Nothing
    
    where 

	  unifyEqs [] = return True
	  unifyEqs ((Eq (t1,r1) (t2,r2) j):bcs) = do
{-	  
		putStrLn ("unifying: " ++ show t1 ++ "\t" ++ show t2)
		q1 <- deRefAll r1
		q2 <- deRefAll r2
		s1 <- readIORef q1
		s2 <- readIORef q2
		putStrLn ("aka     : " ++ show s1 ++ "\t" ++ show s2)
-}		
		b  <- unify r1 r2
--		showStore
--		putStrLn ""
		if b then unifyEqs bcs
		     else return False

	
	  showStore = do strs1 <- mapM showBC b1
			 strs2 <- mapM showUC u1
			 strs3 <- mapM showUC u2
--			 putStrLn (unwords strs2 ++ " (" ++ unwords strs3 ++ 
--				   ") " ++ unwords strs1) 
--			 putStrLn ("tokens: " ++ show tokens)
--			 putStrLn (unwords strs2 ++ unwords strs3)
			 putStrLn (unwords strs1 ++ " (" ++ unwords strs2 ++ 
				    ", " ++ unwords strs3 ++ ")")
	    where showBC (Eq (_,r1) (_,r2) _) = do
			r1' <- deRef r1
			r2' <- deRef r2
			s1  <- formatTypeRef r1'
			s2  <- formatTypeRef r2'
			return (s1 ++ "=" ++ s2)

		  showUC (UC x p ts _) = do 
		    dts <- mapM (\(_,r) -> do r' <- deRef r
					      formatTypeRef r') ts
		    return (show p ++ "(" ++ unwords dts ++ ")") --_" ++ show x)


showStore (_,ucs,bcs,_,_,_) = do 
    strs1 <- mapM showBC bcs
    strs2 <- mapM showUC ucs
    putStrLn "STORE"
    putStrLn ("UCS: " ++ unwords strs2)
    putStrLn ("BCS: " ++ unwords strs1)
    where 
	showBC (Eq (_,r1) (_,r2) _) = do
	    r1' <- deRef r1
	    r2' <- deRef r2
	    s1  <- formatTypeRef r1'
	    s2  <- formatTypeRef r2'
	    return (s1 ++ "=" ++ s2)

	showUC (UC x p ts j) = do 
	    dts <- mapM (\(_,r) -> do r' <- deRefAll r
				      formatTypeRef r') ts
	    return (show p ++ "(" ++ unwords dts ++ ")." ++ show x)


showStoreGoal ((ucs,bcs),_,_,_,_,_) = do 
    strs1 <- mapM showBC bcs
    strs2 <- mapM showUC ucs
    putStrLn "GOAL"
    putStrLn ("UCS: " ++ unwords strs2)
    putStrLn ("BCS: " ++ unwords strs1)
    where 
	showBC (Eq (_,r1) (_,r2) _) = do
	    r1' <- deRef r1
	    r2' <- deRef r2
	    s1  <- formatTypeRef r1'
	    s2  <- formatTypeRef r2'
	    return (s1 ++ "=" ++ s2)

	showUC (UC x p ts j) = do 
	    dts <- mapM (\(_,r) -> do r' <- deRefAll r
				      formatTypeRef r') ts
	    return (show p ++ "(" ++ unwords dts ++ ")." ++ show x)



unifyEqs [] = return True
unifyEqs ((Eq (t1,r1) (t2,r2) j):bcs) = do
    b  <- unify r1 r2
    if b then unifyEqs bcs
	 else return False


goalToStore ((u1,b1),u2,b2,tokens,tvars,vm) =
	    (([],[]),u1++u2,b1++b2,tokens,tvars,vm)


----------------------------------------
-- The version of "reduce" used by the type debugger functions.
-- Key differences:
--  - takes a substitution as an additional argument
--  - does not expect store to be normalized
--  - instead, applies subst at when attempting to match
--  - moreover, the "matching substitution" (unifying the matching types) is
--    created in terms of the original types - without the substitution applied

-- the JustPolicy determines how new justifications are determined
td_reduce :: (SOpt, JustProp) -> CHR -> Store -> TD (E Store)

-- simprule applicable only if goal is empty
td_reduce (opt, jp) simp@(Simp _ yuc@(Std.UC ucix p ts j1) [] _) 
	      (store@(([],[]),ucs,bcs,tokens,tvs,vm)) =
     let 
     
	 -- find a particular match
         findmatch :: [UCons] -> 
		      TD (E ([Std.Type],[TypeAndRef],Index,Just,[Std.Type]))
         findmatch [] = return (Err "td_reduce: no ucons!")
         findmatch (uc@(UC i p1 ts1 _):ucs)
           | p == p1   = do (b,j,t) <- ruleapp (ts,ts1)
			    if b then return (return (ts,ts1,i,j,t))
				 else findmatch ucs

           | otherwise = findmatch ucs
	   
	 ruleapp :: ([Std.Type],[TypeAndRef]) -> TD (Bool, Just, [Std.Type])
	 ruleapp (tr1, tr2) = do mt <- matchRule (dbgOpt opt) tr1 tr2 bcs
				 let mtb = ruleMatches mt
				     (mtj,mtt) = if mtb 
					    then (matchJust mt,matchedTypes mt)
					    else error "no match anyway"
				 return (mtb, mtj, mtt)
	 
         removeAtom :: Index -> [UCons] -> ([UCons], Just)
         removeAtom _ [] = ([], noJust)
         removeAtom i ((a@(UC j p ts jst)):ucs)
            | i == j    = (ucs, jst)
            | otherwise = let (as, j) = removeAtom i ucs
			  in  (a:as, j)

	 isDefRule = isValCHR simp
	 
     in do m <- findmatch ucs
	   case m of

	-- removeAtom should also return the justs on the removed cons... we 
	-- then add them to the constraints in the body (bodyuc, bodybc)
	-- justifications:
	--  j1: rule
	--  j2: user constraints
	--  j3: builtin constraints (implicants)
	    Succ (_ts1,ts2,idx,j3,mt) -> do
		let (ucs', j2) = removeAtom idx ucs
		    (fu, fb, j) = 
			if isDefRule 
			then if isLocalJust jp
			       then (newUCJust j2, newBCJust j2, j2)
			       else (addUCJust j2, addBCJust j2, j2)
			else case opt of
			     SDebug -> let j' = nubJust $ (j2 `addJust` 
						    nubJust (j1 `addJust` j3))
				       in  (newUCJust j', newBCJust j', j')
				       
			     SEvidence -> (addUCJust j2, addBCJust j2, j2)

			     SInference -> let j' = noJust
					   in (newUCJust j', newBCJust j',j')
							    
		(vm',simp') <- internaliseCHR simp
		(vm'', mtr) <- internaliseTypes (vm++vm') mt
		let RSimp _ (UC _ _ ts1 _) _ (bodyuc,bodybc) = simp'
		    matchSubst = reJust noJust (buildEP ts1 mtr)
		    newUCs = map fu bodyuc
		    newBCs = matchSubst ++ map fb bodybc

{-	    
		trace ("\nrule: " ++ show simp ++
		       "\nj1: " ++ show j1 ++ " j2: " ++ show j2 ++ 
			" j3: " ++ show j3 ++
		       "\nbodyuc: " ++ show bodyuc ++
		       "\nnewUCs: " ++ show newUCs) $ --  ++ 
--		       "\nnewBCs: " ++ show newBCs ++
--		       "\nmatchSubst: " ++ show matchSubst ++
--		       "\nmt: " ++ show mt ++
--		       "\nts2: " ++ show (map stripType ts2) ++
--		       "\nopt: " ++ show opt ++
--		       "\nisDefRule: " ++ show isDefRule ++
--		       "\nisLocalJust jp: " ++ show (isLocalJust jp)) $ 
-}		       
	        return (return ((newUCs, newBCs), ucs', bcs, tokens, tvs, vm''))

	    Err _ -> return (Err "rule not applicable")


-- proprule only applicable if goal is empty
td_reduce (opt, jp) prop@(Prop _ i ucs1 [] _) 
	      (store@(([],[]),ucs,bcs,tokens,tvs,vm)) =
     let -- find a particular match
         -- build permutation of lhs-constraints
         -- assumption, ucs1 is not empty
         findmatch'::	     [Std.UCons] -> [UCons] -> 
			    ([Std.Type],[TypeAndRef],[Index]) -> 
		      TD (E ([Std.Type],[TypeAndRef],[Index],Just,[Std.Type]))
         findmatch' ucs1 ucs (ts1,ts2,idxs) =
              let iterate [] = return (Err "")
                  iterate (ucs':ucss) = do -- initially noJust
		    m <- findmatch ucs' ucs (ts1,ts2,idxs,noJust)
		    case m of
                        Succ s -> return (Succ s)
                        _      -> iterate ucss
              in iterate (permute ucs1) 
            
	-- note that the following code reflects our assumption that
        -- commutativity of atoms is handled by permuting heads in prop-rules  
         findmatch ::	     [Std.UCons] -> [UCons] -> 
			    ([Std.Type],[TypeAndRef],[Index],Just) -> 
		      TD (E ([Std.Type],[TypeAndRef],[Index],Just,[Std.Type]))
         findmatch [] _ (ts1,ts2,idxs,j) = do
	    (b,j',t) <- ruleapp (ts1, ts2, idxs)
	    if b then return (Succ (ts1,ts2,idxs,addJust j j',t))
		 else return (Err "can't match ucons types")
         findmatch _ [] _ = return (Err "no match - ran out of store ucons")
         findmatch (cs@((Std.UC i1 p1 t1 j1):cs1)) ucz@((UC i2 p2 t2 j2):ucs) 
		   (acc@(ts1,ts2,idxs,j))
	   | p1 /= p2  = findmatch cs ucs acc
           | otherwise = do let ts1' = t1 ++ ts1
				ts2' = t2 ++ ts2
				j'   = addJust j j2 -- addJust j2 (addJust j1 j)
			    fmMore <- findmatch cs1 ucs (ts1',ts2',i2:idxs,j')
			    
			     -- check whether t1 and t2 can possibly match
--			    doIO (putStr "individual match: ")
			    m <- matchRule False t1 t2 bcs
			    let canMatch = ruleMatches m
			    if canMatch
				then case fmMore of
					Succ res -> return (Succ res)
					Err _    -> findmatch cs ucs acc
			       -- p1 p2 don't match, continue
				else  findmatch cs ucs acc

	 -- rule index already included in list below
         ruleapp :: ([Std.Type],[TypeAndRef],[Index]) -> 
		    TD (Bool, Just, [Std.Type])
         ruleapp (ts1,ts2,idxs) 
           | notElem (sort idxs) tokens = do
--			    puts ("trying rule: " ++ show prop)
--			    doIO $ (putStr "altogether now  : ")
			    mt <- matchRule (dbgOpt opt) ts1 ts2 bcs
			    let mtb = ruleMatches mt
				(mtj,mtt) = if mtb 
					    then (matchJust mt,matchedTypes mt)
					    else error "no match anyway"
			    return (mtb,mtj,mtt)

           | otherwise = return (False, noJust, error "no match anyway")
	   
	-- i: justification for rule
	-- j: justification on matching ucons (and rule?!)
	   
     in -- don't forget the index of the prop rule
	do ix <- next
	   m  <- findmatch' ucs1 ucs ([],[],[i])
	   case m of 
	
	    Succ (_ts1, ts2, idxs, j',mt) -> do
		let j = case opt of 
		       SDebug     -> nubJust (j' `addJust` (locToJust i))
		       SEvidence  -> nubJust (j' `addJust` (locToJust i))
		       SInference -> noJust
		
		(vm',prop') <- internaliseCHR prop
		(vm'', mtr) <- internaliseTypes (vm++vm') mt
		let RProp _ i lucs [] (bodyuc,bodybc) = prop'
		    ts1 = concatMap ucTypes lucs
		    matchSubst = reJust noJust (buildEP ts1 mtr) -- mtr {-ts2-})
		    newUCs = map (newUCJust j) bodyuc
		    newBCs = matchSubst ++ reJust j' bodybc
			     -- reJust j' bodybc ++ matchSubst

{-
		trace ("\nrule: " ++ show prop ++
		       "\ni: " ++ show i ++ " j': " ++ show j' ++ 
		       " j: " ++ show j ++
		       "\nnewUCs: " ++ show newUCs ++
		       "\nnewBCs: " ++ show newBCs ++
		       "\nmatchSubst: " ++ show matchSubst ++
		       "\nmt: " ++ show mt ++
--		       "\nxs: " ++ show xs ++
		       "\nts2: " ++ show (map stripType ts2)) $ 
-}		       
		return (Succ ((newUCs, newBCs), ucs, bcs, 
			          (sort idxs):tokens, tvs, vm''))

	    Err _  -> return (Err "rule not applicable")


-- proprule only applicable if goal is empty
td_reduce (opt, jp) msimp@(MSimp _ ucs1 [] _) 
	      (store@(([],[]),ucs,bcs,tokens,tvs,vm)) =
     let -- find a particular match
         -- build permutation of lhs-constraints
         -- assumption, ucs1 is not empty
         findmatch'::	     [Std.UCons] -> [UCons] -> 
			    ([Std.Type],[TypeAndRef]) -> 
		      TD (E ([Std.Type],[TypeAndRef],Just,[Index],[Std.Type]))
         findmatch' ucs1 ucs (ts1,ts2) =
              let iterate [] = return (Err "")
                  iterate (ucs':ucss) = do -- initially noJust
		    m <- findmatch ucs' ucs (ts1,ts2,noJust,[])
		    case m of
                        Succ s -> return (Succ s)
                        _      -> iterate ucss
              in iterate (permute ucs1) 
            
	-- note that the following code reflects our assumption that
        -- commutativity of atoms is handled by permuting heads in prop-rules  
         findmatch ::	     [Std.UCons] -> [UCons] -> 
			    ([Std.Type],[TypeAndRef],Just,[Index]) -> 
		      TD (E ([Std.Type],[TypeAndRef],Just,[Index],[Std.Type]))
         findmatch [] _ (ts1,ts2,j,ixs) = do
	    (b,j',t) <- ruleapp (ts1, ts2)
	    if b then return (Succ (ts1,ts2,addJust j j',ixs,t))
		 else return (Err "can't match ucons types")
         findmatch _ [] _ = return (Err "no match - ran out of store ucons")
         findmatch (cs@((Std.UC i1 p1 t1 j1):cs1)) ucz@((UC i2 p2 t2 j2):ucs) 
		   (acc@(ts1,ts2,j,ixs))
	   | p1 /= p2  = findmatch cs ucs acc
           | otherwise = do let ts1' = t1 ++ ts1
				ts2' = t2 ++ ts2
				j'   = addJust j j2 -- addJust j2 (addJust j1 j)
			    fmMore <- findmatch cs1 ucs (ts1',ts2',j',i2:ixs)
			    
			     -- check whether t1 and t2 can possibly match
			    m <- matchRule False t1 t2 bcs
			    let canMatch = ruleMatches m
			    if canMatch
				then case fmMore of
					Succ res -> return (Succ res)
					Err _    -> findmatch cs ucs acc
			       -- p1 p2 don't match, continue
				else  findmatch cs ucs acc

         ruleapp :: ([Std.Type],[TypeAndRef]) -> 
		    TD (Bool, Just, [Std.Type])
         ruleapp (ts1,ts2) = do
	    mt <- matchRule (dbgOpt opt) ts1 ts2 bcs
	    let mtb = ruleMatches mt
		(mtj,mtt) = if mtb 
			    then (matchJust mt,matchedTypes mt)
			    else error "no match anyway"
	    return (mtb,mtj,mtt)
	   
         removeAtoms :: [Index] -> [UCons] -> ([UCons], Just)
         removeAtoms _ [] = ([], noJust)
         removeAtoms ixs ((uc@(UC ix p ts j)):ucs) =
	    let ujs@(ucs',js) = removeAtoms ixs ucs
	    in  if ix `elem` ixs
		then ujs
		else (uc:ucs, j `addJust` js)


	   
	-- i: justification for rule
	-- j: justification on matching ucons (and rule?!)
	   
     in -- don't forget the index of the prop rule
	do ix <- next
	   m  <- findmatch' ucs1 ucs ([],[])
	   case m of 
	    Succ (_ts1, ts2, j',ixs,mt) -> do
		let (fu,fb,j) = case opt of 
		       SDebug     -> (nubUCJust . (addUCJust j'), 
				      nubBCJust . (addBCJust j'), j')
		       SEvidence  -> (nubUCJust . (addUCJust j'), 
				      nubBCJust . (addBCJust j'), j')
		       SInference -> (id,id,noJust) 
		

		let (ucs',uc_js) = removeAtoms ixs ucs
		
		(vm',msimp') <- internaliseCHR msimp
		(vm'', mtr) <- internaliseTypes (vm++vm') mt
		let RMSimp _ lucs [] (bodyuc,bodybc) = msimp'
		    ts1 = concatMap ucTypes lucs
		    matchSubst = reJust noJust (buildEP ts1 mtr) -- mtr {-ts2-})
		    newUCs = map fu bodyuc
		    newBCs = matchSubst ++ map fb bodybc
			     -- reJust j' bodybc ++ matchSubst
{-
		trace ("\nrule: " ++ show msimp ++
		       "\nixs: " ++ show ixs ++
		       "\nuc_js: " ++ show uc_js ++
		       " j': " ++ show j' ++ 
		       " j: " ++ show j ++
		       "\nnewUCs: " ++ show newUCs ++
		       "\nnewBCs: " ++ show newBCs ++
		       "\nmatchSubst: " ++ show matchSubst ++
		       "\nmt: " ++ show mt ++
--		       "\nxs: " ++ show xs ++
		       "\nts2: " ++ show (map stripType ts2)) $ 
-}		       
		return (Succ ((newUCs, newBCs), ucs', bcs, tokens, tvs, vm''))

	    Err _  -> return (Err "rule not applicable")

td_reduce _ r s = error ("td_reduce:\n" ++ show r ++ "\n")



td_reduces :: (SOpt, JustProp) -> [CHR] -> Store -> TD (Store,Maybe CHR)
td_reduces opts [] s = return (s,Nothing)
td_reduces opts (chr:chrs) s = do 
    res <- td_reduce opts chr s
    case res of
	 Succ s' -> do -- puts ("\nRULE FIRED: " ++ show chr) 
		       return (s',Just chr)
	 Err _   -> do -- puts ("\nRULE CAN'T FIRE: " ++ show chr)
		       td_reduces opts chrs s


-- gets called from externally, with standard constraints
-- 1) convert these to the internal representation (with IORefs)
-- 2) run the solver
-- 3) convert back to standard representation
derive :: SOpt -> JustProp -> [CHR] -> StdGoal -> TD FinalStore
derive opt jp chrs sgoal = do
    (vm, goal) <- stdGoalToGoal emptyVarMap sgoal
    (b, (g, ucs, bcs, ts, vs, _)) <- deriver opt jp chrs goal vm
    
    let tag | b		= Sat
	    | otherwise = Unsat

        (ucs',bcs') = goalToStdGoal (ucs,bcs)

--    puts ("#bcs: " ++ show (length bcs'))
    
--    puts "FINAL STORE:"
--    puts ("ucs: " ++ show ucs)
--    puts ("bcs: " ++ show bcs)
    return $ tag (g, ucs', bcs', ts, vs, filler "last arg")



-- performs a derivation
-- returns True if the store is satisfiable
deriver :: SOpt -> JustProp -> [CHR] -> Goal -> VarMap -> TD (Bool, Store)
deriver opt jp chrs (goal@(guc,gbc)) vm =
    let init = (goal,[],[],[],nub ((Std.fv guc) ++ (Std.fv gbc)),vm)
        iterate :: Store -> TD (Bool, Store)
        iterate st@(goal,bcs,ucs,ts,vs,vm) = do 
--	    puts "\nUNNORMALISED"
--	    doIO $ showStoreGoal st
	    st0 <- td_normalizeOpt opt st 
	    case st0 of
		Nothing	 -> do st' <- finalStore st
			       return (False, goalToStore st')
		Just st1 -> do
		    let (_,uc,bc,_,_,vm) = st1
--		    puts "\nNORMALISED"
--		    doIO $ showStore st1
		    (st2,bchrs) <- td_reduces (opt, jp) chrs st1
		    case bchrs of
			Nothing     -> do st2' <- finalStore st2
					  return (True, st2')
			Just appchr -> iterate st2
   in  iterate init

    where
	finalStore :: Store -> TD Store
	finalStore st@(goal,ucs,bcs,ts,vs,vm) 
	    | opt == SDebug = return st
	    | otherwise	    = do let trs = concatMap ucTypes ucs ++ 
					   concatMap bcTypes bcs
				     vm' = vm ++ genVarMap (map fst trs)
				 bcs' <- doIO $ project vm'
--				 puts "FINALISING STORE"
--				 puts ("had: " ++ show (length bcs) ++  " bcs")
--				 puts ("now: " ++ show (length bcs') ++ " bcs")
				 return (goal,ucs,bcs',ts,vs,vm)
    
	-- generate the varmap for these Vars
	genVarMap :: [Type] -> VarMap
	genVarMap ts = concatMap gen ts
	    where gen (TyCon _) = []
		  gen (Var n r) = [(n,r)]
		  gen (TApp (t1,_) (t2,_)) =
		    let vm1 = gen t1
			vm2 = gen t2
		    in  vm1 ++ vm2
    
	-- project constraint on variables in VarMap (the original goal vars)
	project :: VarMap -> IO [BCons]
	project = mapM mkEq
	    where mkEq (v,r) = do r' <- deRefAll r
				  t  <- readIORef r'
				  let t1 = Var v r'
				  return (Eq (t1,r') (t,r') noJust)




solve = deriveInference LocalJust

deriveDebug     = derive SDebug
deriveEvidence  = derive SEvidence
deriveInference = derive SInference

--------------------------------------------------------------------------------

-- internalise a goal written in terms of standard (external) constraints
stdGoalToGoal :: VarMap -> StdGoal -> TD (VarMap, Goal)
stdGoalToGoal vm (ucs,bcs) = do (vm', ucs')  <- internaliseUCs vm ucs
			        (vm'', bcs') <- internaliseBCs vm' bcs
			        return (vm'',(ucs',bcs'))
    
-- "internalise" the user and built-in constraints (add IORefs)
internaliseUC :: VarMap -> Std.UCons -> TD (VarMap, UCons)
internaliseUC vm (Std.UC i p ts j) = do (vm',ts') <- internaliseTypes vm ts
			                return (vm', UC i p ts' j)

internaliseBC :: VarMap -> Std.BCons -> TD (VarMap, BCons)
internaliseBC vm (Std.Eq t1 t2 j) = do (vm',t1')  <- internaliseType vm t1
				       (vm'',t2') <- internaliseType vm' t2
				       return (vm'', Eq t1' t2' j)

internaliseBCs = internalise internaliseBC
internaliseUCs = internalise internaliseUC
internaliseTypes = internalise internaliseType

internaliseType :: VarMap -> Std.Type -> TD (VarMap, TypeAndRef)
internaliseType vm (Std.Var v@(_n,_l)) = 
	case lookupVarMap v vm of
		Nothing -> do r <- doIO $ newTVarRef v
			      let vm' = insertVarMap (v,r) vm
			      return (vm', (Var v r, r))
		Just r  -> return (vm, (Var v r, r))

internaliseType vm (Std.TyCon t) = do r <- doIO $ newTConRef t
				      return (vm, (TyCon t, r))

internaliseType vm (Std.TApp t1 t2) = do (vm',tr1)  <- internaliseType vm  t1
					 (vm'',tr2) <- internaliseType vm' t2
					 r <- doIO $ newTAppRef tr1 tr2
					 return (vm'',(TApp tr1 tr2, r))

internalise f vm [] = return (vm, [])
internalise f vm (c:cs) = do (vm',c') <- f vm c
			     (vm'',cs') <- internalise f vm' cs
			     return (vm'',c':cs')


-- strip internal goal representation to get standard goal representation
goalToStdGoal :: Goal -> StdGoal
goalToStdGoal (ucs,bcs) = let ucs' = map stripUC ucs
			      bcs' = map stripBC bcs
			  in  (ucs',bcs')

stripUC :: UCons -> Std.UCons
stripUC (UC i p tr j) = Std.UC i p (map stripType tr) j

stripBC :: BCons -> Std.BCons
stripBC (Eq t1 t2 j) = Std.Eq (stripType t1) (stripType t2) j

stripType :: TypeAndRef -> Std.Type
stripType (Var v _,_) = Std.Var v
stripType (TyCon t,_) = Std.TyCon t
stripType (TApp t1 t2,_) = Std.TApp (stripType t1) (stripType t2)

----------------------------------------

-- renames the CHR, also converts it to the internal format 
internaliseCHR :: CHR -> TD (VarMap, RCHR)
internaliseCHR chr = do
    chr'   <- rnIdx chr
    echr'' <- doTP (Std.rename chr')
    let chr'' = unSucc' "internaliseCHR" echr''
    internal chr''
    where 
	internal (Simp i uc bcs (rucs, rbcs)) = do 
	    (vm0,uc')  <- internaliseUC emptyVarMap uc
	    (vm1,bcs') <- internaliseBCs vm0 bcs
	    (vm2,rucs')<- internaliseUCs vm1 rucs
	    (vm3,rbcs')<- internaliseBCs vm2 rbcs
	    return (vm3, RSimp i uc' bcs' (rucs', rbcs'))
	    
	internal (MSimp i ucs bcs (rucs, rbcs)) = do 
	    (vm0,ucs') <- internaliseUCs emptyVarMap ucs
	    (vm1,bcs') <- internaliseBCs vm0 bcs
	    (vm2,rucs')<- internaliseUCs vm1 rucs
	    (vm3,rbcs')<- internaliseBCs vm2 rbcs
	    return (vm3, RMSimp i ucs' bcs' (rucs', rbcs'))

	internal (Prop i x ucs bcs (rucs, rbcs)) = do 
	    (vm0,ucs') <- internaliseUCs emptyVarMap ucs
	    (vm1,bcs') <- internaliseBCs vm0 bcs
	    (vm2,rucs')<- internaliseUCs vm1 rucs
	    (vm3,rbcs')<- internaliseBCs vm2 rbcs
	    return (vm3, RProp i x ucs' bcs' (rucs', rbcs'))


--------------------------------------------------------------------------------

-- if flag is True, then just find if a match exists and return no
-- justification for it, else attempt to find the minimal implicants
-- NOTE: assumption is that first list come from the rule (second from store)

matchRule :: Bool -> [Std.Type] -> [TypeAndRef] -> [BCons] -> 
	     TD (Maybe ([Loc], [Std.Type]))
matchRule findImps ts1 trs2 bcs = do
    let rs2 = map snd trs2
    ts2 <- mapM (\r -> do r' <- doIO $ deRefAll r 
			  t  <- doIO $ readIORef r'
			  return $ stripType (t,r)) rs2

--    puts ("matchRule  ts1: " ++ show ts1 ++ "\t" ++ "ts2: " ++ show ts2)

    let ts2' = map stripType trs2
		
    let match = Std.matchTypes ts1 ts2
    if not match
     then return Nothing
     else if findImps 
	    then -- find minimal implicants, use the deRef path to limit the
		 -- search. i.e. we only care about those that played a part in
		 -- the above match
		 let p bcs = let ms = Std.unify [] Std.idSubst (Std.buildUP bcs)
				 s = unSucc ms
				 m = Std.matchTypes ts1 (Std.apply s ts2')
			     in  not (isErr ms) && m
			     
		     bcs' = map stripBC bcs	-- FIXME: pretty inefficient!!!
		     res  = Std.oneMinSubsetMaybe p bcs'
		     ts2''= let s = unSucc (Std.unify [] Std.idSubst 
						(Std.buildUP (fromJust res)))
			    in  Std.apply s ts2'
		     ls   = nub (unJust res)
		 in  if Std.matchTypes ts1 ts2'
		     then return (Just ([], ts2'))
		     else return (Just (ls, ts2''))
		 
	    else  -- just report that there was a match
		 return (Just ([],ts2'))


ruleMatches :: Maybe a -> Bool
ruleMatches Nothing  = False
ruleMatches (Just _) = True

matchJust :: Maybe ([Loc], [Std.Type]) -> Just
matchJust (Just (ls,_)) = J (filter (/=noLoc) ls)

matchedTypes :: Maybe ([Loc], [Std.Type]) -> [Std.Type]
matchedTypes (Just (_,ts)) = ts
	  
--------------------------------------------------------------------------------

-- takes a list of flags and returns the inital solver options
genSolverOpt :: [Flag] -> IO SOpt
genSolverOpt fs = do
    let ls = concatMap (\f -> case f of SolverOpt l -> [l]; _ -> []) fs
	l  = if null ls then 0 else head ls
        
    l' <- if l < 0 || l > 2 
	     then putStr ("WARNING: solver mode " ++ show l ++ " undefined" ++ 
			  "  (defaulting to mode 0)\n") >>
		  return 0
	     else return l 
    return (toEnum l')



--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

solveFor = error "solveFor undefined!"


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

runTest = runTD test initTDState

test = do
    let t1 = Std.Var (name "a" 1, 1)
	t2 = Std.Var (name "b" 2, 2)
	t3 = Std.Var (name "c" 3, 3)

	tc1 = Std.TyCon (name "T" 4)


	t4 = Std.TApp (Std.TApp tc1 t1) t2
	t5 = Std.TApp (Std.TApp tc1 t3) t3

	
	eq1 = Std.Eq t4 t5 noJust
	eq2 = Std.Eq tc1 t2 noJust

    (vm,bcs) <- internaliseBCs emptyVarMap [eq1, eq2]
    doIO $ showBCs bcs
    doIO $ unifyEqs bcs
    doIO $ showBCs bcs

project :: VarMap -> IO [BCons]
project = mapM mkEq
	    where mkEq (v,r) = do r' <- deRefAll r
				  t  <- readIORef r'
				  let t1 = Var v r'
				  return (Eq (t1,r') (t,r') noJust)

showBCs bcs = do strs <- mapM showBC bcs
		 putStrLn (unlines strs)
showBC (Eq (_,r1) (_,r2) _) = do
			r1' <- deRef r1
			r2' <- deRef r2
			s1  <- formatTypeRef r1'
			s2  <- formatTypeRef r2'
			return (s1 ++ "=" ++ s2)

