
module TDErrReport (typeErrorExplanation, 
		    typeConflictExplanation,
		    typeConflictExplanations)
where

import Chr
import Core
import Misc
import Justify
import SolverIO
import TD_Monad
import Ann_InputAST
import Constraint
import ConsReason
import Pretty

import List
import Monad

import IOExts

-- provides a step-by-step explanation of the inference process leading to the
-- error
-- arguments are:
--  - the initial solver goal
--  - a minimal unsatisfiable sets of locations
--  - a set of interesting locations (ones we want to examine)

typeErrorExplanation :: CHR -> Constraint -> [Loc] -> [Loc] -> TD ()
typeErrorExplanation chr c@(C ucs bcs) uls ils =
    -- first split the constraint up, based on location
    let lcs = [ lc | lc@(l,_) <- singGroupCons c, l `elem` uls ]
    -- now find the type variable corresponding to each interesting location
	ivs = nub [ vl | vl@(_,l) <- fv c, l `elem` ils ]
    
    in  do puts ("processing locations: " ++ show (map fst lcs))
	   puts ("constraint: " ++ show c)
	   proc lcs trueCons idSubst ivs

    where
	-- perform an inference step:
	--  - consider the next constraint:
	--	- if it's implied by the constraints we already have: ignore it
	--	- else:
	--	    - print an explanatory message about the next constraint
	--	    - add it to the store
	--	    - solve the new store, and builds the substitution
	--	    - display the new state of the interesting variables
	--	      (the consequence of that constraint)
	proc :: [(Loc, Constraint)] -> Constraint -> Subst -> [TVar] -> TD ()
	proc [] _ _ _ = puts "that's all folks"
	proc ((l,c):lcs) k rho ivs = do
	    describe
	    let k0 = c `cand` k
	    chrs <- getCHRs
	    k1 <- deriver chrs k0
	    let ePhi = let bcs = consToBs k1
		       in  unify [] idSubst (buildUP bcs)
		phi  = unSucc ePhi
		divs = zip ivs (apply phi (vars ivs))
	    if isErr ePhi
	     then puts "... and there's the error!\n\n"
	     else do puts "current variable values: "
		     showVars divs
		     proc lcs k1 phi ivs

	    where
	    
		-- prints a textual message about the current loc/constraint
		describe = do p <- getProg
			      puts ("now examining location: " ++ show l)
			      puts (printProgHLGreen [l] p)
			      puts ("the constraints arising from this " ++
				    "location are: \n" ++ show c ++
				    "\nwhich by now are: \n" ++
				    show (apply rho c))
			      
		showVars vs = let out = unlines (map showVar vs)
			      in  puts out
		    where showVar (v@(x,l),t) = show v ++ ": " ++ show t


--------------------------------------------------------------------------------


typeConflictExplanations c ls = do
    puts ("ls: " ++ show ls)
    puts ("c: " ++ show c)
    mapM (typeConflictExplanation c) ls


-- idea is to pick a location involved in a minimal conflict and expose the 
-- two types it could possibly have
typeConflictExplanation :: Constraint -> Loc -> TD ()
typeConflictExplanation c@(C ucs bcs) il =

    let	-- (bc:rest) = bcs
	(bc:rest) = headify [ b:bs | (b,bs) <- allLeft bcs, il `elem` unJust b ]
	phi = unSucc (unify [] idSubst (buildUP rest))
	bc' = apply phi bc
	l   = unJust bc
	bc_l = Eq (bcLeft bc) (bcLeft bc') noJust
	bc_r = Eq (bcRight bc) (bcRight bc') noJust

	impL = headify $ entailSubset (C [] rest) bc_l
	impR = headify $ entailSubset (C [] rest) bc_r

	ident bc = bcLeft bc == bcRight bc

	bc_l_ident = ident bc_l
	bc_r_ident = ident bc_r

    in  do p <- getProg
   
{-    
--	   puts ("constraints: " ++ show c)
           puts ("bc_l: " ++ show bc_l)
	   puts ("bc_r: " ++ show bc_r)
	   puts ("unsats: " ++ show (bc_l_ident, bc_r_ident))
-}

	   puts ("\nat location " ++ show l ++ 
		 " the following must hold: " ++  
		    (showTypeLoc (bcLeft bc)) ++ " = " ++ 
		    (showTypeLoc (bcRight bc)))
	   puts ("or, in other words: " ++ 
		    (showTypeLoc (bcLeft bc')) ++ " = " ++ 
		    (showTypeLoc (bcRight bc')))
{-	   
	   unless (bc_l_ident) (
	    puts ("but according to these locations: " ++ show impL ++ " " ++
	    showTypeLoc (bcLeft bc) ++ " must have type " ++ showTypeLoc (bcLeft bc')) >>
	    puts (printProgHLBlue impL p) )
	   
	   unless (bc_r_ident) (
	    puts ("and according to these locations: " ++ show impR ++ " " ++
	    showTypeLoc (bcRight bc) ++ " must have type " ++ showTypeLoc (bcRight bc')) >>
	    puts (printProgHLBlue impR p) )
-}	   
	   

--------------------------------------------------------------------------------
-- invokes the solver in mode 0, with local justification propagation

deriver :: [CHR] -> Constraint -> TD Constraint
deriver chrs c = do
    let goal = consToGoal c
    der <- derive SInference LocalJust chrs goal
    return $ mixedStoreToCons $ case der of
	        Sat   s -> s
		Unsat s -> s



headify (x:_) = x
headify [] = trace "can't headify an empty list!" []
