--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------



-- type inference (in the TD monad) using pre-generated CHRs

module TDInference where

import Core
import Chr
import TD_Monad
import TD_General
import Constraint
import ConsReason
import Ann_InputAST
import Justify
import Terminal
import Misc
import List
import Maybe
import ASTs
import Pretty (printProgHL, pprintHL, pprintAnn, toString)

import ErrorMsg
import TypeScheme hiding (tdDeriver)

import TDErrReport

import TD_Translate (transExp)

import Monad
import IOExts

import SolverIO 
import Solver (Goal, goalToCons)

-- import Solver

--------------------------------------------------------------------------------

inferTypeRef :: VarRef -> TD ()
inferTypeRef vr = do 
    p  <- getProg
    let vt = vtProg p
	midl = lookupVarTree vr vt
--    puts ("vr: " ++ show vr)
--    puts ("vt: " ++ show vt)
    case midl of
	Right id    -> puts ("Undefined variable `" ++ id ++ "'")
	Left (id,l) -> do inferTypeLoc l (False, True, False) undefined
			  reportLastResult

--------------------------------------------------------------------------------

reportLastResult = do
    (r:_) <- getResults
    case r of 
	Result ((t,l,v), _, _, (cons,scons),_) -> do
	    let ts  = buildTypeSchemeCons cons (snd t, snd l, snd v)
		str = toString (fst (showTypeScheme ts))
	    puts str
	Conflict (sp,_,cs,c) -> do 
	    puts "Type error: "
	    reportConflicts True c cs



--------------------------------------------------------------------------------

-- note: the following 3 steps may not be strictly correct (adhered to)
-- 1) find the CHR that contains variables with this ix (stop if none found)
-- 2) find the indices of all subexpressions of the annotated exp.
-- 3) remove from the body of the CHR constraints with other index

-- if b is True the function may write to stdout (else it won't)
-- if r is True the result will be stored in the state
-- if e if True then only write error reports to stdout!
inferTypeLoc :: Loc -> (Bool, Bool, Bool) -> JustProp -> TD ()
inferTypeLoc l (b, r, e) _jp = 
    do prog <- getProg
       chrs <- getCHRs
       jp   <- getJProp
       let pos  = when b
           posr = when r
           pose = when (not e)
           ix   = unkIx l
           tcs  = chrWithVarIx (ixLoc ix) chrs
       case tcs of
	[]	     -> do pos (puts ("* location " ++ show l ++ " not found"))
			   posr (newResult NoResult)

	((tv,chr,rest):_) -> 
		do let keyIxs = [ix] 
		       nonkeyIxs = []
		    
		       srcpos = chrSrcPos chr
		
		   -- do the derivation (we know the ixs of interest)
		   -- (mInitGoal, deriv) 
		   mderiv <- doDerivLocJP jp l b
		   if isNothing mderiv
		      then pose (puts "derivation failed")
		      else do let (initGoal, c) = fromJust mderiv
			          ucs = consToUs c
			          bcs = consToBs c

				  -- initGoal = fromJust mInitGoal

			      -- find the vars to solve for
			          tvs = map (flip constraintWithVarIx c)
					    (map ixLoc keyIxs)

			          tvs'= map head tvs
			      -- non-key (non-id) sub ixs
			          nk_tvs = map (flip constraintWithVarIx c)
					       (map ixLoc nonkeyIxs)
			          nk_tvs' = map head nk_tvs
			          all_tvs = tvs' ++ nk_tvs'

				  lts = map (\v-> (snd v,v)) tvs'
				  lt1  = head lts
				  lts' = tail lts
				  
				  ucts = ucTypes (simpUCons chr)
			      (tl,tv) <- 
				let f [_,l,v] = return ((noLoc, unVar l), 
							(noLoc, unVar v))
				    f _ = do l <- newLTVar
					     v <- newLTVar
					     return (l, v)
				in  f ucts
				 
			      let ts = (lt1, tl, tv)
				  
			      -- now project store onto these vars:
			      let eqp = buildUP bcs
			          mgu = unify [] idSubst eqp
			      case mgu of
				Err _  -> 
				    do  let ix = unkIx l
					-- doCommand (SetAnnFlag ix Fail)
					
					(cs,unsatCons) <- findConflict c
--					cs <- findConflict2 c
					
					let uls = head (snd cs)
					
					pos (reportConflicts (not e) c cs)


					-- typeErrorExplanation chr initGoal uls uls
					-- typeConflictExplanation unsatCons 6 
					let ls = unJust unsatCons
					--typeConflictExplanations unsatCons ls
					
					let conf = Conflict (srcpos, lt1, cs, c)
					posr (newResult conf)

				    
				Succ s -> 
				    do let ucs' = apply s ucs
				           bcs' = td_projBCons s all_tvs bcs

				           cu   = usToCons ucs'
				           cb   = bsToCons bcs'
				           c'   = cu `cand` cb
				          
				           nk_lts = map (\v-> (snd v,v)) nk_tvs'
				           res = Result 
						-- (lt1,lts',nk_lts,(c,c'),prog)
						   (ts,lts',nk_lts,(c,c'),prog)
				       -- store result
				       posr (newResult res)
				       pose (pos (doIO (putStr (show res))))				       


--------------------------------------------------------------------------------

-- takes a location, a boolean flag and, if possible, returns first the initial
-- solver goal which was run, and the store containing the derivation 
-- (starting at location l), and also the CHR whose body was used to generate 
-- the goal
-- if the flag is True, it will report any errors (produce output)
-- JP version takes a flag which directs the way justifications are propagated
-- during the derivation
doDerivLoc l b = do jp <- getJProp
		    doDerivLocJP jp l b

doDerivLocJP :: JustProp -> Loc -> Bool -> 
		TD (Maybe (Constraint, Constraint))
doDerivLocJP jp l b =
    do prog  <- getProg
       chrs  <- getCHRs
       schrs <- getSCHRs
       let ix  = unkIx l
       let tcs = chrWithVarIx (ixLoc ix) chrs
       case tcs of
	[]  -> do when b (puts ("location " ++ show l ++ " not found"))
		  return Nothing
	
	((tv,chr,rest):_) -> 
			do n <- next
		           let id_ix = findIxProg ix prog
			       ixs_bs = id_ix
			       ixs = map fst ixs_bs
			       -- new chr with unrelated constraints removed
			   let chr'' = remConsWithoutIxsNonNeg (ixsToJusts ixs) chr
			       -- we want its rhs - the initial goal
			       goal  = chrRhs chr''
			       goal' = goal
			       initCons = goalToCons goal'
			   
--			   puts ("goal': " ++ showAnn goal)
--			   puts ("chr: " ++ pprintAnn chr)
--			   puts ("chr'': " ++ pprintAnn chr'')
--			   puts ("ixs: " ++ show ixs)
{-
			   puts ("ucs: " ++ show 
                                (let C ucs bcs = goalToCons goal'
                                     phi = unSucc (unify [] idSubst 
                                                    (buildUP bcs))
                                 in  apply phi ucs))
-}
			   let uchrs = if isLocalJust jp then schrs else chrs

			   deriv <- deriverSt jp uchrs goal'
--			   puts $ "deriv: " ++ show (storeToCons $ fromJust deriv)
			   -- if the store is empty, add: tv = alpha
			   -- (otherwise the InferTypeLoc code will fail
			   -- looking for variables - it's easier this way)
			   if isJust deriv
			    then let st = fromJust deriv
				     c  = storeToCons st
				 in  if c /= trueCons
				     then return (Just (initCons, c))
				     else do a <- newVar
					     let tv' = Var tv
						 bc = Eq tv' a noJust
						 c' = bToCons bc
					     return (Just (initCons, c'))
			    else return Nothing
			   -- return (Just (initCons, deriv'))


--------------------------------------------------------------------------------
-- functions for finding sources of conflict

-- we know that the derivation for location l failed with the given store, c
-- we want to find the set of sets of conflicting locations 
-- it's possible that the derivation failed before all the user constraints
-- were dealt with - as functions being applied, they're also a potential
-- source of conflict
-- Bool flag set if ls (set of common problematic constraints) is itself unsat. 
findConflict :: Constraint -> TD ((([Loc], Bool), [[Loc]]),Constraint)
findConflict c = 
    let bcs  = consToBs c
	ubcs = minUnsatSubsets bcs
	(cbcs, b) = commonUnsatCons bcs
	lss = map unJust ubcs
	-- minUnsatSubsets returns minimal sets of constraints - we're actually
	-- interested in minimal sets of *locations*
	lss' = elimSupersets lss
	ls = unJust cbcs
    in	do -- puts ("unsat cons.:" ++ show c)
	   -- puts ("min. unsats:" ++ show ubcs)
	   -- puts ("common bcs.:" ++ show cbcs)
	   return (((ls, b), lss'), C [] (head ubcs))

-- version of findConflict which first attempts to find a single error, then
-- checks if any of the constraints involved in that error appear in all other
-- errors
findConflict2 :: Constraint -> TD (([Loc], Bool), [[Loc]]) 
findConflict2 c = 
    let bcs = consToBs c
	(ubcs, rs) = minUnsatSubsetsAndRest bcs
	(cbcs, b)  = commonUnsatConsLimit rs (concat ubcs)
	lss = map (nub . unJust) ubcs
	ls  = unJust cbcs

	as = -- allMinUnsatSubsets [] bcs
	     dpMinUnsatSubsets bcs
	
	
    in  -- do puts ("allMinUnsatSubsets: " ++ show as)
	   return ((ls, b), lss)
	

-- gets rid of constraits that have come from a prop. rule application
dropPropCons :: Constraint -> TD Constraint
dropPropCons cons = 
    do Prog decs <- getProg
       let ucs = consToUs cons
	   bcs = consToBs cons
	   decs' = decsToRules decs
	   locs  = map ixLoc (anns decs')
	   --rules = items decs'
	   ucs'  = filter (\u -> all (`notElem` locs) (unJust (ucJust u))) ucs
	   bcs'  = filter (\b -> all (`notElem` locs) (unJust (bcJust b))) bcs
	   cons' = usToCons ucs' `cand` bsToCons bcs'
       puts ("decs': " ++ show decs)
       puts ("locs : " ++ show locs)
       return cons'
		    

minIllLocs :: (Constraint, Loc) -> TD [(Constraint, Loc)]
minIllLocs (c, l) 
    = do ls  <- subLocs True l
	 its <- mapM locIllTyped ls
	 if all isNothing its
	    then return [(c, l)]
	    else do let ls' = [ (c, l) | (l, t) <- zip ls its, 
				         isJust t, let c = fromJust t ]
		    cls <- mapM minIllLocs ls'
		    return (concat cls)
		     

-- returns Nothing if l is well-typed, else Just cons - the constraint at the
-- end of the derivation
locIllTyped :: Loc -> TD (Maybe Constraint)
locIllTyped l = do mderiv <- doDerivLoc l False
		   let (_,c) = fromJust mderiv
		       bcs = cToBCons c
		   if consistentBCons bcs
		    then return Nothing
		    else return (Just c)


-- first, throw away negative locations (corresponding to constraints we don't
-- want to report anything about, eg. variable scoping constraints)
-- [BCons] is a list of constraints which appear in *all* unsat. subsets
--
-- if verb is True, then report is more verbose! (else it's quite minimal -
-- just the highlighted text)
reportConflicts :: Bool -> Constraint -> (([Loc], Bool), [[Loc]]) -> TD ()
reportConflicts verb c sd@((ls0, b), lss0) = 
    do let ls  = filter (>0) ls0
	   lss = map (filter (>0)) lss0
	
       prog <- getProg
       -- puts ("c: " ++ show c ++ "\nsd: " ++ show sd)

       let lss' = map (filter (>=0)) lss
           adls = map (partitionF (isAppLoc prog) id) lss'
           (_als, _dls) = unzip adls

           commonReport = do printCommon ls
			     mapM_ (reportConflict c) lss'

           confReport   = do printConf lss'
                             mapM_ (highlightConfStyle verb) lss'
			     mapM_ (reportConflict c) lss'
       if null (concat lss')
	  then do puts "TYPE ERROR: Insufficient information to pinpoint error"
		  puts "            Try a higher solver mode (e.g. -s 2  OR set solver 2)"

	  else if verb then do when (not (null ls)) commonReport
			       when (not b) confReport
	     
		       else if b then commonReport
				 else confReport

    where printConf lss = do 
	    pos (puts "The following locations have conflicting type information:")
--	    pos (mapM_ (puts . ("  "++) . show . sort) lss)
	    -- puts ("ls: " ++ show ls)	    
	  
	  printCommon ls = do 
	    pos (puts $ if b then "One type conflict found."
			     else "These locations are part of all conflicts:")
	    -- puts ("ls: " ++ show ls)
	    -- filter out administrative locations
	    let ls' = filter (>0) ls
	    highlightConfStyle verb ls'

	  pos  = when verb
	   

reportConflict :: Constraint -> [Loc] -> TD ()
reportConflict c ls = return ()

{-
-- CONTINUE HERE
reportConflict :: Constraint -> [Loc] -> TD ()
reportConflict c ls = do 
    let lls = allLeft ls
    tls <- mapM (uncurry (flip (breakConf c))) lls
    let tlss  = [ (l, nub $ consolidateTypes tl) | (l, tl) <- tls ]
	tlss' = [ lls | lls@(l, ls) <- tlss, not (singleton ls) ]
    puts ("tls:  \n" ++ unlines (map show tls))
    puts ("tlss: \n" ++ unlines (map (\x -> show x ++ "\n") tlss'))
-}

-- what we need to do now is take the result of breakConf, and consolidate all
-- the alternative types it gave
-- so, we split the list of types into lists of those types which are unifiable
-- eg. [Char -> c, d -> Bool, Bool -> e, Int]
--     becomes [[Char -> c, d -> Bool], [d -> Bool, Bool -> e], [Int]]
{-
consolidateTypes :: [(Type, Loc)] -> [[(Type, Loc)]]
consolidateTypes tls = [ tl:utls | (tl@(t,l), tls') <- allLeft tls,
				   let utls = filter (unifiable t) tls' ]

    where unifiable t1 (t2,l) = isSucc $ unify [] idSubst [(t1, t2)]
-}
consolidateTypes :: [(Type, Loc)] -> [(Type, [Loc])]
consolidateTypes tls = [ utls | (tl@(t,l), tls') <- allLeft tls, 
			        let utls = foldl unifiable (t,[l]) tls' ]
			       
    where unifiable (t1,ls) (t2,l) = let es = unify [] idSubst [(t1, t2)]
					 s  = unSucc es
				     in  if isSucc es 
					 then (apply s t1, l:ls)
					 else (t1, ls)
				    

-- Given a minimal, unsatisfiable constraint, and a location, infers a number
-- of `potential' types for the location by taking out one constraint from each 
-- location in the given list (thereby making the constraint satisfiable).
-- The type returned is the most specific of these types. The location is that
-- whose constraint we removed to break the conflict.
breakConf :: Constraint -> [Loc] -> Loc -> TD (Loc, [(Type, Loc)])
breakConf c ls l = do chrs <- getCHRs
		      n    <- next
		      let ls' = filter (/= l) ls
		          cs  = map (remConsLoc c) ls'
		          tvs = constraintWithVarIx l c
		          tv  = heady "c" tvs
		          typs   = map (\c -> solve chrs c tv n) cs
		          
		          tl = [ (fromJust t, l) | (t,l) <- zip typs ls', 
		                		   isJust t ]
		          _spec = mostSpecific tl
		      -- return (spec, tl)
		      if null tvs then return (l, []) else return (l, tl)

    where solve :: [CHR] -> Constraint -> TVar -> Integer -> Maybe Type
	  solve chrs c tv n = do let goal  = convGoal c
			         store <- solverFor chrs goal [tv] n
				 let bcs   = consToBs (storeToCons store)
				     eq    = head (withVar tv bcs)
				     typ   = bcRight eq
				 return typ
	
	  mostSpecific :: [(Type, Loc)] -> (Type, Loc)
	  mostSpecific []     = error "TypeDebug.breakConf.mostSpecific"
	  mostSpecific (t:ts) = mostSpecific' ts t
	    where mostSpecific' []	  a = a
		  mostSpecific' (tl@(t,l):ts) a@(s,_)
		    | s `moreGeneral` t	= mostSpecific' ts tl
		    | otherwise		= mostSpecific' ts a
	  moreGeneral t1 t2 = matchTypes [t1] [t2] && 
			      not (matchTypes [t2] [t1])


{-   
-- given a set of locations which are in conflict, comes up with a clever
-- report for the user
reportConflict :: Constraint -> [Loc] -> TD ()
reportConflict c ls = do prog <- getProg
			 p <- parentLoc ls
--			 doIO $ putStr "cause: "
--			 reportCause prog p
--			 breakUp
			 return ()
{-
    where reportCause prog p
	    | isAppLoc  prog p  = puts "app"
	    | isListLoc prog p	= puts "list"
	    | isIfLoc	prog p	= do subs <- subLocs False p
				     let cond = heady "n" subs   -- always the case
				     if cond `elem` ls
					then puts "if 1"
					else puts "if 2"
	    | any (isAnnLoc prog) ls = puts "ann"
	    | otherwise		= puts "dunno"


	  breakUp = do let c'  = consFromLoc c ls
--		       tls <- mapM (breakConf c' ls) ls
		       tls <- mapM (breakConf c ls) ls
		       puts $ show (zip ls tls)
-}
-}

			      
	    
-- given a set of conflicting locations, prints the problem areas highlighted
-- if verb is True, then display more information (like rules involved)
-- (by default, these possibly needn't be shown except from the debugger)
highlightConf = highlightConfStyle True

highlightConfStyle :: Bool -> [Loc] -> TD ()
highlightConfStyle verb ls0 = do
    let ls = filter (>0) ls0
    prog <- getProg
    chrs <- getCHRs
    let chrLs = getCHRLocs prog
        (rls, pls) = List.partition (`elem` chrLs) ls

    when (not (null pls)) (puts $ printProgHL pls prog) -- pprintHL pls prog)
{-    
			  (do
	let dls  = decLocs prog
	    strs = showHLLocs dls pls prog
	    rep = [ str | (l, str) <- zip dls strs, str /= "", hasHL str ]
	puts (unlines rep)
	)
-}		      
    when (not (null rls) && verb) (do
	puts ("rule" ++ (if singleton rls then "" else "s") ++ " involved:")
	puts (unlines $ map (("  "++) . showCHRLoc chrs) rls)
	)
    
    doIO $ putStr defaultBG

    where showHLLocs dls ls prog = map (\l -> showHLLoc ls l prog) dls
    
	  decLocs (Prog adecs) = map ixLoc (anns adecs)
    
	  -- returns True if no part of the given string is highlighted
	  -- HACK that depends on the Terminal module
	  hasHL str = hasBGEsc str
    
{-
	  -- given a list of locations, returns a new list containing the
	  -- locations which are parents of any of the original (such that all
	  -- the originals are accounted for)
	  --allParentLocs :: [Loc] -> TD [Loc]
	  allParentLocs [] _  = return []
	  allParentLocs ls ms
	    | ls == ms  = return []
	    | otherwise = do
		p  <- parentLoc ls
		cs <- allSubLocs False p
		let ls' = filter (`notElem`(p:cs)) ls
		ps <- allParentLocs ls' ls
		return (p:ps)
-}



-- attempts to come up with a textual explanation for an error
-- loc is the clash point
-- list contains all involved in the problem
explainError :: Loc -> [Loc] -> TD (Maybe String)
explainError p sls =
    do prog <- getProg
       if isIfLoc prog p
	   then return (Just "branches of an if-then-else must have same type")
	   else return Nothing

{-
-- takes a parent location, and a list of sub-locations
-- we infer types for the subs, but first get rid of any constraints justified
-- by the parent location (we break the conflict path)
--breakConf :: Loc -> [Loc] -> TD () -- TD Type
breakConf :: Loc -> [Loc] -> TD [(Loc, Type)]
breakConf p ls =
    do chrs  <- getCHRs
       deriv <- doDerivLoc p False
       n     <- next
       puts ("breakConf.p: " ++ show p)
       puts ("breakConf.ls: " ++ show ls)
       if isNothing deriv
	    then error "TypeDebug.breakConf: deriv. failed"
	    else let store = fromJust deriv
		     cons  = storeToCons store
		     bcs   = consToBs cons
		     ucs   = consToUs cons
--		     bcs'  = filter (\b -> p `notElem` unJust (bcJust b)) bcs
--		     ucs'  = filter (\u -> p `notElem` unJust (ucJust u)) ucs
		     (bcs_in, _bcs_out) = filt (\b -> p `notElem` 
						     unJust (bcJust b)) bcs
		     (ucs_in, _ucs_out) = filt (\u -> p `notElem`
						     unJust (ucJust u)) ucs
		     goal  = convGoal (bsToCons bcs_in `cand` usToCons ucs_in)
		     tcs   = map (\l -> chrWithVarIx (locToJust l) chrs) ls
		     tvs   = map (\((tv,_,_):_) -> tv) tcs
		     res   = solverFor chrs goal tvs n
		 in  if isNothing res
			then do puts $ "TypeDebug.breakConf: " ++ 
				       "\ngoal:  " ++ show goal ++
				       "\ngoal': " ++ showAnn goal ++
--				       "\ncons': " ++ showAnn cons
				       "\ntvs: " ++ show tvs
				-- return []
				error ""
			else let store = fromJust res
				 cons  = storeToCons store
				 bcs   = consToBs cons
				 typs  = [ (head j, bcRight b) |  b <- bcs,
					  let j= unJust (typeJust (bcLeft b))]
			     in  return typs
		
	where filt p []	    = ([], [])
	      filt p (x:xs) = let (t, f) = filt p xs
			      in  if p x then (x:t, f)
					 else (t, x:f)

-}


--------------------------------------------------------------------------------

-- takes a list of locations, and returns the location which is the lowest
-- common parent of all of them in the AST
-- if any of them is an MValDec, we return that
-- eliminates locations which do not belong to a def/exp/pat!
parentLoc :: [Loc] -> TD Loc
parentLoc [] = error "TypeDebug.parentLoc: empty list!"
parentLoc ls = do 
    prog <- getProg
    let ls' = filter (not . (isNonDefLoc prog)) ls
    if null ls' -- none of ls in program
       then return (foldr max 0 ls)
       else do start <- defLoc (heady "a" ls')
	       ps <- trySubs start
	       case ps of
		[]    -> do puts "ERROR:no parent (ignore result)"
			    return (head ls')
		(x:_) -> do return x
		    
   where 
    trySubs l = do sls <- subLocs True l
		   isP <- mapM (flip isParent ls) sls
		   let ps = [ l | (l, b) <- zip sls isP, b ]
		   if null ps
			then return [l]
			else do rs <- mapM trySubs ps
				return (concat rs)
		  
-- true if l is a common parent of ls
isParent l ls = do subs <- allSubLocs False l
	           let isP = all (`elem` subs) ls
		   return isP
		  

-- given a location, finds the outer-most location of the def in which it
-- appears (could be a dec!)
-- NOTE: does this screw up if there's more than one CHR with a variable
-- annotated with a given location? (I think so)
defLoc :: Loc -> TD Loc
defLoc l = do 
    chrs <- getCHRs
    let tcs = chrWithVarIx l chrs
    case tcs of
	[]            -> do -- puts ("*** location " ++ show l ++ " not found")
			    return 0
	((_,chr,_):_) -> do (Prog decs) <- getProg
			    let id  = chrTypeId (chrType chr) -- ucNameStr (simpUCons chr)
			        lis = valDecsToLocIds decs
				loc = heady "x" [ l | (l, id') <- lis, id == id' ]
			    return loc
	

--------------------------------------------------------------------------------

-- returns True if there is an id for a standard function at the given loc
locExpStd :: Loc -> TD Bool
locExpStd l = do prog <- getProg
		 let pos = getLoc l prog
		 case pos of
		    Q3 exp -> case item exp of
				    VarExp id ->  do nms <- getStdDef
						     let ids = map nameStr nms
						     return (id `elem` ids)
				    _	      ->  return False

		    _	      -> return False


-- returns True if the given location belongs to an id (Var Exp/Pat, ConExp)
isIdLoc :: Prog -> Loc -> Bool
isIdLoc prog loc = case getLoc loc prog of
		    Q2 pat  -> case item pat of
				VarPat _    -> True
--				ConPat _ _  -> True
				_	    -> False
		    Q3 exp  -> case item exp of
				VarExp _    -> True
				ConExp _    -> True
				_	    -> False
		    _	    -> False


-- True if the given location is an if-then-else expression
isIfLoc :: Prog -> Loc -> Bool
isIfLoc prog loc = case getLoc loc prog of
		    Q3 exp  -> case item exp of
				IfExp _ _ _ -> True
				_	    -> False
		    _	    -> False


removeAppLocs :: Prog -> [Loc] -> [Loc]
removeAppLocs p ls = filter (not . (isAppLoc p)) ls

isAppLoc :: Prog -> Loc -> Bool
isAppLoc prog loc = case getLoc loc prog of
		     Q3 exp -> case item exp of
				AppExp _ _ -> True
				_	   -> False
		     _	    -> False

isListLoc :: Prog -> Loc -> Bool
isListLoc prog loc = case getLoc loc prog of
		      Q3 exp -> case item exp of
				 ListExp _ -> True
				 _	   -> False
		      _	     -> False

isAnnLoc :: Prog -> Loc -> Bool
isAnnLoc prog loc = case getLoc loc prog of
		     Q3 exp -> case item exp of
				AnnExp _ _ -> True
				_	   -> False
		     _	    -> False


isNonDefLoc prog loc = case getLoc loc prog of
			NoQuad -> True
			Q4 dec -> True
			_    -> False

isDefLoc prog loc = case getLoc loc prog of
			Q1 _ -> True
			_    -> False


-- given a Prog, returns the list of locations which belong to RuleDecs,
-- OverDecs or both (combined)
getCHRLocs, getRuleLocs, getOverLocs :: Prog -> [Loc]
getCHRLocs  prog = getRuleLocs prog ++ getOverLocs prog ++ getOver0Locs prog
getRuleLocs (Prog decs) = map ixLoc $ anns (decsToRules decs)
getOverLocs (Prog decs) = map ixLoc $ anns (decsToAnnOverDecs decs)
getOver0Locs (Prog decs) = map ixLoc $ anns (decsToAnnOverDec0s decs)

-- WARNING: will FAIL if location doesn't exist
getLoc :: Loc -> Prog -> Quad (Ann Def) (Ann Pat) (Ann Exp) (Ann Dec)
getLoc loc prog = let hd [] = NoQuad -- error "chooo!"
		      hd xs = heady "e" xs
		  in  hd (getLocProg loc prog)

-- NOTE: only looks into defs!
getLocProg loc (Prog decs) =
    let defs  = concat (valDecsToDefs decs)
    in  concatMap (getLocDef loc) defs

-- WARNING: only matches first def, in a multi-def function
getLocDef loc def =
    case item def of
	Def id pats rhs w -> 
		if locMatches loc def
		then [Q1 def]
		else let ps = getLocPats loc pats
			 es = concatMap (getLocExp loc) (rhsExps rhs)
			 hd [] = error "arrrghhh!"
			 hd xs = heady "g" xs
		     in  if null es
			 then if null ps
			      then []
			      else [Q2 (hd ps)]
			 else [Q3 (hd es)]

-- returns the expression at the given location ([] means loc not found!)
getLocExp loc exp = 
    if locMatches loc exp
	then [exp]
	else let gle = getLocExp loc
	     in  case item exp of
		    AbsExp id e1    -> gle e1
                    AppExp e1 e2    -> gle e1 ++ gle e2
                    IfExp e1 e2 e3  -> gle e1 ++ gle e2 ++ gle e3
                    AnnExp e1 t	    -> gle e1 
                    TupExp es	    -> concatMap gle es
                    ListExp es	    -> concatMap gle es
                    _		    -> []

getLocPats loc pats = concatMap (getLocPat loc) pats
getLocPat loc pat =
    if locMatches loc pat
	then [pat]
	else let glp = getLocPat loc
	     in  case item pat of
		    ConPat id ps    -> concatMap glp ps
		    TupPat ps	    -> concatMap glp ps
		    ListPat ps	    -> concatMap glp ps 
		    _		    -> []

--------------------------------------------------------------------------------

locMatches loc x = ixLoc (ann x) == loc
locFlag x = ixFlag (ann x)

--------------------------------------------------------------------------------

solverFor = error "solverFor" 

{-
solverFor :: [CHR] -> Goal -> [TVar] -> Integer -> Maybe Store
solverFor chrs goal vs init = let solve = solveFor chrs goal vs
			      in  case unTP solve init of
					Err  _	-> Nothing
					Succ (_, res)	-> Just res
-}
--------------------------------------------------------------------------------

-- finds *all* sub-locations of the given loc. (not just immediate ones)
-- (flag is passed through to subLocs - same meaning)
allSubLocs :: Bool -> Loc -> TD [Loc]
allSubLocs f l = deeper l []
    where deeper l ls = do ss <- subLocs f l
			   if null ss
				then return [l]
				else do let ss' = filter (`notElem` ls) ss
					let ls' = l : ls
					ms <- mapM (flip deeper ls') ss'
					return (l : concat ms)
			   

-- takes a loc, returns a list of the immediate sub-loc.s
-- if the flag is set to True, we decend through VarExps into other decs.
-- else, we return []
subLocs = dd_descend
dd_descend :: Bool -> Loc -> TD [Loc]
dd_descend f l = do
    prog <- getProg
    let ibs = findIxProg (unkIx l) prog
	ls  = map (ixLoc . fst) ibs
	ls' = filter (/=l) ls
    puts ("sublocs of " ++ show l ++ " are " ++ show ibs)
    return ls'
    
	  
{--	  
subLocs = dd_descend
dd_descend :: Bool -> Loc -> TD [Loc]
dd_descend f l =
    do prog <- getProg
       let pos   = getLoc l prog
       let loc x = ixLoc (ann x)
       let locs xs = return (map loc xs)
       case pos of 
	NoQuad	  -> return []
	Q1 def -> 
	    case item def of
		Def _ ps rhs w	-> do lps <- return [] -- locs ps
				      les <- locs (rhsExps rhs) -- [e]
				      return (lps ++ les)
	Q2 pat -> 
	    case item pat of
		ConPat id ps	-> locs ps >>= \ls -> return (loc id : ls)
		TupPat  ps	-> locs ps
		ListPat ps	-> locs ps
		_		-> return []
	Q3 exp -> 
	    case item exp of
		AbsExp id e1    -> locs [e1]
		AppExp e1 e2    -> locs [e1, e2]
		IfExp e1 e2 e3  -> locs [e1, e2, e3] 
		AnnExp e1 t	-> locs [e1]
		TupExp es	-> locs es
		ListExp es	-> locs es
		VarExp id	-> 
			if not f 
			 then return []
			 else do chrs <- getCHRs
				 let (chrs', _) = chrWithId id chrs
				 case chrs' of
				    [chr]  -> do let uc = simpUCons chr
					         return (unJust (ucJust uc))
				    _      -> return []	

		_		 -> return []
	Q4 dec ->
	    case item dec of
		_		-> error "TypeDebug.dd_descend: can't happen"
-}



--------------------------------------------------------------------------------


tdDeriver = deriverSt
--deriverSt = error "deriverSt"

-- return the store regardless of whether its satisfiable
-- deriverSt :: JustProp -> [CHR] -> Goal -> TD (Maybe Store)
deriverSt jp chrs goal = do 
--    puts ("jp: " ++ show jp)
    opt <- getSolvOpt
--    puts ("opt: " ++ show opt)
    der <- derive opt jp chrs goal
    return $ case der of
	Sat   s -> Just s
	Unsat s -> Just s


{-
tdDeriver = deriverSt
deriverSt :: JustProp -> [CHR] -> Goal -> TD (Maybe Store)
deriverSt jp chrs goal = do 
    opt <- getSolvOpt
    n <- next
    let der = derive opt jp chrs goal
    case unTP der n of
	Err e		-> trace e $ return Nothing
	Succ (c,res)	-> do setCounter c
			      return (Just res)
-}

--------------------------------------------------------------------------------

-- turns an ValueInfAnn CHR into a ValueAnn CHR by simplifying it and changing
-- the CHR type tag, also rejustifies the user constraints
-- returns both the new CHR and a mapping from old valdec CHR calls to new
-- annotation CHR calls
revealAnnCHR :: CHR -> TD (CHR, (UCons, UCons))
revealAnnCHR chr | isInfAnnCHR chr = do
    chrs <- getCHRs
    let g = chrRhs chr
    ms <- tdDeriver LocalJust chrs g
    let cons = mixedStoreToCons (fromJust ms)
	rhs = consToRhs cons
        chr' = chrNewRhs rhs chr
	ren  = (infAnnUCons chr, simpUCons chr)
    return (infAnnToAnnCHR chr', ren)
  
revealAnnCHR chr = return (chr, error "revealAnnCHR")

-- reveals the annotation CHR for each ValDec and reroutes calls to it
revealAnnCHRs :: [CHR] -> TD [CHR]
revealAnnCHRs chrs = do
    let (iachrs, rest) = partition isInfAnnCHR chrs
    rev <- mapM revealAnnCHR iachrs
    let (achrs, rens) = unzip rev
	rest' = map (rerouteRhs rens) rest
	chrs' = achrs ++ rest'
    puts ("\n\nrens: " ++ show rens ++ "\n")
    return chrs'
    
    where 
	rerouteRhs rens chr = let (ucs, bcs) = chrRhs chr
				  ucs' = map (reroute rens) ucs
			      in  chrNewRhs (ucs', bcs) chr
			
	
	reroute rens uc@(UC i _ ts j) = case lookupBy eq uc rens of
	    Nothing  -> uc
	    Just uc' -> UC i (uconsPSym uc') ts j
			
	    
	eq uc1 uc2 = (uconsPSym uc1, length (ucTypes uc1)) ==
		     (uconsPSym uc2, length (ucTypes uc2))



-- revealAnnCHRs = mapM revealAnnCHR


