--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- contains functions, data structures which are common to a number of the TD
-- (type debugger) modules


module TD_General (module TD_General, module Solver, noIx)
where

import Solver (solve)
import Chr  hiding (Index)
import Ann_InputAST
import Core
import Justify
import Misc

import List
import Maybe

import IOExts

--------------------------------------------------------------------------------

-- NOTE: this should live in TD_StdDefs, but can't for the time being since
-- that module imports this one (would lead to a circular dependency)
stdConstants = ["True", "False"]


data Quad a b c d = Q1 a | Q2 b | Q3 c | Q4 d | NoQuad	deriving Show


--------------------------------------------------------------------------------
-- function to perform a type check

-- whenever we infer a type for Loc, we start with the given env, and pass the 
-- result to the TCheck fun

type TCheck  = (Loc, Bool)
type TChecks = [TCheck]

lookupTCheck l cs = lookup l cs
addTCheck tc tcs = tc : tcs
noTChecks = []


--------------------------------------------------------------------------------
-- removes from the chr those constraints in its body with an ix not in list
remConsWithoutIxs ixs (Simp n luc lbc (ruc, rbc)) = 
    let ruc' = [ p | p <- ruc, (ucIx p) `elem` ixs ]
	rbc' = [ p | p <- rbc, (bcIx p) `elem` ixs ]
    in  Simp n luc lbc (ruc', rbc')
remConsWithoutIxs ixs _	= error "TD_General.remConsWithoutIxs"

-- removes from the chr those constraints in its body with an ix not in list
remConsWithoutIxsNonNeg ixs (Simp n luc lbc (ruc, rbc)) = 
    let ns   = concatMap unJust ixs
	elem' j = any (`elem`ns) (unJust j)
	ruc' = [ p | p <- ruc, elem' (ucIx p) || any (<0) (unJust $ ucIx p) ]
        rbc' = [ p | p <- rbc, elem' (bcIx p) || any (<0) (unJust $ bcIx p) ]
    in  Simp n luc lbc (ruc', rbc')
remConsWithoutIxsNonNeg ixs _	= error "TD_General.remConsWithoutIxsNonNeg"



-- as above, but discards constraints over types that aren't labelled with one
-- of the specified ixs
remConsWithoutVarIxs ixs (Simp n luc lbcs (rucs, rbcs)) = 
    let	lbcs' = [ p | p <- lbcs, bconsHasVarIxs ixs p ]
	rucs' = [ p | p <- rucs, uconsHasVarIxs ixs p ]
	rbcs' = [ p | p <- rbcs, bconsHasVarIxs ixs p ]
    in  Simp n luc lbcs' (rucs', rbcs')
remConsWithoutVarIxs ixs _ = error "TD_General.remConsWithoutVarIxs"



-- splits a list of CHRs into a pair of those with id, and those without
chrWithId id cs = chrWithId' cs [] []
    where chrWithId' []     m o = (m, o)
	  chrWithId' (c:cs) m o
	    | chrHasId id c = chrWithId' cs (c:m) o
	    | otherwise	    = chrWithId' cs m (c:o)

chrWithIdPrefix id cs = chrWithId' cs [] []
    where chrWithId' []     m o = (m, o)
	  chrWithId' (c:cs) m o
	    | chrHasIdPrefix id c = chrWithId' cs (c:m) o
	    | otherwise		  = chrWithId' cs m (c:o)
				      

-- given a CHR, returns the list of ids of preds on the lhs
chrIds :: CHR -> [Id]
chrIds chr = case chr of
    (Simp _ (UC _ (Name (id,_)) _ _) _ _) -> [id]
    (MSimp _ ucs _ _)  -> map ucId ucs
    (Prop _ _ ucs _ _) -> map ucId ucs
 where ucId (UC _ (Name (id,_)) _ _) = id


chrHasId id chr = id `elem` chrIds chr 

-- FIXME: using this to find chrs from multi-line defs may backfire!!!
chrHasIdPrefix id chr = any (isPrefix id) (chrIds chr)

isPrefix []	_      = True
isPrefix (x:xs) (y:ys) = x == y && isPrefix xs ys
isPrefix _	[]     = False


--fvCHR :: CHR -> [(Index, Name)]
fvCHR (Simp _ luc lbcs (rucs, rbcs)) = 
    nub ((fvUCons luc) ++ (concatMap fvBCons lbcs) ++ 
	 (concatMap fvUCons rucs) ++ (concatMap fvBCons rbcs))
fvCHR (Prop _ _ lucs lbcs (rucs, rbcs)) 
    = nub ((concatMap fvUCons lucs) ++ (concatMap fvBCons lbcs) ++ 
	   (concatMap fvUCons rucs) ++ (concatMap fvBCons rbcs))

--fvCHR (Simp p  c)  = nub ((fvPrimitive p) ++ (fvConstraint c))
--fvCHR (Prop c1 c2) = nub ((fvConstraint c1) ++ (fvConstraint c2))
--fvConstraint (C ps)  = concatMap fvPrimitive ps
--fvPrimitive (B bc _) = fvBCons bc
--fvPrimitive (U uc _) = fvUCons uc
--fvBCons (Eq  t1 t2) = (fvType t1) ++ (fvType t2)
--fvBCons (NEq t1 t2) = (fvType t1) ++ (fvType t2)
fvBCons bc = let t1 = bcLeft bc
		 t2 = bcRight bc
	     in  fvType t1 ++ fvType t2
fvUCons (UC _ _ ts _) = concatMap fvType ts
fvType  (Var (tv, ix))  = [(ix, tv)]
fvType	(TyCon _)    = []
fvType  (TApp t1 t2) = (fvType t1) ++ (fvType t2)


chrsToPrims :: [CHR] -> ([UCons], [BCons])
chrsToPrims [] = ([], [])
chrsToPrims (c:cs) = let (ucs,  bcs)  = chrToPrim c
			 (ucss, bcss) = chrsToPrims cs
		     in  (ucs ++ ucss, bcs ++ bcss)
    where chrToPrim (Simp _ luc  lbcs (rucs, rbcs)) = (luc:rucs, lbcs++rbcs)
	  chrToPrim (Prop _ _ lucs lbcs (rucs, rbcs)) = (lucs++rucs, lbcs++rbcs)


getIxPrimList (ucs, bcs) ix = (doUCs ucs, doBCs bcs)
    where   doUCs [] = []
	    doUCs (uc:ucs)
		| ix == ucIx uc	= uc : doUCs ucs
		| otherwise	= doUCs ucs
	    doBCs [] = []
            doBCs (bc:bcs)
	        | ix == bcIx bc	= bc : doBCs bcs
                | otherwise	= doBCs bcs


eq = Eq

--------------------------------------------------------------------------------


uconsHasVarIx ix uc = (uconsWithVarIx ix uc) /= []
bconsHasVarIx ix bc = (bconsWithVarIx ix bc) /= []
uconsHasVarIxs ixs uc = any (\x -> uconsHasVarIx x uc) ixs
bconsHasVarIxs ixs bc = any (\x -> bconsHasVarIx x bc) ixs


-- finds the CHR whose body contains a variable with the given index 
-- (and all the others)
-- ignores mdef and ann chrs
-- note that it will preferentially select a CHR with the variable in its head
chrWithVarIx :: Loc -> [CHR] -> [(TVar, CHR, [CHR])]
chrWithVarIx ix chrs = 
    case chrWithVarIx' ix chrs [] [] of
	([], _)		 -> []
	(tcs@((t, c):_), cs) -> [(t, c, cs)]
    where 
	chrWithVarIx' _  []		         as cs = (as, cs)
	chrWithVarIx' ix ((Prop _ _ _ _ _):chrs) as cs = 
		chrWithVarIx' ix chrs as cs		    
	chrWithVarIx' ix (chr:chrs) as cs
	 | isMDefCHR chr || isAnnCHR chr = chrWithVarIx' ix chrs as cs
	 | otherwise = 
	    let (ucs, bcs) = simpGetCons chr
		ucs_t = concatMap (uconsWithVarIx ix) ucs
		bcs_t = concatMap (bconsWithVarIx ix) bcs
		ts  = ucs_t ++ bcs_t
	    in  case ts of
		 []	-> chrWithVarIx' ix chrs as (chr:cs)
		 (v:_)	-> chrWithVarIx' ix chrs ((v, chr):as) cs

	simpGetCons (Simp  _ luc  lbcs (rucs, rbcs)) = (luc:rucs, lbcs++rbcs)
	simpGetCons (MSimp _ lucs lbcs (rucs, rbcs)) = (lucs++rucs, lbcs++rbcs)


constraintWithVarIx :: Loc -> Constraint -> [TVar]
constraintWithVarIx ix c = let ucs = consToUs c
			       bcs = consToBs c
			   in  concatMap (uconsWithVarIx ix) ucs ++ 
			       concatMap (bconsWithVarIx ix) bcs

--primitiveWithVarIx :: Index -> ([UCons], [BCons]) -> [TVar]
primitiveWithVarIx ix ([uc], _) = uconsWithVarIx ix uc
primitiveWithVarIx ix (_, [bc]) = bconsWithVarIx ix bc
primitiveWithVarIx _  _ = error "TD_General.primitiveWithVarIx"

goalWithVarIx = primitiveWithVarIx

uconsWithVarIx ix uc = concatMap (typeWithIx ix) (ucTypes uc)
bconsWithVarIx ix bc = let t1 = bcLeft bc
			   t2 = bcRight bc
		       in  (typeWithIx ix t1) ++ (typeWithIx ix t2)

-- FIXME!!! That should probably be a set-equality check (ie. ix `sameSet` ix')
-- (since we consider lists of justifications on constraints and vars)
typeWithIx ix (TApp t1 t2) = (typeWithIx ix t1) ++ (typeWithIx ix t2)
typeWithIx _  (TyCon _)    = []
typeWithIx ix (Var t@(_, ix'))
    | ix == ix'	= [t]
    | otherwise	= []


--------------------------------------------------------------------------------
showCList []     = ""
showCList (c:cs) = (replicate 80 '-') ++ "\n" ++ (show c) ++ "\n" ++
                   (showCList cs)

--------------------------------------------------------------------------------

datDecsToDats = concatMap toDat
    where toDat d = case item d of
	    (DatDec dat)    -> [dat]
	    _		    -> []

nonValDecs :: [Ann Dec] -> [Ann Dec]
nonValDecs = concatMap toNonDef
    where toNonDef d = case item d of
	    (ValDec def)    -> []
	    dec		    -> [d]

-- given a program, returns the list of locations belonging to rule and
-- overload declarations
ruleOverLocs :: Prog -> [Loc]
ruleOverLocs (Prog decs) =
    let o0ds = ((map ixLoc).anns.over0Decs) decs
	ods  = ((map ixLoc).anns.(map overDecToSig).items.overDecs) decs
	rds  = ((map ixLoc).anns.(map ruleDecToRule).items.ruleDecs) decs

    in  o0ds ++ ods ++ rds


----------------------------------------

-- converts an instance declaration into an overload declaration
instToOverDec :: Ann Dec -> Ann Dec
instToOverDec (Ann x0 (InstDec (Ann x1 (InstSig ctxt (Ann x2 (Pred id t)))) 
				adefs)) =
    let sig = Ann x1 (Sig Univ (Ann x2 (varId id)) ctxt t)
    in  Ann x0 (OverDec sig adefs)
instToOverDec adec = adec

----------------------------------------
-- given a list of annotated decs, splits them up into their categories:
-- ([DatDec], [RuleDec], [OverDec], [OverDec0], [ValDec], [AValDec])
splitADecs :: [Ann Dec] -> ([Ann Dec],[Ann Dec],[Ann Dec],
			    [Ann Dec],[Ann Dec],[Ann Dec])
splitADecs ads = split (zip ads (items ads)) ([],[],[],[],[],[])
    where split [] r = r
	  split ((n,e):es) (d,r,o,o0,v,a) 
	    | isDatDec   e = split es (n:d,r,o,o0,v,a)
	    | isRuleDec  e = split es (d,n:r,o,o0,v,a)
	    | isOverDec  e = split es (d,r,n:o,o0,v,a)
	    | isOver0Dec e = split es (d,r,o,n:o0,v,a)
	    | isValDec   e = split es (d,r,o,o0,n:v,a)
	    | isAValDec  e = split es (d,r,o,o0,v,n:a)
	    | otherwise    = split es (d,r,o,o0,v,a)

splitDecs :: [Dec] -> ([Dec],[Dec],[Dec],[Dec],[Dec],[Dec],[Dec])
splitDecs ds = split ds ([],[],[],[],[],[],[])
    where split [] r = r
	  split (e:es) (d,r,o,o0,v,a,i) 
	    | isDatDec   e = split es (e:d,r,o,o0,v,a,i)
	    | isRuleDec  e = split es (d,e:r,o,o0,v,a,i)
	    | isOverDec  e = split es (d,r,e:o,o0,v,a,i)
	    | isOver0Dec e = split es (d,r,o,e:o0,v,a,i)
	    | isValDec   e = split es (d,r,o,o0,e:v,a,i)
	    | isAValDec  e = split es (d,r,o,o0,v,e:a,i)
	    | isInstDec  e = split es (d,r,o,o0,v,a,e:i)
	    | otherwise    = split es (d,r,o,o0,v,a,i) 
	    -- don't care about extern decs


-- returns the number of patterns in a def
defNumPats :: Def -> Int
defNumPats d = length (defPats d)

defPats :: Def -> [Pat]
defPats (Def _ ps _ _) = items ps


datDecs   ds = filter (isDatDec . item) ds
valDecs   ds = filter (isValDec . item) ds
avalDecs  ds = filter (isAValDec . item) ds
hconDecs  ds = filter (isHConDec . item) ds
instDecs  ds = filter (isInstDec . item) ds
externDecs ds = filter (isExternDec . item) ds
interfaceDecs ds = filter (isInterfaceDec . item) ds
ruleDecs  ds = filter (isRuleDec . item) ds
overDecs  ds = filter (isOverDec . item) ds
over0Decs ds = filter (isOver0Dec . item) ds
overDec0s = over0Decs


decDefs (ValDec defs) = defs
decDefs (AValDec _ defs) = defs
decDefs (OverDec _ defs) = defs
decDefs _ = error "decDefs"

hconDecIds (HConDec ids) = ids
hconDecIds _		 = []

decId (ValDec (def:_))  = adefId def
decId (AValDec _ (def:_)) = adefId def
decId (OverDec s _) = case item s of Sig _ id _ _ -> item id
decId (OverDec0 id) = id
decId (InterfaceDec id) = item id
decId (ExternDec (Left s)) = case item s of Sig _ id _ _ -> item id
decId (ExternDec (Right id)) = item id
decId _ = error "decId"

isDatDec  (DatDec _) = True
isDatDec  _	     = False

isValDec  (ValDec _) = True
isValDec  _	     = False

isAValDec (AValDec _ _) = True
isAValDec _		= False

isInstDec (InstDec _ _) = True
isInstDec _		= False

isPatDec (PatDec _ _ _) = True
isPatDec _		= False

isHConDec (HConDec _) = True
isHConDec _	      = False

isExternDec (ExternDec _) = True
isExternDec _		  = False

isInterfaceDec (InterfaceDec _) = True
isInterfaceDec _		= False

isRuleDec (RuleDec _) = True
isRuleDec _	      = False

isOverDec (OverDec _ _) = True
isOverDec _		= False

isOver0Dec (OverDec0 _) = True
isOver0Dec _		= False
isOverDec0 = isOver0Dec

decToSig (AValDec s _) = s
decToSig (OverDec s _) = s
decToSig _	       = error "decToSig"

avalDecToSig (AValDec s _) = s
avalDecToSig _ = error "avalDecToSig"

overDecToSig  (OverDec s _) = s
overDecToSig _ = error "overDecToSig" 

ruleDecToRule (RuleDec r) = r
ruleDecToRule _ = error "ruleDecToRule" 

avalDecToDefs (AValDec _ ds) = ds
avalDecToDefs _ = error "avalDecToDefs" 

valDecToDefs (ValDec ds) = ds
valDecToDefs _           = error "valDecToDefs"

overDecsToDefs = concatMap toDef
    where toDef d = case item d of
			(OverDec _ ds) -> ds
			_	       -> []

valDecsToSigs = concatMap toSig
    where toSig d = case item d of
	    (AValDec sig _) -> [sig]
	    (OverDec sig _) -> [sig]
	    _		    -> []

-- takes a list of Decs and returns a list of the ids belonging to overloaded
-- identifiers
overDecsToIds0 :: [Ann Dec] -> [Id]
overDecsToIds0 [] = []
overDecsToIds0 (ad:ads) = 
    let id = case item ad of
		OverDec _ defs -> [adefId (head defs)]
		OverDec0 id'   -> [id']
		_	       -> []
    in  id ++ overDecsToIds0 ads

-- takes a list of Decs and returns a list of ids belonging to fun. ids
valDecsToIds :: [Ann Dec] -> [Id]
valDecsToIds [] = []
valDecsToIds (ad:ads) = 
    let id = case item ad of
		ValDec defs    -> [adefId (head defs)]
		AValDec _ defs -> [adefId (head defs)]
		_	       -> []
    in  id ++ valDecsToIds ads
    

-- takes a list of annotated decls and for each returns its id and prog location
-- used by inferTypeExp to go straight to the definition of an id
valDecsToLocIds :: [Ann Dec] -> [(Loc, Id)]
valDecsToLocIds ds = [(l, id) | (x, id) <- concatMap toDef ds, let l = ixLoc x ]
    where toDef d = case item d of
	   (ValDec es@(e:_))    -> 
		    [(if singleton es then ann e else ann d, adefId e)]

	   (AValDec _ es@(e:_)) -> [(ann d, adefId e)]
		    
	   (OverDec _ es@(e:_)) -> [(ann d, adefId e)]

	   (OverDec0 id)	-> [(ann d, id)]

	   _ -> []

defId  (Def aid _ _ _) = item aid
defAId (Def aid _ _ _) = aid
adefId  d = case item d of (Def aid _ _ _) -> item aid
adefAId d = case item d of (Def aid _ _ _) -> aid
sigId  (Sig _ aid _ _) = item aid
sigAId (Sig _ aid _ _) = aid
instSigId (InstSig _ p) = case item p of Pred id _ -> id

sigPredLocs (Sig _ _ (Ann _ (Ctxt aps)) _) = map ixLoc (anns aps)

defRHS ad = case item ad of (Def _ _ rhs _) -> rhs
newDefRHS ad rhs = let ix = ann ad
		   in  case item ad of
			Def id ps _ w -> Ann ix (Def id ps rhs w)


valDecsToLocIdsNo0 :: [Ann Dec] -> [(Loc, Id)]
valDecsToLocIdsNo0 ds = 
	[(l, id) | (x, id) <- concatMap toDef ds, let l = ixLoc x ]
    where toDef d = case item d of
	   (ValDec es@(e:_))    -> 
		    [(if singleton es then ann e else ann d, defId e)]

	   (AValDec _ es@(e:_)) -> [(ann d, defId e)]
		    
	   (OverDec _ es@(e:_)) -> [(ann d, defId e)]

	   _ -> []

	  defId d = case item d of (Def aid _ _ _) -> item aid
    
-- True if the named id is an overloaded id within the given decs
isOverloadedId :: [Ann Dec] -> Id -> Bool
isOverloadedId ads id = 
    let ds = items ads
	ods  = filter isOverDec ds
	od0s = filter isOverDec0 ds
	oids  = map decId ods
	o0ids = map decId od0s
	ids   = oids ++ o0ids
    in  id `elem` ids

valDecsToDefs = concatMap toDef
    where toDef d = case item d of
	    (ValDec defs)    -> [defs]
	    (AValDec _ defs) -> [defs]
	    (OverDec _ defs) -> [defs]
	    _		     -> []

-- doesn't include overloads
justValDecsToDefs = concatMap toDef
    where toDef d = case item d of
	    (ValDec defs)    -> [defs]
	    (AValDec _ defs) -> [defs]
	    _		     -> []

overDecsToIds ds = nub (concatMap toId ds)
    where toId d = case item d of
	    (OverDec (Ann _ (Sig _ aid _ _)) _) -> [item aid]
	    _				      -> []

valDecsToLocs :: [Ann Dec] -> [Loc]
valDecsToLocs ds = map ixLoc (concatMap toDef ds)
    where toDef :: Ann Dec -> [Index]
	  toDef d = case item d of
	   (ValDec defs)    -> if singleton defs then anns defs else [ann d]
	   (AValDec _ defs) -> if singleton defs then anns defs else [ann d]
	   (OverDec _ defs) -> if singleton defs then anns defs else [ann d]
	   _		    -> []


decsToAnnOverDec0s = concatMap toOverDec0
    where toOverDec0 d = case item d of
	    (OverDec0 _) -> [d]
	    _		 -> []

decsToAnnOverDecs = concatMap toOverDec
    where toOverDec d = case item d of
            (OverDec _ _) -> [d]
	    _             -> []

decsToOverDecs = concatMap toOverDec
    where toOverDec d = case item d of
	    o@(OverDec _ _) -> [o]
	    _		    -> []

decsToOverDec0s = concatMap toOverDec0
    where toOverDec0 d = case item d of
	    o@(OverDec0 _) -> [o]
	    _		   -> []


decsToRuleDecs = concatMap toRuleDec
    where toRuleDec d = case item d of
	    r@(RuleDec _)   -> [r]
	    _		    -> []

decsToRules = concatMap toRuleDec
    where toRuleDec d = case item d of
	    RuleDec r   -> [r]
	    _		-> []

overDefs :: Ann Dec -> [Ann Def]
overDefs adec = 
    case item adec of
	 (OverDec _ ad) -> ad
	 _		-> error "TD_General.overDefs: not an OverDec"

annDefId :: Ann Def -> String
annDefId (Ann _ (Def (Ann _ id) _ _ _)) = id

newADefId :: Id -> Ann Def -> Ann Def
newADefId id (Ann ix1 (Def (Ann ix2 _) ps rhs w)) = 
	     (Ann ix1 (Def (Ann ix2 id) ps rhs w))

addPatsToADef :: [Ann Pat] -> Ann Def -> Ann Def
addPatsToADef ps' (Ann ix1 (Def id ps rhs w)) =
		  (Ann ix1 (Def id (ps' ++ ps) rhs w))

-- replace all occurrences of locations in list to corresponding expression
-- note that we only consider replacing VarExps (since this routine is invoked
-- as part of the evidence translation and we're just interested in replacing
-- occurrences of overloaded ids)
applyExpUpdates :: [(Loc, Exp)] -> Ann Def -> Ann Def
applyExpUpdates les def = chDef def		    
    where
	chRHS (RHS1 ae)  = RHS1 (chExp ae)
	chRHS (RHSG ges) = RHSG [ (chExp g, chExp e) | (g,e) <- ges ]

	chDecs ads = map chDec ads
	chDec  ad  = let d  = item ad
			 ix = ann ad
		     in  case d of
			ValDec afs    -> Ann ix (ValDec (chDefs afs))
			AValDec s afs -> Ann ix (AValDec s (chDefs afs))
			_	      -> ad
			
	chDefs ads = map chDef ads
	chDef  ad  = let d  = item ad 
			 ix = ann ad
		     in  case d of
			Def id ps rhs w -> Ann ix (Def id ps (chRHS rhs)
							     (chDecs w))
	chStmt stmt = let ix = ann stmt
		      in  Ann ix $ case item stmt of
	    GenStmt p e -> GenStmt p (chExp e)
	    QualStmt e  -> QualStmt e
	    LetStmt ds  -> LetStmt (chDecs ds)

	chAlt (CaseAlt p e w) = CaseAlt p (chExp e) (chDecs w)
	chAlt (CaseAltGuard p ges w) = CaseAltGuard p (map chGE ges) (chDecs w)
	    where chGE (g,e) = (chExp g, chExp e)
	    
	chExp aexp =
	    let exp = item aexp
		ix  = ann aexp
		fnd = lookup (ixLoc ix) les
	        exp' = if not (isIxLoc ix) 
		       then exp
		       else case exp of
		    e@(VarExp id) -> if isJust fnd then fromJust fnd
						   else e
		    e@(ConExp id) -> e
		    e@(LitExp _)  -> if isJust fnd then fromJust fnd
						   else e
		    AbsExp id e1  -> AbsExp id (chExp e1)
		    AppExp e1 e2  -> AppExp (chExp e1) (chExp e2)
		    IfxAppExp e1 e2 e3 -> IfxAppExp (chExp e1) (chExp e2) 
						     (chExp e3)
		    IfExp  e1 e2 e3 -> IfExp (chExp e1) (chExp e2) (chExp e3)
		    AnnExp e1 t	  -> AnnExp (chExp e1) t
		    TAddExp e1 t  -> TAddExp (chExp e1) t
		    TupExp es	  -> TupExp (map chExp es)
		    LetExp ds e1  -> LetExp (chDecs ds) (chExp e1)
		    CaseExp e as  -> CaseExp (chExp e) (map chAlt as)
		    ListExp es	     -> ListExp (map chExp es)
		    ListCompExp p ss -> ListCompExp p (map chStmt ss)

		    SecExpLeft  e1 e2 -> SecExpLeft  (chExp e1) (chExp e2)
		    SecExpRight e1 e2 -> SecExpRight (chExp e1) (chExp e2)
		    
		    EnumExpFrom e1 -> EnumExpFrom (chExp e1)
		    EnumExpFromTo e1 e2	-> EnumExpFromTo (chExp e1) (chExp e2)
		    EnumExpFromThen e1 e2 -> EnumExpFromThen (chExp e1) 
							     (chExp e2)
		    EnumExpFromThenTo e1 e2 e3 -> EnumExpFromThenTo (chExp e1)
							 (chExp e2) (chExp e3)

	    in  Ann ix exp'

--------------------------------------------------------------------------------

-- finds the first (only) program location with the provided annotation and all
-- sub-locations... also returns the name of the def within which it occurs
-- (note new addition: the Bool indicates whether the index belongs to a var...
--  True if so, False otherwise)
findIxProg :: Index -> Prog -> [(Index, Bool)]
findIxProg ix (Prog decs) = concatMap (findIxDec ix) decs

findIxDecs ix decs = concatMap (findIxDec ix) decs
findIxDec :: Index -> Ann Dec -> [(Index, Bool)]
findIxDec ix dec
    | ix == ann dec = subAnnsDec dec
    | otherwise	    =
	case item dec of
	    ValDec defs	    -> concatMap doDef defs

	    AValDec s defs  -> concatMap doDef defs ++ subAnnsSig s

	    OverDec s defs  -> concatMap doDef defs ++ subAnnsSig s

	    dec		    -> []

    where doDef def = let (_id, ixs) = findIxDef ix def
		      in  if null ixs then []
				      else ixs -- [(id, ixs)]

getDecId (ValDec def)   = let Def id _ _ _ = item (head def) in item id
getDecId _	        = error "TD_General.getDecId"

getDefId (Def id _ _ _) = item id

findIxDef :: Index -> Ann Def -> (Id, [(Index, Bool)])
findIxDef ix def
    | ix == ann def = (getDefId (item def), subAnnsDef def)
    | otherwise	    =
	case item def of
	    Def id ps rhs w -> (item id, findIxPats ix ps ++ findIxRHS ix rhs ++
					 findIxDecs ix w)

findIxRHS ix (RHS1 ae)  = findIxExp ix ae
findIxRHS ix (RHSG ges) = concatMap (findIxGE ix) ges

findIxGE ix (g,e) = findIxExp ix g ++ findIxExp ix e

-- finds the first (only!) subexpression with the given annotation, and returns
-- a list of the annotations on all its subexpressions
--
findIxExps ix = concatMap (findIxExp ix)
findIxExp :: Index -> Ann Exp -> [(Index, Bool)]
findIxExp ix exp 
    | ix == ann exp = subAnnsExp exp
    | otherwise	    =
	case item exp of
            VarExp id	    -> []
            ConExp id	    -> []
	    LitExp lit	    -> []
            AbsExp ps e1    -> findIxPats ix ps ++ findIxExp ix e1	      
            AppExp e1 e2    -> findIxExp ix e1 ++ findIxExp ix e2
	    IfxAppExp e1 e2 e3 -> findIxExp ix e1 ++ findIxExp ix e2 ++
				  findIxExp ix e3
            IfExp  e1 e2 e3 -> findIxExp ix e1 ++ findIxExp ix e2 ++ 
			       findIxExp ix e3
            AnnExp e1 ts    -> findIxExp ix e1 ++ findIxTSchm ix ts
	    TAddExp e1 ts   -> findIxExp ix e1 ++ findIxTSchm ix ts
            TupExp  es      -> findIxExps ix es
	    LetExp ds e1    -> findIxExp ix e1 ++ findIxDecs ix ds
	    CaseExp e as    -> findIxExp ix e ++ findIxAlts ix as

            ListExp es       -> findIxExps ix es
	    ListCompExp e ss -> findIxExp ix e ++ findIxStmts ix ss
	    
	    SecExpLeft  e1 e2 -> findIxExp ix e1 ++ findIxExp ix e2
	    SecExpRight e1 e2 -> findIxExp ix e1 ++ findIxExp ix e2

	    EnumExpFrom       e1	-> findIxExp ix e1
	    EnumExpFromTo     e1 e2	-> findIxExp ix e1 ++ findIxExp ix e2
	    EnumExpFromThen   e1 e2	-> findIxExp ix e1 ++ findIxExp ix e2
	    EnumExpFromThenTo e1 e2 e3  -> findIxExp ix e1 ++ findIxExp ix e2 ++
					   findIxExp ix e3


findIxAlts ix = concatMap (findIxAlt ix)
findIxAlt ix (CaseAlt p e w) = findIxPat ix p ++findIxExp ix e ++findIxDecs ix w
findIxAlt ix (CaseAltGuard p ges w) = findIxPat ix p ++ 
				concatMap (findIxGE ix) ges ++ findIxDecs ix w
				      
findIxStmts ix = concatMap (findIxStmt ix)
findIxStmt ix stmt 
    | ix == ann stmt = subAnnsStmt stmt
    | otherwise	     = case item stmt of
	    GenStmt p e -> findIxPat ix p ++ findIxExp ix e
	    QualStmt e  -> findIxExp ix e
	    LetStmt ds  -> findIxDecs ix ds
				      
findIxPats ix = concatMap (findIxPat ix)
--findIxPat :: Index -> Ann Pat -> [Index]
findIxPat ix pat
    | ix == ann pat = subAnnsPat pat
    | otherwise	    =
	case item pat of
	    JokPat	 -> []
	    VarPat id	 -> []
	    LitPat  l	 -> []
	    ConPat id ps -> findIxId ix id ++ findIxPats ix ps
	    TupPat  ps	 -> findIxPats ix ps
	    ListPat ps	 -> findIxPats ix ps

findIxTSchm ix tschm   
    | ix == ann tschm = subAnnsTSchm tschm
    | otherwise	      =
	case item tschm of
	    TSchm c t -> []	-- can't look into type schemes


findIxTyps ix = concatMap (findIxTyp ix)
--findIxTyp :: Index -> Ann Typ -> [Index]
findIxTyp ix typ
    | ix == ann typ = subAnnsTyp typ
    | otherwise	    =
	case item typ of
	    VarTyp id	 -> []
	    ConTyp id	 -> []
	    AppTyp t1 t2 -> findIxTyp ix t1 ++ findIxTyp ix t2
	    TupTyp  ts	 -> findIxTyps ix ts
	    ListTyp t1	 -> findIxTyp ix t1

--findIxId :: Index -> Ann Id -> [Index]
findIxId ix id 
    | ix == ann id  = [(ix, True)]
    | otherwise	    = []



----------------------------------------

-- NOTE: also include the location of the annotation (if any)
subAnnsDec dec =
    let a  = ann dec
	af = (a, False)
    in  case item dec of
	    ValDec defs	   -> af : concatMap subAnnsDef defs
	    AValDec s defs -> af : subAnnsSig s ++ concatMap subAnnsDef defs
	    OverDec s defs -> af : subAnnsSig s ++ concatMap subAnnsDef defs
	    InstDec s defs -> af : subAnnsInstSig s ++ concatMap subAnnsDef defs

	    OverDec0 _	   -> [af]
	    ExternDec _	   -> [af]
	    x	-> error ("TD_General.subAnnsDec: missing case for: " ++ show x)

subAnnsDecs = concatMap subAnnsDec

subAnnsDefs = concatMap subAnnsDef
subAnnsDef :: Ann Def -> [(Index, Bool)]
subAnnsDef def =
    let a  = ann def
	af = (a, False)
    in	case item def of
	    Def id ps rhs w -> af : subAnnsPats ps ++ subAnnsRHS rhs ++
				    subAnnsDecs w

subAnnsRHS (RHS1 ae)  = subAnnsExp ae
subAnnsRHS (RHSG ges) = subAnnsGEs ges

subAnnsGEs = concatMap subAnnsGE
subAnnsGE (g, e) = subAnnsExp g ++ subAnnsExp e

-- takes an annotated expression and returns the list of indices in its
-- subexpressions
-- NOTE: does not descend into sub-declarations of let expression!
subAnnsExps = concatMap subAnnsExp
--subAnnsExp :: Ann Exp -> [Index]
subAnnsExp exp =
    let a  = ann exp
	af = (a, False)
	at = (a, True)
    in	case item exp of
	    VarExp id	    -> if id `notElem` stdConstants then [at] else [af]
	    ConExp id	    -> if id `notElem` stdConstants then [at] else [af]
	    LitExp lit	    -> [af]
	    AbsExp ps e1    -> af : subAnnsPats ps ++ subAnnsExp e1
	    AppExp e1 e2    -> af : subAnnsExp e1 ++ subAnnsExp e2
	    IfxAppExp e1 e2 e3 -> af : subAnnsExp e1 ++ subAnnsExp e2 ++
				       subAnnsExp e3
	    IfExp  e1 e2 e3 -> af : subAnnsExp e1 ++ subAnnsExp e2 ++
				   subAnnsExp e3
	    AnnExp e1 t     -> af : subAnnsExp e1 ++ subAnnsTSchm t
	    TAddExp e1 t    -> af : subAnnsExp e1 ++ subAnnsTSchm t
	    TupExp  es      -> af : subAnnsExps es
	    LetExp ds e1    -> af : subAnnsExp e1 ++ subAnnsDecs ds
	    CaseExp e as    -> af : subAnnsExp e ++ subAnnsAlts as
	    ListExp es       -> af : subAnnsExps es
	    ListCompExp e ss -> af : subAnnsExp e ++ subAnnsStmts ss

	    SecExpLeft  e1 e2 -> af : subAnnsExp e1 ++ subAnnsExp e2
	    SecExpRight e1 e2 -> af : subAnnsExp e1 ++ subAnnsExp e2

	    EnumExpFrom       e1	-> af : subAnnsExp e1
	    EnumExpFromTo     e1 e2	-> af : subAnnsExp e1 ++ subAnnsExp e2
	    EnumExpFromThen   e1 e2	-> af : subAnnsExp e1 ++ subAnnsExp e2
	    EnumExpFromThenTo e1 e2 e3  -> af : subAnnsExp e1 ++ subAnnsExp e2
							      ++ subAnnsExp e3 

subAnnsAlts = concatMap subAnnsAlt
subAnnsAlt (CaseAlt p e w) = subAnnsPat p ++ subAnnsExp e ++ subAnnsDecs w
subAnnsAlt (CaseAltGuard p ges w) = subAnnsPat p ++ subAnnsDecs w ++ 
				    subAnnsGEs ges

subAnnsStmts = concatMap subAnnsStmt
subAnnsStmt stmt =
    let a  = ann stmt
	af = (a, False)
    in  case item stmt of
	    GenStmt p e -> af : subAnnsPat p ++ subAnnsExp e
	    QualStmt e  -> af : subAnnsExp e
	    LetStmt ds  -> af : subAnnsDecs ds


subAnnsPats = concatMap subAnnsPat
--subAnnsPat :: Ann Pat -> [Index]
subAnnsPat pat =
    let a = ann pat
	af = (a, False)
	at = (a, True)
    in  case item pat of
	    JokPat	 -> [af]
            VarPat id	 -> [at]
	    LitPat l	 -> [af]
            ConPat id ps -> af : (ann id, False) : subAnnsPats ps
	    TupPat  ps	 -> af : subAnnsPats ps
            ListPat ps	 -> af : subAnnsPats ps


subAnnsTyps = concatMap subAnnsTyp
subAnnsTyp :: Ann Typ -> [(Index, Bool)]
subAnnsTyp typ =
    let a = ann typ
	af = (a, False)
	at = (a, True)
    in	case item typ of
	    VarTyp id	 -> [at]
            ConTyp id	 -> [af]
            AppTyp t1 t2 -> af : subAnnsTyp t1 ++ subAnnsTyp t2
            TupTyp  ts	 -> af : subAnnsTyps ts
            ListTyp t1	 -> af : subAnnsTyp t1

subAnnsTSchm tschm =
    case item tschm of
	TSchm ctxt typ -> (ann tschm, False) : 
			  subAnnsCtxt ctxt ++ subAnnsTyp typ

subAnnsSig :: Ann Sig -> [(Index, Bool)]
subAnnsSig sig = 
    let a  = ann sig
	af = (a, False)
    in  case item sig of
	    Sig _ id ctxt t   -> af : subAnnsCtxt ctxt ++ subAnnsTyp t

subAnnsInstSig :: Ann InstSig -> [(Index, Bool)]
subAnnsInstSig isig =
    let a  = ann isig
	af = (a, False)
    in  case item isig of
	    InstSig ctxt p -> af : subAnnsCtxt ctxt ++ subAnnsPred p

subAnnsCtxt :: Ann Ctxt -> [(Index, Bool)]
subAnnsCtxt ctxt = 
    let af = (ann ctxt, False)
    in  case item ctxt of
	    Ctxt aps -> af : concatMap subAnnsPred aps

subAnnsPred :: Ann Pred -> [(Index, Bool)]
subAnnsPred pred =
    let af = (ann pred, False)
    in  case item pred of
	    Pred _ at -> af : subAnnsTyp at

--------------------------------------------------------------------------------

-- given a Prog, returns all the decs within it (including within expressions)
allProgDecs (Prog adecs) = concatMap allDecDecs adecs

allDecsDecs = concatMap allDecDecs
allDecDecs adec =
    case item adec of
	    ValDec defs	   -> adec : allDefsDecs defs
	    AValDec _ defs -> adec : allDefsDecs defs
	    OverDec _ defs -> adec : allDefsDecs defs
	    _		  -> [adec] 

allDefsDecs = concatMap allDefDecs
allDefDecs :: Ann Def -> [Ann Dec]
allDefDecs def =
    case item def of
	    Def _ _ rhs w -> allRHSDecs rhs ++ allDecsDecs w

allRHSDecs (RHS1 ae)  = allExpDecs ae
allRHSDecs (RHSG ges) = allGEsDecs ges

allGEsDecs = concatMap allGEDecs
allGEDecs (g, e) = allExpDecs g ++ allExpDecs e

allExpsDecs = concatMap allExpDecs
allExpDecs exp =
       case item exp of
	    LetExp ds e -> ds ++ allExpDecs e
	    _		-> []
	    

--------------------------------------------------------------------------------
-- program post-processing
-- (merge clauses of the same def into common dec)
-- call after parsing, and before any additional processing


mergeDefs :: [Ann Dec] -> [Ann Dec]
mergeDefs ads = mergeValDefs $ mergeOverDefs $ mergeAValDefs ads

mergeOverDefs :: [Ann Dec] -> [Ann Dec]
mergeOverDefs []       = []
mergeOverDefs (ad:ads) =
    let dec = item ad
        ix  = ann ad
    in  case dec of
         (OverDec s ds) ->
            let id = annDefId (head ds)
                vals = takeWhile isValDec (items ads)
                same = takeWhile (\d -> id == (annDefId . head . valDecToDefs) d) vals
                od = Ann ix (OverDec s (ds ++ (concatMap valDecToDefs same)))
            in  od : mergeOverDefs (drop (length same) ads)

         _              -> ad : mergeOverDefs ads

mergeValDefs :: [Ann Dec] -> [Ann Dec]
mergeValDefs []       = []
mergeValDefs (ad:ads) =
    let dec = item ad
	ix  = ann ad
    in  case dec of
	 (ValDec ds) -> 
	    let id = annDefId (head ds)
		vals = takeWhile isValDec (items ads)
		same = takeWhile (\d -> id == (annDefId . head . valDecToDefs) d) vals
		vd = Ann ix (ValDec (ds ++ (concatMap valDecToDefs same)))
	    in  vd : mergeValDefs (drop (length same) ads)

	 _	    -> ad : mergeValDefs ads


mergeAValDefs :: [Ann Dec] -> [Ann Dec]
mergeAValDefs []       = []
mergeAValDefs (ad:ads) =
    let dec = item ad
	ix  = ann ad
    in  case dec of
	 (AValDec s ds) -> 
	    let id = annDefId (head ds)
		vals = takeWhile isValDec (items ads)
		same = takeWhile (\d -> id == (annDefId . head . valDecToDefs) d) vals
		vd = Ann ix (AValDec s (ds ++ (concatMap valDecToDefs same)))
	    in  vd : mergeAValDefs (drop (length same) ads)

	 _	    -> ad : mergeAValDefs ads


-- if we find an AValDec or OverDec with a mismatching definition, unpack the
-- definition and put it in its own ValDec (the AValDec or OverDec will later be
-- found to be a dangling annotation)
unpackMismatchedDecs :: [Ann Dec] -> [Ann Dec]
unpackMismatchedDecs = concatMap unpack
    where
	-- note: null annotations are okay here since this'll be caught as an
	-- error anyway
	unpack av@(Ann x1 (AValDec s f@[Ann x2 d])) 
	    | sigId (item s) /= defId d = 
				   [Ann x1 (AValDec s []), nullAnn (ValDec f)]
	    
	unpack av@(Ann x1 (OverDec s f@[Ann x2 d])) 
	    | sigId (item s) /= defId d = 
				   [Ann x1 (OverDec s []), nullAnn (ValDec f)]

	unpack ad = [ad]

--------------------------------------------------------------------------------

withVar v [] = []
withVar v (b@(Eq (Var t) _ _):bs)
  | v == t      = [b]
  | otherwise   = withVar v bs
withVar _ _     = error "TypeDebug.doCommand.withVar"


--------------------------------------------------------------------------------


