--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- contains definition for the type debugger monad (used to live in TypeDebug,
-- but this lead to circular module dependencies - TypeDebug imports a lot of
-- support modules... some of which need the TD defintion

module TD_Monad (module TD_Monad, module Ann_InputAST)
where

import qualified TD_Translate as Trans 
import Ann_InputAST (Prog(..), Def(..), Pat(..), Exp(..), RHS(..),
		     Ann(..), item, ann, Dec(..), GuardedExp,
		     ixLoc, Loc, noLoc)
import Justify
import TD_General
import Pretty
import Core hiding (newTVar)

import Maybe
import IOExts
import List (nub)
import Chr
--import Solver
import Misc

import OptFlags

import System (exitFailure)


--------------------------------------------------------------------------------
-- 
data State = St {prog::Prog, chrs::[CHR], schrs::[CHR], jprop::JustProp,
		 counter::Integer, environment :: Trans.Env, 
		 results::[Result], stddef::[Name], solvopt::SOpt,
		 optionFlags::[Flag], filename::String,
		 haskellConNames :: [Name]}
    deriving Show

-- used to store the result of the last inferred type
-- * note that locations are always paired with a type variable... solving the
--   constraint for this variable will reveal the type of that location
-- * first loc is the loc whose type was requested (Actually the (l,t,v) head
--   of the corresponding CHR, where the l is the type of the specific
--   expression of interest - but the t and v come from the CHR)
-- * list of locs represents sub-locations which correspond to vars
-- * first constraint if the result of the derivation (unsimplified)
-- * second is simplified so result can be easily read off
type LTVar = (Loc, TVar)

data Result = Result ((LTVar, LTVar, LTVar), [LTVar], [LTVar],
		      (Constraint,Constraint), Prog) 
	    | Conflict (SrcPos, (Loc,TVar), (([Loc], Bool), [[Loc]]), Constraint) -- inconsistent constraint
	    | Rejected Result
	    | NoResult
	    
isResult (Result _) = True
isResult _	    = False
isConflict (Conflict _)	= True
isConflict _		= False
isRejected (Rejected _)	= True
isRejected _		= False

noResult NoResult = True
noResult _	  = False

locResult (Result r) = let (((l,_),_,_),_,_,_,_) = r in l
locResult _	     = error "TD_Monad.locResult"
locConflict (Conflict c) = let (_, l, _, _) = c in l
locConflict _            = error "TD_Monad.locConflict"

unResult (Result r) = r
unResult _	    = error "TD_Monad.unResult"
unConflict (Conflict c) = c
unConflict _		= error "TD_Monad.unConflict"


hedd []     = error "HEDD"
hedd (x:xs) = x

instance Show Result where

    show (Result ((lt1,_,_), lts, nk_lts, (_,c), p)) = 
{-	let (_, v) = lt1
	    ucs = consToUs c
	    bcs = consToBs c
	    bc  = bcs `withLHSVar` v 
	    typ = bcRight bc
	    vs  = nub (fv typ ++ fv ucs)
	    phi = simpTVarSubst vs

	    typStr  = show (apply phi typ)
	    ctxtStr = if null ucs then ""
				  else showSpaceList (apply phi ucs) ++ " => "
	    
	in  ctxtStr ++ typStr 

	where withLHSVar [] _ = error "TD_Monad.show.withLHSVar"
	      withLHSVar (b@(Eq (Var t) _ _):bs) v
		| v == t      = b
		| otherwise   = bs `withLHSVar` v
	      withLHSVar _ _  = error "TD_Monad.show.withLHSVar"
-}
    
	let pil (tv, c) = 
	      let v   = tv
		  res = let ps   = consToUs c 
			    eqs  = consToBs c  
			    eqs' = withVar v eqs
			    c1 = usToCons ps
			    c2 = bsToCons eqs'
			    c' = c1 `cand` c2
			in  c' 
		  
	      in ("", res)	-- just display the type, not the def/exp

	    withVar v []     = []
	    withVar v (b@(Eq (Var t) _ _):bs)
	      | v == t	    = [b]
	      | otherwise   = withVar v bs
	    withVar _ _	    = error "TD_Monad.show.withVar"

	    formatRes subst (C ucs [bc]) =
		let typ  = bcRight bc
		    ucs' = apply subst ucs -- cullUCs ucs (map fst (fv typ))
		    ctxt_str = let str = showCommaList ucs' 
			       in  if ucs' /= [] then "(" ++ str ++ ") => "
					         else ""
		    typ_str = show (apply subst typ)
		in  ctxt_str ++ typ_str
	    formatRes _ c = trace ("TD_Monad.formatRes failed on constraint" ++
				 show c ++ "\n") (show c)

	    cullUCs []     _  = []
	    cullUCs (u:us) vs
		| any (`elem` vs) (map fst (fv u)) = u : cullUCs us vs
		| otherwise		     = cullUCs us vs

	    print res = let form = map pil res
			    (_strs, typs) = unzip form
			    cons = foldr cand trueCons typs
			    -- get rid of types on the lhs of Eqs
			    rhs = map bcRight (consToBs cons)
			    fvs = nub (fv (consToUs cons) ++ fv rhs)
			    phi = -- idSubst 
				  simpTVarSubst fvs
			in  unlines [ s ++ formatRes phi t | (s, t) <- form ]

	in  print (zip (map snd ((lt1:lts) )) (repeat c))


    show (Conflict ls)	= "conflicts: " ++ show ls

    show (Rejected r)	= "rejected " ++ show r

    show NoResult	= "no result"


-- Type debugger monad (contains state for debugger)... can also do IO
newtype TD a = TD (State -> IO (E (a, State)))

instance Monad TD where

    return v	 = TD (\st -> return (Succ (v, st)))
    
    (TD a) >>= b = TD (\st -> do res <- a st
				 case res of
				    Err e	  -> return (Err e)
				    Succ (a',st') -> let TD b' = b a'
						     in  b' st')

    fail s	 = TD (\st -> return (Err s))


-- not sure if this works
-- FIXME: test!
catchTD td a = do st <- getState
		  res <- doIO $ runTD td st
		  case res of
		       Err e	      -> return a
		       Succ (a', st') -> do setState st'
					    return a'

-- takes a TP calculation and returns the same calculation in TD
doTP = tdRunTP
tdRunTP :: TP a -> TD (E a)
tdRunTP tp = do n <- getCounter
		let res = unTP tp (fromIntegral n)
		case res of
		     Err e	  -> return (Err e)
		     Succ (st, a) -> do setCounter (fromIntegral st)
					return (Succ a)

-- as above, but for a TR action
doTR = tdRunTR
tdRunTR :: Trans.TR a -> TD (E a)
tdRunTR tr = do n <- getCounter
		res <- Trans.runTRInit (fromIntegral n) tr
		case res of
		     Err e	  -> return (Err e)
		     Succ (a, st) -> do setCounter (fromIntegral (Trans.num st))
					return (Succ a)


doIO io = TD (\st -> do res <- io
			return (Succ (res, st)))

puts str = doIO (putStrLn str)

bailToOS str = puts str >> doIO exitFailure

runTD :: TD a -> State -> IO (E (a, State))
runTD (TD p) st = p st

------------------------------------------------------------
---- functions for examining and updating the state
trGet    f = TD (\st -> return (Succ (f st, st)))
trUpdate f = TD (\st -> return (Succ ((), f st)))

getState    = TD (\st -> return (Succ (st, st)))
setState st = TD (\_ -> return (Succ ((), st)))

prog'	 f st = st {prog = f (prog st)}
chrs'	 f st = st {chrs = f (chrs st)}
schrs'	 f st = st {schrs = f (schrs st)}
jprop'   f st = st {jprop = f (jprop st)}
stddef'	 f st = st {stddef = f (stddef st)}
counter' f st = st {counter = f (counter st)}
results' f st = st {results = f (results st)}
solvopt' f st = st {solvopt = f (solvopt st)}
filename' f st= st {filename = f (filename st)}
environment' f st = st {environment = f (environment st)}

getProg    = trGet prog
getCHRs    = trGet chrs
getSCHRs   = trGet schrs
getJProp   = trGet jprop
getCounter = trGet counter
getEnv	   = trGet environment
getResults = trGet results
getSolvOpt = trGet solvopt
getStdDef  = trGet stddef
getFilename = trGet filename
getOptionFlags = trGet optionFlags
getHaskellConNames = trGet haskellConNames

setProg	p = trUpdate $ prog' (const p)
setSolvOpt s = trUpdate $ solvopt' (const s)

addCHR  chr  = trUpdate $ chrs' (chr:)
setCHRs chrs = trUpdate $ chrs' (const chrs)
setSCHRs schrs = trUpdate $ schrs' (const schrs)

setJProp jp  = trUpdate $ jprop' (const jp)

-- assume there's always a previous result... too much checking otherwise
newResult res = trUpdate $ results' (res:)
clrResults    = trUpdate $ results' (const [])
setResults ress = trUpdate $ results' (const ress)

lastResult = do rs <- getResults 
		if null rs then error "TD_Monad.lastResult: no last result!"
			   else return (head rs)
		    
replaceLastResult r = do rs <- getResults
			 if null rs 
			    then trace ("TD_Monad.replaceLastResult:" ++ 
					"no last result!") newResult r
			    else let rs' = r : tail rs
				 in  setResults rs'


incCounter = trUpdate $ counter' (+1)
setCounter n = trUpdate $ counter' (const n)
next = do n <- getCounter
	  incCounter
	  return n

newVar = do v <- newTVar
	    return (Var (v, noLoc))

newLTVar = do v <- newTVar
	      return (noLoc, (v, noLoc))

newTVar    = do n <- next
		let id = 'x':(show n)
		let nm = name id n
		return nm

newTypeVar id = do n <- next
		   let nm = name id n
		   return (Var (nm, noLoc))

lookupPred id = do env <- getEnv
		   return (Trans.lookupV id env)

newEnv env = trUpdate (environment' (const env))

initTDState = St {prog = Prog [], chrs = [], schrs = [], counter = 0, 
		  jprop = LocalJust, environment = Trans.emptyEnv,results = [], 
		  stddef = [], solvopt = SInference, optionFlags = [], 
		  filename = "", haskellConNames = [] }

genTDState p c sc e st sd so jp fs fn hcs = 
    St {prog = p, counter = count, environment = e, results = [],
	chrs = c, schrs = sc, stddef = sd, jprop = jp,
	solvopt = so, optionFlags = fs, filename = fn,
	haskellConNames = hcs }

    where count = Trans.num st


--------------------------------------------------------------------------------

-- takes a list of locations (to highlight) a location (to display) and the
-- program itself, returns the formatted string
dispLoc :: [Loc] -> Loc -> Prog -> String
dispLoc ls loc prog = hd (getLocProg loc prog)

    where   hd [] = "" 
		  -- trace ("dispLoc: "++show loc++" not found in program!") ""
	    hd xs = head xs
	    
	    getLocProg loc (Prog decs) =
		let mdecs = filter (\d -> let i = item d 
				          in isValDec i || isAValDec i ||
					     isOverDec i) decs
		    mdecstrs = concatMap (getMDecLoc loc) mdecs
		    defs = concat (valDecsToDefs decs)
		    sigs = valDecsToSigs decs
		    sigstrs = concatMap (getSigLoc loc) sigs
		    odefs= overDecsToDefs decs
		    o0decs = (overDec0s decs)
		    o0strs = map (decId . item)
				 (filter (\x-> ixLoc (ann x)==loc) o0decs)
		    rs	 = decsToRules decs
		    rstrs  = map show (filter (\mv-> ixLoc (ann mv) == loc) rs)
		    opt xs exp = if not (null xs) then xs 
						  else exp
		in (opt o0strs
		    (opt sigstrs
		     (opt mdecstrs
		      (opt rstrs
		       (concatMap (getLocDef loc) (defs ++ odefs))))))
	    
	    getMDecLoc loc dec =
		if locMatches loc dec
		   then [showit' dec]
		   else []
		
	    getSigLoc loc sig =
		if locMatches loc sig
		   then [showit sig]
		   else []

	    getLocDef loc def =
	        case item def of
		   Def id pats rhs w	-> if locMatches loc def
					    then [showit def]
					    else concatMap (getLocPat loc) pats
						   ++ getLocRHS loc rhs

	    getLocRHS loc (RHS1 exp) = getLocExp loc exp
	    getLocRHS loc _ = [] -- exit "dispLoc: can't look into RHS"
	    
	    getLocExp loc exp = 
	        if locMatches loc exp
		   then [showit exp]
		   else let gle = getLocExp loc
			in  case item exp of
				AbsExp id e1    -> gle e1
				AppExp e1 e2    -> gle e1 ++ gle e2
				IfExp e1 e2 e3  -> gle e1 ++ gle e2 ++ gle e3
				AnnExp e1 t	-> gle e1
				TupExp es	-> concatMap gle es
				ListExp es	-> concatMap gle es
				_		-> []

	    getLocPat loc pat = 
	       if locMatches loc pat
	   	then [showit pat]
	   	else let glp = getLocPat loc                    
           	     in  case item pat of
           		    ConPat id ps    -> getLocId loc id ++ 
					       concatMap glp ps
           		    TupPat ps	    -> concatMap glp ps
	   		    ListPat ps	    -> concatMap glp ps
			    _		    -> []
	    
	    getLocId loc id = if locMatches loc id
				then [show (ConExp (item id))] -- hacky
				else []
	    
	    locMatches loc x = ixLoc (ann x) == loc				

	    showit x = if null ls then show x
				  else pprintHL ls x

	    -- version of `showit' that removes the annotation first
	    -- use this when we don't want the outer-most statement to be
	    -- highlighted (eg. for whole declarations)
	    showit' x = if null ls then show x
				   else pprintHL ls x 

----------------------------------------
dispCHRLoc :: [CHR] -> Loc -> String
dispCHRLoc chrs l =
    let (ss, ps) = partitionF isSimp id chrs
	ssl = zip (map (heady "1" . unJust') ss) ss
	psl = zip (map (heady "2" . unJust') ps) ps
	slu = lookup l ssl
	plu = lookup l psl
    in  if isJust slu 
	   then show (presentableCHR (fromJust slu))
	   else if isJust plu
		   then show (presentableCHR (fromJust plu))
		   else trace ("dispCHRLoc: location " ++ show l ++ 
			       " not found!") "" 	
	-- ugly hack ... only necessary for "standard library" CHRs (which are
	-- not justified at all)
    where unJust' x = let j = unJust x 
		      in  if null j
			     then [-1]
			     else j
			
--------------------------------------------------------------------------------
-- functions for showing program content at a given location

-- given a location, returns the string
showLoc :: Loc -> Prog -> String
showLoc = dispLoc []

showProgLoc :: Prog -> Loc -> String
showProgLoc = flip showLoc

showHLLoc = dispLoc 

showCHRLoc = dispCHRLoc

-- as above, but also has the location label in front
showLocLabel l p = show l ++ ") " ++ (showLoc l p)

--------------------------------------------------------------------------------

-- merges adjacent declarations into each other where they represent the same
-- value (does this for all sub-declarations as well i.e. let/where)
mergeDefsProg :: Monad m => Prog -> m Prog
mergeDefsProg (Prog adecs) = do 
    adecs' <- mapM (processSubDecs (\ds -> return $ mergeDefs ds)) adecs
    return (Prog (mergeDefs adecs'))

unpackMismatchedDecsProg :: Monad m => Prog -> m Prog
unpackMismatchedDecsProg (Prog adecs) = do
    adecs' <- mapM (processSubDecs (\ds -> return $ unpackMismatchedDecs ds))
		   adecs
    return (Prog (unpackMismatchedDecs adecs'))

-- returns all the AValDecs and OverDecs which contain no definition (just an
-- annotation)
findDanglingAnnotations :: Prog -> [Ann Dec]
findDanglingAnnotations (Prog ads) = concatMap lone ads 
    where 
	lone ad = case item ad of
	    AValDec _ [] -> [ad]
	    OverDec _ [] -> [ad]
	    _		 -> []

--------------------------------------------------------------------------------

processEachSubDec :: Monad m => (Ann Dec -> m (Ann Dec)) -> 
				 Ann Dec -> m (Ann Dec)
processEachSubDec f = processSubDecs (mapM f)

-- given a dec., processes all the nested decs (in let expressions, and where
-- clauses), by applying the given transformation function
processSubDecs :: Monad m => ([Ann Dec] -> m [Ann Dec]) -> 
			      Ann Dec -> m (Ann Dec)
processSubDecs f adec = pDec adec
    where 
	
	pDec ad = case item ad of 
		ValDec ds    -> do ds' <- pDefs ds
				   return (Ann (ann ad) (ValDec ds'))
		AValDec s ds -> do ds' <- pDefs ds
				   return (Ann (ann ad) (AValDec s ds'))
		OverDec s ds -> do ds' <- pDefs ds
				   return (Ann (ann ad) (OverDec s ds'))
		_ -> return ad

	pDefs = mapM pDef
	pDef ad = case item ad of
		Def id ps rhs w -> do w' <- f w
				      rhs' <- pRHS rhs
				      return (Ann (ann ad) (Def id ps rhs' w'))
				      
	pRHS (RHS1 ae)  = do ae' <- pExp ae
			     return (RHS1 ae')
	pRHS (RHSG ges) = do ges' <- pGEs ges
			     return (RHSG ges')
				
	pGEs [] = return []
	pGEs ((g,e):ges) = do g' <- pExp g
			      e' <- pExp e
			      ges' <- pGEs ges
			      return ((g',e'):ges')

	pExp2 (e1,e2) = do e1' <- pExp e1
			   e2' <- pExp e2
			   return (e1',e2')

	pExp3 (e1,e2,e3) = do e1' <- pExp e1
			      e2' <- pExp e2
			      e3' <- pExp e3
			      return (e1',e2',e3')
	pExps = mapM pExp
	pExp ae = do
	   e <- case item ae of
		AbsExp ps e -> pExp e >>= \x -> return (AbsExp ps x)
		AppExp e1 e2 -> do (x1,x2) <- pExp2 (e1,e2)
				   return (AppExp x1 x2)
		IfxAppExp e1 e2 e3 -> do (x1,x2,x3) <- pExp3 (e1,e2,e3)
					 return (IfxAppExp x1 x2 x3)
		-- CaseExp (Ann Exp) [CaseAlt]
		IfExp  e1 e2 e3 -> do (x1,x2,x3) <- pExp3 (e1,e2,e3)
				      return (IfExp x1 x2 x3)
		AnnExp e t  -> pExp e >>= \x -> return (AnnExp x t)
		TAddExp e t -> pExp e >>= \x -> return (TAddExp x t)
		TupExp  es -> pExps es >>= \xs -> return (TupExp xs)
		ListExp es -> pExps es >>= \xs -> return (ListExp xs)
		-- ListCompExp (Ann Exp) [Ann Stmt]
		SecExpLeft  e1 e2 -> do (x1,x2) <- pExp2 (e1,e2)
				        return (SecExpLeft x1 x2)
		SecExpRight e1 e2 -> do (x1,x2) <- pExp2 (e1,e2)
					return (SecExpRight x1 x2)
		EnumExpFrom  e -> pExp e >>= \x ->return (EnumExpFrom x)
		EnumExpFromTo e1 e2 -> do (x1,x2) <- pExp2 (e1,e2)
					  return (EnumExpFromTo x1 x2)
		EnumExpFromThen e1 e2-> do (x1,x2) <- pExp2 (e1,e2)
					   return(EnumExpFromThen x1 x2)
		EnumExpFromThenTo e1 e2 e3 -> do 
				    (x1,x2,x3) <- pExp3 (e1,e2,e3)
				    return (EnumExpFromThenTo x1 x2 x3)
		LetExp ds e -> do x <- pExp e
				  ds' <- f ds 
			          return (LetExp ds' x)
		e -> return e
	   return (Ann (ann ae) e)




--------------------------------------------------------------------------------

-- maps pairs of terms to a new term (used for anti-unification)
type AntiMap = [((Type, Type), Type)]

-- finds the anti-unifier of the given types
antiUnifier :: [Type] -> TD Type
antiUnifier ts = bld ts 
    where
	bld []  = newVar
	bld [t] = return t
	bld (t1:t2:ts) = do (_, t) <- gen [] (t1, t2)
			    bld (t:ts)
	
        -- anti-unifies the two terms, with the given termmap, returning the
        -- anti-unifier and new term map
        gen :: AntiMap -> (Type, Type) -> TD (AntiMap, Type)
        gen am (TyCon t1, TyCon t2) | t1 == t2 = return (am, TyCon t1)
        gen am (TApp t1 t2, TApp s1 s2) = do (am', r1) <- gen am (t1,s1)
					     (am'',r2) <- gen am (t2,s2)
					     return (am, TApp r1 r2)
        gen am (t1, t2) = look am (t1,t2) 

	look :: AntiMap -> (Type, Type) -> TD (AntiMap, Type)
        look am (t1, t2) = case lookup (t1,t2) am of
                            Nothing -> do v <- newVar
					  let am' = ((t1,t2),v):((t2,t1),v):am
					  return (am', v)
			    Just v  -> return (am, v)


--------------------------------------------------------------------------------

-- returns a list of all the locations on the RHS of any prop. rule
propRuleLocs :: TD [Loc]
propRuleLocs = do
    chrs <- getCHRs
    let props = filter isProp chrs
	rhss  = map chrRhs props
	ucs   = concatMap fst rhss
    return (unJust ucs)

-- as above, but only for Chameleon constraints
chameleonPropRuleLocs :: TD [Loc]
chameleonPropRuleLocs = do
    chrs <- getCHRs
    hns  <- getHaskellConNames
    let props = filter isProp chrs
	rhss  = map chrRhs props
	ucs   = concatMap fst rhss
	ucs'  = filter (\uc -> ucName uc `notElem` hns) ucs
    return (unJust ucs)

