--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- contains definitions of CHRs and constraints for "standard" functions and 
-- constructors (eg. (:), (++), (.), etc.)
--
-- To add a new definition:
--  1) create a new, unique Name
--  2) create a new, unique UCons
--  3) write a simp. rule to give the value its type 
--	(note that one of the variables in the body should be the same as that
--	 in the UCons)
--  4) update "stdCHRs" and "stdNames" (below)
--  5) add entries to `initEnv' in TD_Translate (add new type constraints too)


module TD_StdDefs
where

import Justify
import Chr
import Core
import Constraint
import Ann_InputAST (noLoc, multiTypeName)

arrow :: Type -> Type -> Type
arrow t1 t2 = let arr = TyCon arrowName
	      in  TApp (TApp arr t1) t2

list = listType
listType :: Type -> Type
listType t = let lst = TyCon nilName
	     in  TApp lst t

listTyCon :: Type
listTyCon = TyCon nilName

tupleType :: [Type] -> Type
tupleType ts = 
    let arity = fromIntegral (length ts)
    in  if arity == 1 
	    then head ts
	    else let cons = TyCon (tupName arity)
		 in  foldl (TApp) cons ts 
tupName n = 
    name ("(" ++ (replicate (fromIntegral (n-1)) ',') ++ ")") ((-100) - n)


multiType ts = let arity = fromIntegral (length ts)
	           cons  = TyCon (multiTypeName arity)
	       in  foldl TApp cons ts

fcName    = name "."  (-400)
addName   = name "+"  (-401)
subName	  = name "-"  (-402)
mulName   = name "*"  (-403)
divName	  = name "/"  (-404)
eqName	  = name "==" (-405)
headName  = name "head" (-406)
tailName  = name "tail" (-407)
nullName  = name "null" (-408)
showName  = name "show" (-409)
readName  = name "read" (-410)
elemName  = name "elem" (-411)
notName	  = name "not"  (-412)
otherwiseName = name "otherwise" (-413)
numName	  = name "Num"  (-420)
numCName  = numName
enumCName = name "Enum" (-421)
showCName = name "Show" (-422)
readCName = name "Read" (-423)
trueCName = name "True" (-424)
fractionalCName = name "Fractional" (-425)

varr n = [Var (name "x" n, noLoc)] 
newCons n nm = UC n nm (var) noJust
    where var = [Var (name "x" (n-1), noLoc)]

consType :: UCons
consType = newCons (-200) consName 
catType	 = newCons (-202) catName  
addType	 = newCons (-204) addName    
subType	 = newCons (-206) subName    
mulType	 = newCons (-208) mulName    
divType	 = newCons (-210) subName    
fcType	 = newCons (-212) fcName	 
eqType	 = newCons (-214) eqName	 
headType = newCons (-216) headName	 
tailType = newCons (-218) tailName	 
nullType = newCons (-220) nullName
showType = newCons (-222) showName
readType = newCons (-224) readName
elemType = newCons (-226) elemName
otherwiseType = newCons (-228) otherwiseName
notType  = newCons (-228) notName

undefinedType = newCons (-298) undefinedName

vnilType  = TyCon vnilName
unitType  = TyCon (tupleName 0)
vconsType = TyCon vconsName
arrowType = TyCon arrowName

--------------------------------------------------------------------------------

numCons  t = numConsJust  t noJust
enumCons t = enumConsJust t noJust
showCons t = showConsJust t noJust
readCons t = readConsJust t noJust
fractionalCons t = fractionalConsJust t noJust

numConsJust  t j = uToCons (UC (-300) (numCName) [t] j)
enumConsJust t j = uToCons (UC (-301) (enumCName) [t] j)
showConsJust t j = uToCons (UC (-302) (showCName) [t] j)
readConsJust t j = uToCons (UC (-303) (readCName) [t] j)
fractionalConsJust t j = uToCons (UC (-304) (fractionalCName) [t] j)

boolType, intType, charType :: Type
boolType = TyCon boolName
intType  = TyCon (name "Int" (-320))
charType = TyCon (name "Char" (-321))
stringType = listType charType

stdTypes = [boolType, intType, charType]

haskellUCNames = [numCName, enumCName, fractionalCName]

--------------------------------------------------------------------------------

stdCHRs = [consCHR, catCHR, fcCHR, undefinedCHR, otherwiseCHR]
	   -- headCHR, tailCHR, nullCHR, readCHR, showCHR, addCHR,
	   -- , elemCHR, notCHR]
stdNames = [consName, catName, fcName, addName, subName, mulName, divName, 
	    eqName, headName, tailName, nullName, readName, showName, 
	    elemName, notName ] -- numName


beq = Eq

simp lhs rhs = Simp stdInfo lhs [] (consToUs rhs, consToBs rhs)

------------------------------------------------------------
-- Prelude (in no particular order)

undefinedCHR 
	= let lhs = undefinedType
	      t   = Var (name "x" (-299), noLoc)
	      a   = Var (name "a" (-606), noLoc)
	      rhs = bToCons (beq t a noJust)
	  in  simp lhs rhs
	  
consCHR = let lhs = consType
	      t   = Var (name "x" (-201), noLoc)
	      a   = Var (name "a" (-600), noLoc)
	      la  = list a
	      rhs = bToCons (beq t (a `arrow` (la `arrow` la)) noJust)
	  in  simp lhs rhs



addCHR  = let lhs = addType
	      t   = Var (name "x" (-205), noLoc)
	      a   = Var (name "b" (-605), noLoc)
	      rhsPred = numCons a
	      rhsEq   = bToCons (beq t (a `arrow` (a `arrow` a)) noJust)
	      rhs = rhsPred `cand` rhsEq
	  in  simp lhs rhs

fcCHR   = let lhs = fcType
	      t   = Var (name "x" (-213), noLoc)
	      a   = Var (name "a" (-602), noLoc)
	      b   = Var (name "b" (-603), noLoc)
	      c   = Var (name "c" (-604), noLoc)
	      f   = a `arrow` b
	      g   = c `arrow` a
	      typ = f `arrow` (g `arrow` (c `arrow` b))
	      rhs = bToCons (beq t typ noJust)
	  in  simp lhs rhs

-- Boolean functions
notCHR = let lhs = notType
	     t   = Var (name "x" (-229), noLoc)
	     b   = boolType
	     rhs = bToCons (beq t (b `arrow` b) noJust)
	 in  simp lhs rhs


-- Read and Show

showCHR = let lhs = showType
	      t   = Var (name "x" (-223), noLoc)
	      a   = Var (name "b" (-999), noLoc)
	      rhsPred = showCons a
	      rhsEq   = bToCons (beq t (a `arrow` stringType) noJust)
	      rhs = rhsPred `cand` rhsEq
	  in  simp lhs rhs

readCHR = let lhs = readType
	      t   = Var (name "x" (-225), noLoc)
	      a   = Var (name "b" (-999), noLoc)
	      rhsPred = readCons a
	      rhsEq   = bToCons (beq t (stringType `arrow` a) noJust)
	      rhs = rhsPred `cand` rhsEq
	  in  simp lhs rhs


------------------------------------------------------------
-- PreludeList  

headCHR = let lhs = headType
	      t   = Var (name "x" (-217), noLoc)
	      a   = Var (name "a" (-999), noLoc)
	      la  = list a
	      rhs = bToCons (beq t (la `arrow` a) noJust)
	  in  simp lhs rhs

tailCHR = let lhs = tailType
	      t   = Var (name "x" (-219), noLoc)
	      a   = Var (name "a" (-999), noLoc)
	      la  = list a
	      rhs = bToCons (beq t (la `arrow` la) noJust)
	  in  simp lhs rhs

nullCHR = let lhs = nullType
	      t   = Var (name "x" (-221), noLoc)
	      a   = Var (name "a" (-999), noLoc)
	      b   = boolType
	      la  = list a
	      rhs = bToCons (beq t (la `arrow` b) noJust)
	  in  simp lhs rhs

catCHR  = let lhs = catType
	      t   = Var (name "x" (-203), noLoc)
	      a   = Var (name "a" (-601), noLoc)
	      la  = list a
	      rhs = bToCons (beq t (la `arrow` (la `arrow` la)) noJust)
	  in  simp lhs rhs

elemCHR = let lhs = elemType
	      t   = Var (name "x" (-227), noLoc)
	      a   = Var (name "a" (-999), noLoc)
	      b   = boolType
	      la  = list a
	      rhs = bToCons (beq t (a `arrow` (la `arrow` b)) noJust)
	  in  simp lhs rhs

otherwiseCHR = let lhs = otherwiseType
		   t   = Var (name "x" (-229), noLoc)
		   rhs = bToCons (beq t boolType noJust)
	       in  simp lhs rhs
