--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


module TD_Translate (progToCHRs, State, num, runTRInit, TR(..),
		     Env(..), emptyEnv, lookupV, lookupT,
		     apply, fvCHR, transTyp, transDef, transTSchm,
		     progHaskellCons, checkUndeclaredConstraints,
		     transExp,
		     simplifyCHR, simplifyCHRs)

where

import Chr  hiding (Index)
import TD_StdDefs
import Core
import Ann_InputAST
import TD_General
import Misc
import Justify
import Terminal (uncommas)

import List
import Maybe
import Monad
import System
import IO
import Char

import IOExts

----------------------------------------

-- representation of uncomputed type information
type Rep  = (Type, Constraint)
type LRep = (Type, Constraint, [TVar])	-- list of all lambda-bound variables
					-- in the given expression

type VEnv = [(Id, Rep)]
type TEnv = [(Id, Rep)]
type PEnv = [(Id, Name)]
type UBEnv = [(Id, VEnv)]

-- venv:  maps values (ids) to Reps
-- tenv:  maps types (ids) to Reps
-- penv:  maps predicate names (ids) to unique Names
-- ubenv: is a list of the preceeding outer name scopes, arranged like a stack, 
--	  and the lambda bound variables within them. The ubenv stack grows by
--        one level each time a let group of definitions is begun. When
--	  processing top-level defs, it is [].
data Env = Env { venv::VEnv, tenv::TEnv, penv::PEnv, ubenv::UBEnv }
    deriving Show

emptyEnv :: Env
emptyEnv = Env [] [] [] []

-- just combines the type and value components of the env
addEnv :: Env -> Env -> Env
addEnv e1 e2 = e1 {venv = (venv e2)++(venv e1), tenv = (tenv e2)++(tenv e1),
		   penv = (penv e2)++(penv e1) }
sumEnvs :: [Env] -> Env
sumEnvs es = foldl addEnv emptyEnv es

extendV, extendT :: Env -> (Id, Rep) -> Env
extendP :: Env -> (Id, Name) -> Env
extendV env v = env {venv = v : venv env}
extendT env t = env {tenv = t : tenv env}
extendP env p = env {penv = p : penv env}
extendUB env u = let ((sid, venv):ubs) = ubenv env
		     ubs' = (sid, u:venv):ubs
		 in  env {ubenv = ubs'}
    

lookupV, lookupT :: Id -> Env -> Maybe Rep
lookupP :: Id -> Env -> Maybe Name
lookupV id env = lookup id (venv env)
lookupT id env = lookup id (tenv env)
lookupP id env = lookup id (penv env)

ubEnvEmpty :: Env -> Bool
ubEnvEmpty env = null (ubenv env)

-- generates a vector of unbound types from the ubenv
ubVector :: Env -> Type
ubVector env = let irs = concatMap snd (ubenv env)
		   rs  = map snd irs
		   ts  = map fst rs
	       in  listify ts	

-- creates a type vector
listify :: [Type] -> Type
listify vs = listifyWithTail vs vnilType

-- puts a don't-care variable on the tail of the list
listifyOpen :: [Type] -> TR Type
listifyOpen vs = do tl <- newVar
		    return (listifyWithTail vs tl)

listifyWithTail :: [Type] -> Type -> Type
listifyWithTail vs tl = foldr (\v t -> TApp (TApp vconsType v) t) tl vs


-- lookup a type constructor in the type env.
lookupTCon :: Env -> Ann Id -> TR Rep
lookupTCon env aid =
    case lookupT (item aid) env of
	Just v  -> return v
	Nothing	-> fail $ formatSrcPos (ixSrcPos (ann aid)) ++
			  " ERROR: undefined type constructor: `" ++ 
			  item aid ++ "'\n"


-- simply lookup the named type variable, fail if it's not there
-- returns Env so that it can be used interchangably with useTVar
lookupTVar :: Env -> Ann Id -> TR (Rep, Env)
lookupTVar env aid =
    case lookupT (item aid) env of
	Just v  -> return (v, env)
	Nothing -> fail $ formatSrcPos (ixSrcPos (ann aid)) ++
			  " ERROR: undefined type variable: `" ++ 
			  item aid ++ "'\n"

-- lookup a type var. in the type env.... if it's not there, add it
useTVar :: Env -> Id -> TR (Rep, Env)
useTVar env id = 
    case lookupT id env of
	 Just r  -> return (r, env)
	 Nothing -> do v <- newVar
		       let r = typeRep v
		       let env' = extendT env (id, r)
		       return (r, env')


-- takes a type (or ucons) and converts it into a rep
typeRep :: Type -> Rep
typeRep t = (t, trueCons)
uconRep :: UCons -> Rep
uconRep uc = let t = head (ucTypes uc)
	     in  (t, usToCons [uc])
--------------------------------------------------------------------------------
-- the monad

data State   = St {num::Integer , rules::[CHR], 
		   parentIds::[Id], childIds::[Id],
		   globalCons::[[UCons]],
		   parentUCons::[(UCons, Bool)], 
			    -- constraint for parent's join CHR
			    -- True if it's annotated, else False
		   stvVecs::[Type],	 -- scoped type variables (stack)
		   delayedCHRs::[DelayedCHR],
		   innerEnv :: Env	 -- environment containing types of all
					 -- sub-defs in program
		  }
newtype TR a = TR (State -> E (a,State))

type DelayedCHR = [CHR] -> CHR

instance Monad TR where

    -- return :: a -> TR a
    return a = TR (\s -> Succ (a,s))

    -- >>=    :: TR a -> (a -> TR b) -> TR b
    (TR a) >>= f = TR (\s -> case a s of
				Succ (a',s') -> let TR f' = f a' in f' s'
				Err t	     -> Err t)
    fail e = failTR e

-- unpacks the translator
runTR tr = runTRInit 0 tr

runTRInit :: Monad m => Integer -> TR a -> m (E (a, State))
runTRInit n (TR f) 
	= case f initState of
		s@(Succ _) -> return s
		e@(Err  _) -> return e	       
    where initState = St n [] [] [] [] [] [] [] emptyEnv

-- introduces an error
failTR e = TR (\s -> Err e)

-- given an action in the TP monad, converts it into a TR action
doTP :: TP a -> TR a
doTP tp = do n <- next
	     let res = unTP tp n
		 (n', a) = unSucc res
	     if isErr res
		then fail (unErr res)
		else do setNum n'
			return a

----------------------------------------

-- functions for examining and updating the state
trGet    f = TR (\st -> Succ (f st, st)) 
trUpdate f = TR (\st -> Succ ((), f st))

num'   f st = st {num = f (num st)}
rules' f st = st {rules = f (rules st)}
stvVecs' f st = st {stvVecs = f (stvVecs st)}
innerEnv' f st = st {innerEnv = f (innerEnv st)}
childIds'  f st = st {childIds  = f (childIds st)}
parentIds' f st = st {parentIds = f (parentIds st)}
globalCons' f st = st {globalCons = f (globalCons st)}
parentUCons' f st = st {parentUCons = f (parentUCons st)}
delayedCHRs' f st = st {delayedCHRs = f (delayedCHRs st)}

next = do n <- getNum ; incNum ; return n
nextList n 
    | n <= 0	= return []
    | otherwise = do x  <- next
		     xs <- nextList (n-1)
		     return (x:xs)

getNum = trGet num
incNum = trUpdate $ num' (+1) 
setNum n = trUpdate $ num' (const n)


getRules  = trGet rules
clrRules  = trUpdate $ rules' (const []) 
addRule r = trUpdate $ rules' (r:) 
addRules rs = trUpdate $ rules' (rs++)

topParentId     = trGet parentIds >>= \ids -> return (head ids)
setParentId  id = trUpdate $ parentIds' (const [id])
clrParentIds    = trUpdate $ parentIds' (const [])
popParentId     = trUpdate $ parentIds' tail
pushParentId id = trUpdate $ parentIds' (id:)

getChildIds = trGet childIds
setChildIds ids = trUpdate $ childIds' (const ids)
addChildIds ids = trUpdate $ childIds' (ids++)
clrChildIds = setChildIds []

getParentUCons = trGet parentUCons
setParentUCons ucs = trUpdate $ parentUCons' (const ucs)
pushParentUCons uc = trUpdate $ parentUCons' (uc:)
topParentUCons = trGet (head . parentUCons)
popParentUCons = do (top:ucs) <- getParentUCons
		    setParentUCons ucs
		    return top

-- returns Nothing if there's no annotated parent
topParentAnnUCons = do ucbs <- getParentUCons
		       let ucs = [ uc | (uc,b) <- ucbs, b ]
		       case ucs of
			[]     -> return Nothing
			(uc:_) -> return (Just uc)
		       

setStvVec vs = trUpdate $ stvVecs' (const vs)
getStvVec = trGet stvVecs
pushStvVec v = trUpdate $ stvVecs' (v:)
topStvVec = trGet (head . stvVecs)
popStvVec = do (top:vs) <- getStvVec
	       setStvVec vs
	       return top

getGlobalCons = trGet globalCons
setGlobalCons  ucs = trUpdate $ globalCons' (const ucs)
pushGlobalCons ucs = trUpdate $ globalCons' (ucs:)
popGlobalCons = do (top:ucs) <- getGlobalCons
		   setGlobalCons ucs
		   return top

currentGlobalCons = do cons <- trGet globalCons
		       return $ case cons of
				     (_:ucss) -> concat ucss
				     _	      -> []
			    
getDelayedCHRs = trGet delayedCHRs
setDelayedCHRs dchrs = trUpdate $ delayedCHRs' (const dchrs)
addDelayedCHR dchr = trUpdate $ delayedCHRs' (dchr:)
clearDelayedCHRs  = setDelayedCHRs []

getInnerEnv :: TR Env
getInnerEnv = trGet innerEnv
setInnerEnv env = trUpdate $ innerEnv' (const env)

isChild id = do ids <- getChildIds 
		return (id `elem` ids)

genId     = trGet (genName.parentIds)
genIdPrefix = trGet (genNamePrefix.tail'.parentIds) 
    where tail' [] = []
	  tail' (_:xs) = xs

-- produce a new variable, of no particular location
newVar :: TR Type
newVar = newVarLoc noLoc

-- generate a new type variable annotated with the given location
newVarLoc :: Loc -> TR Type
newVarLoc l = do n <- next
		 let tv = name ('v':(show n)) n
		 return (Var (tv, l))

-- generates n new types, no location
newVars :: Int -> TR [Type]
newVars 0 = return []
newVars n = do v  <- newVar
	       vs <- newVars (n-1)
	       return (v:vs)

-- given a list of locations, returns a list of type vars annotated with each
newVarsLocs :: [Loc] -> TR [Type]
newVarsLocs []     = return []
newVarsLocs (x:xs) = do v  <- newVarLoc x
		        vs <- newVarsLocs xs
		        return (v:vs)

--------------------------------------------------------------------------------
-- miscellaneous

nameScopeSep = "/"

-- generates a name based on the given list of (nested) ids
genName []  = ""
genName ids = 
    let ids' = (reverse ids)
    in	foldr (\s -> ((++) (s ++ nameScopeSep))) (last ids') (init ids')

genNamePrefix []  = ""
genNamePrefix ids = 
    let ids' = reverse ids
    in  gen ids'
	where gen [] = "" 
	      gen (x:xs) = x ++ nameScopeSep ++ gen xs



-- returns a constraint enforcing all given types as equal
eqList :: Just -> [Type] -> [BCons]
eqList j ts = [ eq t1 t2 j | (t1, t2) <- zip ts (tail ts) ]

-- version of rename, but for TR
renameTR x = doTP (rename x)

-- used to rename all but the third argument of user constraints (the stv arg)
-- NOW renames the whole enchilada (why didn't it do that before?)
renameCalled :: Rep -> TR Rep
renameCalled tc@(_, (C [uc] [])) = 
	case ucTypes uc of
	    (t:l:s:vs) -> do (t',l',s') <- renameTR (t,l,s)
			     let uc' = newUCTypes (t':l':s':vs) uc
			     return (t', (C [uc'] []))
	    _	       -> renameTR tc
renameCalled c = renameTR c 
    

renameKeepLocTR :: TC a => a -> TR a
renameKeepLocTR x = 
    doTP (do s <- renameSubst (fv x)
	     let s' = map (\(o@(_,l), Var (v,_)) -> (o, Var (v,l))) s
	     return (apply s' x))

-- given a Constraint, changes the ix of each ucons
renameConsIx :: Constraint -> TR Constraint
renameConsIx (C ucs bcs) = do
    ucs' <- mapM changeUCIx ucs
    return (C ucs' bcs)

changeUCIx uc = do n <- next
		   return (newUCIx n uc)

-- renames variables and uc ixs
renameCHR :: CHR -> TR CHR
renameCHR chr = do chr' <- renameTR chr
		   let rhs = rhsToCons (chrRhs chr')
		   rhs' <- renameConsIx rhs
		   let rhs'' = consToRhs rhs'
		   return (chrNewRhs rhs'' chr')

-- an initial translator environment
initEnv :: Env
initEnv = Env v t p u
    where   t = [("->",		typeRep arrowType),
		 ("Bool",	typeRep boolType),
		 ("Int",	typeRep intType),
		 ("Char",	typeRep charType),
		 ("[]",		typeRep listTyCon)] ++ mts

		where mts = [ (nameStr mt, typeRep (TyCon mt)) | 
				 mt <- stdMultiTypeNames ]

	    v = [("True",	typeRep boolType),
		 ("False",	typeRep boolType), 
		 (":",		uconRep consType),
		 ("undefined",	uconRep undefinedType),
		 (".",		uconRep fcType),
		 ("otherwise",	uconRep otherwiseType)]

	    p = [("Num",	numCName),
		 ("Enum",	enumCName),
		 ("Fractional",	fractionalCName)]

	    u = [("<top>", [])]

-- creates a uc with given id and a args
idToUC :: Int -> Id -> TR UCons
idToUC a id = idToUCWithLoc a noLoc id

idToUCWithLoc :: Int -> Loc -> Id -> TR UCons
idToUCWithLoc a l id = do 
    n <- next
    x <- next
    ts <- newVarsLocs (l : replicate (a-1) noLoc) -- (replicate a l)
    let ps = name id n
	uc = UC x ps ts (locToJust l) -- noJust
    return uc


-- suffix to add to the name of the chr representing the declared type of an id
annSuffix = ":a"
infSuffix = ":i"
hasAnnSuffix id = take 2 (reverse id) == reverse annSuffix
addAnnSuffix x = x ++ annSuffix
addInfSuffix x = x ++ infSuffix
dropAnnSuffix x = take (length x - length annSuffix) x
isAnnUCons uc = hasAnnSuffix (ucNameStr uc)

-- changes the name of the given CHR (must be a simp rule!) by adding its 
-- location to the end 
-- do this when generating the unifying ValDec CHR
-- NOTE: any recursive calls will remain unchanged!!
renameCHRByLoc :: CHR -> TR CHR
renameCHRByLoc chr = do
    i <- next
    let l    = chrLoc chr
	str  = nameStr $ simpName chr
	str' = str ++ ":" ++ show l
	chr' = chSimpName (name str' i) chr
    return chr'

-- flattens all occurrences of self-recursion in the given CHR (replaces
-- calls to itself with eqs to unify types)
-- NOTE: We throw away the global variable vector in recursive calls. The
--	 reason is that this could be larger at the recursive call site than at
--	 the definition site. By unifying them we force the size at the
--	 definition site up, which is a problem if the number of globals
--	 doesn't match this length.
--	 eg. f x = f x; -- f(t,v) <==> v would be [], except that at the
--				       recursive call site it's [x]
--	     g = f; -- f is called with [], clashes with the above!
--
-- NOTE: Actually, we discard everything but the function's type, the first
--	 component of the recursive call
unfoldSelfRecursion :: CHR -> TR CHR
unfoldSelfRecursion chr = do
    let ps   = simpNameStr chr
	ts   = ucTypes $ simpUCons chr
	rhs  = chrRhs chr
	rhs' = procUCs ps (dropTypes ts) rhs
	chr' = chrNewRhs rhs' chr
    return chr'
    where
	procUCs ps ts ([], bcs) = ([], bcs)
	procUCs ps ts ((uc:ucs), bcs)
	    | ps == ucNameStr uc = 
				let ss = dropTypes (ucTypes uc)
				    j  = ucJust uc
				    eqs = [ Eq t s j | (t, s) <- zip ts ss ]  
				    (ucs', bcs') = procUCs ps ts (ucs, bcs)
				in  (ucs', eqs ++ bcs')
	    | otherwise = let (ucs', bcs') = procUCs ps ts (ucs, bcs)
			  in  (uc:ucs', bcs')

	dropTypes (x:_) = [x]
	dropTypes xs    = xs

predId :: Id -> Id
predId ""     = error "TD_Translate.predId"
predId (c:cs) = toUpper c : cs


-- given types represented by two CHRs, r1 and r2, inserts into r1 a call to
-- r2, and unifies r1s type with the result
-- use this to propagate the effect of a type declaration into individual
-- function definitions
unifyCHRTypes :: CHR -> CHR -> TR CHR
unifyCHRTypes r1 r2 = do 
    let uc1 = simpUCons r1
	uc2 = simpUCons r2
	rhs = chrRhs r1
	c1  = rhsToCons rhs
	j   = J (unJust uc1)	-- should be justified by both	
        uc2' = reJust j uc2
    uc2''  <- renameTR uc2'
    uc2''' <- changeUCIx uc2''
    let	ts1 = ucTypes uc1
	ts2 = ucTypes uc2'''
	eqs = [ Eq t1 t2 j | (t1, t2) <- zip ts1 ts2 ]
	c2  = C [uc2'''] eqs `cand` c1
	rhs'= consToRhs c2
	r1' = chrNewRhs rhs' r1
    return r1'

-- as above, but the call to the second chr is unfolded
unifyCHRTypesUnfold :: CHR -> CHR -> TR CHR
unifyCHRTypesUnfold r1 r2 = do
    r2' <- return r2 -- renameTR r2
    let uc1 = simpUCons r1
	uc2 = simpUCons r2'
	c1  = rhsToCons (chrRhs r1)
	(ucs2, bcs2) = chrRhs r2'
	c2  = C ucs2 bcs2
		
	j   = J (unJust uc1)	-- should be justified by both	
    	ts1 = ucTypes uc1
	ts2 = ucTypes uc2
	eqs = [ Eq t1 t2 j | (t1, t2) <- zip ts1 ts2 ]
	c1'  = bsToCons eqs `cand` c2 `cand` c1
	rhs1'= consToRhs c1'
	r1'  = chrNewRhs rhs1' r1
    return r1'

-- given a type representing the types of lambda-bound variables, and a chr, 
-- introduces the vector into the chr, unifying it with the chr's second arg
addLambdaVarVec :: Type -> CHR -> CHR
addLambdaVarVec lv chr =
    let uc = simpUCons chr
	v  = ucTypes uc !! 1
	j  = ucJust uc
	eq = Eq v lv j
	(ucs, bcs) = chrRhs chr
	bcs' = eq : bcs
	rhs' = (ucs, bcs')
    in	chrNewRhs rhs' chr


-- given a type representing the scoped variable vector, and a chr, introduces
-- the vector into the chr, unifying it with the chr's third arg
addScopedTVarVec :: Type -> CHR -> CHR
addScopedTVarVec sv chr =
    let uc = simpUCons chr
	v  = ucTypes uc !! 2
	j  = ucJust uc
	eq = Eq v sv j
	(ucs, bcs) = chrRhs chr
	bcs' = eq : bcs
	rhs' = (ucs, bcs')
    in	-- trace ("adding scoped var vec: " ++ show sv)
		chrNewRhs rhs' chr

-- given an environment, generates the scoped variable vector pair
-- all the variables in the environment go into the `global' portion, while the
-- `local' part is unconstrained
genScopedVarVec :: Env -> TR Type
genScopedVarVec env = do
    let gvs = takeWhile isVarRep (map snd (tenv env))
    lvec <- newVar
    gvec <- listifyOpen (reverse (map fst gvs))
    return (tupleType [gvec, lvec])

-- version of the above used to project out variables in the body of a valdec 
-- CHR (assumption is that there are no longer vars)
genScopedVarVecProj :: Env -> TR Type
genScopedVarVecProj env = do
    let gvs = takeWhile isVarRep (map snd (tenv env))
    gvec <- listifyOpen (reverse (map fst gvs))
    return (tupleType [gvec, vnilType])


-- given an environment, generates a vector of the types of lambda-bound
-- variables
-- importantly, the vector generated from env will always be a prefix of the
-- vector generated from an extended version of env!
genLambdaVarVec :: Env -> Type
genLambdaVarVec env = genLambdaVarVecWithTail env vnilType
	
genLambdaVarVecWithTail :: Env -> Type -> Type
genLambdaVarVecWithTail env tl =
	-- entry represents a lambda-bound var if it has a variable id, and the
	-- constraint component is empty
    let ts = [ (id,t) | (id, r@(t, _)) <- venv env, isVarId id, isVarRep r ]
    in  listifyWithTail (reverse (map snd ts)) tl

genLambdaVarVecPair :: Env -> TR Type
genLambdaVarVecPair env = do v <- newVar
			     let vec = genLambdaVarVec env
			     return (tupleType [vec,v])

isVarId "" = False
isVarId (c:_) = isLower c 

isVarRep :: Rep -> Bool
isVarRep (Var _, c) = c == trueCons 
isVarRep _ = False

isLamVarRep tc@(t, c) = isVarRep tc

-- replaces the second argument of the user constraint with the given type
-- if it doesn't have a second argument, it is simply returned
replace2ndUCArg :: Type -> UCons -> UCons
replace2ndUCArg t2 uc = 
    let ts' = case ucTypes uc of
		(t1:_:ts) -> t1:t2:ts
		ts	  -> ts
    in  newUCTypes ts' uc

-- replaces the third argument of the user constraint with the given type
replace3rdUCArg :: Type -> UCons -> UCons
replace3rdUCArg t3 uc = 
    let ts' = case ucTypes uc of
		(t1:t2:t:ts) -> t1:t2:t3:ts
		ts	     -> ts
    in  newUCTypes ts' uc

replace2ndUCArgs lv = map (replace2ndUCArg lv)
replace3rdUCArgs sv = map (replace3rdUCArg sv)

--------------------------------------------------------------------------------

-- Given a program, finds the names of all the *declared* user constraints
-- A constraint is considered *declared* if it arises from an overload, rule or
-- hconstraint declaration
-- returns the Chameleon, then Haskell constraints
declaredConstraints :: Prog -> ([Id],[Id])
declaredConstraints (Prog adecs) = 
    let adecs' = map instToOverDec adecs
    in  flatten (map decCons (items adecs'))

 where decCons (OverDec0 id) = ([conId id], [])
       decCons (HConDec ids) = ([],ids)
       decCons (OverDec s _) = case item s of Sig _ id _ _ -> 
					       ([conId (item id)], [])
       decCons _ = ([],[])

       flatten ps = let (cs,hs) = unzip ps
		    in  (concat cs, concat hs)

	  
-- Finds all the occurrences of user constraints within type signatures
-- We can then check if they've all been declared
usedConstraints :: Prog -> [(Id, SrcPos)]
usedConstraints (Prog adecs) = concatMap usedCons (items adecs)
 
 where usedCons (OverDec as _) = usedSig (item as)
       usedCons (AValDec as _) = usedSig (item as)
       usedCons (RuleDec ar) = case item ar of
	    PropRule ax ac -> usedCtxt (item ax) ++ usedCnst (item ac)
	    SimpRule ax ac -> usedCtxt (item ax) ++ usedCnst (item ac)
       usedCons (ExternDec (Left as)) = usedSig (item as)
       usedCons _ = []
       
       usedSig  (Sig _ _ ac _) = usedCtxt (item ac)
       usedCtxt (Ctxt aps)   = map usedPred aps
       usedCnst (Cnst aps)   = concatMap usedPrim (items aps)
       usedPrim (PredPrim ap) = [usedPred ap]
       usedPrim _	      = []
       usedPred ap = case item ap of
		Pred id _ -> (id, ixSrcPos (ann ap))

-- looks for use of undeclared constraints
-- returns True if the check succeeds, else False
checkUndeclaredConstraints :: Prog -> IO Bool
checkUndeclaredConstraints (Prog decs) =
    let prog = Prog (map instToOverDec decs)
    
	std   = map nameStr haskellUCNames 
	(cs,hs) = declaredConstraints prog
	dec   = std ++ (cs ++ hs)
	used  = usedConstraints prog
	
	undec = [ idp | idp@(id, _) <- used, id `notElem` dec ]
	multi = filter (`elem`cs) hs

    in  do mapM_ reportUndec undec
	   mapM_ reportMulti multi
	   return (notNull undec || notNull multi)
    where 
	reportMulti id = putStrLn $ " ERROR: Haskell constraint used as " ++
				    "Chameleon constraint: `" ++ id ++ "'"
	
	reportUndec (id, pos) = 
		putStrLn $ formatSrcPos pos ++ 
			   " ERROR: undeclared constraint `" ++ id ++ "'"

--------------------------------------------------------------------------------

-- translates an entire program into a list of CHRs and a final env, given an
-- initial environment
-- NOTE: delayed CHRs cannot be woken yet, since they may depend on CHRs which
-- haven't been generated yet - hence we return them to the caller
--
-- also returns the lambda-bound variables within the program 
--
-- also, converts instance declarations into overload declarations
trProg :: Env -> [TVar] -> Prog -> TR (Env, [CHR], [TVar])
trProg env lvs (Prog adecs0) = do
    let adecs = map instToOverDec adecs0
    env0  <- trExternDecs env adecs
    env1  <- trDatDecs env0 adecs
    env2  <- valDecsEnv env1 adecs
    (env3, rchrs) <- trRuleDecs env2 adecs
    (vchrs,lvs') <- trValDecs env3 lvs adecs
    vchrs' <- mapM (unfoldRecursion env3 vchrs) vchrs
    let chrs = rchrs ++ vchrs'
    return (env3, chrs, lvs')

-- wakes delayed CHRs in the translator state
wakeDelayedCHRs :: [CHR] -> TR [CHR]
wakeDelayedCHRs chrs = do dchrs <- getDelayedCHRs  
			  return [ dchr chrs | dchr <- dchrs ]

-- Go through all ValDecs/AValDecs/OverDecs and add entries for them to the 
-- environment. In the case of ValDecs, just create a new predicate. For
-- AValDecs, create separate predicates for inferred and declared types. For
-- OverDecs, add a new pred which will point to the generic typing.
valDecsEnv :: Env -> [Ann Dec] -> TR Env
valDecsEnv env adecs = do
    let vds = valDecs adecs
	ads = avalDecs adecs
	ods = overDecs adecs
	o0ds = over0Decs adecs
	
    idp' <- genId
    let idp = if null idp' then idp' else idp' ++ nameScopeSep
    
    let vids  = map decId (items vds)
    vucs <- mapM (idToUC 3) (map (idp++) vids)
    let vrs  = zip vids (map uconRep vucs)

    let aids  = map decId (items ads)
	aids' = map addAnnSuffix aids
    aucs <- mapM (idToUC 3) (map (idp++) aids')
    let ars  = zip aids (map uconRep aucs)

    let oids0 = map decId (items o0ds) ++ map decId (items ods) 
	oids1 = nub oids0
	oids2 = map predId oids1
    oucs <- mapM (idToUC 1) oids2  -- overload constraints!
    let pss = map uconsPSym oucs
	ops = zip oids2 pss
        ors = zip oids1 (map uconRep oucs)
	
	env' = env { venv = vrs ++ ars ++ ors ++ venv env, 
		     penv = penv env ++ ops }

	ids = map (decId . item) (vds ++ ads ++ ods ++ o0ds)

    return env'


-- handles 'extern' declarations, which are simply type declarations, by adding
-- them to the type environment
trExternDecs :: Env -> [Ann Dec] -> TR Env
trExternDecs env adecs = do
    let exts = items (externDecs adecs)
    (venv',tenv') <- trExternDecs exts
    let env' = env { venv = venv' ++ (venv env), tenv = tenv' ++ (tenv env) }
    return env'

    where isLeft (ExternDec (Left _)) = True
	  isLeft _ = False

	  trExternDecs :: [Dec] -> TR (VEnv, TEnv)
	  trExternDecs decs = do
		let (left,right) = partition isLeft decs
		tenvs <- mapM trTyCon right
		let tenv' = concat tenvs
		    env'  = env { tenv = tenv' ++ tenv env }
		venvs <- mapM (trSig env') left
		let venv' = concat venvs
		return (venv', tenv')
	 
	  trTyCon (ExternDec (Right aid)) = do
			   x <- next
			   let id = item aid
			       nm = name id x
			       tc = TyCon nm
			       rep = [(id, typeRep tc)]
			   return rep
	    
	  trSig env (ExternDec (Left asig)) = case item asig of
		Sig _ id ctxt t -> do (env1, cc) <- trCtxt env ctxt
				      (env2, (tt, tc)) <- trTypNew env1 t
				      let c  = cc `cand` tc
					  bcs = buildUP (consToBs c)
					  phi= unSucc (unify [] idSubst bcs)
					  tt'= apply phi tt
					  rep = (tt', trueConsAlt `cand` c)
									      
				      return [(item id, rep)]


-- also returns the lambda-bound variables from the declarations
-- (and sub-declarations)
trValDecs :: Env -> [TVar] -> [Ann Dec] -> TR ([CHR], [TVar])
trValDecs env olvs adecs = do
    let isVal = (\d -> isValDec d  || isAValDec d 
		    || isOverDec d || isInstDec d {- || isPatDec d -} ) . item
	adecs' = filter isVal adecs

    (chrs,lvs) <- trans olvs adecs'
    return (chrs, lvs)

    where
	trans lvs []	 = return ([],lvs)
	trans lvs (d:ds) = do clrRules
			      (chrs1, lvs') <- trValDec env olvs d
			      chrs2 <- getRules 
			      popStvVec
			      let chrs = chrs1 ++ chrs2
			      (chrs', lvs'') <- trans lvs' ds
			      return (chrs++chrs', lvs'')


-- takes an (A)ValDec and returns a list of corresponding CHRs
-- produce joining chr simply by combining bodies of individual def chrs
-- also rename those sub-chrs
-- NOTE: always returns join chr at head of list 
--	 (the CHR which unifies inferred and annotated types)
-- 
-- also returns the lambda-bound variables from the definition
-- (and sub-definitions)
trValDec :: Env -> [TVar] -> Ann Dec -> TR ([CHR], [TVar])
trValDec env olvs adec = do
    let dec = item adec
	ix  = ann adec
	j   = ixToJust ix
	l   = ixLoc ix
	id  = decId dec
    pushParentId id
    chrs <- case dec of

	    -- similar to trDef
	    (PatDec ap arhs aw)	 -> 
		trace ("PATTERN BINDING TRANSLATION NOT FINALISED YET!") (return ([],[]))
{-				    do
			    (env1, wchrs) <- trWhere env0 w
			    (t_e, c_e, lv_e) <- trRHS env1 ix rhs
			    addRules wchrs

			    -- find all pattern variables, then patterns
			    let pvars = patAVars (item ap)
				pats = map (flip replacePatVars ap) pvars
				

			    -- (env0, (t_ps, c_p, lv_p)) <- trPat env ap
-}
		where -- pass in the type/constraint from the RHS, and the
		      -- pattern variable of interest
{-		      procVar (te, ce) av = do
			let v = item av
			    l = ixLoc (ann av)
			
			rn' <- next
			let rn = fromIntegral rn' 
			   
			    pat = replacePatVars v ap
			(_, (tp, cp, lvps)) <- trPat env ap
			    -- ptv is the type variable representing the
			    -- pattern variable v
			let ptv = Var (heady "TD_Translate:procVar" lvps)
			    bc  = Eq ptv te noJust

			    -- now generate the CHR for the pattern variable
			    rep = fromJust $ lookupV v env
			    uc  = head $ consToUs $ snd rep
			    
			    chr = Simp (mdefInfo rn id l)
-}			
	
		      -- replace all other pattern vars with "don't-cares"
		      replacePatVars v ap = Ann (ann ap) $ case item ap of
			    p@(VarPat id) | item v == id -> p
					  | otherwise	 -> JokPat
			    ConPat id ps-> ConPat id (map (replacePatVars v) ps)
			    TupPat ps   -> TupPat (map (replacePatVars v) ps)
			    ListPat ps  -> ListPat (map (replacePatVars v) ps)
			    p -> p
			    
    
	    (ValDec adefs)	 -> do 
			    let aid = adefAId $ head adefs
				pos = ixSrcPos (ann aid)
				setPos = newCHRSrcPos pos
	    
			    -- identifying number, used to group all the CHRs 
			    -- from this declaration
			    rn' <- next
			    let rn = fromIntegral rn'

			    -- record the constraint representing the join CHR
			    -- arising from this value declaration
			    let c   = fromJust (lookupV id env)
				uc  = head $ consToUs $ snd c
			    pushParentUCons (uc,False)
			    -- just push a free var, it won't be used anyway
			    sv <- newVar
			    pushStvVec sv
	    
			    -- translate each definition clause
			    -- chrs <- mapM (trNonOverDef env) adefs
			    (chrs,lvs) <- trNonOverDefs env olvs adefs
			    -- unfold self recursion (self-recursion)
			    chrs0 <- mapM unfoldSelfRecursion chrs
			    -- rename each chr by its location number
			    chrs1 <- mapM renameCHRByLoc chrs0
			
			    -- add scoped type variable vectors
			    -- (note: all the type vars are global!)
			    sv <- genScopedVarVecProj env 
			    let chrs2 = map (addScopedTVarVec sv) chrs1

			    -- now build the (unifying) mdef. and join CHRs
			    -- NOTE: should probably just call mdefCHRs below!
			    --	     (this would make it consistent with the
			    --	      translation for over/ann decs)
			    t <- newVarLoc l
			    v <- newVarLoc noLoc 
			    s <- newVarLoc noLoc
			    x <- next
			    let tss = map (ucTypes . simpUCons) chrs2
				[ts, vs, ss] = transpose tss
				-- unify types and vectors of individual defs
				cts = listEq t j ts trueCons
				css = listEq s j ss trueCons
				cvs = listEq v j vs trueCons
				-- "unfold calls"
			    let crhs = map (rhsToCons . chrRhs) chrs2
				crhs' = candList crhs
				-- all rhs constraints
				cs  = cts `cand` cvs `cand` css `cand` crhs'
				uc' = newUCTypes [t, v, s] uc
			    uc'' <- do let str  = ucNameStr uc'
					   str' = str ++ ":a"
				       n <- next
				       let nm = name str' n
				       return (newUCName nm uc')
					
			    let rhs = consToRhs cs
				minf = mdefInfo rn id l
				jinf = joinInfo rn id l
				ainf = iannInfo rn id l [] uc'

				-- rejustify ucons for ann chr
			    ann_rhs <- rejustifyRHSUCs rhs

				-- generate identical mdef/join chrs
			    let mchr = setPos (Simp minf uc' [] rhs)
				jchr = setPos (Simp jinf uc' [] rhs)
				achr = setPos (Simp ainf uc'' [] ann_rhs)
{-				
			    trace ("\nchrs2: " ++ show chrs2 ++ 
				   "\nuc': " ++ show uc' ++
				   "\ncrhs': " ++ show crhs' ++ 
				   "\ncts: " ++ show cts ++
				   "\ncvs: " ++ show cvs ++
				   "\ncss: " ++ show css) $ -}
			    return (jchr:mchr:chrs2, lvs)

			where
			    rejustifyRHSUCs (ucs,bcs) = do 
				       ucs' <- mapM rej ucs
				       return (ucs',bcs)

				where rej (UC i p ts _) = do 
						 l <- next
						 let j = locToJust l
						 return (UC i p ts j)

	    -- as above, but recursive calls should reference the declared type
	    -- and when generating the mdef/join chr, include the content of 
	    -- the ann chr
	    (AValDec asig adefs) -> do 
			    let aid = adefAId $ head adefs
				pos = ixSrcPos (ann aid)
				setPos = newCHRSrcPos pos
	    
			    -- identifying number, used to group all the CHRs 
			    -- from this declaration
			    rn' <- next
			    let rn = fromIntegral rn'

			    t <- newVarLoc l
			    v <- newVarLoc noLoc 
			    s <- newVarLoc noLoc
			    x1 <- next
			    x2 <- next
			    x3 <- next
			    x4 <- next

			    -- record the constraint representing the join CHR
			    -- arising from this value declaration
			    let ps = let c   = snd $ fromJust (lookupV id env)
				         str = ucNameStr $ head (consToUs c)
				     in  dropAnnSuffix str
				ts   = [t, v, s]
			        psj  = name ps x3
				ucj  = UC x4 psj ts j

			    (env',achr,sv) <- trAValDecSig env rn annInfo asig
			    pushParentUCons (replace3rdUCArg sv ucj,True)
			    -- now add global constraints to the ann CHR
			    gucs <- currentGlobalCons
			    let gcs = usToCons gucs
				achr' = chrAddCons achr gcs
			    
			    pushStvVec sv
			    -- chrs0 <- mapM (trNonOverDef env') adefs
			    (chrs0,lvs) <- trNonOverDefs env' olvs adefs

			    chrs1 <- mapM renameCHRByLoc chrs0
			    let chrs2 = map (addScopedTVarVec sv) chrs1

			    lvt <- newVar
			    let achr'' = achr'
				chrs3  = achr'' : chrs2

			    -- generate it on the spot, nobody except for a 
			    -- debugger user will ever call it 
			    -- (they'll call the _a chr)
			    let psm  = name (addInfSuffix ps) x1
				ucm  = UC x2 psm ts j
				infm = mdefInfo rn id l
				mchr0 = Simp infm ucm [] emptyRhs
			    mchr <- mdefCHRs (mdefInfo rn) mchr0 chrs2

			    -- now generate the join chr 
			    -- (unifying inf. and ann. types)
			    let infj = joinInfo rn id l
				jchr0 = Simp infj ucj [] emptyRhs
			    jchr <- mdefCHRs (joinInfo rn) jchr0 chrs3

			    -- last cons. added are no longer in scope
			    popGlobalCons

			    let mchr' = setPos mchr
				jchr' = setPos jchr
{-
			    trace ("\nachr: " ++ show achr ++
				   "\n\nchrs1: " ++ showlist chrs1 ++
				   "\nchrs2: " ++ showlist chrs2 ++
				   "\nsv: " ++ show sv ++
				   "\nachr'': " ++ show achr'' ++
				   "\n\njchr': " ++ show jchr') $ -}
			    return (jchr':mchr':chrs3, lvs)

	    -- translate OverDecs just like AValDecs, except we also generate an
	    -- additional rule from the sig
	    (OverDec asig adefs) -> do 

			    let aid = adefAId $ head adefs
				pos = ixSrcPos (ann aid)
				setPos = newCHRSrcPos pos
	    
			    t <- newVarLoc l
			    v <- newVarLoc noLoc 
			    s <- newVarLoc noLoc
			    x1 <- next
			    x2 <- next
			    x3 <- next
			    x4 <- next
			    -- record the constraint representing the join CHR
			    -- arising from this value declaration
			    let ts = [t, v, s]
				psj  = name id x3
				ucj  = UC x4 psj ts j

			    rn' <- next
			    let rn = fromIntegral rn'
			    (env',ochr,achr,sv) <- trOverSig env rn asig
			    pushParentUCons (replace3rdUCArg sv ucj, True)
			    pushStvVec sv
			    let trdefs = trDefs (odefInfo rn)
			    -- chrs0 <- mapM (trdef env') adefs
			    (chrs0, lvs) <- trdefs env' olvs adefs
			    -- gen a uniquely-named chr for each def
			    chrs1 <- mapM renameCHRByLoc chrs0
			    let chrs2= map(addScopedTVarVec sv) chrs1

			    lvt <- newVar
			    let -- lv = genLambdaVarVecWithTail env lvt
				achr' = achr -- addLambdaVarVec lv achr
				chrs3 = achr' : chrs2
				      
			    -- generate it on the spot, nobody except for a 
			    -- debugger user will ever call it 
			    -- (they'll call the F chr)
			    let psm = name (id ++ ":i-" ++ show l) x1
				ucm = UC x2 psm ts j
				infm = omdefInfo rn id l
				mchr0 = Simp infm ucm [] emptyRhs

			    mchr <- mdefCHRs (mdefInfo rn) mchr0 chrs2

			    -- now generate the join chr
			    -- (unifying inf. and ann. types)
			    let infj = ojoinInfo rn id l
				jchr0 = Simp infj ucj [] emptyRhs
			    jchr <- mdefCHRs (joinInfo rn) jchr0 chrs3
	    
			    -- last cons. added are no longer in scope
			    popGlobalCons

			    let mchr' = setPos mchr
				jchr' = setPos jchr

			    return (jchr':mchr':ochr:chrs3, lvs)
{-
	    -- translate instance declarations by converting them into overload
	    -- declarations, and re-invoking
	    InstDec (Ann x1 (InstSig ctxt (Ann x2 (Pred id t)))) adefs ->
			    let sig = Ann x1 (Sig Univ (Ann x2 id) ctxt t)
				dec = Ann ix (OverDec sig adefs)
			    in  do pushParentUCons (error "InstDec")
				   trValDec env olvs dec
-}			    
    popParentUCons
    popParentId
    return chrs

    where 
     -- generates all the chrs for an annotated (inc. overloaded) dec
     -- takes: info fun (mdefInfo/omdefInfo), an mdef CHR skeleton, def CHRs
     mdefCHRs :: CHRInfoFun -> CHR -> [CHR] -> TR CHR
     mdefCHRs inf mchr chrs = mkMdef mchr chrs
	where mkMdef mchr [] = return mchr
	      mkMdef mchr (chr:chrs) = do
		mchr' <- unifyCHRTypesUnfold mchr chr
		mkMdef mchr' chrs

     
-- translate the sig from an overload dec into a CHR  
-- eg. overload f :: C => t
--     becomes: F(t) <==> C	    (first CHR)
-- also returns f:a(t') <==> t' = t (second CHR)
-- returns scoped variable vector and new type env, as per trAValDecSig
trOverSig :: Env -> Int -> Ann Sig -> TR (Env, CHR, CHR, Type)
trOverSig env rn asig = 
    let sig = item asig
	ix  = ann asig
	l   = ixLoc ix
    in  case sig of
	 -- should always be a universal annotation!
	 (Sig _ aid actxt at) -> do 
		let id = item aid
		x1 <- next
		x2 <- next
		(env1, cc)       <- trCtxt env actxt
		(_, (ott', otc)) <- trTypNew env1 at
		(env2, (tt, tc)) <- trTypNewCons env1 at
		(ps, env3)	 <- trPredId env2 (predId id)

		let tc_phi = unSucc (unify [] idSubst (buildUP (consToBs otc)))
		    ott = apply tc_phi ott'
	
		let pos = ixSrcPos (ann aid)

		-- add context constraints to the set of global constraints
		let ucs = consToUs cc
		pushGlobalCons ucs
		
		tv <- newVar
		vv <- newVar
		as <- newVar
		let j   = ixToJust ix
		    oinf = oruleInfo id l
		    ouc  = UC x1 ps [ott] j
		    orhs = consToRhs cc
		    ochr = Simp oinf ouc [] orhs

		    -- generate scoped variable vector pair
		    (lvs,gvs) = let l1 = length (tenv env)
				    l2 = length (tenv env3)
				    l  = l2 - l1
				    tr = map snd (tenv env3)
				    vars = map Var . nub . fv
				in  (vars $ take l tr, 
				     vars $ drop l (takeWhile isVarRep tr))

		    lvec = listify lvs
		    gvec = listify gvs
		    vec = tupleType [gvec, lvec]

--		    vec = listify ts
		    
		    c    = fromJust (lookupV id env)
		    auc  = head $ consToUs $ snd c
		    auc' = newUCTypes [tv, vv, as] auc
		    auc''= newUCName (name (varId (ucNameStr auc ++ annSuffix)) 
					   x2) auc'
		    cons = tc `cand` cc `cand` bsToCons [Eq tv tt j,Eq as vec j]
		    arhs = consToRhs cons
		    ainf = oannInfo rn id (ixLoc ix)
		    achr = newCHRSrcPos pos (Simp ainf auc'' [] arhs)
				

		return (env3, ochr, achr, vec)


--   given sig :  f :: C => t
-- returns chr :  f(t',_ ,v) <==> t' = t, v = as, C
--       where :  as is a list of free vars in C, t
--
-- also return the new (type) environment, and the shared variable vector
-- update the env so that local recursive calls refer to the declared type
--trAValDecSig :: Env -> (Id -> Loc -> [UCons] -> CHRInfo) -> Ann Sig -> TR (Env, CHR, Type)
trAValDecSig env rn ainfo asig = 
    let sig = item asig
	ix  = ann asig
    in  case sig of
	 -- here we care about the annotation type (whether its universal or
	 -- existential, since that will affect how we classify the ann. chr)
	 Sig st id ctxt t -> do 
	    -- translate context and type of sig
	    (env1, cc)	     <- trCtxt env ctxt
	    (env2, (tt, tc)) <- trTypNewCons env1 t

	    let pos = ixSrcPos (ann id)

	    -- add context constraints to the set of global constraints
	    let ucs = consToUs cc
	    pushGlobalCons ucs
	    gucs <- currentGlobalCons

	    -- work out the free vars in cxt, t
	    -- first are the local tvars, second are global
	    let (lvs,gvs) = let l1 = length (tenv env)
				l2 = length (tenv env2)
				l  = l2 - l1
				tr = map snd (tenv env2)
				get = vars . nub . fv
				local  = get (take l tr)
				global = get (drop l tr)
			    in  (local, global) 
		-- safe because only vars could have been introduced
		j   = ixToJust ix
		
	    lvec <- return $ listify (reverse lvs)
	    gvec <- listifyOpen (reverse gvs)
	    let vec = tupleType [gvec, lvec]
	    tv <- newVar
	    vv <- newVar
	    as <- newVar
	    n1 <- next
	    n2 <- next

		-- get the _a form name that'll be used by callers
	    let c   = fromJust (lookupV (item id) env)
		uc  = head $ consToUs $ snd c
		uc' = newUCTypes [tv, vv, as] uc
		cons = tc `cand` cc `cand` bsToCons [Eq tv tt j, Eq as vec j]
		rhs  = consToRhs cons 
		info = (case st of 
			 Univ  -> ainfo 
			 Exist -> taddInfo) rn (item id) (ixLoc ix) gucs
		chr  = newCHRSrcPos pos (Simp info uc' [] rhs)

	    let uc'  = replace3rdUCArg vec uc
		(t:_)= ucTypes uc'
		env3 = extendV env2 (item id,(t,uToCons uc'))
	    
	    return (env3, chr, vec)


-- given an id that refers to a specific defintion, we return the id that
-- refers to the entire value declaration to which the definition belongs
-- def. ids are of the form id-loc
undefId :: Id -> Id
undefId = takeWhile (/=':')

-- given a CHR, unfolds all calls which lead to cycles 
-- (to eliminate problems when dealing with mutually recursive defs)
-- ids tracks the ids of the defs we've already been through
unfoldRecursion :: Env -> [CHR] -> CHR -> TR CHR
unfoldRecursion env chrs orig_chr = do

    -- add the inital entry for calls back to this chr
    let uc   = simpUCons orig_chr
	id   = undefId (ucNameStr uc)
	ut   = -- mkPair (take 2 (ucTypes uc))   
	       -- head (ucTypes uc)
	       mkTriple (ucTypes uc)
        ve   = (id, typeRep ut)
        env' = extendV env ve

  
    -- NOTE: we don't want to necessarily affect the type of the CHR (because
    -- the cycle may just be due to calls unifying lambda-bound vars), so we
    -- begin unfolding from the constraints in its body - i.e. we never
    -- break the cycle using the constraint in the head of the rule.
    (cyc, cons') <- unfold 1 False env orig_chr []   
		    -- unfold env' chr [id]

    cons <- renameConsIx cons'
    chr' <- if cyc 
       then let (_, bcs) = chrRhs orig_chr
		cons' = cons `cand` bsToCons bcs
		rhs'  = consToRhs cons'
	    in  return (chrNewRhs rhs' orig_chr)
       else return orig_chr
    
--    trace ("\noriginal CHR: " ++ show orig_chr ++
--	   "\nunfolded CHR: " ++ show chr') $ return chr'

    return chr'	   
    
    where 

	orig_id  = undefId (ucNameStr (simpUCons orig_chr))

	-- returns True if there is a cycle, as well as the constraint which
	-- should replace the existing ucons
	-- admin is True if we've encountered any administrative constraints
	-- (in which case we break the cycle by removing the last call)
	
--	unfold :: Env -> CHR -> [Id] -> TR (Bool, Constraint)
	unfold count admin env chr ids = do
	    let (ucs, _) = chrRhs chr
	    if null ucs 
		-- trivially not a cycle
		then return (False, usToCons ucs)
		-- for each call: 
		-- if it completes a cycle:
		--   * replace it, with a type
		--   * get the type from the env
		-- if not:
		--   * add an entry to the env for monomorphisation later
		--   * find the chr it represents 
		--   * process all sub-calls by calling unfold
	        else
		 let proc uc = 
		      let id = ucNameStr uc
			  ts = ucTypes uc
			  ut = mkTriple ts
			  j  = ucJust uc
			  admin' = admin || isAdminUCons uc
		      in  if makesCycle ids uc
			  then let r = fromJust (lookupV id env)
				   et = fst r
				   ut = mkTriple ts
				   eq = Eq et ut j
				   cons = if admin' 
					  then trueCons
					  else bsToCons [eq]

				   msg = ("\n" ++ show admin' ++ "." ++ show count ++ ":" ++ id ++ ": " ++ show uc ++ " is the last call in a cycle, so replacing it with " ++ show cons)
			       in  -- trace msg $ 
				    return (True, cons)

			  else do let ve = (id, typeRep ut)
				      env' = extendV env ve
				      mdchr = findCHR id
				      foundCHR = isJust mdchr
				  
				  if not foundCHR
				   then return (False, error "*")
				   else do 
				     dchr <- renameKeepLocTR (fromJust mdchr)
				     let ts = ucTypes (simpUCons dchr)
					 ht = mkTriple ts 
					 eq = Eq ht ut j
					 ids' = id : ids
				     (cyc, c) <- unfold (count+1) admin' env' dchr ids'
				     let rcs = rhsToCons (chrRhs dchr)
					 eqs = eq : (consToBs rcs)
					 cons = bsToCons eqs `cand` c
					 msg  = ("\n" ++ show admin' ++ "." ++ 
						 show count ++ ":" ++ id ++ ": " ++ show uc ++ " is part of a cycle, so replacing it with " ++ show cons ++
						"\nbtw, added " ++ show ve ++ 
						" to the environment") 
				     if cyc
					then -- trace msg $
					      return (True, cons)
					else return (False, error "!")

		 in  do let (ducs, oucs) = partition (isVarId . ucNameStr) ucs
			-- do not unfold calls to overload defs
			ps <- mapM proc ducs
			let bps = zip ps ducs
			    cs  = [ k | ((b, c), u) <- bps,
					let j = ucJust u
					    C ucs bcs = c
					    ucs' = map (addUCJust j) ucs
					    bcs' = map (addBCJust j) bcs
					    c' = C ucs' bcs',
					let k = if b then varsNoLoc c'
						     else usToCons [u] ]
			    cons = usToCons oucs `cand` candList cs
			    b = any fst ps
			return -- (trace ("\nenv: " ++ show env ++ "\nreplacements:\n" ++ unlines (map show (zip ducs cs)))
					(b, cons)
		    
	mkTriple :: [Type] -> Type
	mkTriple (t:v:s:_) = tupleType [t,v,s]	
	mkTriple _ = error "TD_Translate.mkTriple"


	makesCycle :: [Id] -> UCons -> Bool
	makesCycle ids uc = ucNameStr uc `elem` ids

	-- returns the chr with the given id, if one exists
	findCHR :: Id -> Maybe CHR
	findCHR id = let chr = filter ((id==) . simpNameStr) (chrs ++ stdCHRs)
		     in  if null chr then Nothing
				     else Just (head chr)

	-- changes the location of all the variables to noLoc 
	varsNoLoc :: Constraint -> Constraint
	varsNoLoc t = let vs = nub (fv t)
			  s  = [ (vl, Var (v, noLoc)) | vl@(v,_) <- vs ]
		      in  apply s t
	
trNonOverDefs = trDefs (defInfo noLoc)
	
trNonOverDef :: Env -> [TVar] -> Ann Def -> TR (CHR, [TVar])
trNonOverDef = trDef (defInfo noLoc)

trDefs :: CHRInfoFun -> Env -> [TVar] -> [Ann Def] -> TR ([CHR], [TVar])
trDefs _ _ lvs [] = return ([],lvs)
trDefs i e lvs (d:ds) = do (chr,lvs') <- trDef i e lvs d
			   (chrs,lvs'') <- trDefs i e lvs' ds
			   return (chr:chrs,lvs'')


-- process:
-- 1) patterns (they're in scope over the whole def)
-- 2) 'where' definitions (in scope of rhs)
-- 3) rhs
--
-- also returns the lambda-bound variables from the definition 
-- (and sub-definitions)
trDef :: CHRInfoFun -> Env -> [TVar] -> Ann Def -> TR (CHR, [TVar])
trDef mkInfo env lvs adef =
    let ix  = ann adef
	j   = ixToJust ix
	l   = ixLoc ix
    in  case item adef of
	    (Def aid aps rhs w)  -> do 
		    (env0, (t_ps, c_p, lv_p)) <- trPats env aps
		    (env1, wchrs,lv_w) <- trWhere env0 (lvs ++ lv_p) w
		    (t_e, c_e, lv_t) <- trRHS env1 lv_w ix rhs
		    addRules wchrs
		    idp <- genIdPrefix
		    let na = 3
			id = item aid
			id' = idp ++ id
		    uc <- idToUCWithLoc na l id'
		    tl1 <- newVar
		    tl2 <- newVar
		    tl3 <- newVar
		    let ts  = ucTypes uc
			vt  = nthElem 2 ts
			vid = nthElem 1 ts

			-- we need at this point, both the lambda-bound
			-- variables from the outer scope, AND the lambda-bound
			-- variables in the expression in this definition
			-- (the ones that'll get passed into inner defs)
			out_vec = genLambdaVarVecWithTail env tl1
			in_vec  = listifyWithTail (map Var lv_t) tl3
			vec = tupleType [out_vec, in_vec]
			c_v = bsToCons [Eq vt vec j]
{-			c_v = trace ("\nid     : " ++ item aid ++
				     "\nout_vec: " ++ show out_vec ++
				     "\nin_vec : " ++ show in_vec ++
--				     "\nenv    : " ++ show env ++
				     "\nlvs    : " ++ show lvs ++
				     "\nlv_p   : " ++ show lv_p ++
				     "\nlv_w   : " ++ show lv_w ++
				     "\nlv_t   : " ++ show lv_t) $ 
				    c_v'
-}			
			t_id = foldr arrow t_e t_ps
			
		    bc_abs <- funType vid (zip (anns aps) t_ps) (ix, t_e)
		    
		    let c_abs = -- trace ("bc_abs: " ++ show bc_abs) $ 
				bsToCons bc_abs
		    
			c_id = bsToCons [Eq vid t_id j]
			
			c = c_p `cand` c_e `cand` c_v `cand` c_id -- c_abs -- c_id
			inf = mkInfo id l
			chr = Simp inf uc [] (consToRhs c)
{-
		    trace ("\nid: " ++ id ++ 
			   "\nvec (out,in): " ++ show vec) {- ++ 
			   "\nlvs: " ++ show lvs ++ 
			   "\nlv_t: " ++ show lv_t)-} $ -}
		    return (chr, lv_t)

    where -- given a list of pattern types and an expression type, builds a
	  -- list of eq. constraints for a function's type. Note that the 
	  -- patterns are treated as lambda-abstractions.
	  funType :: Type -> [(Index, Type)] -> (Index, Type) -> TR [BCons]
	  funType t [] (x,et) = return [Eq t et (ixToJust x)]
	  funType t ((x,p):nps) et = do t' <- newVar
					let bc= Eq t (p `arrow` t') (ixToJust x)
					bcs <- funType t' nps et
					return (bc:bcs)
	  

-- translates the right-hand-side of a definition
-- for multiple bodies, translate:
--  1) all guards, and unify with Bool
--  2) all expressions, and unify with each other (ix has loc for unifying var)
trRHS :: Env -> [TVar] -> Index -> RHS -> TR LRep
trRHS env lvs _  (RHS1 ae)  = trExp env lvs ae

trRHS env lvs ix (RHSG ges) = do
    let j = ixToJust ix
        (gs, es) = unzip ges
    (tgs,v1) <- trExps env lvs gs
    (tes,v2) <- trExps env v1 es
    tv  <- newVar

    let beqs = [ cg' | (tg, cg) <- tgs, 
		       let eq  = bToCons (Eq tg boolType j)
			   cg' = eq `cand` cg ]
			      
	reqs = [ ce' | (te, ce) <- tes, 
		       let eq  = bToCons (Eq tv te j)
			   ce' = eq `cand` ce ]
	cons = candList (beqs ++ reqs)

    return (tv, cons, v2)

	
trWhere :: Env -> [TVar] -> [Ann Dec] -> TR (Env, [CHR], [TVar])
trWhere env lvs []  = return (env, [], lvs)
trWhere env lvs ads = trProg env lvs (Prog ads)

-- looks up an id in the tenv
-- its annotation should reveal its source location!
trAId :: Env -> Ann Id -> TR Rep
trAId env aid = case lookupV (item aid) env of
		Nothing -> fail (formatSrcPos (ixSrcPos (ann aid)) ++
				 "ERROR: undefined variable `" ++ 
				 item aid ++ "'\n")

		Just tc@(t, c) -> -- don't rename if its just a type variable
				  if isVarRep tc
				     then return tc

				     else do (t, c') <- renameCalled tc
					     c'' <- renameConsIx c'
					     return (t, c'')
				     
-- process data declarations
trDatDecs :: Env -> [Ann Dec] -> TR Env
trDatDecs env adecs = do
    let dats = items $ datDecsToDats adecs
    env0 <- trTypeCons env dats
    let env0' = env0 `addEnv` env
    env1 <- trDatCons env0' dats
    let env1' = env1 `addEnv` env0'
    return env1'

-- process rule declarations
trRuleDecs :: Env -> [Ann Dec] -> TR (Env, [CHR])
trRuleDecs env adecs = do
    let rules = (map ruleDecToRule) $ items $ ruleDecs adecs
    trRules env rules 

-- given a list of data decls, generate new type constructors and insert them
-- into the environment (this must be done for all decls before data
-- constructors can be handled)
trTypeCons		    :: Env -> [Dat] -> TR Env
trTypeCons env dats	    = do envs <- mapM (trTypeCon env) dats
				 return (sumEnvs envs)

trTypeCon env (Dat aid as _)= do x <- next
				 vs <- newVars (length as)
				 let id = item aid
				     nm = name id x
				     tc = TyCon nm
				     typ = tc
				     env0 = extendT emptyEnv (id, typeRep typ)
				 return env0
   

-- takes current environment and a list of data declarations, and returns a new
-- environment with updated venv to include new data constructors
trDatCons		    :: Env -> [Dat] -> TR Env
trDatCons env dats	    = foldM (trDatCon env) emptyEnv dats

trDatCon		    :: Env -> Env -> Dat -> TR Env
trDatCon outEnv env (Dat id as cons)
			    = do vs <- newVars (length as)
				 x  <- next
				 let rs = zip vs (repeat trueCons)
				     tcon = fst $ fromJust $ 
						lookupT (item id) outEnv
				     t = foldl TApp tcon vs
				     env1 = foldl extendT env
						  (zip (items as) rs)
				 envs <- mapM (trCon outEnv env1 t) (items cons)
				 return (addEnv env (sumEnvs envs))

trCon			    :: Env -> Env -> Type -> Con -> TR Env
trCon outEnv env res (Con id ts)   
			    = do etcs <- trTypsCur (env `addEnv` outEnv) ts 
				 let types = map fst (snd etcs)
				     t = foldr arrow res types
				     env' = extendV emptyEnv (item id,typeRep t)
				 return env'


trExps :: Env -> [TVar] -> [Ann Exp] -> TR ([Rep], [TVar])
trExps _   v0 []     = return ([],v0)
trExps env v0 (e:es) = do
    (t, c, v1) <- trExp  env v0 e
    (rs,v2) <- trExps env v1 es
    let r = (t,c)
    return (r:rs,v2)

-- translates an expression into a rep
-- set of CHRs which arise from local declarations (ie. let expressions) get
-- hidden in the monad state and passed implicitly
-- NOTE: need to return a list of the lambda-bound variables encountered within
-- this expression
trExp :: Env -> [TVar] -> Ann Exp -> TR LRep
trExp env v0 aexp =   do (t,c,e_lvs) <- rep
{-			 trace ("\nexp: " ++ show (item aexp) ++
				"\nv0: " ++ show v0 ++
				"\ne_lvs: " ++ show e_lvs)
			    $ -}
			 return (t,c, e_lvs)
    where
	addLV = (v0++)
	ix  = ann aexp
	exp = item aexp
	j   = ixToJust ix
	l   = ixLoc ix
    
	rep = case exp of
	 VarExp id	 -> do r@(t, c) <- trAId env (Ann ix id)
			       t1 <- newVarLoc l
			       sd <- newVar
			       sv <- genScopedVarVec env 
--			       is <- trGet parentIds
			       isC <- isChild id
			       lv  <- genLambdaVarVecPair env
			           -- if this entry belongs to a lambda-bound
				   -- var, justify the eq. constraint by both
				   -- the expression, and binding location
				   -- so that if this constraint appears in an
				   -- error, the pattern location is implicated
				   -- as well as the use
			       let j' = if isLamVarRep r
					   then let (_, l) = unVar t
						in  locToJust l `addJust` j
					   else j
				   
				   eq = bsToCons [Eq t1 t j']
				   k = if c == trueConsAlt then trueCons else c
			           c'  = eq `cand` reJust j k
			           c'' = if not $ isVarRep r 
					 then -- must add lambda-bound to call
					 let C ucs bcs = c'
					     ucs' = replace2ndUCArgs lv ucs
					     ucs''= if not isC -- id `elem` is 
						    -- don't replace on calls up
						    then ucs' 
						    else replace3rdUCArgs sv ucs'
					 in  C ucs'' bcs
					 else c'
			       return (t1, c'', v0)
	
	 ConExp id	 -> do (t, c) <- trAId env (Ann ix id)
			       t1 <- newVarLoc l
			       let eq = bsToCons [Eq t1 t j]
			           c' = reJust j (eq `cand` c)
			       return (t1, c', [])

	 LitExp i	 -> do (t, c) <- trLit i
			       tl <- newVarLoc l
			       let eq = bsToCons [Eq tl t j]
				   c' = reJust j (eq `cand` c)
			       return (tl, c', [])

	 AbsExp ps e     -> do (env', (tps, pc, vp)) <- trPats env ps
                	       (te, c, ve) <- trExp env' (addLV vp) e
			       t <- newVarLoc l
			       let tp = foldr arrow te (reverse tps)
				   c' = bToCons (eq t tp j)
                	       return (t, c `cand` c' `cand` pc, ve)

	 AppExp e1 e2	 -> do (t1, c1, v1) <- trExp env v0 e1 
			       (t2, c2, v2) <- trExp env v1 e2 
			       t3  <- newVarLoc l
			       let c' = bToCons (eq t1 (t2 `arrow` t3) j)
			       return (t3, c' `cand` (c1 `cand` c2), v2)
             
	 IfxAppExp e1 e2 e3 ->
			    do (t1, c1, v1) <- trExp env v0 e1
			       (t2, c2, v2) <- trExp env v1 e2
			       (t3, c3, v3) <- trExp env v2 e3
			       r <- newVarLoc l
			       let c = bToCons
					(Eq t2 (t1 `arrow` (t3 `arrow` r)) j)
			       return (r, candList [c1,c2,c3,c], v3)

	     
	 IfExp b e1 e2   -> do (t1, c1, v1) <- trExp env v0 b 
			       (t2, c2, v2) <- trExp env v1 e1
			       (t3, c3, v3) <- trExp env v2 e2
			       t  <- newVarLoc l
			       let c4 = eq t1 boolType j
			           c5 = eq t t2 j
				   c6 = eq t t3 j
				   cs = bsToCons [c4, c5, c6]
			       return (t, candList [c1, c2, c3, cs], v3)

	 -- type addition - just constrains e's type by t
	 TAddExp e ts	 -> do (t1, c1, v1) <- trExp env v0 e
			       (t2, c2) <- trTSchm env ts -- trTypNew env t
			       v <- newVarLoc l
			       let c  = bToCons (eq v t2 j)
				   c' = bToCons (eq v t1 j)
				   cs = c `cand` c' `cand` c1 `cand` c2
			       return (v, cs, v1)

	 -- annotated expression, e in: f = ... e_j :: t
	 -- is translated as though it 
	 --    were a local definition: f = ... let hj :: t
	 --					    hj = e_j
	 --					in  hj
	 -- i.e. inferred, and annotated chrs are generated
	 -- (which are then used for the entailment/ambiguity checks)
	 -- NOTE: these CHRs are generated *in addition* to the normal
	 --	  translation. i.e. we generate constraints for the current CHR
	 --	  (of the def. being translated) as for type addition

	 -- this version actually generates the corresponding LetExp and calls
	 -- trExp recursively (so that the ann-let code doesn't need to be 
	 -- duplicated)
	 -- PROBLEM: calling trExp on the LetExp returns a call to the
	 --	     annotation CHR, when we really want a call to the join CHR
	 --	     (so that we unify inferred and annotated types at this
	 --	     point)
	 --	     So, maybe we just use the let to generate the CHRs, and go
	 --	     ahead and directly generate the constraints for local use
	 --	     right here (duplication of work, but meh)
	 -- SOLVED!
	 AnnExp e ts	 -> let at = Ann ix
				ae = Ann (ann e)
				id = "ann:" ++ show j
				sig = tschmToSig (item ts) (ae id)
				def = Def (ae id) [] (RHS1 e) []
				dec = AValDec (at sig) [at def]
				exp = LetExp [at dec] (at (VarExp id))
			    in  do -- generate CHRs as though 'twere a let bind
				   trExp env v0 (at exp)

				   -- now generate constraints for local use
				   (te, ce, v1) <- trExp env v0 e
				   (tt, ct) <- trTSchm env ts
				   t <- newVarLoc l
				   let teq1 = Eq t te j
				       teq2 = Eq t tt j
				       cteq = bsToCons [teq1, teq2]
				       c = cteq `cand` ce `cand` ct 
				   return (t, c, v1)
			    
{-
	 AnnExp e ts	 -> do -- generate the CHRs for annotation checking
			       let at = Ann ix -- use the loc of the whole ann 
				            -- (ann t)
				   ae = Ann (ann e)
			           id = "ann:" ++ show j
				   sig = tschmToSig (item ts) (ae id)
				   def = Def (ae id) [] (RHS1 e) []
				   dec = AValDec (at sig) [ae def]
				   adec = Ann ix dec
			       env' <- valDecsEnv env [adec]
			       (chrs,_v1) <- trValDec env' v0 adec

			       gucs <- currentGlobalCons
			       let (achrs, rest) = partition isAnnCHR chrs
				   (jchrs, rest')= partition isValJoinCHR rest
				   -- add global constraints to annotation CHR
				   achr  = head achrs
				   jchr  = head jchrs
				   gcs   = usToCons gucs
				   achr' = chrAddCons achr gcs
				-- add call to parent to join CHR
				-- (to unify lambda-bound vars and type vars)
			       id <- topParentId
			       out_stvs <- popStvVec
			       let [_,j_l,j_stvs] = ucTypes (simpUCons jchr)
			       
			       mpuc <- topParentAnnUCons
			       lvp <- newVar
			       lv1 <- newVar
			       lv2 <- newVar
			       puc_t <- newVar
			       let puc  = fromJust mpuc
				   puc' = reJust noJust puc
				   [_puc_t,_,stvs']= ucTypes puc'
				   puc'' = newUCTypes [puc_t, lvp, stvs'] puc'
			   
				   eq_stvs = [ j_stvs `njEq` out_stvs,
					      -- pass back to the parent the
					      -- list of lambda-bound vars
					      -- coming into the child (reverse
					      -- the lambda-vec pair)
					      j_l `njEq` tupleType [lv1,lv2],
					      lvp `njEq` tupleType [lv2,lv1]
					      ]
			       
			       let pc = if isJust mpuc 
					then usToCons [puc'']
					else trueCons
				   pc'= pc `cand` (bsToCons eq_stvs)
				   jchr' = chrAddCons jchr pc'
				   jchr''= reJust noJust jchr'
			       
			       let chrs' = jchr' : achr' : rest'
{-				   chrs'' = trace (
				    "\noriginal CHR:\n" ++ show jchr ++
				    "\nupdated CHR:\n" ++ show jchr' ++ 
				    "\npc': " ++ show pc')
				    chrs' -}				    
			       addRules chrs'

			       -- generate the constraints for the current def.
			       (t1, c1, v1) <- trExp env v0 e
			       (t2, c2) <- trTSchm env ts -- trTypNew env t
			       t <- newVarLoc l
			       let c  = bToCons (eq t t2 j)
				   c' = bToCons (eq t t1 j)
				   cs = c `cand` c' `cand` c1 `cand` c2
			       return (t, cs, v1)
-}

	 TupExp es       -> if singleton es
			       then trExp env v0 (head es)
			       else do (rs,vs) <- trExps env v0 es
				       v   <- newVarLoc l
				       let (ts,cs) = unzip rs
					   t = tupleType ts
					   b = bToCons (eq v t j)
				       return (v,b `cand` candList cs, vs)
					  
	 ListExp es      -> if null es 
			       then do v <- newVar
				       t <- newVarLoc l
				       let c = bToCons (eq t (listType v) j)
				       return (t, c, v0)
			       else do (rs, vs) <- trExps env v0 es
				       let (ts,cs) = unzip rs
				       v1  <- newVar
				       v2  <- newVarLoc l
				       let eqc = listEq v1 j ts trueCons
					   eql = bToCons
						    (eq v2 (listType v1) j)
					   cs' = candList cs
					   cs2 = cs' `cand` eqc `cand` eql
				       return (v2, cs2, vs)

	 EnumExpFrom e   -> do t <- newVarLoc l
			       (te,ce,v1) <- trExp env v0 e
			       let ec = enumConsJust te j
			           bt = bToCons (Eq t (listType te) j)
				   cons = bt `cand` ec `cand` ce
			       return (t, cons, v1)
			       
	 EnumExpFromThen e1 e2 -> 
			    do v <- newVarLoc l
			       (t1,c1,v1) <- trExp env v0 e1
			       (t2,c2,v2) <- trExp env v1 e2
			       let ec = enumConsJust t1 j
			           bt = bToCons (Eq v (listType t1) j)
				   eq = bToCons (Eq t1 t2 j)
				   cons = candList [bt,eq,ec,c1,c2]
			       return (v, cons, v2)

	 -- same as EnumExpFromThen
	 EnumExpFromTo e1 e2 -> 
			    do v <- newVarLoc l
			       (t1,c1,v1) <- trExp env v0 e1
			       (t2,c2,v2) <- trExp env v1 e2
			       let ec = enumConsJust t1 j
			           bt = bToCons (Eq v (listType t1) j)
				   eq = bToCons (Eq t1 t2 j)
				   cons = candList [bt,eq,ec,c1,c2]
			       return (v, cons, v2)

	 EnumExpFromThenTo e1 e2 e3 -> 
			    do v <- newVarLoc l
			       (t1,c1,v1) <- trExp env v0 e1
			       (t2,c2,v2) <- trExp env v1 e2
			       (t3,c3,v3) <- trExp env v2 e3
			       let ec = enumConsJust t1 j
			           bt = bToCons (Eq v (listType t1) j)
				   eq = bsToCons [Eq t1 t2 j, Eq t2 t3 j]
				   cons = candList [bt,eq,ec,c1,c2,c3]
			       return (v, cons, v3)


	 -- 1) translate statements from left to right, threading env
	 -- 2) translate expression
	 -- 3) result is list type of expression
	 ListCompExp e ss -> 
			    do (env0, rs, v1) <- trStmts env v0 ss
			       r@(te, ce, v2) <- trExp env0 v1 e
			       t <- newVarLoc l
			       let eq  = bToCons (Eq t (listType te) j)
				   cs0 = candList (map snd rs)
				   cs1 = eq `cand` ce `cand` cs0
			       return (t, cs1, v2)


	 CaseExp e as -> do (te, ce, v1) <- trExp env v0 e
			    (ps, as, v2) <- trCaseAlts env v0 j as
			    at  <- newVar
			    let (pts, pcss) = unzip ps
				(ats, acss) = unzip as
				acs = candList acss
				pcs = candList pcss
				eqp = listEq te j pts trueCons
				eqa = listEq at j ats trueCons
				cs  = eqp `cand` eqa `cand` ce `cand`
				      pcs `cand` acs
			    return (at, cs, v2)

	 SecExpLeft  e1 e2 -> 
			    do (t1, c1, v1) <- trExp env v0 e1
			       (t2, c2, v2) <- trExp env v1 e2
			       t <- newVarLoc l
			       a <- newVar
			       b <- newVar
			       let eqs = bsToCons [Eq t2 (t1 `arrow` t) j,
						   Eq t (a `arrow` b) j]
			       return (t, eqs `cand` c1 `cand` c2, v2)
				

	 SecExpRight e1 e2 ->
			    do (t1, c1, v1) <- trExp env v0 e1
			       (t2, c2, v2) <- trExp env v1 e2
			       t <- newVarLoc l
			       a <- newVar
			       b <- newVar
			       let eqs = bsToCons 
					  [Eq t1 (a `arrow` (t2 `arrow` b)) j,
					   Eq t (a `arrow` b) j]
			       return (t, eqs `cand` c1 `cand` c2, v2)


	 -- `let' begins a nested block of declarations (just definitions atm)
	 -- we generate chrs as per top-level decls., as well as add entries
	 -- for them to the environment
	 -- second chr arg represents a list of the types of lambda-bound in
	 -- the above scope
	 LetExp ds e	-> do let subIds = map (decId.item) ds
			      (env', chrs, v1) <- trProg env v0 (Prog ds)
			      cids <- getChildIds
			      addChildIds subIds

			      rep@(t_e,c_e,v2) <- trExp env' v1 e

			      let (achrs, rest) = partition isAnnCHR chrs
				  (jchrs, rest')= partition isValJoinCHR rest
				  

			      -- adds necessary administrative constraints to
			      -- the join CHR:
			      -- 1) need to call enclosing CHR
			      -- mpuc <- topParentUCons -- topParentAnnUCons
			      mpuc' <- topParentUCons
			      let mpuc = Just (fst mpuc')
			      lv1 <- newVar
			      lv2 <- listifyOpen (vars v2)
			      -- v_l <- newVar
			      vv_new <- newVar
			      t_new  <- newVar
			      let puc = fromJust mpuc
				  [puc_t,_,puc_v] = ucTypes puc
				  mvgl = tupleTypes puc_v
				  ~[v_g, v_l] = fromJust mvgl
				  vv = if isNothing mvgl 
				       then vv_new -- v_l
				       else puc_v -- tupleType [v_g,v_l]
				  lv = tupleType [lv1,lv2]
			
			      -- NOTE: we justify the call up by the location
			      --       of the head of the CHR!
			      --       (should be okay for now)
			      let trJCHR chr = do
				    ix <- next
				    let j = ucJust (simpUCons chr)
					puc' = -- reJust j $ 
						reJust adminJust $ 
						newUCIx (-ix) $
						newUCTypes [t_new,lv,vv] puc
					cons = if isJust mpuc
						then uToCons puc'
						else trueCons
					chr1 = chrAddCons chr cons
{-					
				    trace ("\nsubIds: " ++ show subIds ++ 
					   "\npuc: " ++ show puc ++
					   "\npuc': " ++ show puc' ++ 
					   "\nv0: " ++ show v0 ++ 
					   "\nv1: " ++ show v1 ++
					   "\nv2: " ++ show v2) $ -}
			            return chr1

			      achrs' <- return achrs -- mapM trACHR achrs
			      jchrs' <- mapM trJCHR jchrs
				  
			      let chrs' = jchrs' ++ achrs' ++ rest'

			      setChildIds cids
			      addRules chrs'
			      ps <- trGet parentIds
			      return (t_e,c_e,v2)

	 e -> error ("no translation for expression: " ++ show e)


trStmts :: Env -> [TVar] -> [Ann Stmt] -> TR (Env, [Rep], [TVar])
trStmts env v0 [] = return (env, [], v0)
trStmts env v0 (s:ss) = do (env', r, v1)  <- trStmt env v0 s
			   (env'',rs,v2) <- trStmts env' v1 ss
			   return (env'', r:rs, v2)

trStmt :: Env -> [TVar] -> Ann Stmt -> TR (Env, Rep, [TVar])
trStmt env v0 astmt = 
    let stmt = item astmt
	ix   = ann astmt
	j    = ixToJust ix
    in  case stmt of

	 GenStmt p e -> do (env0, (tp, cp, vp)) <- trPat env p
			   (te, ce, v1) <- trExp env (v0 ++ vp) e
			   let eq = bToCons (Eq te (listType tp) j)
			       cs = eq `cand` cp `cand` ce
			   return (env0, (te, cs), v1)
			   
	 QualStmt e  -> do (t, c, v1) <- trExp env v0 e
			   let eq = bToCons (Eq t boolType j)
			   return (env, (t, eq `cand` c), v1)
	 
	 LetStmt ds  -> exit "trStmt: let in list comp not handled yet"


trCaseAlts :: Env -> [TVar] -> Just -> [CaseAlt] -> TR ([Rep], [Rep], [TVar])
trCaseAlts env v0 j [] = return ([],[],v0)
trCaseAlts env v0 j (c:cs) = do
    (rg, re, v1) <- trCaseAlt  env v0 j c
    (rgs,res,v2) <- trCaseAlts env v1 j cs
    return (rg:rgs, re:res, v2)

-- returns a rep for both the pattern and result
-- FIXME: translate the where clauses as well!
trCaseAlt :: Env -> [TVar] -> Just -> CaseAlt -> TR (Rep, Rep, [TVar])
trCaseAlt env v0 j (CaseAlt p e w) = do 
    (env0, (t,c,v1)) <- trPat env p
    (t',c',v2) <- trExp env0 (v0++v1) e
    let rp = (t,c)
	re = (t',c')
    return (rp, re, v2)

{-
trCaseAlt env j (CaseAltGuard p ges w) = do
    (env0, rp) <- trPat env p
    rs <- mapM (trGEs env0) ges
    t  <- newVar
    let (ts, cs, vs) = unzip3 rs
	cs0 = candList cs
	eq  = listEq t j ts trueCons
    return (rp, (t, eq `cand` cs0, concat vs))

    where trGEs env (g,e) = do (tg, cg, vg) <- trExp env g
			       (te, ce, ve) <- trExp env e
			       let eq = bToCons 
					    (Eq tg boolType (ixToJust (ann g)))
				   cs = eq `cand` cg `cand` ce
			       return (te, cs, vg++ve)

-}

listEq _ _ []     c = c
listEq v j (t:ts) c = let c'  = bToCons (eq v t j)
			  c'' = c `cand` c'
		      in  listEq v j ts c''


listEqSnd :: Type -> Just -> [Type] -> Constraint -> TR Constraint
listEqSnd v j ts c = do v1 <- newVar
			v2 <- newVar
			let bc = bToCons (eq v (tupleType [v1,v2]) j)
			c' <- list v1 c ts
			return (bc `cand` c')
    where list _  c [] = return c
	  list v1 c (t:ts) = do v2 <- newVar
				let bc = bToCons (eq (tupleType [v1,v2]) t j)
				list v1 (bc `cand` c) ts



trLit :: Lit -> TR Rep
trLit lit = case lit of 
    IntLit _ _   -> do v <- newVar
		       return (v, numCons v)
    FloatLit _ _ -> do v <- newVar
		       return (v, fractionalCons v)
    StrLit _  -> return (typeRep stringType)
    CharLit _ -> return (typeRep charType)


trTypNew = trTyp (True, False)
trTypCur = trTyp (False, False)
trTypsNew = trTyps (True, False)
trTypsCur = trTyps (False, False)
trTypNewCons = trTyp (True, True)

-- FLAGS:
--  1. if True, unbound variables can be added to the type environment if not 
--     there already, else their use results in failure
--  2. if True, return types as constraints, else just return the type directly
--     (constraint component can be discarde by caller)
trTyp :: (Bool, Bool) -> Env -> Ann Typ -> TR (Env, Rep)
trTyp fs@(fNew, fCons) env atyp = 
    let ix  = ann atyp
	typ = item atyp
	l   = ixLoc ix
	j   = ixToJust ix
	trTyp'  = trTyp fs
	trTyps' = trTyps fs
	getTVar = if fNew then useTVar
			  else \env id -> lookupTVar env (Ann ix id)
    in  case typ of
	     (VarTyp id) 
				  
		| head id == '_' -> do t1 <- newVarLoc l
				       return (env, (t1, trueCons))
			    
		| otherwise -> do t1 <- newVarLoc l
				  (r@(t,c), env') <- getTVar env id
				  let eq = bToCons (Eq t1 t j)
				  if fCons then return (env', (t1, eq `cand` c))
					   else return (env', r)

	     (ConTyp id)    -> do t1 <- newVarLoc l
				  r@(t, c) <- lookupTCon env (Ann ix id)
				  let eq = bToCons (Eq t1 t j)
				  if fCons then return (env, (t1, eq `cand` c))
					   else return (env, r)	

	     (AppTyp t1 t2) -> do v <- newVarLoc l
				  (env1, (t1', c1)) <- trTyp' env  t1
                                  (env2, (t2', c2)) <- trTyp' env1 t2
				  let t  = TApp t1' t2'
				      eq = bToCons (Eq v t j)
				      c  = c1 `cand` c2
				      c' = eq `cand` c
				  if fCons then return (env2, (v, c'))
					   else return (env2, (t, c))

	     (TupTyp ts)    -> do (env', tcs) <- trTyps' env ts
				  t <- newVarLoc l
                                  let (ts, cs) = unzip tcs
				      cs' = candList cs
				      c = bToCons (Eq t (tupleType ts) j)
				      cs'' = c `cand` cs'
                                  return (env', (t, cs''))

	     (ListTyp t)    -> do (env', (tt, ct)) <- trTyp' env t
				  t <- newVarLoc l
				  let c = bToCons (Eq t (listType tt) j)
				      tc = (t, c `cand` ct)
                                  return (env', tc)


-- calls trTyp for each type in the list, threading the environment through
-- from left to right (ie. types are in the same scope)
trTyps :: (Bool, Bool) -> Env -> [Ann Typ] -> TR (Env, [Rep])
trTyps f env [] = return (env, [])
trTyps f env (at:ats) = do
    (env0, r)  <- trTyp f env at
    (env1, rs) <- trTyps f env0 ats
    return (env1, r:rs)



-- returns a list of types, one per pattern element (I think) and a constraint
trPats :: Env -> [Ann Pat] -> TR (Env, ([Type], Constraint, [TVar]))
trPats env []       = return (env, ([], trueCons, []))
trPats env (ap:aps) = do (env1, (t, c, v))   <- trPat  env  ap
			 (env2, (ts, c', vs)) <- trPats env1 aps
			 return (env2, (t:ts, c `cand` c', v++vs))

trPat :: Env -> Ann Pat -> TR (Env, LRep)
trPat env apat =
    let ix  = ann apat
	pat = item apat
	j   = ixToJust ix
	l   = ixLoc ix
    in  case pat of
	 JokPat		-> do v <- newVarLoc (ixLoc ix)
                              return (env, (v, trueCons, []))

	 VarPat id	-> do v <- newVarLoc (ixLoc ix)
                              let env'  = extendV env (id, typeRep v)
				  env'' = env' -- extendUB env' (id, typeRep v)
                              return (env'', (v, trueCons, [unVar v]))

	 ConPat id ps	-> do let aid = ann id 
				  idj = ixToJust aid
                              (t, c1) <- trAId env id
			      cv <- newVarLoc (ixLoc aid)
			      let bc  = bToCons (Eq cv t idj)
			      let c1' = reJust idj c1

                              (env', (ts, c2, vs)) <- trPats env ps
                              v <- newVarLoc (ixLoc ix)
                              let pt = foldr arrow v ts
                                  c3 =c1' `cand` c2 `cand` bc `cand` 
					  bToCons (eq cv pt j)
                              return (env', (v, c3, vs))

	 LitPat lt	-> do (t, c) <- trLit lt
			      tl <- newVarLoc l
			      let eq = bsToCons [Eq tl t j]
			          c' = reJust j (eq `cand` c)
			      return (env, (tl, c', []))

	 TupPat ps	-> do (env',(ts, cs, vs)) <- trPats env ps
                              v <- newVarLoc (ixLoc ix)
                              let c   = bToCons (eq v (tupleType ts) j)
                                  cs' = c `cand` cs
                              return (env', (v, cs', vs))

	 ListPat ps	-> case items ps of
                            [] -> do v1 <- newVar -- Loc l
                                     v2 <- newVarLoc l
                                     let c = bToCons (eq v2 (listType v1) j)
                                     return (env, (v2, c, []))
                            qs -> do (env',((t:ts), cs, vs)) <- trPats env ps
                                     v <- newVarLoc l
                                     let eqs = bsToCons (eqList j ts)
                                         cl  = bToCons (eq v (listType t) j)
                                         cs' = cs `cand` eqs `cand` cl
                                     return (env', (v, cs', vs))

trRules :: Env -> [Ann Rule] -> TR (Env, [CHR])
trRules env [] = return (env, [])
trRules env (ar:ars) = do (env0, chr)  <- trRule env ar
			  (env1, chrs) <- trRules env0 ars
			  return (env1, chr:chrs)

trRule :: Env -> Ann Rule -> TR (Env, CHR)
trRule env arule = 
    let rule = item arule
	ix   = ann arule
	j    = ixToJust ix
    in  case rule of
	 (SimpRule ctxt cnst) ->
			      do (env1, c1) <- trCtxt env  (Ann ix (item ctxt))
				 (env2, c2) <- trCnst env1 ix (item cnst)
				 let lucs = reJust j (consToUs c1)
				     luc  = head lucs
				     rhs  = (consToUs c2, consToBs c2)
				     info = ruleInfo (ixLoc ix)
				     simp = Simp info luc [] rhs
				     msimp= MSimp info lucs [] rhs
				     env3 = env {penv = penv env2}
				 if length lucs > 1 then return (env3, msimp)
						    else return (env3, simp)
	 (PropRule ctxt cnst) ->
			      do let (Ctxt ps) = item ctxt
				     ctxt' = Ann ix (Ctxt (map (Ann ix) (items ps)))
				 (env1, c1) <- trCtxt env ctxt' 
				 (env2, c2) <- trCnst env1 ix (item cnst)
				 let n = fromIntegral (ixLoc ix)
				     lucs = consToUs c1
				     rhs  = (consToUs c2, consToBs c2)
				     info = ruleInfo (ixLoc ix)
				     chr  = Prop info n lucs [] rhs
				     env3 = env {penv = penv env2}
				 return (env3, chr)


trCnst			    :: Env -> Index -> Cnst -> TR (Env, Constraint)
trCnst env ix (Cnst prims)  = trPrims env ix (items prims)

trPrims			    :: Env -> Index -> [Prim] -> TR (Env, Constraint)
trPrims env ix []	    = return (env, trueCons)
trPrims env ix (p:ps)	    = do (env1, c1) <- trPrim env ix p
				 (env2, c2) <- trPrims env1 ix ps
				 return (env2, c1 `cand` c2)
				 
trPrim			    :: Env -> Index -> Prim -> TR (Env, Constraint)
trPrim env ix (PredPrim p)  = do (env', uc) <- trPred env ix p
				 return (env', uToCons uc)
trPrim env ix (EqPrim t1 t2)= do (env1, (tt1, c1)) <- trTypNew env  t1
				 (env2, (tt2, c2)) <- trTypNew env1 t2
				 let j   = ixToJust ix
				     t1' = apply (unSucc (buildMGU c1)) tt1
				     t2' = apply (unSucc (buildMGU c2)) tt2
				     c3  = bToCons (eq t1' t2' j)
				 return (env2, c3)

buildMGU :: Constraint -> E Subst
buildMGU c = unify [] idSubst (buildUP (consToBs c))

-- generate a bunch of UCons (that can be put in the body of a simp. rule)
trCtxt :: Env -> Ann Ctxt -> TR (Env, Constraint)
trCtxt env actxt = 
    let ctxt = item actxt
    in  case ctxt of
	 (Ctxt ps) -> do (env', ucs) <- trPreds env ps
			 let cons = usToCons ucs
			 return (env', cons)

trPred :: Env -> Index -> Ann Pred -> TR (Env, UCons)
trPred env ix pred  = do (env',[uc])<- trPreds env [Ann ix (item pred)]
		         return (env', uc)

trPreds :: Env -> [Ann Pred] -> TR (Env, [UCons])
trPreds env []	= return (env, [])
trPreds env (ap:aps) =
    let pred = item ap
	ix   = ann ap
    in  case pred of
	 (Pred id t) -> do (nm, env1) <- trPredId env id
			   (env2, (tt', tc)) <- trTypNew env1 t
			   ix' <- next
			   let tc_phi = unSucc (unify [] idSubst 
						(buildUP (consToBs tc)))
			       tt  = apply tc_phi tt'
			       uc  = UC ix' nm [tt] (ixToJust ix) 
			   (env3, ucs) <- trPreds env2 aps
			   return (env3, uc:ucs)

-- lookup the given pred id in the penv, if it's not there, add an entry for it
trPredId 		    :: Env -> Id -> TR (Name, Env)
trPredId env id		    = do let res = lookupP id env
				 if isJust res
				    then return (fromJust res, env)
				    else do n <- next
					    let nm   = name id n
					    let env' = extendP env (id, nm)
					    return (nm, env')


trTSchm :: Env -> Ann TSchm -> TR (Type, Constraint)
trTSchm env ats =
    case item ats of
	 (TSchm ctxt typ) -> do (env',c1)  <- trCtxt env ctxt
				(_,(t,c2)) <- trTypNewCons env' typ
				let l = ixLoc (ann ats)
				    j = ixToJust (ann ats)
				v <- newVarLoc l
				let c  = c1 `cand` c2
				    b  = bToCons (eq v t j)
				    c' = b `cand` c

				    ret = (v,c')
				return ret

--------------------------------------------------------------------------------

-- given a single def, translates it into a CHR
-- not used during initial CHR generation, only later to infer types
-- of one-off expressions in the debugger
transDef :: Env -> Ann Def -> TR CHR
transDef env adef = do
    env' <- valDecsEnv env [nullAnn (ValDec [adef])]
    (chr0,_) <- trNonOverDef env' [] (adef)
    unfoldSelfRecursion chr0
    
    

transTSchm :: Env -> Ann TSchm -> TR (Type, Constraint)
transTSchm env ats = trTSchm env ats

transTyp :: Env -> Ann Typ -> TR Type
transTyp env typ = do (_, r) <- trTypNew env typ
		      return (fst r)

transExp :: Env -> Ann Exp -> TR (Type, Constraint)
transExp env exp = do (t,c,_) <- trExp env [] exp
		      return (t,c)

--------------------------------------------------------------------------------
-- functions for translating types/defs/exprs

-- integer comes from the parser (ensures all unique numbers are actually
-- unique - always)
-- also passed in are entries for the pred. env (from hconstrain decl.s)
progToCHRs :: Integer -> PEnv -> Prog -> IO ((Env, [CHR]), State)
progToCHRs i pe prog = do res <- doProg i pe prog
			  case res of
			    Succ as@(a, s) -> return as
			    Err  e	   -> putStr e >> exitFailure

doProg n pe prog = runTRInit n (do clrParentIds 
				   (env, chrs, _) <- trProg env [] prog
				   wchrs <- wakeDelayedCHRs chrs
				   inenv <- getInnerEnv 
				   return (inenv `addEnv` env, chrs ++ wchrs))
    where env = initEnv { penv = penv initEnv ++ pe }


-- generates names for all the constraints appearing in hconstraint decl.s
-- also, returns a version of the program without  hconstraint decl.s
progHaskellCons :: Prog -> IO (PEnv, Prog)
progHaskellCons (Prog adecs) = 
    let ids  = concatMap (hconDecIds . item) adecs
	ids' = nub ids 
	is  = [-1000,-1001..]
	ns  = [ name id n | (id, n) <- zip ids' is ]
	
	adecs' = filter (not . isHConDec . item) adecs
	
    in  do warn ids 
	   return (zip ids ns, Prog adecs')

    where warn ids = 
	    let ms = multiples ids
	    in  if null ms 
		then return ()
		else putStr ("WARNING: multiply declared Haskell " ++ 
			     "constraints: " ++ uncommas ms ++ "\n")
			    

--------------------------------------------------------------------------------
-- simplify rules for use in local mode

simplifyCHRs :: [CHR] -> [CHR]
simplifyCHRs = map simplifyCHR

-- simplifies any given simplification rule, justifications are discarded
simplifyCHR :: CHR -> CHR
simplifyCHR chr 
    | isSimp chr = let ts = ucTypes $ simpUCons chr
		       vs = fv ts
		       (ucs,bcs) = chrRhs chr
		       mgu = unify vs idSubst (buildUP bcs)
		       rhs' = case mgu of
			    Succ phi -> 
				let ucs' = apply phi ucs
				    bcs' = zipWith (\a b -> eq a b noJust)
						   ts (apply phi ts)
				in  (ucs', bcs')
			    Err _    -> consToRhs falseCons
				    
		   in  chrNewRhs rhs' chr
    | otherwise	= chr
		       
