--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


module TypeDebug (debugger, doSrcCommands)
where

import TD_Monad
import qualified TD_Translate as Trans
import Ann_InputAST 
import Pretty
import Parser
import Justify
import TD_General
import Misc

import Solver
import Chr  hiding (Index)
import ConsReason
import Core

import TDInference (inferTypeLoc, inferTypeRef, reportConflicts,
		    doDerivLoc, removeAppLocs, highlightConf, 
		    dd_descend, locExpStd, subLocs, solverFor, reportLastResult)
import TDExplain   (explainLocType, explainLocTypeOpt, 
		    explainExpType, explainRefType)

import TypeScheme hiding (tdDeriver)

import ASTs

import ErrorMsg


import List
import Monad
import Maybe
import System (exitWith, exitFailure, ExitCode(..), system)
import IO
import Char
import IOExts (trace)
import Array

-- import {-# SOURCE #-} Control

--------------------------------------------------------------------------------

data Command = NoCmd | UnrecCmd String
	     | PrintAllCHRs | PrintAnnAllCHRs | PrintProg | PrintAnnProg
	     | PrintCHR String | PrintAnnCHR String
	     | PrintIxCons Index | PrintEnv
	     | Deriv String | AnnDeriv String 
	     | DerivLoc Loc Bool | AnnDerivLoc Loc Bool
	     | InferTypeId NestedRef -- String 
	     | InferTypeLoc Loc (Bool, Bool, Bool) JustProp
	     | InferTypeRef VarRef
	     | InferTypeExp (Ann Exp)
	     | SetAnnFlag Index Flag 
	     | SetJustProp JustProp
	     | SetSOpt SOpt
	     | NarrowLoc Loc (Ann TSchm)
	     | NarrowId Id (Ann TSchm)
	     | DDebug (Maybe Loc) | TDebug Loc
	     | PrintResults
	     | ExplainLoc Bool Bool Loc Loc  
	     | ExplainLocType Loc Loc (Ann TSchm)
	     | ExplainExpType (Ann Exp) (Ann TSchm)
	     | ExplainRefType VarRef (Ann TSchm)
	     | PrintHLProg [Loc]
	     | Evaluate (Ann Exp)
	     | Reload  | Load String
	     | Help | Quit
	     | MISC String

data DDCommand = Yes | No | Explain Loc | Jump Loc | Back





--------------------------------------------------------------------------------
-- interactive environment

-- print initial state description, then enter eval-print
debugger :: TD ()
debugger = do jp <- getJProp
	      sm <- getSolvOpt
	      str <- getFilename
	      puts ("Solver Mode   : " ++ show (fromEnum sm))
	      puts ("Just'n Policy : " ++ show jp)
	      evalPrint str





----------------------------------------
-- process the source-based debugger commands

doSrcCommands :: [SrcCommand] -> TD ()
doSrcCommands cs = mapM_ (\c -> do puts (show c) 
				   proc c)  cs
    where 
	proc (SCTypeLoc l) = do
		p <- getProg
		inferTypeLoc l (True, True, True) LocalJust
		res <- lastResult
		puts (pprintAndHLLoc l p ++ "   :: " ++ show res)

	proc (SCExplainLoc l ts) = do
		p <- getProg
		so <- getSolvOpt
		setSolvOpt SDebug
		ls <- explainLocTypeOpt False l l (nullAnn ts)
		setSolvOpt so
		puts (pprintAndHLLocs ls p ++ "   :: " ++ show ts) 
		
	proc (SCTypeVar i l) = proc (SCTypeLoc l)

	proc (SCExplainVar i l ts) = proc (SCExplainLoc l ts)
	
	proc c = puts ("unhandled command: " ++ show c)


--------------------------------------------------------------------------------
-- top-level debugger functions


-- the interactive eval-print loop
-- fn - the prompt string
evalPrint :: String -> TD ()
evalPrint fn = loop
    where loop = do prompt fn
		    input <- doIO getInput
		    cmd   <- parseInput input
		    doCommand cmd
		    doIO (hFlush stdout)
		    loop

-- displays a user prompt
prompt :: String -> TD ()
prompt str = doIO (putStr (str ++ "> ") >> hFlush stdout)

-- reads user input
getInput :: IO String
getInput = do l <- getLine `catch` handleInputError
	      return l

-- simple error handler (allows for clean ^D exit)
handleInputError :: IOError -> IO a
handleInputError _ = do putStrLn "\nbye!"
			exitFailure


--------------------------------------------------------------------------------
-- command parser
-- commands are always of the form: <ws> cmd <ws> [arg_1, ... , arg_n]

parseInput :: String -> TD Command
parseInput str =
    let str' = dropWhile isSpace str
    in  case str' of
	    (':':cs)	-> parseCommand cs
	    _		-> parseCommand str'
--	    _		-> parseExpression str'

parseExpression :: String -> TD Command
parseExpression str = do n <- next
			 let pexp = parseExp (fromIntegral n) str
			 if isErr pexp
			    then return (UnrecCmd "invalid expression")
			    else do let (_, _, ix, exp) = unSucc pexp
					n' = fromIntegral (ixNum ix)
				    setCounter (n'+1)
				    return (Evaluate exp)

parseCommand :: String -> TD Command
parseCommand str =
    let str'  = dropWhile isSpace str
	str'' = preProcessCommandString str'
    in  case str'' of
	    ('q':_)	    -> return Quit
	    ('p':'a':_)	    -> return PrintAnnProg
	    ('p':'h':ls)    -> return (PrintHLProg ((read ls)::[Loc]))
	    ('p':_)	    -> return PrintProg
	    
	    ('c':'a':cs)-> case readStr cs of
				[]  -> return PrintAnnAllCHRs
				id  -> return (PrintAnnCHR id)
    			
	    ('c':cs)    -> case readStr cs of
                                []  -> return PrintAllCHRs
                                id  -> return (PrintCHR id)

	    ('z':cs)	-> return (MISC (readStr cs))

	    ('d':'a':' ':cs)-> parseDCommand True  cs
	    ('d':' ':cs)    -> parseDCommand False cs

	    ('d':'d':cs)    -> let str = readStr cs
			       in  if null str || not (isNum str)
				    then return (DDebug Nothing)
				    else return (DDebug (Just (read str)))

	    ('t':'d':cs)    -> let str = readStr cs
			       in  if null str || not (isNum str)
				    then return (UnrecCmd "")
				    else return (TDebug (read str))

	    ('t':' ':cs)    -> parseTCommand cs
	    ('T':' ':cs)    -> parseTTCommand cs
	    ('f':' ':cs)    -> parseFCommand cs

	    ('s':' ':cs)    -> parseSCommand cs

	    ('x':'c':cs)    -> let str = dropWhile isSpace cs
			       in  if not (isNum str)
				    then return (UnrecCmd "")
				    else let (numstr, str') = span isDigit str
					     l1    = read numstr
					     _str'' = dropWhile isSpace str'
					 in  return (ExplainLoc True True l1 l1)

	    ('x':' ':cs)    -> parseXCommand cs
	    
	    ('n':' ':cs)    -> parseNCommand cs
	    
	    ('e':_)	    -> return PrintEnv
	    
--	    ('r':_)	    -> return Reload -- PrintResults
--	    ('l':cs)	    -> return (Load (readStr cs))
	    ('h':_)	    -> return Help
	    (_:_)	    -> return (UnrecCmd "")
	    _		    -> return NoCmd

-- converts long version of command names into abbreviated version
preProcessCommandString :: String -> String
preProcessCommandString str = 
    let (key, rest) = span (not . isSpace) str
    in  case lookup key trans of
	     Just new	-> new ++ rest
	     Nothing	-> str
    
    where trans = [("explain", "x"), ("type", "t"), ("Type", "T"),
		   ("declare", "n"),
		   ("annotate", "n"), ("help", "h"), ("quit", "q"),
		   ("set", "s"), ("print", "p"), ("chr", "c"), ("deriv", "d")]
 

----------------------------------------
-- command-specific parsers

-- if flag is True then parse as a 'da' command
parseDCommand :: Bool -> String -> TD Command
parseDCommand da cs = let str = readStr cs
			  d   = if da then AnnDeriv    else Deriv
			  dl  = if da then AnnDerivLoc else DerivLoc
		      in  if isNum str 
			     then return (dl (read str) True)
			     else return (d str)

parseSCommand :: String -> TD Command
parseSCommand cs = 
    let str = dropWhile isSpace cs
	(tok, rst) = span (not . isSpace) str
    in  case tok of
	  "local"  -> return (SetJustProp LocalJust)
	  "global" -> return (SetJustProp GlobalJust)
	  "solver" -> let str = dropWhile isSpace rst
		      in  if isNum str
			     then let n = read str
				      n' = if n < 0 || n > 2 then 0 else n
				  in  do when (n /= n') 
					    (puts $ "value out of range; " ++
						    "assuming value `0'")
					 return (SetSOpt (toEnum 0))
			     else return (UnrecCmd "solver mode must be an \
						    \integer (0-2)")
	  str      -> return (UnrecCmd ("unknown flag `" ++ str ++ "'"))

parseTCommand :: String -> TD Command
parseTCommand cs = 
    let str = readStr cs
    in  if isNum str 
	then return (InferTypeLoc (read str) (True, True, False) LocalJust )
	else do n <- next
		let mvr  = parseVarRef cs
		    pexp = parseExp (fromIntegral n) cs

		if isErr pexp
		 then if isNothing mvr
			then return (UnrecCmd "Invalid expression/reference")
			else -- it's a reference
			     return (InferTypeRef (fromJust mvr))
			     
		 else -- it's an expression
		      do let (_, _, ix, exp) = unSucc pexp
			     n' = fromIntegral (ixNum ix)
			 setCounter (n'+1)
			 return (InferTypeExp exp)

parseTTCommand :: String -> TD Command
parseTTCommand cs = let mvr = parseVarRef cs
		    in  return $ if isJust mvr
				 then InferTypeRef (fromJust mvr)
				 else UnrecCmd "Could not parse reference"

-- we only support two kinds of explain queries now! (Explain(Loc/Exp)Type)
parseXCommand :: String -> TD Command
parseXCommand cs =
    let str = dropWhile isSpace cs
    in  if isNum str
	   then let (numstr, str') = span isDigit str
		    num = read numstr
		in  readTSchm (ExplainLocType num num) str'
	   else do n <- next
		   let (cs', rst) = bracketed cs
		       mvr  = parseVarRef (stripOuterParens cs')
		       pexp = parseExp (fromIntegral n) cs'
		   if isNothing mvr && isErr pexp
		      then return (UnrecCmd "invalid expression")
		      else do n2 <- next
			      let (ix,f) = if isJust mvr 
				      then (unkIx (fromIntegral n2), 
					    ExplainRefType (fromJust mvr))
				      else let (_,_,ix',exp) = unSucc pexp
					   in  (ix', ExplainExpType exp)

				  n' = fromIntegral (ixNum ix)
				  (cs'', rst') = bracketed rst
				  cs''' = stripOuterParens cs''
			      setCounter (n'+1)
			      readTSchm f cs''' 
    where -- invokes parser to read a type scheme
	  readTSchm f str = do
	    n <- next
	    let pts  = parseTSchm (fromIntegral n) str
	    if isErr pts 
	       then return (UnrecCmd "invalid typescheme")
	       else do let (_, _, ix, ts) = unSucc pts
		           n' = fromIntegral (ixNum ix)
		       setCounter (n'+1)
		       return (f ts)

	    

-- breaks the string into two parts, fst is the left-most bracketed
-- substring, snd is the rest	  
bracketed :: String -> (String, String)
bracketed str =
    let str'  = dropWhile (/='(') str
	str'' = if null str' then "" else tail str'
    in  match 1 "(" str''
	    where match 0 a cs	   = (reverse a, cs)
		  match _ a []	   = (reverse a, "")
		  match n a (c:cs) = case c of
		    '('	-> match (n+1) (c:a) cs
		    ')' -> match (n-1) (c:a) cs
		    _	-> match n     (c:a) cs


stripOuterParens :: String -> String
stripOuterParens str = 
    case dropWhile isSpace str of
	('(':cs) -> case dropWhile isSpace (reverse cs) of
	    (')':cs) -> reverse cs
	    _	-> str
	_ -> str


parseNCommand :: String -> TD Command
parseNCommand cs = 
    let str = dropWhile isSpace cs
    in  if isNum str
	   then let (numstr, str') = span isDigit str
		    loc = read numstr
		in  readTSchm (NarrowLoc loc) str'
		    
{-
		in  if isErr pts
		       then return (UnrecCmd "invalid type scheme")
		       else let (_, ts) = unSucc pts
			    in  return (NarrowLoc loc ts)
-}			    
	   else let (cs', rst) = bracketed cs
		    (_, cs2) = span (not . isAlpha) cs'
		    id = takeWhile isAlphaNum cs2
		in  readTSchm (NarrowId id) rst
{-		       
		       pts = parseTyp rst
		   if isErr pts
		      then return (UnrecCmd "invalid type scheme")
		      else let (_, ts) = unSucc pts
			   in  return (NarrowId id ts)
-}
    where -- invokes parser to read a type scheme
	  readTSchm f str = do
	    n <- next
	    let str' = stripOuterParens str
		pts  = parseTSchm (fromIntegral n) str'
	    if isErr pts 
	       then return (UnrecCmd "invalid typescheme")
	       else do let (_, _, ix, ts) = unSucc pts
		           n' = fromIntegral (ixNum ix)
		       setCounter (n'+1)
		       return (f ts)


parseFCommand :: String -> TD Command
parseFCommand cs = case read2Str cs of
			Nothing	    -> return (UnrecCmd "")
			Just (l, f) -> if not (isNum l)
					  then return (UnrecCmd "")
					  else let f' = readFlag f
						   l' = unkIx (read l)
					       in  return (SetAnnFlag l' f')



----------------------------------------
-- command parsing helper functions

isEmpty :: String -> Bool
isEmpty str = all isSpace str

-- true if the string begins with a digit
isNum :: String -> Bool
isNum ""    = False
isNum (x:_) = isDigit x

readFlag :: String -> Flag
readFlag ('p':_) = Pass
readFlag ('f':_) = Fail
readFlag _	 = Unknown

-- returns the first two non-whitespace tokens in the string, or Nothing if
-- they don't exist
read2Str :: String -> Maybe (String, String)
read2Str cs = let str = dropSpace cs
	      in  if null str
		    then Nothing
		    else let (str1, str') = span (not.isSpace) str
			     str'' = dropSpace str'
			 in  if null str'' 
				then Nothing
				else let str2  = takeWhile (not.isSpace) str''
				     in  Just (str1, str2)
		  

-- returns the first non-whitespace token from the input string
readStr :: String -> String
readStr cs = let str  = dropWhile isSpace cs
		 str' = takeWhile (not.isSpace) str
	     in  str'

dropSpace = dropWhile isSpace


--------------------------------------------------------------------------------
-- command dispatch
--
-- NOTE: the bigger definitions (eg. InferTypeLoc) should be separated from
-- doCommand - which should be a fairly minimal, "switch" function

doCommand :: Command -> TD ()
doCommand NoCmd	      = return ()
doCommand (UnrecCmd "") = puts "Unrecognised command. Type :h for help\n"
doCommand (UnrecCmd s)  = puts ("Unrecognised command: " ++ s)

doCommand Help	= printHelp
doCommand Quit	= do puts "bye!"
 		     doIO (exitWith ExitSuccess)

doCommand (SetJustProp jp) = setJProp jp >> puts "done"

doCommand (SetSOpt s)	   = setSolvOpt s >> puts "done"

doCommand (PrintCHR id)	   = do chrs <- getCHRs
				let chr = filter (chrHasId id) chrs
				puts (showCHRs chr)

doCommand (PrintAnnCHR id) =  do chrs <- getCHRs
                                 let chr = filter (chrHasId id) chrs
                                 puts (showAnnCHRs chr)
			 
doCommand PrintAllCHRs = do chrs <- getCHRs
			    schrs <- getSCHRs
			    let chrs' = map presentableCHR chrs
			    puts (showCHRs chrs')
--			    puts (showCHRs schrs)

doCommand PrintAnnAllCHRs = do chrs <- getCHRs
			       puts (showAnnCHRs chrs)


doCommand PrintProg = do prog <- getProg
                         puts (show prog)

doCommand PrintAnnProg = do prog <- getProg
			    puts (pprintAnn prog)

doCommand (PrintHLProg ls) = do prog <- getProg
			        puts -- (pprintLocs ls prog)
				     (printProgHL ls prog)
				     -- (pprintLocs ls prog)
				     -- (printProgLoc (head ls) prog)

doCommand PrintEnv	= do env <- getEnv
			     puts (show env)

doCommand PrintResults	= do rs  <- getResults
			     let rs' = map (\r -> show r ++ "\n") rs
			     let rs''= unlines (reverse rs')
			     doIO (putStr rs'')

doCommand (PrintIxCons ix) = do chrs <- getCHRs
			        let prims  = chrsToPrims chrs
			        let prims' = getIxPrimList prims (ixToJust ix)
				let (ucs, bcs) = prims'
				puts (showAnnList ucs ++ showAnnList bcs)
				

-- FIXME: get this to work again (involves messing with getProg and module
-- namespaces)
doCommand (SetAnnFlag ix f) = do return () {-
				 prog <- getProg
				 let prog' = changeFlag prog ix f 
				 setProg prog' -}


doCommand (Deriv id)	 = do der <- doDeriv id
			      case der of
				Nothing	-> puts "something went bad"
				Just st	-> let c = storeToCons st
					   in  puts (show c)

doCommand (AnnDeriv id)	 = do der <- doDeriv id
			      case der of
				Nothing	-> puts "something went bad"
				Just st	-> let c = storeToCons st
					   in  puts (pprintAnn c)

doCommand (InferTypeId id) = inferTypeId id


doCommand (DerivLoc l b) = do mderiv <- doDerivLoc l b
			      if isNothing mderiv
			       then puts "unexpected problem"
			       else do let (_,c) = fromJust mderiv
				       puts (show c)

doCommand (AnnDerivLoc l b) = do mderiv <- doDerivLoc l b
				 if isNothing mderiv
				    then puts "unexpected problem"
				    else do let (_,c) = fromJust mderiv
					    puts (pprintAnn c)

doCommand (InferTypeLoc l (b,r,e) jp)	= do inferTypeLoc l (False,r,e) jp
					     reportLastResult
					     -- when (b) reportLastConflict

doCommand (InferTypeExp exp)		= inferTypeExp exp

doCommand (ExplainLoc b co l1 l2)	= explainLoc b co l1 l2

doCommand (ExplainLocType l1 l2 ts)	= explainLocType l1 l2 ts >> return ()

doCommand (ExplainExpType exp ts)	= explainExpType exp ts

doCommand (ExplainRefType ref ts)	= explainRefType ref ts

doCommand (NarrowLoc l ts)		= narrowLoc l ts

doCommand (NarrowId id ts)		= narrowId id ts

doCommand (Evaluate exp)		= evaluateExpression exp

{-
doCommand Reload			= do fs <- getOptionFlags
					     fn <- getFilename
					     doIO $ start (fs, [fn])

doCommand (Load fn)			= do fs <- getOptionFlags
					     doIO $ start (fs, [fn])
-}
doCommand (DDebug ml)			= -- ddebugger ml
					  -- return ()
					  -- prologCons
					
					  do p <- getProg
					     puts (show (ltProg p))
					
--					  do chrs <- getCHRs
--					     let chrs' = map Trans.simplifyCHR chrs
--					     puts (showlist chrs')
					

{-
				          do x <- getEvidenceCons (fromJust ml)
					     puts (showAnn x) 
-}					
{-					
					  do j <- getJProp
					     puts (show j) -}
{-					     
					  -- do c <- classifyIds
					  --   puts (show c)
					  do -- p <- getProg
					     -- puts (show (transAnnDecs p))
					     evidenceTranslation
-}					     -- ddebugger ml

doCommand (TDebug l)   = do okay <- locExists l
			    if not okay
				then puts ("location " ++ show l ++ 
					   " can't be queried\n")
				else do traceDebug l 

doCommand (InferTypeRef vr) = inferTypeRef vr

--------------------------------------------------------------------------------

evaluateExpression :: Ann Exp -> TD ()
evaluateExpression aexp = do
    inferTypeExp aexp
    (res:_) <- getResults
    if not (isResult res)
       then return ()
       else do
	prog <- getProg
	n    <- next
	let sprog = strip prog
	    exp   = stExp (item aexp)
	    file  = "/tmp/chameleon-temp-" ++ show n
	    progTxt = -- "import qualified Prelude as P\n\n" ++
		       show sprog ++ 
		       "\n\nmain = putStrLn (show (" ++ show exp ++"))"
	h <- doIO $ openFile file WriteMode
	doIO $ hPutStr h progTxt
	doIO $ hClose h
	doIO $ system ("runhugs " ++ file)
	return ()


--------------------------------------------------------------------------------

-- print out the constraints in a form appropriate for prolog
prologCons :: TD ()
prologCons = do
    let pred = "test"
	var  = Var (name "T" (-2), (-2))
    prog <- getProg
    chrs <- getCHRs

    chrs' <- relocCHRs chrs
    
    eqs <- unfoldCHR chrs' pred var
    let ps = comment $ pprintAnn prog
--	cs = pred ++ "(" ++ show var ++ "):- " ++ 
--		(unwordsc $ map showEq eqs) ++ "."
--		
--	vs = "show(M):- M = [" ++ 
-- --		unwordsc (map showMap (nub $ concatMap freeVars eqs)) 
--		unwordsc (procShow eqs)
--		++ "]."

--    puts (ps ++ "\n\n" ++ cs ++ "\n\n" ++ vs)
    let cs = pred ++ "([" ++ ({-unwordsc-} format $ map showEq eqs) ++ "],\n" ++
	     "[" ++ {-unwordsc-} format (procShow eqs) ++ "])."
    puts (ps ++ "\n\n" ++ cs)  

    where comment = unlines . (map ("% "++)) . lines

	  format [] = ""
	  format [x] = x
	  format (x:xs) = x ++ ",\n" ++ format xs

	  procShow eqs =
	    let vs  = nub $ fv eqs
		vs' = map trans vs
		svs = map (\(x,i) -> "T" ++ show x ++ "_" ++ show i ) vs
	    in  [ "p("++o++",'"++n++"')" | (o,n) <-  zip svs vs' ]
	    where trans (v,l)
		    | l < 100   = "T" ++ show l
		    | otherwise = "nonlocal"

	  showEq (Eq t1 t2 j)
	    | j == noJust = "dontcare(" ++ showType t1 ++ " = " ++
					    showType t2 ++ ")"
	    | otherwise   = showType t1 ++ " = " ++ showType t2
	  
	  
	  lower (x:xs) = toLower x : xs
	  lower "" = ""

	  -- find all variables with don't care locations; and buld a
	  -- substitution to replace them with variables that have other
	  -- locations not actually present in the program
	  -- this way at least the variable names will be of the same format
	  -- (instead of the '-1' location being used)
	  relocCHRs :: [CHR] -> TD [CHR]
	  relocCHRs chrs = do
	    let vs  = nub $ concatMap fv chrs
	    ts <- mapM rep vs
	    let s = zip vs ts
	    return (map (apply s) chrs)
	    
	    where rep vl@(v,l)
		    | l == noLoc = do l' <- next
				      return (Var (v, (fromIntegral l')))
		    | otherwise  = return (Var vl)

	  ren x = do let vs = fv x
		     ts <- mapM rep vs
		     let s = zip vs ts
		     return (apply s x)
	    where rep (_,l) = do v <- TD_Monad.newTVar
				 return (Var (v, l))
		    

	  -- given the name of a chr (lhs uc) fully unfolds its rhs
	  -- assume no ucons in result (will crash if not the case)
	  unfoldCHR :: [CHR] -> String -> Type -> TD [BCons]
	  unfoldCHR chrs id t1 = do
	    let chrs' = dropWhile (\c -> id /= simpNameStr c) chrs
	    let ~(chr':_) = chrs'
	    when (null chrs') (error ("couldn't find: " ++ id))
	    
	    chr <- ren chr'
	    
	    let (ucs, bcs) = chrRhs chr
		ts = ucTypes $ simpUCons chr
		t = head ts
		
		trip [_,_,_] = True
		trip _ = False
		bcs' = if trip ts
			  then filter (\eq -> let lv = ts !! 1
						  tv = ts !! 2
						  l = bcLeft eq
						  r = bcRight eq
					      in  l /= lv && l /= tv &&
						  r /= lv && r /= tv) bcs
			  else bcs
		ucIds = map ucNameStr ucs
		ucTyps = map (head . ucTypes) ucs
		eq = Eq t1 t noJust
	    bcss <- mapM (\(i,t) -> unfoldCHR chrs i t) (zip ucIds ucTyps)
	    return (eq : bcs' ++ concat bcss)


	  unwordsc [] = ""
	  unwordsc [x] = x
	  unwordsc (x:xs) = x ++ "," ++ unwordsc xs

	  showType :: Type -> String
	  showType t = showsType 0 t ""

	  showsType :: Int -> Type -> String -> String
	  showsType n (Var (x,i)) ss = "T" ++ show x ++ "_" ++ show i ++ ss -- (show (x)) ++ ss
	  showsType n (TyCon x)   ss = lower (show x) ++ ss
          showsType n (TApp t@(TApp (TyCon name) t1) t2) ss
	    | name == arrowName =
		if n > 0
		then '(' : showsType 1 t1 (" --> " ++ showsType 0 t2 (')' : ss))
		else showsType 1 t1 (" --> " ++ showsType 0 t2 ss)
--		"a(" ++ showsType 0 t1 "" ++ "," ++ showsType 0 t2 "" ++ ")"
	    
	    | name == (tupleName 2) = 
		("(" ++ (showType t1) ++ ", " ++ (showType t2) ++ ")") ++ ss
	    | otherwise =
		if n > 1
		 then '(' : showsType 1 t (" " ++ showsType 2 t2 (')' : ss))
		 else showsType 1 t (" " ++ showsType 2 t2 ss)
	  showsType n (TApp t1 t2) ss =
	    let non_list =
	         if n > 1
		  then '(' : showsType 1 t1 (" " ++ showsType 2 t2 (')' : ss))
		  else showsType 1 t1 (" " ++ showsType 2 t2 ss)
	    in case t1 of
	     (TyCon name)   ->	if name == nilName 
				    then ("[" ++ (showType t2) ++ "]") ++ ss
--				    then ("l(" ++ (show t2) ++ ")") ++ ss
				    else non_list
	     _		    ->  non_list

	  

{-
	  showType (Var (_, l)) = "T" ++ show l
	  showType (TyCon nm) 
	    | nm == arrowName = "a"
	    | nm == tupleName 2 = "p"
	    | otherwise = lower (show nm)
	  showType (TApp t1 (TApp t2 t3)) = showType t2 ++ "(" ++ showType t3 ++ "," ++ 
							      showType t1 ++ ")"
	  showType (TApp t1 t2) = showType t1 ++ "(" ++ showType t2 ++ ")"
-}

-------------------------------------------------------------------------------


----------------------------------------
-- version of inferTypeLoc, which returns the strongest type for l, implied by
-- all the maximal satisfiable subsets of the resulting store
--
-- if b is True the function may write to stdout (else it won't)
-- if r is True the result will be stored in the state
-- if e if True then only write error reports to stdout!
inferTypeLoc2 :: Loc -> (Bool, Bool, Bool) -> JustProp -> TD ()
inferTypeLoc2 l (b, r, e) _jp = 
    do prog <- getProg
       chrs <- getCHRs
       jp   <- getJProp
       let pos  = when b
           posr = when r
--           pose = when (not e)
           tcs  = chrWithVarIx l chrs
       case tcs of
	[]	     -> do pos (puts ("* location " ++ show l ++ " not found"))
			   posr (newResult NoResult)

	((tv,chr,rest):_) -> 
	    do -- do the derivation (we know the ixs of interest)
	       n <- next
	       let rhs = chrRhs chr
--	       mst <- tdDeriver LocalJust chrs rhs
	       st <- unfoldCHR (rhsToCons rhs)
	       if False -- isNothing mst
		then puts "derivation failed ... somehow"
		else let -- st  = storeToCons (fromJust mst)
			 bcs = consToBs st
			 ucs = consToUs st
			 tv  = Var (head (constraintWithVarIx l st))
			 ephi = unify [] idSubst (buildUP bcs)
		     in  case ephi of
			  Succ phi -> do -- satisfiable
				let phi = unSucc ephi
				    t = apply phi tv
				puts (show l ++ ") " ++ show t)
			  Err _	-> do -- unsatisfiable
				let maxs = maxSatisfiable bcs
				    ts = map (`solveFor`tv) maxs
				puts "type error"
				puts ("constraint: " ++ show st)
				puts ("type var: " ++ show tv)
				puts ("maximal satisfiable subsets:\n" ++
				      unlines (map show maxs))
				puts ("types:\n" ++
				      unlines (map show ts))

				t <- antiUnifier ts
				puts ("anti-unifier: " ++ show t)

			    where solveFor bcs t = 
				    let phi = unSucc $ 
					      unify [] idSubst (buildUP bcs)
				    in  apply phi t

--------------------

-- unfolds the given CHR as far as possible, does not detect cycles!
-- (essentially performs a derivation with local justifications)
-- first unfold UCons alone, then reintroduce all BCons and rerun
unfoldCHR :: Constraint -> TD Constraint 
unfoldCHR (C ucs bcs) = do
	chrs <- getCHRs
	st <- tdDeriver LocalJust chrs (consToGoal (C ucs []))
	let C ucs' bcs' = storeToCons (fromJust st)
	st' <- tdDeriver LocalJust chrs (consToGoal (C ucs' (bcs' ++ bcs)))
	return (storeToCons (fromJust st'))
		 			 

----------------------------------------

-- construct a new ValDec, add it to the program temporarily and
-- do an inferTypeLoc for it
-- (if the expression is just an identifier, look it up and do an inferTypeLoc)
inferTypeExp :: Ann Exp -> TD ()
inferTypeExp exp = do
    jp <- getJProp
    case item exp of
     (VarExp id) -> do -- lookup the corresponding def, infer its type
	(Prog decs) <- getProg
	chrs <- getCHRs
	let lis  = valDecsToLocIds decs
	    locs = [ l | (l, id') <- lis, id == id' ]
	    loc  = head locs
	if null locs
	   then -- puts ("`" ++ id ++ "' undefined!")
		-- there's no CHR reprentation of the type of id
		-- try looking directly in the type environment for it
		-- strip any herbrand constraints from the context
		do mtc <- lookupPred id
		   let (t,c@(C ucs _)) = presentableVars (fromJust mtc)
		       c' = C ucs []
		   unless (isNothing mtc) $
			puts ((if c == trueCons 
			       then ""
			       else "(" ++ show c' ++ ") =>") ++ show t)
	   else	if isOverloadedId decs id
		   then inferTypeId (idToNestedRef id)
		   else do inferTypeLoc loc (False, True, False) jp
			   reportLastResult

     _		 -> do -- convert to a def, then a chr, then infer type
        n1 <- next
	n2 <- next
	n3 <- next
	n4 <- next
	let id   = "(" ++ show n3 ++ ")"
	    def  = Def (Ann (unkIx (fromIntegral n1)) id) [] (RHS1 exp) []
	    adef = Ann (unkIx (fromIntegral n2)) def
	    dec  = ValDec [adef]
	    adec = Ann (unkIx (fromIntegral n3)) dec
	prog <- getProg
        chrs <- getCHRs
	env  <- getEnv

	let Prog ds = prog
            prog' = Prog (adec:ds)
	res <- doTR $ Trans.transDef env adef
	case res of
	 Err e -> do newResult NoResult
		     let e' = tail (dropWhile (/=':') e)
		     puts e'
	 Succ chr -> do
	    let chrs' = chr : chrs
    
	    setProg prog'
	    setCHRs chrs'
    
	    inferTypeLoc (ixNum (ann exp)) (False, True, False) jp
	    reportLastResult
				
	    setProg prog
	    setCHRs chrs

{-
reportLastResult = do 
    (r:_) <- getResults
    if isConflict r
     then do let (Conflict (sp,_,cs,c)) = r
	     doIO $ putStr "Type error: " 
	     reportConflicts True c cs
     else doIO $ putStr (show r)
-}



----------------------------------------
 -- given an id name, infers its type 
inferTypeId :: NestedRef -> TD () -- Id -> TD ()
inferTypeId [(id, Nothing)] = do 
    st <- inferFunType id
    case st of
	Nothing  -> puts "ill-typed"
	Just st' -> do let c    = storeToCons st'
			   disp = id ++ " :: " ++ formatRes c
		       puts disp
inferTypeId _ = puts "support for more advanced references not implemented yet"


----------------------------------------
					       
-- consider the type of l2 and sub-locations (including l1)
-- explain the type for l1
-- an "explanation" provides the following information:
--  a) a justification for each constraint appearing in the inferred context
--  b) a justification for each sub-type of the inferred type. ie. why each part
--     of the type was inferred as it was
--
-- we fulfill these requirements as follows:
--  a) simply get the justification we already have assigned to each constraint
--  b) we generate the list of all sub-types*, and for each we ask which
--     minimal subset of all the constraints entails that type at the location
--     (in this sense, a "sub-type" is a generalisation of a type)

-- if b is True then explain the type, else just the context
explainLoc b co l1 l2 = 
    do doCommand (InferTypeLoc l2 (False, True, False) LocalJust)
       res <- lastResult
       if noResult res || isConflict res
	then do doIO (putStr ("could not infer a type for " ++ show l2))
		if noResult res 
		    then puts " (it doesn't exist!)"
		    else puts " (ill-typed)"
	else
	     do	let ((lt1,_,_), k_lts, nk_lts, (cons, s_cons), _) = unResult res
		-- the variable representing the type of interest
		let lts = lt1 : k_lts ++ nk_lts
		let mtv = lookup l1 lts
--		puts (showAnn cons)
		if isNothing mtv 
		    then puts ("could not find " ++ show l1 ++ 
			       " within " ++ show l2 ++ "\n")
		    else do prog <- getProg
			    let tv = fromJust mtv
			    let typ = bcRight (heady "j" 
					    (withVar tv (cToBCons s_cons)))
			    -- first explain the type (using bcons)
			    texps <- expType tv typ cons
			    let disp1 = unlines (map (formType prog) texps)

			    -- now explain the context
			    let s_ucs = cToUCons s_cons
			    let pucs  = zip (cToUCons cons) s_ucs
			    exps <- mapM (flip expCons cons) pucs
			    let cexps = zip s_ucs exps
			    let disp2 = unlines (map (formCtxt prog) cexps)
			    
			    if null texps
			     then when b (puts "inferred type not an instance")
			     else when b ((when (not co) (puts disp1)) >> 
					   puts disp2)

    where
	formType p (t, e) = let e' = map filterLocs e 
			    in  "type of form: " ++ show t ++ "\ndue to: " ++ 
				show e' ++ "\n" 
				++ ieUnlines (map (shl p) e) 
	ieUnlines = foldr (\xs ys -> xs ++ "\n" ++ ys) ""

	formCtxt p (c, (j, tes)) = 
	    let repNL "" = ""
		repNL ('\n':cs) = "\n              " ++ repNL cs
		repNL (c:cs)    = c : repNL cs
		ft (t, e) = "       where: a = " ++ show t ++ 
			    "\t\tdue to " ++ show e ++ "\n" ++ 
			    "      source: " ++ repNL (unwords $ map (shl p) e)
		ts = map ft tes
		ay = Var (name "a" 0, noLoc)
		c' = newUCTypes [ay] c
		
	    in  "  constraint: " ++ show c' ++ "\t\tdue to " ++ show j ++ 
		"\n" ++ unlines ts

	-- gets rid of locations we don't care about (like negative ones)
	filterLocs :: [Loc] -> [Loc]
	filterLocs ls = filter (>0) ls

	-- shows the contributing locations highlighted
	shl :: Prog -> [Loc] -> String
	shl p ls = let ls' = removeAppLocs p ls
		       l   = foldr max 0 ls
		   in  showHLLoc ls' l p
    
	-- takes a pair of Ucons (unsimplified, simplified) and all the other
	-- constraints, and returns a pair containing a justifiation for the
	-- existence of the user constraint, as well as an explanation for the
	-- type being predicated
	expCons :: (UCons,UCons)->Constraint->TD (Just, [(Type, [[Loc]])])
	expCons (uc, s_uc) cons = 
		do let js = ucIx uc
		   let tv  = unVar (heady "k" (ucTypes uc))
		   let typ = heady "l" (ucTypes s_uc)
--		   gens  <- genType typ
		   texps <- expType tv typ cons
--		   return (js, []) -- texps)
		   return (js, texps)		   
	
	expType :: TVar -> Type -> Constraint -> TD [(Type, [[Loc]])]
	expType tv typ cons = 
		do gens <- genType typ
		   let eqs  = map (\t -> Eq (Var tv) t noJust) gens
		       ents = map (\bc -> entailSubsetExp cons bc) eqs
		       exps = zip gens ents
		   return exps
    
	-- given a type, returns a list of generalisations of this type. ie. 
	-- types which can be instantiated such that they become the original
	genType t = do gen <- genType' t False -- True -- False
		       return (procGen gen)
	genType' t@(Var _)   _ = return [t]
	genType' t@(TyCon _) l = if l then return [t]
					else do tv0 <- newVar
						return [t, tv0]
	genType' t0@(TApp t1 t2) l = 
		    do tv0 <- newVar
		       g1  <- genType' t1 True
		       g2  <- genType' t2 False
		       let g0 = [TApp x y | x <- g1, y <- g2]
		       if l then return (nub (t0 : g0))
			    else return (nub (tv0 : t0 : g0))

	procGen []   = []
	procGen [ts] = [ts]
	procGen ts   = tail ts				       

----------------------------------------	    

-- takes a string with the name of a function identifier and returns the
-- location of its definition in the program
idToLoc :: String -> TD (Maybe Loc)
idToLoc id = do
	(Prog decs) <- getProg
	chrs <- getCHRs
	let lis  = valDecsToLocIds decs
	    locs = [ l | (l, id') <- lis, id == id' ]
	    loc  = head locs
	if null locs
	   then return Nothing 
	   else return (Just loc)

narrowId :: String -> Ann TSchm -> TD ()
narrowId id ts = do
    ml <- idToLoc id
    case ml of
	 Nothing -> puts ("`" ++ id ++ "' not found")
	 Just l  -> narrowLoc l ts
   

-- first find the CHR and var. using the loc.
-- then add the constraint, var = t to that CHR
-- we generate and report a fresh justification
narrowLoc :: Loc -> Ann TSchm -> TD ()
narrowLoc l ts = 
    do prog <- getProg
       chrs <- getCHRs
       let ix  = unkIx l
           tcs = chrWithVarIx (ixLoc ix) chrs
       case tcs of
	[]	     -> do puts ("** location " ++ show l ++ " not found")

	((tv,chr,rest):_) -> do
		env <- getEnv
		n   <- next
		l'  <- next
		let l = negate l'
		let jst = locToJust l
		tres <- doTR (Trans.transTSchm env ts)
		case tres of
		     Err e   -> puts e
		     Succ tc -> do
			let (t,c) = tc
			    bc = Eq (Var tv) t jst	-- noJust?
			    c' = bToCons bc `cand` c
			    chr'  = chrAddCons chr c'
			    chrs' = chr' : rest
			setCHRs chrs'
			puts ("new ann has location: " ++ show l ++ "\n") 
		   
----------------------------------------
		   

--------------------------------------------------------------------------------



-- takes a constraint which represents a type, and formats it as such
formatRes (C ucs [bc]) = 
	let vars = nub (fv ucs ++ fv (bcRight bc))
	    phi  = simpTVarSubst vars
	    ctxt = showCommaList (apply phi ucs) ++ 
		   if ucs /= [] then " => " else ""
	    typ  = show (apply phi (bcRight bc))
	in  ctxt ++ typ
formatRes (C [uc] []) = 
	let vars = fv uc
	    phi  = simpTVarSubst vars
	    ctxt = show (apply phi uc) ++ " => "
	    typ  = show (apply phi (head (ucTypes uc)))
	in  ctxt ++ typ
formatRes c = error ("TypeDebug.formatRes: constraint" ++ show c)

showAnnUConsList []  = ""
showAnnUConsList [uc] = pprintAnn uc
showAnnUConsList (uc:ucs) = pprintAnn uc ++ ", " ++ showAnnUConsList ucs


eqRight t [] = error "TypeDebug.eqRight: didn't find tv - impossible!"
eqRight t ((Eq t1 t2 _):cs)
    | (Var t) == t1	= t2
    | otherwise		= eqRight t cs
eqRight _ _ = error "TypeDebug.eqRight"




showCHRs chrs = (replicate 80 '-') ++ "\n" ++ (showCHRs' chrs)
    where showCHRs' []         = ""
	  showCHRs' (chr:chrs) = (show chr) ++ "\n" ++ (showCHRs' chrs)

showAnnCHRs chrs = (replicate 80 '-') ++ "\n" ++ (showCHRs' chrs)
    where showCHRs' []         = ""
	  showCHRs' (chr:chrs) = (pprintAnn chr) ++ "\n" ++ (showCHRs' chrs)

----------------------------------------

printHelp =
    let (l, r) = unzip help
	lw  = foldl max 0 (map length l)
	lw' = (8 - (lw `mod` 8)) + lw
	l'  = map (pad lw') l
	help'  = [ l1 ++ r1 | (l1, r1) <- zip l' r ]
	msg = "\n" ++ rule ++ unlines help' ++ "\n" ++ general ++ "\n" ++ 
	      rule ++ "\n"
    in	doIO (putStr msg)
	
    where 
	  rule = replicate 80 '-' ++ "\n"
	  pad  n s = let gap = n - length s
		     in  s ++ replicate gap ' '
	  help = [("print, p", "print program"),
		  ("chr, c [id]", "print the named CHR, or all CHRs"),
--		  ("e", "print environment (from translator)"),
		  ("deriv, d (id | loc)", 
		    "show derivation for given id or location"),
		  ("type, t (exp | loc)", "infer the type of an id, or expression"),

		  ("explain, x (exp | loc) tschm", 
		    "explain why exp/loc type is of form tschm"),


--		  ("s loc (p | f)", "set the flag for a location"),
		  ("set, s (global | local)", "set just. policy"),


--		  ("x loc1 loc2", "explain type of loc1, within loc2"),
--		  ("x loc", "same as `x loc loc'"),
--		  ("x loc  typ","explain why loc's type is an instance of typ"),
--		  ("xc loc", "explain constraint in type of loc"),
		  
		  ("declare, n loc type", "constrain type of loc (permanent)"),
--		  ("",""),
--		  ("dd [loc]", "run the declarative debugger"),
--		  ("td loc", "run the trace debugger"),
		  ("",""),
		  ("help, h", "display help"),
		  ("quit, q",	"quit")]

	  general = "Note: ?a version of p,c,d commands displays the result" ++
		    " with annotations."
	    

--------------------------------------------------------------------------------
-- The declarative debugger
--
-- see `td_dd.txt' for details

-- top-level declarative debugger function
--  - prints the annotated program
--  - prompts for a starting location
--  - enters the main loop
ddebugger :: Maybe Loc -> TD ()
ddebugger ml = do l <- if isNothing ml
			then do lineBreak
				doCommand PrintAnnProg
				puts ""
				readLoc
			else return (fromJust ml)
		  dd_loop l
		  return ()


-- returns True if the location is okay
dd_loop :: Loc -> TD Bool
dd_loop l = 
    do prog <- getProg 
       let flag = getFlag l prog
       failed <- if isUnknown flag
		    then do doCommand (InferTypeLoc l 
					(False, True, False) LocalJust)
			    lastFailed
		    else return (isFail flag)

       (okay, ml) <- if not failed then dd_result
			           else return (False, Nothing)
       let l' = if isJust ml then fromJust ml else l
       
       if okay then puts ("okay: " ++ show l) >> return True
	       else do puts ("not okay: " ++ show l)
		       subs_okay <- dd_conflict l'
		       if subs_okay 
			  then puts ("error at: " ++ show l) >> return False
			  else puts ("error within: " ++ show l) >> return False
				    

-- returns True if the inferred types at Loc's sublocations are acceptable
-- (also if the location is a leaf in the AST)
dd_conflict :: Loc -> TD Bool
dd_conflict l =
    do 
       
       res <- lastResult
       err_subs <- 
	 if isConflict res 
	    then do let (_,_,(_,lss),_) = unConflict res
		    let lss' = map (delete l) lss
		    puts ("conflicting locations: " ++ show lss')
		    --let ls = commonLocs lss'
		    --puts ("most likely sources of error: " ++ show ls)
		    return (concat lss')
	    else return []
       imm_subs <- dd_descend True l
       --let all_subs = err_subs ++ subs
       let subs = if null err_subs then imm_subs else err_subs
       if null subs 
	 then return True
	 else do puts ("sub-locs: " ++ show subs) 
		 dd_untilFalse subs

-- process given sub-locations, until one is found to be wrong
-- returns True if they're all okay, else False
dd_untilFalse :: [Loc] -> TD Bool
dd_untilFalse []     = return True
dd_untilFalse (l:ls) = do l_okay <- dd_loop l
			  if l_okay 
			    then dd_untilFalse ls
			    else return False 


-- returns True if the inferred type is acceptable
-- if the location belongs to a standard id, then automatically return True
-- user has a chance to respond in one of the following ways:
--   y / n	accept or reject the inferred type(s)
--   x loc	ask for an explanation of a type inferred for a location
--   j loc	jump to the given location after leaving 
--  
-- (note that only a y/n response allows us to leave the node... until one is
--  given we remain at the same spot.)
dd_result :: TD (Bool, Maybe Loc)
dd_result = 
    do res <- lastResult
       let l  = locResult res
       isStd <- locExpStd l
       if isStd 
	then return (True, Nothing)
        else do lineBreak
		puts (show res)
		handleResponse l Nothing

    where   handleResponse l n =
		do inp <- getResponse
		   case inp of
		    Yes	      -> return (True, n)
		    No	      -> return (False, n)
		    Jump x    -> handleResponse l (Just x)
		    Explain x -> doCommand (ExplainLoc True False x l) >>
				    handleResponse l n
		    Back      -> do res <- getResults
				    case res of
					[]	-> handleResponse l n
					(r:rs)	-> do let rs' = dropTil1 rs
						      when (not (null rs'))
							     (setResults rs')
						      res <- lastResult
						      let l = locResult res
						      handleResponse l n
						      
	    dropTil1 [] = []
	    dropTil1 [r] = [r]
	    dropTil1 (r:rs) = dropWhile (not.isResult) rs

-- returns the list of failed locations in list
-- if we don't know anything about a location, descend
trySubLocs :: [Loc] -> TD [Loc]
trySubLocs []	  = return []
trySubLocs (l:ls) = 
    do prog <- getProg
       let flag = getFlag l prog
       failed <- if isUnknown flag
		      then dd_loop l >> lastFailed
		      else return (isFail flag)
       let l' = if failed then [l] else []
       ls' <- trySubLocs ls
       return (l' ++ ls')


lastFailed = do res <- lastResult
		return (isConflict res || isRejected res)
lastSucceeded = do res <- lastResult
		   return (isResult res)

commonLocs lss = let ls  = concat lss
		     ls' = nub ls
		     lfs = zip ls' (map (freq ls) ls')
		     sorted = sort lfs
		     mfreq  = snd (head sorted)
		     maxs   = takeWhile (\lf -> snd lf == mfreq) sorted
		 in  map fst maxs
		 
    where freq ns n = length (filter (n==) ns)
	  sort xs = sortBy (\(_, f1) (_, f2) -> compare f2 f1) xs




getBoolInput = do prompt "Is the intended type an instance? (y/n)"
		  inp <- doIO getInput
		  let inp' = map toUpper inp
		  case inp' of
			('Y':_)	-> return True
			('N':_) -> return False
			_	-> getBoolInput


-- reads user's response at the dd_result prompt
-- user can answer: y/n, x loc, or j loc
-- (no response - eg. just hitting enter - is interpreted as a "yes")
getResponse :: TD DDCommand
getResponse = do prompt "response"
		 inp <- doIO getInput
		 let inp' = dropWhile isSpace inp
		 case map toUpper inp' of
		    ""	     -> return Yes
		    ('Y':_)  -> return Yes
		    ('N':_)  -> return No
		    ('B':_)  -> return Back
		    ('X':cs) -> readLoc Explain cs
		    ('J':cs) -> readLoc Jump cs 
		    _	     -> puts "unrecognised command" >>
				getResponse

    where readLoc f cs = let cs' = dropWhile isSpace cs
		         in  if null cs'
				then getResponse
				else if isDigit (head cs')
					then return (f (read cs'))
					else getResponse
		    
rejectResult = do res <- lastResult 
		  replaceLastResult (Rejected res)

lineBreak = puts ('\n' : replicate 40 '-')

-- returns Nothing for whitespace, else Just the number
readLocEnter =
    do prompt "enter location"
       str <- doIO getInput
       let str' = dropWhile isSpace str
       case str' of
        	""	    -> return Nothing
        	ccs@(c:cs)  -> if isDigit c 
        			    then return (Just (read ccs))
        			    else puts "not a location" >>
        				 readLocEnter
readLoc = 
    do prompt "enter location"
       str <- doIO getInput
       let str' = dropWhile isSpace str
       case str' of
		""	    -> readLoc
		ccs@(c:cs)  -> if isDigit c 
				    then return (read ccs)
				    else puts "not a location" >>
					 readLoc
{-
readOptLoc =
    do prompt "next"
       str <- doIO getInput
       let str' = dropWhile isSpace str
       case str' of
	    ""		-> return Nothing
	    ccs@(c:cs)	-> if isDigit c
				then return $ Just (read ccs)
				else doIO (putStrLn "not a location") >>
				     readOptLoc
-}

--------------------------------------------------------------------------------
-- functions for interfacing with the CHR solver


-- note that the solver must be supplied with an initial state - an Integer
-- (this must be greater than all other values used for uniqueness so far!)
solver :: [CHR] -> Goal -> Integer -> Maybe Store
solver chrs goal init = let slv = td_solve chrs goal
			in  case unTP slv init of
				Err  x	-> Nothing
				Succ (_, res)	-> Just res





deriver :: SOpt -> [CHR] -> Goal -> Integer -> Maybe Store
deriver opt chrs goal init = 
    let der = derive opt LocalJust chrs goal
    in  case unTP der init of
	    Err  _	  -> Nothing
	    Succ (_, res) -> Just res


-- takes a string, fun, generates a goal: fun(x), and solves for x
-- currently a HACK (assumes constraint component is just a uc)
inferFunType :: String -> TD (Maybe Store)
inferFunType id = 
    do rep  <- lookupPred id
       case rep of
	Nothing		-> do puts ("`" ++ id ++ "' undefined\n")
			      return Nothing
	Just (_,  c) -> do chrs <- getCHRs
			   n <- next
		           let uc = head (consToUs c)
			       init = fromIntegral n
			       goal = convGoal (uToCons uc)
			   return (solver chrs goal init)


-- HACK: same problem as above!
doDeriv :: String -> TD (Maybe Store)
doDeriv id = 
     do rep  <- lookupPred id
        case rep of
	 Nothing	   -> do puts ("`" ++ id ++ "' undefined\n")
			         return Nothing
	 Just (_,  c) -> do chrs <- getCHRs
			    let uc = head (consToUs c)
				goal = convGoal (uToCons uc)
			    n <- next
			    opt <- getSolvOpt
			    let init = fromIntegral n
			    return (deriver opt chrs goal init)   



----------------------------------------

-- standard 4-port, box-model trace debugger...
--  call -> .... -> exit
--  fail <- .... <- redo
--
-- (allows user to do a prolog-style traversal of the derivation)
--  o  the AST is traversed, accumulating all the constraints on types as they 
--     are encountered
--  o  on entry to a node (call) we add its constraints to the store
--  o  if the constraints are unsatisfiable, we fail
--  o  if constraints are consistent, we prompt the user to decide on a node to
--     exit into
--  o  if failure was encountered at a child node, we re-enter (redo) and
--     prompt the user again to pick a sub-node... if all sub-nodes have
--     failed, we fail



traceDebug l = do td <- traceDebugger (l, [], trueCons)
	          when (isJust td) (puts ("location " ++ show l ++ " okay"))

-- Loc is the current location
-- list contains Locs already visited (to prevent loops)
-- Constraint is the accumulated constraints of parent nodes visited
-- returns: Nothing   - if the derivation is unsatisfiable when entering l
--	    Just cons - where cons is the accumulated constraint, if sat.
-- Extend this so that whenever we visit a node (on the way in, or out), we
-- display the type as it has been constrained so far.
traceDebugger :: (Loc, [Loc], Constraint) -> TD (Maybe Constraint)
traceDebugger (l, ns, cons) = 
    do mc <- doDerivStep l True
       let c = fromJust mc 
       let cons' = c `cand` cons
       unsat <- derivUnsat cons'
       if unsat 
	then do et <- typeAt l cons
		p  <- getProg
		puts ("type error at: " ++ show l ++ ") " ++ showLoc l p)
		puts ("expected type: " ++ show et)
--		puts ("inferred type: " ++ show it)
		return Nothing

	else do ls  <- subLocs True l
		let ls' = filter (`notElem` ns) ls
		res <- traceSubLocs ls' ns cons'
		return res 
	     
    where traceSubLocs [] _  cons = return (Just cons)
	  traceSubLocs ls ns cons = do l <- pickLoc cons ls
				       r <- traceDebugger (l, l:ns, cons)
				       if isNothing r
					then return Nothing
					else do let ls' = delete l ls
						traceSubLocs ls' ns (fromJust r)
						
	  pickLoc cons [l] = return l
	  pickLoc cons ls  = 
	    do  p <- getProg
		let ch l = do -- t <- typeAt l cons
			      return ("  " ++ show l ++ ") " ++ showLoc l p ) 
--				    ++ " :: " ++ showAnn t)
		chs <- mapM ch ls
		let chs' = unlines chs
		puts ("pick a sub-location:\n" ++ chs')
		ml <- readLocEnter
		let l = if isNothing ml then head ls else fromJust ml
		if l `notElem` ls then pickLoc cons ls
				  else return l

	  typeAt l cons = do v <- newVar
			     n <- next
			     chrs <- getCHRs
			     let tv  = constraintWithVarIx l cons
--			     let bc  = Eq v (T.Var (head tv)) noJust
--			     let goal  = convGoal (bsToCons [bc])
			     let goal  = convGoal cons
			     let store = solverFor chrs goal tv n
			     let res = if isNothing store
					then error "typeAt"
					else consToBs (storeToCons 
							    (fromJust store))
			     let typ = bcRight (head res)
			     return typ 

-- True if the given constraint (representing the result of a derivation) is
-- unsatisfiable
derivUnsat :: Constraint -> TD Bool
derivUnsat cons = do n <- next
		     chrs <- getCHRs
		     let goal = convGoal cons
		     let store = solver chrs goal n
		     return (isNothing store)
		     

-- finds the constraints associated with the location, does not further descend
-- into sub-locations
doDerivStep :: Loc -> Bool -> TD (Maybe Constraint)
doDerivStep l b = 
    do prog <- getProg
       chrs <- getCHRs
       let tcs = chrWithVarIx l chrs
       case tcs of
	    []	-> do when b (puts ("location " ++ show l ++ 
				    " not found"))
		      return Nothing
		   
	    ((tv,chr,rest):_) ->
		   do let chr' = remConsWithoutVarIxs [l] chr
		      -- let chr' = remConsWithoutIxs [jst] chr
		      let goal = chrRhs chr'
		      let cons = goalToCons goal
		      return (Just cons)


-- True if the given location is used somewhere in the translated CHRs
locExists :: Loc -> TD Bool
locExists l =
    do prog <- getProg
       chrs <- getCHRs
       let tcs = chrWithVarIx l chrs
       return (not (null tcs))
   

--------------------------------------------------------------------------------
-- functions for flipping flags

{-
addAnn :: Index -> a -> Ann a
addAnn ix a = Ann ix a

caf = changeAnnFlag

changeFlag :: Prog -> Index -> Flag -> Prog
changeFlag (Prog decs) ix f = 
    let defs  = valDecsToDefs decs
	other = nonValDecs decs
	defs' = changeFlagDefs defs ix f
	decs' = other ++ [ Ann a (ValDec d) | (d, a) <- zip defs' (anns defs) ]
    in  Prog decs'

changeFlagDefs :: [Ann Def] -> Index -> Flag -> [Ann Def]
changeFlagDefs []     _  _ = []
changeFlagDefs (d:ds) ix f
    | (ann d) == ix = (caf d f) : ds
    | otherwise	    = (changeFlagDef d ix f) : (changeFlagDefs ds ix f)

changeFlagDef :: Ann Def -> Index -> Flag -> Ann Def
changeFlagDef def ix f = 
    case item def of
	(Def id pats exp)   -> if ix == ann def
				then caf def f
				else let id'   = changeFlagId id ix f
				         pats' = changeFlagPats pats ix f
					 exp'  = changeFlagExp exp ix f
					 def'  = Def id' pats' exp'
				     in  addAnn (ann def) def'
	
changeFlagId id ix f = case item id of
                       	str	-> if ix == ann id
                       		    then caf id f
                       		    else id

changeFlagExp exp ix f = 
    if ix == ann exp 
     then caf exp f
     else 
	let aa = addAnn (ann exp)
	in  case item exp of
	    AbsExp id e1    -> let id' = changeFlagId id ix f
				   e1' = changeFlagExp e1 ix f
			       in  aa (AbsExp id' e1')

	    AppExp e1 e2    -> let e1' = changeFlagExp e1 ix f
                                   e2' = changeFlagExp e2 ix f
                               in  aa (AppExp e1' e2')

	    IfExp  e1 e2 e3 -> let e1' = changeFlagExp e1 ix f
                                   e2' = changeFlagExp e2 ix f
                                   e3' = changeFlagExp e3 ix f
                               in  aa (IfExp e1' e2' e3')

	    AnnExp e1 t	    -> let e1' = changeFlagExp e1 ix f
			       in  aa (AnnExp e1' t)

	    TupExp  es	    -> let es'= map (\x-> changeFlagExp x ix f) es
			       in  aa (TupExp es')
	    
	    ListExp es	    -> let es'= map (\x-> changeFlagExp x ix f) es
                               in  aa (ListExp es')

	    _		    -> exp

--changeFlagPats = undefined

changeFlagPats pats ix f = map (\p -> changeFlagPat p ix f) pats

changeFlagPat pat ix f =
    if ix == ann pat
     then caf pat f
     else
	let aa = addAnn (ann pat)
	in  case item pat of
	
		ConPat id ps -> let id' = changeFlagId id ix f
				    ps' = changeFlagPats ps ix f
				in  aa (ConPat id' ps')
				
		TupPat  ps   -> let es' = changeFlagPats ps ix f
				in  aa (TupPat es')
				
		ListPat ps   ->	let es' = changeFlagPats ps ix f
				in  aa (ListPat es')

		_	     -> pat

-}
------------------------------------------------------------

locMatches loc x = ixLoc (ann x) == loc
locFlag x = ixFlag (ann x)

-- WARNING: will FAIL if location doesn't exist
getFlag loc prog = case getFlagProg loc prog of
		    []	  -> error ("TypeDebug.getFlag: no flag for"++show loc)
		    (x:_) -> x
		    

-- NOTE: only looks into defs!
getFlagProg loc (Prog decs) =
    let defs = concat (valDecsToDefs decs)
    in  concatMap (getFlagDef loc) defs

getFlagDef loc def =
    if locMatches loc def
     then [locFlag def]
     else case item def of
	    Def id pats rhs w -> getFlagId loc id ++ 
				 concatMap (getFlagPat loc) pats ++
				 getFlagRHS loc rhs

getFlagId loc id = if locMatches loc id then [locFlag id] else []

getFlagRHS loc rhs = concatMap (getFlagExp loc) (rhsExps rhs)

getFlagExp loc exp = 
    if locMatches loc exp 
     then [locFlag exp] 
     else let gfe = getFlagExp loc
	  in  case item exp of
		AbsExp ps e1    -> concatMap (getFlagPat loc) ps ++ gfe e1
		AppExp e1 e2    -> gfe e1 ++ gfe e2
		IfExp e1 e2 e3  -> gfe e1 ++ gfe e2 ++ gfe e3
		AnnExp e1 ts	-> gfe e1 -- ++ getFlagTSchm loc ts
		TupExp es	-> concatMap gfe es
		ListExp es	-> concatMap gfe es
		_		-> []


getFlagTyp loc typ =
    if locMatches loc typ
     then [locFlag typ]
     else let gft = getFlagTyp loc
	  in  case item typ of
		AppTyp t1 t2 -> gft t1 ++ gft t2
		TupTyp ts    -> concatMap gft ts
		ListTyp t    -> gft t
		_	     -> []

getFlagPat loc pat =
    if locMatches loc pat
     then [locFlag pat]
     else let gfp = getFlagPat loc
	  in  case item pat of
		ConPat id ps -> getFlagId loc id ++ concatMap gfp ps
		TupPat  ps   -> concatMap gfp ps
		ListPat ps   ->	concatMap gfp ps
		_	     -> []

------------------------------------------------------------

--------------------------------------------------------------------------------

-- as "solverSt" above
tdDeriver = deriverSt
deriverSt :: JustProp -> [CHR] -> Goal -> TD (Maybe Store)
deriverSt jp chrs goal = do 
    opt <- getSolvOpt
    n <- next
    let der = derive opt jp chrs goal
    case unTP der n of
	Err e		-> trace e $ return Nothing
	Succ (c,res)	-> do setCounter c
			      return (Just res)


