--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------

-- contains code for building type schemes from CHRs

module TypeScheme where

import Core
import Chr
import Constraint
import Misc
import TD_Monad
import SolverIO
import Justify
import TD_StdDefs (tupleType)


import List
import Maybe
import Monad

import IOExts

--------------------------------------------------------------------------------

type TypeScheme = ([TVar], Constraint, Type)
type DebugTypeScheme = ([TVar], (Constraint, Constraint), Type)

-- placeholder
nullTS = error "no null type scheme"

--------------------------------------------------------------------------------

buildTypeSchemeCheap = buildTypeSchemeSubConf True False True
buildTypeSchemeSub = buildTypeSchemeSubProp False
buildTypeSchemeSubCheap cheap = buildTypeSchemeSubConf cheap False
buildTypeSchemeSubProp = buildTypeSchemeSubConf False 


--------------------------------------------------------------------------------

-- given a simp. CHR, builds the type scheme it represents
-- (as per the paper)
-- if `simplify' is True, constraints in the CHR are simplified 
--	(else, they're taken as is)
-- if `dropProp' is True, user constraints arising from propagation rules are
--	dropped from the type scheme
-- if `cheap' is True, use the fastest CHR solver available 
--  (warning: error diagnosis based on these results could be wrong!)
buildTypeSchemeSubConf :: Bool -> Bool -> Bool -> CHR -> TD (Maybe DebugTypeScheme)
buildTypeSchemeSubConf cheap dropProp simplify chr = do
    -- puts ("chr: " ++ show chr)
    let uc = simpUCons chr
	[t,l,v] = ucTypes uc
	rhs = consToGoal (rhsToCons (chrRhs chr))
    -- chr is uc(t,l,s) <==> ...
    chrs <- getCHRs
    mst  <- -- tdSolver chrs (consToGoal (uToCons uc))
	    do opt <- getSolvOpt
	       if cheap then setSolvOpt SInference
			else setSolvOpt SDebug
	       r <- tdDeriver GlobalJust chrs rhs
	       setSolvOpt opt
	       return r

    if isNothing mst 
     then -- type error
	  return Nothing
     else do  -- uc(t,l,s) -->* cons
	      -- build mgu phi of cons
	      -- avec = fv(phi cons) \ ( (fv (phi l)) cup (fv (phi sg)))
	      --        (where sg are global scoped vars.)
	      -- ts = forall avec. phi cons => phi t

	  cons0 <- if simplify 
		  then -- possibly get rid of the Chameleon propagation ucons
		       -- definitely get rid of administrative ucons
		       let c@(C ucs bcs) = storeToCons (fromJust mst)
		       in  if dropProp
			   then do ls <- chameleonPropRuleLocs 
				   let ucs' = [ uc | uc <- ucs, 
					        all (`notElem`ls) (unJust uc) ]
				   return (C ucs' bcs)
			   else return (C ucs bcs)
		  
		  else return (rhsToCons (chrRhs chr))

	  -- always get rid of admin. ucons
	  let cons = dropAdmin cons0

	  let bcs  = consToBs cons
	      -- phi is mgu of cons
	      phi  = unSucc' "HEY!" (unify [] idSubst (buildUP bcs))
	      t' = apply phi t
	      l' = apply phi l
	      v' = apply phi v
	      vl = locals v'
	      vg = globals v'
	      lg = globals l'
	      ll = globals l'
	      cons'  = apply phi cons
	      cons'' = remRedundantCons cons'

	      vgs = [ v | v <- fv vg, v `notElem` fv vl ]

	      vs = -- fv( phi C, phi t) \ fv(phi v', phi l_l)
		   fv (cons', t') `without` (vgs ++ fv ll)
		   
	      -- throw out qualified vars not appearing in type scheme
	      vs' = map Var [ v | v <- vs, v `elem` fv (cons'', t') ]


	      ts = (nub vs', (cons, cons''), tupleType [t', getGlobals v'])
	  erts <- doTP (rename ts)
	  let (vs,cs,ts) = unSucc erts
	      ts' = (fv vs, cs, ts)

	  return (Just ts')


--------------------------------------------------------------------------------

-- builds a type scheme directly from the given constraint
-- variables represent type, lambda-bound and scoped variable components
-- no simplification is performed (other than removal of administrative
-- constraints)
buildTypeSchemeCons :: Constraint -> (TVar, TVar, TVar) -> TypeScheme
buildTypeSchemeCons c@(C ucs bcs) (t,l,v) = 

-- uc(t,l,s) -->* cons
-- build mgu phi of cons
-- avec = fv(phi cons) \ ( (fv (phi l)) cup (fv (phi sg)))
--        (where sg are global scoped vars.)
-- ts = forall avec. phi cons => phi t

    let C ucs' _ = dropAdmin c

        phi  = unSucc' "HEY!" (unify [] idSubst (buildUP bcs))
	t' = apply phi (Var t)
	l' = apply phi (Var l)
	v' = apply phi (Var v)
	vl = locals v'
	vg = globals v'
	lg = globals l'
	ll = globals l'
	cons   = C ucs' []
	cons'  = apply phi cons
	cons'' = remRedundantCons cons'

	vgs = [ v | v <- fv vg, v `notElem` fv vl ]

	vs = fv (cons', t') `without` (vgs ++ fv ll)
		   
	-- throw out qualified vars not appearing in type scheme
	vs' = [ v | v <- vs, v `elem` fv (cons'', t') ]

    in  {- trace ("\n(t,l,v): " ++ show (t,l,v) ++
	       "\nc: " ++ show c ++
	       "\ncons: " ++ show cons ++
	       "\ncons'': " ++ show cons'' ++
	       "\n(vg,vl): " ++ show (vg,vl) ++
	       "\n(lg,ll): " ++ show (lg,ll) ++
	       "\nvs': " ++ show vs' ++
	       "\nnub vs': " ++ show (nubTVars vs')) $ -}
	(nubTVars vs', cons'', t')



--------------------------------------------------------------------------------

  -- given a simp. CHR, builds the type scheme it represents
  -- the first constraint is unsimplified
buildDebugTypeScheme :: CHR -> TD (Maybe DebugTypeScheme)
buildDebugTypeScheme chr = do
    let uc = simpUCons chr
	[t,l,s] = ucTypes uc
	rhs = consToGoal (rhsToCons (chrRhs chr))
    -- chr is uc(t,l,s) <==> ...
    chrs <- getCHRs
    mst  <- do opt <- getSolvOpt
	       setSolvOpt SDebug
	       r <- tdDeriver GlobalJust chrs rhs
	       setSolvOpt opt
	       return r

    if isNothing mst 
     then -- type error
	  return Nothing
     else     -- uc(t,l,s) -->* cons
	      -- build mgu phi of cons
	      -- avec = fv(phi cons) \ ( (fv (phi l)) cup (fv (phi sg)))
	      --        (where sg are global scoped vars.)
	      -- ts = forall avec. phi cons => phi t

	  let cons = storeToCons (fromJust mst)
	      bcs  = consToBs cons
	      -- phi is mgu of cons
	      phi  = unSucc (unify [] idSubst (buildUP bcs))
	      t' = apply phi t
	      l' = apply phi l
	      s' = apply phi s
	      sg = globals s'
	      sl = locals s'

	      sg_vs = fv sg `without` fv sl

	      cons' = -- remRedundantCons (apply phi cons)
		      (apply phi cons)
	      cons'' = remRedundantCons cons'
	      vs = fv cons' `without` (fv l'  ++ 
				      -- fv sg ++
				      sg_vs ++
				      fv [l, s])
	    
	      -- throw out qualified vars not appearing in type scheme
	      vs' = nub [ v | v <- vs, v `elem` fv (cons'', t') ]

	  in  return (Just (vs', (dropAdmin cons, dropAdmin cons''), t'))

dropAdmin (C ucs bcs) = 
    let isAdmin uc = case unJust (ucJust uc) of
	    [] -> False
	    ms@(m:_) -> any (==adminLoc) ms
	ucs' = filter (not.isAdmin) ucs
    in  C ucs' bcs



-- projects out the global/local component of the scoped type 
-- variables
getGlobals (TApp (TApp _ vg) _vl) = vg
globals (TApp (TApp _ vg) _vl) = map Var (fv vg)
globals _ = []
locals (TApp (TApp _ _vg) vl) = map Var (fv vl)
locals _ = []

globalParam (TApp (TApp _ sg) _sl) = Just sg
globalParam _ = Nothing


--------------------------------------------------------------------------------

tdDeriver jp chrs goal = do 
    opt <- getSolvOpt
    der <- derive opt jp chrs goal
    return $ case der of
	Sat   s -> Just s
	Unsat s -> Nothing

