--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- version of the Types module, with annotated type variables

module Types(TVar, TCon, Type(Var,TyCon,TApp), 
	     arrowName, tupleName, listName, nilName, consName, undefinedName,
	     boolName, trueName, falseName, catName, vnilName, vconsName,
	     Subst, idSubst, TP, unTP, TC, 
             apply, fv, occurs, rename, format, 
	     renameSubst, renameSubst2, lookupVar, updateSubst, 
             newTVar, newTCon, newInteger, errorTP, errorE,
             newTVar, newTCon, newInteger, unify, unifyTypes, matchTypes, 
	     matchTypes2, matchTypes3,
             errorTP, errorE,
	     isTyCon, isTApp, 
	     varLoc, typeLocs, locToJust,
	     unVar, isVar, JustProp, isOrigJust,
	     unJust, unifyWithIds,
	     simpTVarSubst, presentableVars, tupleArity,
	     isTupleCon, tupleTypes, vars,
	     showTypeLoc, nubTVars
             ) where


import Name
import List
import Maybe
import Monad
import Justify
import Misc
import Ann_InputAST (noLoc, Loc, stdMultiTypeNames, multiTypeName)
import Pretty
import qualified Terminal as Term
 
import Char
 
-------------------------------
-- Type and data definitions --
-------------------------------


-- Type language:
-- Variables and type constructors are represented by Int's

type TVar = (Name, Loc)
type TCon = Name
data Type = Var TVar | TyCon TCon | TApp Type Type -- deriving (Eq)


unVar (Var tv) = tv
unVar _	       = error "TD_Types.unVar: not a var"
isVar (Var tv) = True
isVar _ = False

vars = map Var

-- ignores location
nubTVars = fv . nub . vars

instance Eq Type where
    (Var (n1,_)) == (Var (n2,_)) = n1 == n2
    (TyCon c1)   == (TyCon c2)   = c1 == c2
    (TApp t1 t2) == (TApp t1' t2') = t1 == t1' && t2 == t2'
    _		 == _		   = False


isTApp (TApp _ _) = True
isTApp _          = False
isTyCon (TyCon _) = True
isTyCon _	  = False

varLoc :: TVar -> Loc
varLoc (_, l) = l

newVarLoc :: Loc -> TVar -> TVar
newVarLoc l (v, _) = (v, l)

typeLocs :: Type -> [Loc]
typeLocs t = map varLoc (fv t)


-- Primitive types

-- note: variable numbers -19 to -8 are left available
arrowName       = name "->" 0
listName	= nilName
nilName         = name "[]" (-1)
consName	= name ":"  (-2)
catName         = name "++" (-3)
trueName	= name "True"  (-4)
falseName	= name "False" (-5)
boolName	= name "Bool"  (-6)
undefinedName	= name "undefined" (-7)

vnilName	= name "|" (-8)
vconsName	= name "\\" (-9)

tupleName arity = name ("(" ++ replicate (arity-1) ',' ++ ")")
		       (fromIntegral (-(arity+20)))


-- True if the given name represents a tuple constructor
isTupleCon :: Name -> Bool
isTupleCon = isJust . tupleArity

tupleArity :: Name -> Maybe Int
tupleArity name = do
    let id = show name
    when (head id /= '(') (fail "")
    when (last id /= ')') (fail "")
    if and [ b | x <- (init.tail) id, let b = x == ',']
       then return (length id - 2)
       else fail ""

isMultiCon :: Name -> Bool
isMultiCon = isJust . multiArity
 
multiArity :: Name -> Maybe Int
multiArity name = do
    case nameStr name of
	'{':xs -> let (n,r) = span isDigit xs
		  in  case r of
			"}" -> return (read n)
			_   -> Nothing
	_ -> Nothing
    
    
-- if the given type is a tuple, return the elements of that tuple type, 
-- else Nothing
tupleTypes :: Type -> Maybe [Type]
tupleTypes t = do lc <- tts t
		  return (lc [])
    where   tts :: Type -> Maybe ([Type] -> [Type])
	    tts (TApp (TyCon tc) t1)
		| isTupleCon tc = return ([t1] ++ )
		| otherwise     = fail ""
	    tts (TApp t1 t2) = do lc <- tts t1
				  return ((lc [t2]) ++ )
	    tts (TyCon tc) | tc == tupleName 0 = return (const [])
	    tts _ = fail ""

-- as above, but for multiple types represented as one
multiTypes :: Type -> Maybe [Type]
multiTypes t = do lc <- mts t
		  return (lc [])
    where   mts :: Type -> Maybe ([Type] -> [Type])
	    mts (TApp (TyCon tc) t1)
		| isMultiCon tc = return ([t1] ++ )
		| otherwise     = fail ""
	    mts (TApp t1 t2) = do lc <- mts t1
				  return ((lc [t2]) ++ )
	    mts (TyCon tc) | tc == multiTypeName 0 = return (const [])
	    mts _ = fail ""



-- Substitutions

type Subst = [(TVar,Type)]

idSubst :: Subst
idSubst = []

-- The state
type State = Integer

-- State monad for new variable supply
data TP a = TP (State -> E (State,a))
 

-------------------------------------
-- Class and instance declarations --
-------------------------------------


class TC a where
   apply :: Subst -> a -> a
   fv :: a -> [TVar]
   occurs :: TVar -> a -> Bool 
   rename :: a -> TP a
   format :: a -> String


unTP (TP a) = a
instance Monad TP where
   m >>= k  = TP ( \ s -> case ((unTP m) s) of 
                            Succ (s',x) -> unTP (k x) s'
                            Err e       -> Err e )
   return x = TP ( \ s -> Succ (s,x) )

instance TC a => TC [a] where
   apply s = map (apply s)
   fv x = foldl f [] x 
           where
            f z y = z ++ (fv y)        -- note there might duplicate variables
   occurs v x = foldl f False x 
                 where
                   f z y = z || (occurs v y)
   rename x = do s <- renameSubst (fv x)
                 return (apply s x)
   format xs = f xs
                    where
		     f [] = ""
		     f [x] = (format x)
		     f (x:xs) = (format x) ++ ", " ++ (f xs)

instance (TC a, TC b) => TC (a, b) where
    apply s (a, b) = (apply s a, apply s b)
    fv (a, b) = fv a ++ fv b
    occurs v (a, b) = occurs v a || occurs v b
    rename ab = renameSubst (fv ab) >>= \s -> return (apply s ab)
    format (a, b) = "(" ++ format a ++ "," ++ format b ++ ")"
    
instance (TC a, TC b, TC c) => TC (a, b, c) where
    apply s (a, b, c) = (apply s a, apply s b, apply s c)
    fv (a, b, c) = fv a ++ fv b ++ fv c
    occurs v (a, b, c) = occurs v a || occurs v b || occurs v c
    rename abc = renameSubst (fv abc) >>= \s -> return (apply s abc)
    format (a, b, c) = "(" ++ format a ++ "," ++ format b ++ "," ++ 
			      format c ++ ")"


instance PPrint Type where
    pprint t = return (text (show t))

instance Show Type where
    -- prec 0: infix
    -- prec 1: application
    -- prec 2: atomic
    showsPrec n (Var (x,i)) ss = show x {- ++ "." ++ show i -} ++ ss
    showsPrec n (TyCon x)   ss = show x ++ ss
    showsPrec n t@(TApp t1 t2) ss =
	let no_con =
	     if n > 1
		then '(' : showsPrec 1 t1 (" " ++ showsPrec 2 t2 (')' : ss))
		else showsPrec 1 t1 (" " ++ showsPrec 2 t2 ss)
	    non_lst =  let mts = tupleTypes t 
			   ts  = fromJust mts
			   mms = multiTypes t
			   ms  = fromJust mms
		       in  if isNothing mts
			      then if isNothing mms
				   then no_con
				   else unwords (map show ms) ++ ss
			      else '(': Term.uncommas (map show ts) ++ ")" ++ ss
	in case t1 of
	    (TApp (TyCon (Name ("->",_))) t1) ->
		if n > 0
		then '(' : showsPrec 1 t1 (" -> " ++ showsPrec 0 t2 (')' : ss))
		else showsPrec 1 t1 (" -> " ++ showsPrec 0 t2 ss)
	    
	    (TyCon name) | name == nilName -> "[" ++ show t2 ++ "]" ++ ss
--			 | otherwise       -> non_lst
	    t1 -> non_lst
		  

showTypeLoc x = showTypeLocPrec 0 x ""
showTypeLocPrec n (Var (x,i)) ss = "t" ++ show i ++ ss
showTypeLocPrec n (TyCon x)   ss = show x ++ ss
showTypeLocPrec n t@(TApp t1 t2) ss =
	let no_con =
	     if n > 1
		then '(' : showTypeLocPrec 1 t1 (" " ++ showTypeLocPrec 2 t2 (')' : ss))
		else showTypeLocPrec 1 t1 (" " ++ showTypeLocPrec 2 t2 ss)
	    non_lst =  let mts = tupleTypes t 
			   ts  = fromJust mts
		       in  if isNothing mts
			      then no_con
			      else '(': Term.uncommas (map (\x -> showTypeLocPrec 0 x "") ts) ++ ")" ++ ss
	in case t1 of
	    (TApp (TyCon (Name ("->",_))) t1) ->
		if n > 0
		then '(' : showTypeLocPrec 1 t1 (" -> " ++ showTypeLocPrec 0 t2 (')' : ss))
		else showTypeLocPrec 1 t1 (" -> " ++ showTypeLocPrec 0 t2 ss)
	    
	    (TyCon name) -> if name == nilName 
			       then "[" ++ showTypeLocPrec 0 t2 "" ++ "]" ++ ss
			       else non_lst
	    t1 -> non_lst
	  
		  

instance TC Type where
   apply s (Var v) = lookupVar s v
   apply s (TyCon tc) = TyCon tc 
   apply s (TApp t1 t2) = TApp (apply s t1) (apply s t2)
   fv (Var v) = [v]
   fv (TyCon _) = []
   fv (TApp t1 t2) = nub $ (fv t1) ++ (fv t2)
   occurs v (Var v') = if (fst v) == (fst v') then True else False
   occurs v (TyCon _) = False
   occurs v (TApp t1 t2) = (occurs v t1) || (occurs v t2)
   rename t = do s <- renameSubst (fv t)
                 return (apply s t)
   format = show


-- we impose an order on types
-- variables less than anything
instance Ord Type where
 (<) (Var (x,_)) (Var (y,_)) = x < y
 (<) (Var _) _ = True
 (<) (TyCon x) (TyCon y) = (x < y) 
 (<) (TApp t1 t2) (TApp t1' t2') = if t1 == t1' then t2 < t2'
                                   else t1 < t1' 
 (<) _ _ = False


---------------
-- Functions --
---------------


-- lookup, update Substitutions 

lookupVar :: Subst -> TVar -> Type
lookupVar [] v = Var v
lookupVar ((v,t):s) v'
   | (fst v) == (fst v')   = t
   | otherwise		   = lookupVar s v'


updateSubst :: Subst -> TVar -> Type -> Subst
updateSubst s v t = (v,t):(map f s)
     where 
       f (v',t') = (v',apply [(v,t)] t')

-- new variable or type name, renaming types

newTVar :: TP TVar
newTVar = TP (\n -> return (n+1, (name ("a"++(show n)) n, noLoc)))

newTCon :: String -> TP TCon
newTCon id = TP (\n -> return (n+1, name id n))

newInteger :: TP Integer
newInteger = TP (\n -> return (n+1,n))



-- compute renaming substitution for set vs of variables
{-
renameSubst :: [TVar] -> TP Subst
renameSubst vs = let rensubst = foldl f (return idSubst) vs
                     f z y = do s <- z
                                n <- newTVar
                                return (updateSubst s y (Var n))
                 in rensubst 
-}

-- maintains the original location of the renamed variable
renameSubst :: [TVar] -> TP Subst
renameSubst vs = let rensubst = foldl f (return idSubst) vs
                     f z y@(n,l) = do s <- z
                                      (n',_) <- newTVar
				      let y' = (n',l)
				      return (updateSubst s y (Var y'))
                 in rensubst 


-- in addition we return the set of all freshly created variables

renameSubst2 :: [TVar] -> TP (Subst,[TVar])
renameSubst2 vs = let rensubst = foldl f (return (idSubst,[])) vs
                      f z y = do (s,ns) <- z
                                 n <- newTVar
                                 return (updateSubst s y (Var n),n:ns)
                  in rensubst 

 
-- substitution will be threaded through
-- tvs is a list of variables which should not be substituted
-- by another type variable not contained in tvs

unify :: [TVar] -> Subst -> [(Type,Type)] -> E Subst
unify tvs subst ts = foldl f (return subst) ts
   where
     f z (t1,t2) = do s <- z
                      s' <- unify' (s, apply s t1, apply s t2)
                      return s'
     -- substitution will be threaded through
     -- assumption: substitution is already applied to types
     tvarNotElem (v, _) vs = v `notElem` (map fst vs)
     unify' :: (Subst, Type, Type) -> E Subst
     unify' (s, Var x, Var y) = 
               if (fst x) == (fst y) then return s
               else if tvarNotElem x tvs then return (updateSubst s x (Var y))   -- this is crucial !!
                    else return (updateSubst s y (Var x))
     unify' (s, Var x, t) =
               if (occurs x t) then errorE ("occurs check " ++ show t) idSubst
               else return (updateSubst s x t)  
     unify' (s, t, Var x) = unify' (s, Var x, t)
     unify' (s, TyCon tc1, TyCon tc2) =
         if tc1 /= tc2 then errorE "no mgu" idSubst
         else return s
     unify' (s, TApp t1 t2, TApp t1' t2') = 
         do s' <- unify' (s, t1, t1')
            unify' (s', apply s' t2, apply s' t2') 
     unify' (s, _, _) = errorE "no mgu" idSubst

     
-- as above, but keeps identity substitutions (ie. x |-> x)
unifyWithIds :: [TVar] -> Subst -> [(Type,Type)] -> E Subst
unifyWithIds tvs subst ts = foldl f (return subst) ts
   where
     f z (t1,t2) = do s <- z
                      s' <- unify' (s, apply s t1, apply s t2)
                      return s'
     -- substitution will be threaded through
     -- assumption: substitution is already applied to types
     tvarNotElem (v, _) vs = v `notElem` (map fst vs)
     unify' :: (Subst, Type, Type) -> E Subst
     unify' (s, Var x, Var y) = 
               -- if (fst x) == (fst y) then return s
               --else 
	       if tvarNotElem x tvs 
		  then let s1 = updateSubst s x (Var y)
		       in  return (updateSubst s1 y (Var y))
                  else let s1 = updateSubst s y (Var x)
		       in  return (updateSubst s1 x (Var x))
     unify' (s, Var x, t) =
               if (occurs x t) then errorE ("occurs check " ++ show t) idSubst
               else return (updateSubst s x t)  
     unify' (s, t, Var x) = unify' (s, Var x, t)
     unify' (s, TyCon tc1, TyCon tc2) =
         if tc1 /= tc2 then errorE "no mgu" idSubst
         else return s
     unify' (s, TApp t1 t2, TApp t1' t2') = 
         do s' <- unify' (s, t1, t1')
            unify' (s', apply s' t2, apply s' t2') 
     unify' (s, _, _) = errorE "no mgu" idSubst


-- variation of unify
-- we assume that variables in tvs are fixed, i.e. are not allowed to be 
-- substituted
unifyTypes :: [TVar] -> Subst -> [(Type,Type)] -> E Subst
unifyTypes tvs subst ts = foldl f (return subst) ts
   where
     f z (t1,t2) = do s <- z
                      s' <- unify' (s, apply s t1, apply s t2)
                      return s'
     -- substitution will be threaded through
     -- assumption: substitution is already applied to types
     unify' :: (Subst, Type, Type) -> E Subst
     unify' (s, Var x, Var y) = 
               if x == y then return s
               else if notElem x tvs then return (updateSubst s x (Var y))   -- this is crucial !!
                    else if notElem y tvs then return (updateSubst s y (Var  x)) -- again
                         else errorE "fail: variable is fixed" idSubst
     unify' (s, Var x, t) =
               if (occurs x t) then errorE "fail: occurs check" idSubst
               else if notElem x tvs then return (updateSubst s x t)  
                    else errorE "fail: variable is fixed" idSubst
     unify' (s, t, Var x) = unify' (s, Var x, t)
     unify' (s, TyCon tc1, TyCon tc2) =
         if tc1 /= tc2 then errorE "fail: no mgu" idSubst
         else return s
     unify' (s, TApp t1 t2, TApp t1' t2') = 
         do s' <- unify' (s, t1, t1')
            unify' (s', apply s' t2, apply s' t2') 
     unify' (s, _, _) = errorE "fail: no mgu" idSubst


-- type matching

-- a version of `apply' on types that applies the subst. at most once, and also 
-- returns the justification on the variable substituted for
applyType s t = snd (subType False s t)

-- returns True if a subst. was made
subType :: Bool -> Subst -> Type -> (Bool, (Type, Loc))
subType b s tc@(TyCon _) = (b, (tc, noLoc))
subType b s (TApp t1 t2) = let (b1, (t1', j1)) = subType b s t1 
			       (b2, (t2', j2)) = subType b s t2
			       b' = b1 || b2
			   in  if b1 -- not (nullJust j1)
				  then (b1, (TApp t1' t2, j1))
				  else if b2 -- not (nullJust j2)
					  then (b2, (TApp t1 t2', j2))
					  else (b', (TApp t1 t2, noLoc))
subType b s (Var v) = sub s v
    where sub [] v = (False, (Var v, noLoc))
	  sub (((n, j), t):s) v
	    | n == (fst v)  = (True, (t, j))
	    | otherwise	    = sub s v
 

--------------------------------------------------------------------------------

-- assumption both lists are of equal size
-- we match elements at the same position
-- matchTypes ts1 ts2 = True if there's phi s.t. phi ts1 = ts2
matchTypes :: [Type] -> [Type] -> Bool
matchTypes [] [] = True
matchTypes ts1 ts2 = let vars (Var v) vs= if elem v vs then vs 
                                          else v:vs
                         vars (TyCon _) vs = vs
                         vars (TApp t1 t2) vs = let vs'= vars t1 vs
                                                in vars t2 vs'
                         varss [t] vs = vars t vs
                         varss (t:ts) vs = let vs' = vars t vs
                                           in varss ts vs'
                         skconst (Var (v,_)) = v
                         skconst (TyCon c) = c
                         skconst (TApp t1 t2) = max (skconst t1) (skconst t2)
                         skconsts [t] = skconst t
                         skconsts (t:ts) = max (skconst t) (skconsts ts)
                         sk = skconsts (ts1++ts2)
                         vs = varss (ts2) []
                         (skolemsubst,_) = foldl f (idSubst, incName sk) vs
                         f (subst,sk) v = (updateSubst subst v (TyCon sk), incName sk)
                         ts2'= apply skolemsubst ts2
                     in case (unify [] idSubst (zip ts1 ts2')) of
                         Succ _ -> True
                         Err  _ -> False 

-- matchTypes2 vs s1 t1 t2 = s2
-- find renaming substitution s2 (for v not in vs) such that t1 = s t2
-- otherwise report failure
matchTypes2 :: [TVar] -> Subst -> [Type] -> [Type] -> E Subst
matchTypes2 vs s [] [] = return s 
matchTypes2 vs s ts1 ts2 = foldl f (return s) (zip ts1 ts2)
  where
    f z (t1,t2) = do s <- z
                     s' <- matchType2 vs s t1 (apply s t2)
                     return s'       
    matchType2 :: [TVar] -> Subst -> Type -> Type -> E Subst
    matchType2 vs s (Var x) (Var y) = 
                     if x == y then return s
                     else if (notElem y vs) then return (updateSubst s y (Var x))
                          else Err "no renaming found (0)"
			      
    matchType2 vs s t (Var y) = 
                     if (notElem y vs) then return (updateSubst s y t)
                     else Err "no renaming found (1)"

    matchType2 vs s (TyCon tc1) (TyCon tc2) = if tc1 == tc2 then return s
                                              else Err "no renaming found (2)"
    matchType2 vs s (TApp t1 t2) (TApp t1' t2') =
                do s' <- matchType2 vs s t1 t1'
                   matchType2 vs s' t2 (apply s' t2')
    matchType2 _ _ _ _ = Err "no renaming found (3)"


-- signal whether there exists renaming phi such that phi ts1 = ts2
matchTypes3 :: [Type] -> [Type] -> Bool
matchTypes3 ts1 ts2 = case (matchTypes2 [] idSubst ts1 ts2) of
                       Succ _ -> True
                       Err _ -> False

-- report error

-- x is default parameter, thrown away in the current implementation
errorTP :: Show a => String -> a -> TP a
errorTP text x = TP (\n -> Err msg) 
    where msg = text ++ "\n" 


-- x is default parameter, thrown away in the current implementation
errorE :: String -> a -> E a
errorE text x = Err text


--------------------------------------------------------------------------------
-- builds a renaming subst. for the given type vars; to give them more
-- palatable names

simpTVarSubst :: [TVar] -> Subst
simpTVarSubst tvs = foldl (\s (v,t) -> updateSubst s v t) idSubst (zip tvs typs)
    where
	typs = map (\id -> Var (name id 0, noLoc)) strs
    
	strs = "a" : map incStr strs
    
	incStr ""  = "a"
	incStr str = let (c:cs) = reverse str
			 str'   = reverse (succ c : cs)
		     in  doCarry str'

	-- while last char. is > 'z', set to 'a' and inc. the prefix
	doCarry str = let (c:cs) = reverse str
		      in  if c > 'z'
			     then reverse ('a' : (incStr (reverse cs)))
			     else str


presentableVars :: TC a => a -> a
presentableVars a = let vs = nub (fv a)
			phi = simpTVarSubst vs
		    in  apply phi a
